// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:28 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include <stdlib.h>
#include <list>
#include "command.h"
#include "dagnode.h"
#include "datablock.h"

// ************************************************************************* //
/// DESCRIPTION: Looks for string _m_name in parameters. 
///  If found, sets parameter to true, else false.
///  No error possible with this, right?
// ************************************************************************* //
bool	ParameterSignal::browse(const String	&parameters,
				String		&)
{
  std::list<String>	words;
  std::list<String>::iterator	wit;

  parameters.chopInWords(words);
  for (wit = words.begin(); wit != words.end(); ++wit)
    if (*wit == String("-") + _m_name)
      {
        _m_value = true;
	return true;
      }
  _m_value = false;
  return true;
}

// ************************************************************************* //
/// DESCRIPTION: Looks for string _m_name in parameters. 
///  If found, gets the following word and uses it as a parameter.
// ************************************************************************* //
bool	ParameterInt::browse(const String	&parameters,
			     String		&errorMsg)
{

  std::list<String>	words;
  std::list<String>::iterator	wit;

  parameters.chopInWords(words);
  for (wit = words.begin(); wit != words.end(); ++wit)
    if (*wit == String("-") + _m_name)
      {
	if (!(++wit)->isInt() || wit == words.end())
	  return false;
	_m_value = wit->asInt();
	return true;
      }
  if (_m_hasDefault)
    {
      _m_value = _m_defaultValue;
      return true;
    }
  errorMsg = String("Parameter \"") + _m_name + "\" needs to be specified";
  return false;
}

// ************************************************************************* //
/// DESCRIPTION:
// ************************************************************************* //
bool	ParameterPnt3d::browse(const String	&parameters,
			       String		&errorMsg)
{
  std::list<String>	words;
  std::list<String>::iterator	wit;

  parameters.chopInWords(words);
  for (wit = words.begin(); wit != words.end(); ++wit)
    if (*wit == String("-") + _m_name)
      {
	if (!(++wit)->isDouble() || wit == words.end())
	  return false;
	_m_value.x = wit->asDouble();
	if (!(++wit)->isDouble() || wit == words.end())
	  return false;
	_m_value.y = wit->asDouble();
	if (!(++wit)->isDouble() || wit == words.end())
	  return false;
	_m_value.z = wit->asDouble();
	return true;
      }
  if (_m_hasDefault)
    {
      _m_value = _m_defaultValue;
      return true;
    }
  errorMsg = String("Parameter \"") + _m_name + "\" needs to be specified";
  return false;
}

// ************************************************************************* //
/// DESCRIPTION:
// ************************************************************************* //
bool	ParameterVec3d::browse(const String	&parameters,
			       String		&errorMsg)
{
  std::list<String>	words;
  std::list<String>::iterator	wit;

  parameters.chopInWords(words);
  for (wit = words.begin(); wit != words.end(); ++wit)
    if (*wit == String("-") + _m_name)
      {
	if (!(++wit)->isDouble() || wit == words.end())
	  return false;
	_m_value.x = wit->asDouble();
	if (!(++wit)->isDouble() || wit == words.end())
	  return false;
	_m_value.y = wit->asDouble();
	if (!(++wit)->isDouble() || wit == words.end())
	  return false;
	_m_value.z = wit->asDouble();
	return true;
      }
  if (_m_hasDefault)
    {
      _m_value = _m_defaultValue;
      return true;
    }
  errorMsg = String("Parameter \"") + _m_name + "\" needs to be specified";
  return false;
}

// ************************************************************************* //
bool	ParameterString::browse(const String	&parameters,
				String		&errorMsg)
{
  std::list<String>	words;
  std::list<String>::iterator	wit;

  parameters.chopInWords(words);
  for (wit = words.begin(); wit != words.end(); ++wit)
    if (*wit == String("-") + _m_name)
      {
	if (++wit == words.end())
	  {
	    errorMsg = String("Parameter \"") + _m_name
	      + "\" needs to be specified";
	    return false;
	  }
	_m_value = *wit;
	return true;
      }
  if (0 < _m_sel)
    {
      std::list<DagNode *>	nodes;
      std::list<DagNode *>::iterator	nit;
      unsigned	i;

      DagNode::getNodesSelected(nodes);
      for (nit = nodes.begin(), i = 1; nit != nodes.end(); ++nit, ++i)
	if (i == _m_sel)
	  {
	    _m_value = (*nit)->path();
	    return true;
	  }
    }
  if (_m_hasDefault)
    {
      _m_value = _m_defaultValue;
      return true;
    }
  errorMsg = String("Parameter \"") + _m_name + "\" needs to be specified";
  return false;
}

// ************************************************************************* //
void	Command::addSignal(const String &name)
{
  _m_parameters[name] = new ParameterSignal(name);
}

// ************************************************************************* //
void	Command::addInt(const int	defaultValue,
			const String	&name)
{
  _m_parameters[name] = new ParameterInt(defaultValue, name);
}

// ************************************************************************* //
void	Command::addInt(const String &name)
{
  _m_parameters[name] = new ParameterInt(name);
}

// ************************************************************************* //
void	Command::addPnt3d(const Pnt3d	&defaultValue,
			  const String	&name)
{
  _m_parameters[name] = new ParameterPnt3d(defaultValue, name);
}

// ************************************************************************* //
void	Command::addPnt3d(const String &name)
{
  _m_parameters[name] = new ParameterPnt3d(name);
}

// ************************************************************************* //
void	Command::addVec3d(const Vec3d	&defaultValue,
			  const String	&name)
{
  _m_parameters[name] = new ParameterVec3d(defaultValue, name);
}

// ************************************************************************* //
void	Command::addVec3d(const String &name)
{
  _m_parameters[name] = new ParameterVec3d(name);
}

// ************************************************************************* //
void	Command::addString(const String		&defaultValue,
			   const String		&name,
			   const unsigned	sel)
{
  _m_parameters[name] = new ParameterString(defaultValue, name, sel);
}

// ************************************************************************* //
void	Command::addString(const String		&name,
			   const unsigned	sel)
{
  _m_parameters[name] = new ParameterString(name, sel);
}

// ************************************************************************* //
bool	Command::readAsSignal(const String &name)
{
  Parameter	*p;

  if ((p = _m_parameters[name]) == 0)
    {
      std::cerr << "Invalid parameter request." << std::endl;
      std::exit(EXIT_FAILURE);
    }
  return p->readAsSignal();
}

// ************************************************************************* //
int	Command::readAsInt(const String &name)
{
  Parameter	*p;

  if ((p = _m_parameters[name]) == 0)
    {
      std::cerr << "Invalid parameter request." << std::endl;
      std::exit(EXIT_FAILURE);
    }
  return p->readAsInt();
}

// ************************************************************************* //
const Pnt3d	&Command::readAsPnt3d(const String &name)
{
  Parameter	*p;

  if ((p = _m_parameters[name]) == 0)
    {
      std::cerr << "Invalid parameter request." << std::endl;
      std::exit(EXIT_FAILURE);
    }
  return p->readAsPnt3d();
}

// ************************************************************************* //
const Vec3d	&Command::readAsVec3d(const String &name)
{
  Parameter	*p;

  if ((p = _m_parameters[name]) == 0)
    {
      std::cerr << "Invalid parameter request." << std::endl;
      std::exit(EXIT_FAILURE);
    }
  return p->readAsVec3d();
}

// ************************************************************************* //
String	Command::readAsString(const String &name)
{
  Parameter	*p;

  if ((p = _m_parameters[name]) == 0)
    {
      std::cerr << "Invalid parameter request." << std::endl;
      std::exit(EXIT_FAILURE);
    }
  return p->readAsString();
}

// ************************************************************************* //
bool	Command::browse(const String	&parameters,
			String		&errorMsg)
{
  std::map<String, Parameter*>::iterator	pit;

  for (pit = _m_parameters.begin(); pit != _m_parameters.end(); ++pit)
    if (!(*pit).second->browse(parameters, errorMsg))
      { 
	errorMsg += String("\nUsage: ") + usage();
	return false;
      }
  return true;
}

// ************************************************************************* //
/// DESCRIPTION: Parses string and chops it for commands.
// ************************************************************************* //
DagNode	*Command::run(const String	&cmd)
{
  /*
  std::list<String>		words;
  std::list<String>::iterator	wit;

  cmd.chopInWords(words);
  for (wit = words.begin(); wit != words.end(); ++wit)
    std::cout << *wit << ' ';
  std::cout << std::endl;
  */
  
  String	cmdStr(cmd);

  cmdStr.skip(' ');
  cmdStr = cmdStr.chopBack(' ');

  const int id = Command::names2id()[cmdStr];

  if (id == 0)
    {
      std::cerr << "Error: unknown command " << cmdStr << std::endl;
      return 0;
    }
  
  const String	paramStr(cmd.chopFront(' '));

  Command	*cmdInst = Command::creators()[id]();
  DagNode	*ret = cmdInst->execute(paramStr);

  delete cmdInst;
  return ret;
}

// ************************************************************************* //
String	Command::usage() const
{
  String	ret;
  std::map<String, Parameter*>::const_iterator	pit;

  ret = getName() + " ";

  for (pit = _m_parameters.begin(); pit != _m_parameters.end(); ++pit)
    ret += String("-") + (*pit).first + " " + (*pit).second->typeStr() + " ";
  return ret;
}
