// ************************************************************************* //
/// AUTHOR: Alexis Angelidis.
/// STARTED: Thu Sep 22 11:47:50 NZST 2005
// ************************************************************************* //
#ifndef	__DATAGRID_HH__
#define	__DATAGRID_HH__

// ************************************************************************* //
#include "plug.h"
#include "plugoctree.h"
#include "gfxtypes.h"

// ************************************************************************* //
template <class T>
class Grid
{
  BBox3d	_m_bbox;
  unsigned	_m_n;
  T		*_m_m;
  Mat4d		_m_local2world;
  Mat4d		_m_world2local;
public:
  Grid(): _m_n(0),  _m_m(0) {}
  Grid(const Grid<T> &g):
    _m_bbox(g._m_bbox), _m_n(g._m_n), _m_m(new T[size()]),
    _m_local2world(g._m_local2world), _m_world2local(g._m_world2local)
  { for (unsigned i(0), n3(size()); i < n3; ++i) _m_m[i] = m._m_m[i]; }
  Grid(const BBox3d	&bbox,
       const unsigned	n):
    _m_bbox(bbox), _m_n(n), _m_m(new T[size()])
  {
    const Vec3d	diag = (bbox.pmax - bbox.pmin) * (1.0 / (_m_n - 1.0));

    _m_local2world = Mat4d(Vec3d(diag.x, 0.0, 0.0),
			   Vec3d(0.0, diag.y, 0.0),
			   Vec3d(0.0, 0.0, diag.z), bbox.pmin);
    if (1.0e-12 < fabs(_m_local2world.det()))
      _m_world2local = _m_local2world.inverse();
  }
  virtual ~Grid() { if (_m_m) delete[] _m_m; }
  Grid<T>	&operator=(const Grid<T> &g)
  {
    resize(g._m_n);
    for (unsigned i(0), n3(g.size()); i < n3; ++i)
      _m_m[i] = g._m_m[i];
    _m_bbox = g._m_bbox;
    _m_local2world = g._m_local2world;
    _m_world2local = g._m_world2local;
    return *this;
  }
  T		&operator()(unsigned i, unsigned j, unsigned k)
  { return _m_m[_m_n * ((_m_n * k) + j) + i]; }
  const T	&operator()(unsigned i, unsigned j, unsigned k) const
  { return _m_m[_m_n * ((_m_n * k) + j) + i]; }

  const T	&query(const Pnt3d &pworld) const
  {
    const Pnt3d	plocal(_m_world2local * pworld);

    return operator()(int(nearbyint(bound(plocal.x, 0.0, _m_n - 1.0))),
		      int(nearbyint(bound(plocal.y, 0.0, _m_n - 1.0))),
		      int(nearbyint(bound(plocal.z, 0.0, _m_n - 1.0))));
  }
  BBox3d	bbox(unsigned i, unsigned j, unsigned k) const
  {
    const Pnt3d	pworld(_m_local2world * Pnt3d(i, j, k));
    const Vec3d	diag = (_m_bbox.pmax - _m_bbox.pmin) * (0.5 / (_m_n - 1.0));

    return BBox3d(pworld - diag, pworld + diag);
  }
  void	place(const BBox3d &b) { _m_bbox = b; }
  void	glRender() const
  {
    _m_bbox.glRender();
  }
  unsigned	size() const { return _m_n * _m_n * _m_n; }
  void		resize(const unsigned n)
  {
    if (size() < n * n * n)
      {
	if (_m_m)
	  delete[] _m_m;
	_m_m = new T[n * n * n];
      }
    _m_n = n;
  }
};

// ************************************************************************* //
#endif//__DATAGRID_HH__
