// ************************************************************************* //
/// AUTHOR: Alexis Angelidis.
/// STARTED: Fri Jul  1 20:49:37 NZST 2005
/// UPDATES: 
// ************************************************************************* //
#ifndef __STRING_HH__
#define __STRING_HH__

// ************************************************************************* //
#include <iostream>
#include <list>

// ************************************************************************* //
class String
{
  char	*_m_m;
public:
  String(): _m_m(0) {}
  String(const char *s)
  { if (s) _m_m = new char[strlen(s) + 1], strcpy(_m_m, s); else _m_m = 0; }
  String(const String &s)
  { if (s._m_m) _m_m = new char[s.len() + 1], cpy(s); else _m_m = 0; }
  String(double f);
  String(int n);
  String(const unsigned int n);
  virtual ~String() { if (_m_m) delete[] _m_m; _m_m = 0; }
  friend std::ostream	&operator<<(std::ostream &os, const String &s)
  { return s._m_m? (os << s._m_m): os; }
  friend std::istream	&operator>>(std::istream &is, String &s);
  const char	*data() const { return _m_m; }
  char		&operator[](const int i) { return _m_m[i]; }
  const char	&operator[](const int i) const { return _m_m[i]; }
  char		&operator()(const int i) { return _m_m[i]; }
  const char	&operator()(const int i) const { return _m_m[i]; }
  int		len() const { return _m_m? strlen(_m_m): 0; }
  void		cpy(const String &src);
  void		resize(const int n)
  { if (_m_m) delete[] _m_m; _m_m = new char[n]; }
  String	&operator=(const String &s);
  String	operator+(const String &s) const;
  String	&operator+=(const String &s);
  bool		operator==(const char *s) const;
  bool		operator!=(const char *s) const;
  bool		operator==(const String &s) const;
  bool		operator!=(const String &s) const;
  bool		operator<(const String &s) const;
  String	chopBack(const char c) const;
  String	chopFront(const char c) const;
  bool		contains(const char c) const;
  int		contains(const String &str) const;
  bool		isDouble() const;
  double	asDouble() const { return atof(_m_m); }
  bool		isInt() const;
  int		asInt() const { return atoi(_m_m); }
  bool		empty() const { return !_m_m || _m_m[0] == 0; }
  void		chopInWords(std::list<String> &words) const;
  void		skip(const char c);
  void		saveAscii(std::ostream &os);
  void		loadAscii(std::istream &is);
};

// ************************************************************************* //
/// DESCRIPTION: Returns true if it encodes a double.
// ************************************************************************* //
inline bool	String::isDouble() const
{
  int i(0);

  if (!_m_m)
    return false;
  if (_m_m[i] == '-')
    ++i;
  for (; _m_m[i] && ('0' <= _m_m[i] && _m_m[i] <= '9'); ++i)
    ;
  if (_m_m[i] == 0)
    return true;
  if (_m_m[i] != '.')
    return false;
  i++;
  for (; _m_m[i] && ('0' <= _m_m[i] && _m_m[i] <= '9'); ++i)
    ;
  if (_m_m[i] == 0)
    return true;
  if (_m_m[i] != 'e' && _m_m[i] != 'E')
    return false;
  ++i;
  if (_m_m[i] == '-' || _m_m[i] == '+')
    ++i;
  if (_m_m[i] < '0' || '9' < _m_m[i])
    return false;
  ++i;
  for (; _m_m[i] && ('0' <= _m_m[i] && _m_m[i] <= '9'); ++i)
    ;
  if (_m_m[i] == 0)
    return true;
  return false;
}

// ************************************************************************* //
/// DESCRIPTION: Returns true if it encodes an int.
// ************************************************************************* //
inline bool	String::isInt() const
{
  if (_m_m)
    for (int i(0); _m_m[i]; ++i)
      if (_m_m[i] < '0' || '9' < _m_m[i])
	return false;
  return true;
}

// ************************************************************************* //
inline String::String(double f)
{
  char	buf[256];
  int	k(0);

  sprintf(buf, "%e", f);
  _m_m = new char[strlen(buf) + 1];
  for (; buf[k] != 0; ++k)
    _m_m[k] = buf[k];
  _m_m[k] = 0;
}

// ************************************************************************* //
inline String::String(int n)
{
  char	buf[256];
  int	k(0);

  sprintf(buf, "%i", n);
  _m_m = new char[strlen(buf) + 1];
  for (; buf[k] != 0; ++k)
    _m_m[k] = buf[k];
  _m_m[k] = 0;
}

// ************************************************************************* //
inline String::String(const unsigned int n)
{
  unsigned int	pad(0);

  for (unsigned int npad(n); npad; npad /= 10)
    pad++;
  if (pad == 0)
    pad = 1;
  _m_m = new char[1 + pad];
  for (unsigned int i(0), npad(n); i < pad; ++i, npad /= 10)
    _m_m[pad - i - 1] = '0' + npad % 10;
  _m_m[pad] = 0;
}

// ************************************************************************* //
inline void	String::cpy(const String &src)
{
  if (src._m_m)
    strcpy(_m_m, src._m_m);
}

// ************************************************************************* //
/// DESCRIPTION: 
// ************************************************************************* //
inline String	&String::operator=(const String &s)
{
  resize(s.len() + 1);
  cpy(s);
  return *this;
}

// ************************************************************************* //
inline String	String::operator+(const String &s) const
{
  String	ret;
  const int	l1(len()), l2(s.len());

  ret.resize(l1 + l2 + 1);
  ret.cpy(*this);
  for (int i(0); i < l2; ++i)
    ret[l1 + i] = s[i];
  ret[l1 + l2] = 0;
  return ret;
}

// ************************************************************************* //
inline String	&String::operator+=(const String &s)
{
  String	ret = *this;
  const int	l1(len()), l2(s.len());

  resize(l1 + l2 + 1);
  cpy(ret);
  for (int i(0); i < l2; ++i)
    operator[](l1 + i) = s[i];
  operator[](l1 + l2) = 0;
  return *this;
}

// ************************************************************************* //
/// DESCRIPTION: Uses dictionnary with ascii character order.
// ************************************************************************* //
inline bool	String::operator<(const String &s) const
{
  const int	l1(len()), l2(s.len());
  const int	l0((l1 < l2)? l1: l2);

  for (int i(0); i < l0; ++i)
    {
      const char	c0(_m_m[i]), c1(s[i]);

      if (c0 < c1)
	return true;
      else if (c1 < c0)
	return false;
    }
  return l1 < l2;
}

// ************************************************************************* //
/// DESCRIPTION: Returns true if strings are identical.
// ************************************************************************* //
inline bool	String::operator==(const String &s) const
{
  int	i;

  for (i = 0; _m_m[i] && s._m_m[i]; ++i)
    if (_m_m[i] != s._m_m[i])
      return false;
  return !_m_m[i] && !s._m_m[i];
}

// ************************************************************************* //
/// DESCRIPTION: Returns true if strings are identical.
// ************************************************************************* //
inline bool	String::operator!=(const String &s) const
{
  return !(*this == s);
}

// ************************************************************************* //
/// DESCRIPTION: Returns true if strings are identical.
// ************************************************************************* //
inline bool	String::operator==(const char *s) const
{
  int	i;

  for (i = 0; _m_m[i] && s[i]; ++i)
    if (_m_m[i] != s[i])
      return false;
  return !_m_m[i] && !s[i];
}

// ************************************************************************* //
/// DESCRIPTION: Returns true if strings are identical.
// ************************************************************************* //
inline bool	String::operator!=(const char *s) const
{
  return !(*this == s);
}

// ************************************************************************* //
/// DESCRIPTION: Chops the back of the string up to character
///  (i.e. keeps the front) c and ads a 0 at the end.
// ************************************************************************* //
inline String	String::chopBack(const char c) const
{
  if (!_m_m)
    return *this;

  String	ret;
  int		l(0);

  for ( ; _m_m[l] != 0 && _m_m[l] != c; ++l)
    ;
  ret.resize(l + 1);
  for (int i(0); i < l; ++i)
    ret[i] = _m_m[i];
  ret[l] = 0;
  return ret;
}

// ************************************************************************* //
/// DESCRIPTION: Chops the front of the string up to character
///  (i.e. keeps the back) c and ads a 0 at the end.
// ************************************************************************* //
inline String	String::chopFront(const char c) const
{
  if (!_m_m)
    return *this;

  String	ret;
  int		l(len()), s(0);

  for ( ; _m_m[s] != 0 && _m_m[s] != c; ++s)
    ;
  if (_m_m[s] == 0)
    return ret;
  ++s;
  ret.resize(l - s + 1);
  for (int i(0); i < l - s; ++i)
    ret[i] = _m_m[s + i];
  ret[l - s] = 0;
  return ret;
}

// ************************************************************************* //
/// DESCRIPTION: Returns true if the string contains the character c.
// ************************************************************************* //
inline bool	String::contains(const char c) const
{
  if (!_m_m)
    return false;
  for (int i(0); _m_m[i]; ++i)
    if (_m_m[i] == c)
      return true;
  return false;
}

// ************************************************************************* //
/// DESCRIPTION: Returns -1 if the string does not contains the string str,
///  else returns the index of str within the string.
// ************************************************************************* //
inline int	String::contains(const String &str) const
{
  if (!_m_m)
    return -1;
  for (int i(0), n(len() - str.len()); i <= n; ++i)
    for (int j(0); _m_m[i + j] == str[j]; )
      if (str[++j] == 0)
	return i;
  return -1;
}

// ************************************************************************* //
/// DESCRIPTION: Produces a list of words, using ' ' as a separator and '\n'
///  or 0 as a terminator.
// ************************************************************************* //
inline void	String::chopInWords(std::list<String> &words) const
{
  char	*cbegin, *cend;

  if (!_m_m || !_m_m[0])
    return;
  for (cbegin = _m_m; *cbegin == ' ' && *cbegin && *cbegin != '\n'; ++cbegin)
    ;
  while (*cbegin && *cbegin != '\n')
    {
      String	curWord;

      for (cend = cbegin; *cend != ' ' && *cend && *cend != '\n'; ++cend)
	;
      curWord.resize(cend - cbegin + 1);
      strncpy(curWord._m_m, cbegin, cend - cbegin);
      curWord._m_m[cend - cbegin] = 0;
      words.push_back(curWord);
      for (cbegin = cend; *cbegin == ' ' && *cbegin && *cbegin != '\n';
	   ++cbegin)
	;
    }
}

// ************************************************************************* //
/// DESCRIPTION: Skips all the characters c at the biginning of the string.
// ************************************************************************* //
inline void	String::skip(const char c)
{
  if (_m_m[0] != c)
    return;

  int	ibegin(0), iend;

  for (; _m_m[ibegin] == c && _m_m[ibegin]; ++ibegin)
    ;
  for (iend = ibegin; _m_m[iend]; ++iend)
    _m_m[iend - ibegin] = _m_m[iend];
  _m_m[iend] = 0;
}

// ************************************************************************* //
inline void	String::saveAscii(std::ostream &os)
{
  const unsigned	l(len());

  os << l << ' ';
  for (unsigned i(0); i < l; ++i)
    os << _m_m[i];
}

// ************************************************************************* //
inline void	String::loadAscii(std::istream &is)
{
  unsigned	l;

  is >> l;
  if (l != 0)
    {
      _m_m = new char[l + 1];
      is.read(_m_m, 1);
      is.read(_m_m, l);
      _m_m[l] = 0;
    }
  else
    _m_m = 0;
}

// ************************************************************************* //
#endif//__STRING_HH__
