// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:29 $
/// $Revision: 1.1 $
// ************************************************************************* //
#ifndef __NODECAMERA_HH__
#define __NODECAMERA_HH__

// ************************************************************************* //
#include "plug.h"
#include "datablock.h"
#include "nodeshape.h"
#include "command.h"

// ************************************************************************* //
/// DESCRIPTION: Root of the scene graph (node that contains the whole thing).
// ************************************************************************* //
class NCamera: public NShape
{
public:
  static int	typeNbr;
  static String	typeStr;
public:
  NCamera(DagNode *parent = 0): NShape(parent) {}
  virtual void		initialize();
  virtual int		getTypeNbr() const { return typeNbr; }
  virtual const String	&getTypeStr() const { return typeStr; }
  static DagNode	*creator() { return new NCamera(); }
  virtual void		compute(Plug &plugOut);
  void			glRender(Plug &plugOut);
  void			buildDagNodes();
};

// ************************************************************************* //
/// DESCRIPTION: Extends a data node. This is stored in the scene graph.
// ************************************************************************* //
class PlugFrustum: public Plug
{
//protected:
public:
  Frustum	*_m_data;
public:
  static int	typeNbr;
  static String	typeStr;
public:
  PlugFrustum(DataBlock &db): Plug(db), _m_data(0) { newData(); }
  virtual ~PlugFrustum() { freeData(); }
  static Plug	*creator(DataBlock &db) { return new PlugFrustum(db); }
  virtual int	getTypeNbr() const { return typeNbr; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual void	freeData() { if (_m_data) delete _m_data; _m_data = 0; }
  virtual void	newData() { if (!_m_data) _m_data = new Frustum; }
  virtual void	saveData(std::ostream &os) { os << *_m_data; }
  virtual void	loadData(std::istream &is) { is >> *_m_data; }
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const Frustum*>(v); }
  virtual void	**rawdata() { return (void**)(&_m_data); }
};

// ************************************************************************* //
#endif//__NODECAMERA_HH__
