// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:29 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include "nodecontrol.h"
#include "plug.h"
#include "datablock.h"
#include "gbb.h"

// ************************************************************************* //
int	NControl::typeNbr(61);
String	NControl::typeStr("Control");

// ************************************************************************* //
void	NControl::initialize()
{
  DataBlock	&db(getDataBlock());

  NShape::initialize();
  db.addInt(0, MININT, MAXINT, "f", false);// frame
  db.addDouble(0.0, 0.0, MAXDOUBLE, "dt", false);
  db.addDouble(0.125, 0.0, MAXDOUBLE, "radius");
  db.addDouble(1.0, 0.0, MAXDOUBLE, "flatness");
  db.addDouble(3.0, 0.0, MAXDOUBLE, "angleVelocity");// rad per second
  db.addPnt3d(Pnt3d(0.0), "src", false);
  db.addVec3d(Vec3d(0.0, 0.0, 1.0), "dstAxis");
  db.addPnt3d(Pnt3d(0.0, 0.0, 1.0), "dst", false);

  db.affect("f", "dst");
}

// ************************************************************************* //
void	NControl::glRender(Plug &plug)
{
  if (readAsInt("renderIn") != Gbb::Rendering)
    {
      pushName();
      glBegin(GL_POINTS);
      readAsPnt3d("dst").glVertex();
      glEnd();//GL_POINTS
      popName();
    }
  plug.writeAsSignal();
}

// ************************************************************************* //
/// DESCRIPTION:
// ************************************************************************* //
void	NControl::compute(Plug &plug)
{
  if (plug.shortName() == "renderOut")
    glRender(plug);
  else if (plug.shortName() == "dst")
    {
      const int	f = readAsInt("f");
      const double	dt = readAsDouble("dt");
      const Pnt3d	&src(readAsPnt3d("src"));
      const Vec3d	&dstAxis(readAsVec3d("dstAxis"));
      const Vec3d	z(dstAxis.unit());
      Vec3d		x, y;
      const double	radius(readAsDouble("radius"));
      const double	flatness(readAsDouble("flatness"));
      const double	angleVelocity(readAsDouble("angleVelocity"));

      if (fabs(z.x) < fabs(z.y) && fabs(z.x) < fabs(z.z))
	y = Vec3d(1.0, 0.0, 0.0) - z * z.x;
      else if (fabs(z.y) < fabs(z.z))
	y = Vec3d(0.0, 1.0, 0.0) - z * z.y;
      else
	y = Vec3d(0.0, 0.0, 1.0) - z * z.z;
      y.eqUnit();
      x = y ^ z;

      const double	alpha(f * dt * angleVelocity);
      const Pnt3d	dst = 
	src + (z + x * sin(alpha) * radius +
	       y * cos(alpha) * radius * flatness).unit();
      // TODO: control speed!!!
      plug.writeAsPnt3d(dst);
    }
}
