// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:29 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include <GL/glu.h>
#include "nodecurrentgenerator.h"
#include "datablock.h"
#include "command.h"
#include "plugcustom.h"
#include "gbb.h"

// ************************************************************************* //
int	NCurrentGenerator::typeNbr(59);
String	NCurrentGenerator::typeStr("CurrentGenerator");

// ************************************************************************* //
void NCurrentGenerator::initialize()
{
  DataBlock	&db(getDataBlock());

  NShape::initialize();
  db.addInt(0, MININT, MAXINT, "f", false);
  db.addSignal("update", false);
  db.addMat4d(Mat4d::identity(), "matParent", false);//position
  db.addBool(true, "displayTexture");
  db.addInt(12, 1, MAXINT, "mod");
  db.addInt(2, 1, MAXINT, "nbMotionHarmonics");//Motion
  db.addInt(12, 3, MAXINT, "nbMotionSamples");
  db.addInt(24, 3, MAXINT, "nbTextureSamples");//
  db.addDouble(1.0, 0.0, MAXDOUBLE, "strength");
  db.addDouble(0.0, MINDOUBLE, MAXDOUBLE, "twist");
  db.addDouble(0.0, 0.0, MAXDOUBLE, "buoyancy");
  db.addDouble(5.0, 0.0, MAXDOUBLE, "thickness");
  db.addDouble(1.0, 0.0, MAXDOUBLE, "radius");
  db.addDouble(0.2, 0.0, MAXDOUBLE, "ttb");// Time to birth
  db.addDouble(4.0, 0.0, MAXDOUBLE, "ttl");// Time to live
  db.addDouble(5.0, 0.0, MAXDOUBLE, "ttd");// Time to die
  db.addPnt3d(Pnt3d(0.0), "src");
  db.addPnt3d(Pnt3d(0.0, 0.0, 1.0), "dst");
  db.addInt(0, 0, MAXINT, "fakeHarmonic");
  db.addDouble(0.0, 0.0, MAXDOUBLE, "fakeStrength");
  db.addBool(false, "split");
  db.addDouble(1.0, 0.0, MAXDOUBLE, "minStrangle");

  db.addDouble(0.0, 0.0, 360.0, "paddleAngle");// TODO: Track
  db.addDouble(0.0, "paddleAmount");

  db.affect("f", "update");
  buildDagNodes();
}

// ************************************************************************* //
void	NCurrentGenerator::buildDagNodes()
{
  DagNode	*oldParent(_m_parent);
  DagNode	*tr = Command::run(String("new -type Transform -path :"));
  DagNode	*clock = DagNode::getNodeByPath(":Clock0");

  if (oldParent)
    DagNode::parent(*oldParent, *tr);

  DagNode::parent(*tr, *this);
  DagNode::connect(*oldParent, "renderIn", *tr, "renderIn");

  DagNode::disconnect(*oldParent, "renderIn", *this, "renderIn");

  DagNode::connect(*tr, "renderIn", *this, "renderIn");
  DagNode::connect(*tr, "matChild", *this, "matParent");
  DagNode::connect(*clock, "f", *this, "f");
}

// ************************************************************************* //
void	NCurrentGenerator::glRender(Plug &plug)
{
  if (readAsInt("renderIn") != Gbb::Rendering)
    {
      Plug		*hndSrc(getPlug("src"));
      Plug		*hndDst(getPlug("dst"));
      const Pnt3d	&src(hndSrc->readAsPnt3d());
      const Pnt3d	&dst(hndDst->readAsPnt3d());
      const double	radius(0.1);
      static GLUquadric	*quad = gluNewQuadric();
      const int		sel(readAsInt("sel"));
      const Vec3d	z((dst - src).unit());
      Vec3d		x, y, color;

      pushName();
      glPushMatrix();
      readAsMat4d("matParent").glMultMatrix();

      if (fabs(z.x) < fabs(z.y) && fabs(z.x) < fabs(z.z))
	y = Vec3d(1.0, 0.0, 0.0) - z * z.x;
      else if (fabs(z.y) < fabs(z.z))
	y = Vec3d(0.0, 1.0, 0.0) - z * z.y;
      else
	y = Vec3d(0.0, 0.0, 1.0) - z * z.z;
      y.eqUnit();
      x = y ^ z;

      if (sel == 1)
	color = Vec3d(1.0, 0.0, 5.0);
      else if (sel)
	color = Vec3d(1.0, 0.5, 0.0);
      else
	color = Vec3d(0.0, 0.0, 1.0);
      color.glColor();

      hndSrc->pushName();
      glPushAttrib(GL_CURRENT_BIT);
      if (hndSrc->isSelected())
	glColor3d(1.0, 1.0, 0.0);
      else
	color.glColor();
      glPushMatrix();
      Mat4d(x, y, z, src).glMultMatrix();
      glTranslated(0.0, 0.0, -radius);
      glPushMatrix();
      glRotated(180.0, 1.0, 0.0, 0.0);
      gluDisk(quad, 0, radius, 16, 1);
      glPopMatrix();
      glPushMatrix();
      glTranslated(0.0, 0.0, radius);
      gluDisk(quad, 0, radius, 16, 1);
      glPopMatrix();
      gluCylinder(quad, radius, radius, radius, 16, 1);
      glPopMatrix();
      glPopAttrib();//GL_CURRENT_BIT
      hndSrc->popName();

      hndDst->pushName();
      glPushAttrib(GL_CURRENT_BIT);
      if (hndDst->isSelected())
	glColor3d(1.0, 1.0, 0.0);
      else
	color.glColor();
      glPushMatrix();
      Mat4d(x, y, z, dst).glMultMatrix();
      gluCylinder(quad, radius, 0.0, 2.5 * radius, 16, 1);
      glPushMatrix();
      glRotated(180.0, 1.0, 0.0, 0.0);
      gluDisk(quad, 0, radius, 16, 1);
      glPopMatrix();
      glPopMatrix();
      glPopAttrib();//GL_CURRENT_BIT
      hndDst->popName();

      glPushAttrib(GL_LIGHTING_BIT);
      glDisable(GL_LIGHTING);
      glBegin(GL_LINES);
      src.glVertex();
      dst.glVertex();
      glEnd();//GL_LINES
      glPopAttrib();//GL_LIGHTING_BIT

      glPopMatrix();
      popName();
    }
  plug.writeAsSignal();
}

// ************************************************************************* //
/// DESCRIPTION: Returns true if the generator is being controlled by a 
///  CurrentControl node.
// ************************************************************************* //
DagNode		*NCurrentGenerator::isControlled()
{
  DagNode	*tr = getNodeParent();
  DagNode	*parent;

  if (tr && (parent = tr->getNodeParent()))
    if (parent->getTypeStr() == "CurrentControl")
      return parent;
  return 0;
}

// ************************************************************************* //
/// DESCRIPTION: Checks if the generator is under control. If true, the 
///  vortex ring is added as a child of the control. Else it is a child 
///  of the set of vortices.
// ************************************************************************* //
void	NCurrentGenerator::update(Plug &plug)
{
  const int	f = readAsInt("f");

  plug.writeAsSignal();
  if (!(f % readAsInt("mod")))
    {
      const int		nbMotionHarmonics(readAsInt("nbMotionHarmonics"));
      const double	radius(readAsDouble("radius"));
      const double	strength(readAsDouble("strength"));
      const double	twist(readAsDouble("twist"));
      DagNode	*rs = DagNode::getNodeByPath(":VortexSet0");
      DagNode	*ctrl = isControlled();
      DagNode	*newparent = ctrl? ctrl: rs;
      DagNode	*n = Command::run(String("new -type VortexRing -path ") +
				  newparent->path());

      if (n->getNodeParent())
	DagNode::disconnect(*n->getNodeParent(), "renderIn", *n, "renderIn");
      DagNode::connect(*newparent, "renderIn", *n, "renderIn");

      const Pnt3d	&src = readAsPnt3d("src");
      const Pnt3d	&dst = readAsPnt3d("dst");

      n->writeAsInt("nbMotionSamples", readAsInt("nbMotionSamples"));
      n->writeAsInt("nbTextureSamples", readAsInt("nbTextureSamples"));
      n->writeAsInt("nbMotionHarmonics", nbMotionHarmonics);
      n->writeAsDouble("strength", strength);
      DagNode::connect(*this, "twist", *n, "twist");
      DagNode::connect(*this, "displayTexture", *n, "displayTexture");
      n->writeAsDouble("temperature", readAsDouble("buoyancy"));
      n->writeAsDouble("thickness", readAsDouble("thickness"));

      DagNode::connect(*this, "split", *n, "split");
      DagNode::connect(*this, "minStrangle", *n, "minStrangle");

      const Vec3d	z((dst - src).unit());
      Vec3d		x, y;

      if (fabs(z.x) < fabs(z.y) && fabs(z.x) < fabs(z.z))
	y = Vec3d(1.0, 0.0, 0.0) - z * z.x;
      else if (fabs(z.y) < fabs(z.z))
	y = Vec3d(0.0, 1.0, 0.0) - z * z.y;
      else
	y = Vec3d(0.0, 0.0, 1.0) - z * z.z;
      y.eqUnit();
      x = y ^ z;

      const double	a(readAsDouble("paddleAngle"));

      n->writeAsDouble("paddleAmount", readAsDouble("paddleAmount"));
      n->writeAsVec3d("paddle", x * cos(DEG2RAD(a)) + y * sin(DEG2RAD(a)));

      // TODO: find freq that put ring in plane.
      const Mat4d	&m = readAsMat4d("matParent");
      const Vec3d	mx(m.getx());
      const Vec3d	my(m.gety());

      n->writeAsPnt3d("center", m * src);
    
      /*
      n->writeAsPnt2dArray("cx", 0, Pnt2d(x * mx, x * my) * 0.5 * radius);
      n->writeAsPnt2dArray("cy", 0, Pnt2d(y * mx, y * my) * 0.5 * radius);
      n->writeAsPnt2dArray("cz", 0, Pnt2d(z * mx, z * my) * 0.5 * radius);
      n->writeAsPnt2dArray("cx", 0, Pnt2d(1.0, 0.0) * 0.5 * radius);
      n->writeAsPnt2dArray("cy", 0, Pnt2d(0.0, 1.0) * 0.5 * radius);
      n->writeAsPnt2dArray("cz", 0, Pnt2d(0.0) * 0.5 * radius);
      */
      n->writeAsPnt2dArray("cx", 0, Pnt2d(0.0, 0.5 * radius));
      n->writeAsPnt2dArray("cy", 0, Pnt2d(M_PI * 0.5, 0.5 * radius));
      n->writeAsPnt2dArray("cz", 0, Pnt2d(0.0));

      n->writeAsDouble("fakeStrength", readAsDouble("fakeStrength"));
      n->writeAsInt("fakeHarmonic", readAsInt("fakeHarmonic"));


      //n->writeAsPnt2dArray("wl", 0, Pnt2d(1.0, 0.0) * strength);
      n->getPlug("cx")->resize(nbMotionHarmonics);
      n->getPlug("cy")->resize(nbMotionHarmonics);
      n->getPlug("cz")->resize(nbMotionHarmonics);
      //n->getPlug("wl")->resize(nbMotionHarmonics);
      for (int k(1); k < nbMotionHarmonics; ++k)
	{
	  n->writeAsPnt2dArray("cx", k, Pnt2d(0.0));
	  n->writeAsPnt2dArray("cy", k, Pnt2d(0.0));
	  n->writeAsPnt2dArray("cz", k, Pnt2d(0.0));
	  //n->writeAsPnt2dArray("wl", k, Pnt2d(0.0));
	}
      n->writeAsVec3d("ex", x);
      n->writeAsVec3d("ey", y);
      n->writeAsVec3d("ez", z);
      n->writeAsDouble("ttb", readAsDouble("ttb"));
      n->writeAsDouble("ttd", readAsDouble("ttd"));
      n->writeAsDouble("ttl", readAsDouble("ttl"));
    }
}

// ************************************************************************* //
void	NCurrentGenerator::compute(Plug &plug)
{
  if (plug.shortName() == "renderOut")
    glRender(plug);
  else if (plug.shortName() == "update")
    update(plug);
}
