// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:29 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include "nodeextrudedshape.h"
#include "nodebezier3d.h"
#include "plug.h"
#include "datablock.h"
#include "command.h"
#include <GL/glu.h>

// ************************************************************************* //
int	NExtrudedShape::typeNbr(67);
String	NExtrudedShape::typeStr("ExtrudedShape");
int	NCurveConstraint::typeNbr(68);
String	NCurveConstraint::typeStr("CurveConstraint");
int	NCoonsPatch::typeNbr(69);
String	NCoonsPatch::typeStr("CoonsPatch");

// ************************************************************************* //
void	NExtrudedShape::initialize()
{
  NShape::initialize();

  DagNode::connect(*_m_parent, "renderIn", *this, "renderIn");
  // TODO: create subdag of elements: curve, patch, ...
  // 
  DagNode *cpath0 = Command::run(String("new -type Bezier3d -path ") + path());
  DagNode::connect(*this, "renderIn", *cpath0, "renderIn");

  DagNode *cpath1 = Command::run(String("new -type Bezier3d -path ") + path());
  DagNode::connect(*this, "renderIn", *cpath1, "renderIn");

  DagNode *cpath2 = Command::run(String("new -type Bezier3d -path ") + path());
  DagNode::connect(*this, "renderIn", *cpath2, "renderIn");

  DagNode *cpath3 = Command::run(String("new -type Bezier3d -path ") + path());
  DagNode::connect(*this, "renderIn", *cpath3, "renderIn");

  const Pnt3d	pnts0[5] = { Pnt3d(0.0, 0.0, 0.0),
			     Pnt3d(0.0, 0.0, 0.25),
			     Pnt3d(0.0, 0.0, 0.50),
			     Pnt3d(0.0, 0.0, 0.75),
			     Pnt3d(0.0, 0.0, 1.0) };
  const Pnt3d	pnts1[5] = { Pnt3d(0.0, 0.0, 1.0),
			     Pnt3d(0.0, 0.25, 1.0),
			     Pnt3d(0.0, 0.50, 1.0),
			     Pnt3d(0.0, 0.75, 1.0),
			     Pnt3d(0.0, 1.0, 1.0) };
  const Pnt3d	pnts2[5] = { Pnt3d(0.0, 1.0, 1.0),
			     Pnt3d(0.0, 1.0, 0.75),
			     Pnt3d(0.0, 1.0, 0.50),
			     Pnt3d(0.0, 1.0, 0.25),
			     Pnt3d(0.0, 1.0, 0.0) };
  const Pnt3d	pnts3[5] = { Pnt3d(0.0, 1.0, 0.0),
			     Pnt3d(0.0, 0.75, 0.0),
			     Pnt3d(0.0, 0.50, 0.0),
			     Pnt3d(0.0, 0.25, 0.0),
			     Pnt3d(0.0, 0.0, 0.0) };

  for (int i(0); i < 5; ++i)
    cpath0->writeAsPnt3dArray("ci", i, pnts0[i]),
      cpath1->writeAsPnt3dArray("ci", i, pnts1[i]),
      cpath2->writeAsPnt3dArray("ci", i, pnts2[i]),
      cpath3->writeAsPnt3dArray("ci", i, pnts3[i]);
  cpath0->writeAsPnt3d("front", pnts0[0]);
  cpath0->writeAsPnt3d("back", pnts0[4]);
  cpath1->writeAsPnt3d("front", pnts1[0]);
  cpath1->writeAsPnt3d("back", pnts1[4]);
  cpath2->writeAsPnt3d("front", pnts2[0]);
  cpath2->writeAsPnt3d("back", pnts2[4]);
  cpath3->writeAsPnt3d("front", pnts3[0]);
  cpath3->writeAsPnt3d("back", pnts3[4]);
}

// ************************************************************************* //
void	NExtrudedShape::compute(Plug &plugOut)
{
  if (plugOut.shortName() == "renderOut")
    {
      const int	sel(readAsInt("sel"));

      pushName();
      if (sel == 1)
	glColor3d(1.0, 0.0, 0.5);
      else if (sel)
	glColor3d(1.0, 0.5, 0.0);
      else
	glColor3d(0.0, 0.0, 1.0);
      
      std::list<DagNode*>::iterator	nit;
      std::list<DagNode*>	nodes;
      
      getNodesAtDepth(1, nodes);
      for (nit = nodes.begin(); nit != nodes.end(); ++nit)
	(*nit)->readAsSignal("renderOut");

      popName();
      readAsSignal("renderIn");
      plugOut.writeAsSignal();
    }
}

// ************************************************************************* //
// TODO: Coons patch node.

// ************************************************************************* //
/// DESCRIPTION:
// ************************************************************************* //
void	NCurveConstraint::initialize()
{
  DataBlock	&db(getDataBlock());

  db.addInt(0, MININT, MAXINT, "f", false);
  db.addSignal("update", false);

  db.addSignal("curve0", false);
  db.addSignal("curve1", false);//constrained.

  db.addBool(true, "front0");//true if front vertex of curve0 is constraint
  db.addBool(true, "front1");//true if front vertex of curve1 is constrained.

  //db.addPnt3d(0.0, "pnt0");
  //db.addPnt3d(0.0, "pnt1");//constrained.

  //TODO: make a point ... connect to output point, for easy graph traversal?
  db.affect("f", "update");

  buildDagNodes();
}

// ************************************************************************* //
void	NCurveConstraint::buildDagNodes()
{
  DagNode	*clock = DagNode::getNodeByPath(":Clock0");

  DagNode::connect(*clock, "f", *this, "f");
}

// ************************************************************************* //
void	NCurveConstraint::compute(Plug &plugOut)
{
  if (plugOut.shortName() == "update")
    {
      readAsInt("f");

      const bool	front0 = readAsBool("front0");
      const bool	front1 = readAsBool("front1");
      DagNode	*curve1 = getPlug("curve1")->getInputNode();
      DagNode	*curve0 = getPlug("curve0")->getInputNode();

      // TODO: call update at every frame.
      if (curve0 && curve1)
	{
	  const Pnt3d	pntwrt = curve0->readAsPnt3d(front0? "front": "back");

	  curve1->writeAsPnt3d(front1? "front": "back", pntwrt);
	}
      plugOut.writeAsSignal();
    }
}

// ************************************************************************* //
void	NCoonsPatch::initialize()
{
  DataBlock	&db(getDataBlock());

  NShape::initialize();

  db.addSignal("curve0", false);
  db.addSignal("curve1", false);
  db.addSignal("curve2", false);
  db.addSignal("curve3", false);

  db.addInt(16, 2, MAXINT, "rows");
  db.addInt(16, 2, MAXINT, "cols");

  buildDagNodes();
}

// ************************************************************************* //
void	NCoonsPatch::buildDagNodes()
{
  DagNode::connect(*_m_parent, "renderIn", *this, "renderIn");
}

// ************************************************************************* //
void	NCoonsPatch::compute(Plug &plugOut)
{
  if (plugOut.shortName() == "renderOut")
    {
      const int	sel(readAsInt("sel"));

      pushName();
      if (sel == 1)
	glColor3d(1.0, 0.0, 0.5);
      else if (sel)
	glColor3d(1.0, 0.5, 0.0);
      else
	glColor3d(0.0, 0.0, 1.0);

      DagNode	*curve0(0), *curve1(0), *curve2(0), *curve3(0);

      if ((curve0 = getPlug("curve0")->getInputNode()) && 
	  (curve1 = getPlug("curve1")->getInputNode()) && 
	  (curve2 = getPlug("curve2")->getInputNode()) && 
	  (curve3 = getPlug("curve3")->getInputNode()))
	{
	  Plug *x0 = curve0->getPlug("u");
	  Plug *f0 = curve0->getPlug("pu");
	  Plug *t0 = curve0->getPlug("tu");
	  Plug *x1 = curve1->getPlug("u");
	  Plug *f1 = curve1->getPlug("pu");
	  Plug *t1 = curve1->getPlug("tu");
	  Plug *x2 = curve2->getPlug("u");
	  Plug *f2 = curve2->getPlug("pu");
	  Plug *t2 = curve2->getPlug("tu");
	  Plug *x3 = curve3->getPlug("u");
	  Plug *f3 = curve3->getPlug("pu");
	  Plug *t3 = curve3->getPlug("tu");
	  
	  const unsigned	rows(readAsInt("rows"));
	  const unsigned	cols(readAsInt("cols"));

	  Vec3d	*ntab = new Vec3d[cols * rows];
	  Pnt3d	*tab = new Pnt3d[cols * rows];

	  x0->writeAsDouble(0.0);
	  const Pnt3d	p00 = f0->readAsPnt3d();
	  x1->writeAsDouble(0.0);
	  const Pnt3d	p01 = f1->readAsPnt3d();
	  x2->writeAsDouble(0.0);
	  const Pnt3d	p11 = f2->readAsPnt3d();
	  x3->writeAsDouble(0.0);
	  const Pnt3d	p10 = f3->readAsPnt3d();

	  for (unsigned i(0); i < rows; ++i)
	    {
	      const double	x = i / (rows - 1.0);

	      for (unsigned j(0); j < cols; ++j)
		{
		  const double	y = j / (cols - 1.0);

		  x0->writeAsDouble(x);
		  x1->writeAsDouble(y);
		  x2->writeAsDouble(1.0 - x);
		  x3->writeAsDouble(1.0 - y);
		  tab[i * cols + j] = 
		      f0->readAsPnt3d() * (1.0 - y)
		    + f1->readAsPnt3d() * x
		    + f2->readAsPnt3d() * y
		    + f3->readAsPnt3d() * (1.0 - x)
		    + p00 * (-(1.0 - x) * (1.0 - y))
		    + p01 * (-x * (1.0 - y))
		    + p11 * (-x * y)
		    + p10 * (-(1.0 - x) * y);
		  ntab[i * cols + j] =
		    ((t0->readAsVec3d() * (1.0 - y) - t2->readAsVec3d() * y)
		     + (f1->readAsPnt3d() - f3->readAsPnt3d())
		     + (p10 - p11) * y + (p00 - p01) * (1.0 - y)) ^ 
		    ((t1->readAsVec3d() * x -  t3->readAsVec3d() * (1.0 - x))
		     + (f2->readAsPnt3d() - f0->readAsPnt3d())
		     + (p01 - p11) * x + (p00 - p10) * (1.0 - x));
		}
	    }

	  GLint	param;

	  glGetIntegerv(GL_RENDER_MODE, &param);
	  if (!sel || (param != GL_SELECT))
	    for (unsigned i(0); i < rows - 1; ++i)
	      {
		glBegin(GL_QUAD_STRIP);
		for (unsigned j(0); j < cols; ++j)
		  {
		    ntab[i * cols + j].glNormal();
		    tab[i * cols + j].glVertex();
		    ntab[(i + 1) * cols + j].glNormal();
		    tab[(i + 1) * cols + j].glVertex();
		  }
		glEnd();//GL_QUAD_STRIP
	      }
	  delete[] ntab;
	  delete[] tab;
	}
      
      popName();
      readAsSignal("renderIn");
      plugOut.writeAsSignal();
    }
}

