// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:29 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include <GL/gl.h>
#include <GL/glu.h>
#include "nodegizmo.h"
#include "datablock.h"
#include "plug.h"
#include "command.h"

// ************************************************************************* //
#ifndef M_2PI
#define M_2PI	6.28318530717958647692
#endif//M_2PI

// ************************************************************************* //
int	NGizmo::typeNbr(48);
String	NGizmo::typeStr("Gizmo");
GLuint	NGizmo::_s_list(0);

// ************************************************************************* //
void	NGizmo::initialize()
{
  NShape::initialize();

  DataBlock	&db(getDataBlock());

  db.addSignal("transform", false);
  db.addBool(false, "visible", false);
  db.addInt(4, 0, MAXINT, "axis", false);
  db.addDouble(0.0, 0.0, MAXDOUBLE, "size", false);
  db.addMat4d(Mat4d::identity(), "model", false);
  db.addPnt3d(0.0, "p0", false);
  db.addPnt3d(0.0, "p1", false);
  db.addSignal("update", false);
  db.addBool(true, "isEasy");
  db.addBool(true, "translate");
  db.addBool(true, "rotate");
  db.addBool(true, "scale");

  db.affect("p0", "update");
  db.affect("p1", "update");
  buildDagNodes();
}

// ************************************************************************* //
void	NGizmo::buildDagNodes()
{
  DagNode::connect(*_m_parent, "renderIn", *this, "renderIn");
}

// ************************************************************************* //
Pnt3d	NGizmo::project(const DagNode	&c,
			const Pnt3d	&v)
{
  const Mat4d	&model(c.readAsMat4d("model"));
  const Mat4d	&proj(c.readAsMat4d("proj"));
  const Mat2i	&view(c.readAsMat2i("view"));
  Pnt3d	ret;
  
  gluProject(v(0), v(1), v(2), model._m_m, proj._m_m,
	     view._m_m, &ret(0), &ret(1), &ret(2));
  return ret;
}

// ************************************************************************* //
Pnt3d	NGizmo::unproject(const DagNode	&c,
			  const Pnt3d	&v)
{
  const Mat4d	&model(c.readAsMat4d("model"));
  const Mat4d	&proj(c.readAsMat4d("proj"));
  const Mat2i	&view(c.readAsMat2i("view"));
  Pnt3d	ret;
  
  gluUnProject(v(0), v(1), v(2), model._m_m, proj._m_m,
	       view._m_m, &ret(0), &ret(1), &ret(2));
  return ret;
}

// ************************************************************************* //
/// DESCRIPTION: Generates two lists: an arrow and a cube.
// ************************************************************************* //
void	NGizmo::glInitLists() const
{
  GLUquadricObj	*qobj;

  if (!_s_list)
    {
      _s_list = glGenLists(2);

      glNewList(_s_list, GL_COMPILE);

      glPushAttrib(GL_TRANSFORM_BIT);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();

      glPushAttrib(GL_LINE_BIT);
      glLineWidth(2.0);
      glPushAttrib(GL_LIGHTING_BIT);
      glDisable(GL_LIGHTING);
      glBegin(GL_LINES);
      Pnt3d(0.0, 0.0, 0.0).glVertex();
      Pnt3d(0.0, 0.0, 0.8).glVertex();
      glEnd();//GL_LINES
      glPopAttrib();//GL_LIGHTING_BIT
      glPopAttrib();//GL_LINE_BIT

      qobj = gluNewQuadric();
      gluQuadricDrawStyle(qobj, GLU_FILL);
      gluQuadricNormals(qobj, GLU_SMOOTH);
      glTranslated(0.0, 0.0, 0.8);
      glScaled(0.2, 0.2, 0.2);
      gluCylinder(qobj, 0.33, 0.0, 1.0, 16, 1);
      gluDeleteQuadric(qobj);

      qobj = gluNewQuadric();
      gluQuadricDrawStyle(qobj, GLU_FILL);
      gluQuadricNormals(qobj, GLU_SMOOTH);
      gluDisk(qobj, 0.33, 0.0, 16, 1);
      gluDeleteQuadric(qobj);
      glPopMatrix();//GL_MODELVIEW
      glPopAttrib();//GL_TRANSFORM_BIT
      glEndList();

      glNewList(_s_list + 1, GL_COMPILE);
      glPushAttrib(GL_LIGHTING_BIT);
      glPushAttrib(GL_TRANSFORM_BIT);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      glTranslated(0.0, 0.0, 1.3);
      glScaled(0.1, 0.1, 0.1);
      glBegin(GL_QUADS);
      Vec3d(-1.0, 0.0, 0.0).glNormal();
      Pnt3d(-1.0, -1.0, -1.0).glVertex();
      Pnt3d(-1.0, -1.0, +1.0).glVertex();
      Pnt3d(-1.0, +1.0, +1.0).glVertex();
      Pnt3d(-1.0, +1.0, -1.0).glVertex();
      Vec3d(1.0, 0.0, 0.0).glNormal();
      Pnt3d(1.0, -1.0, -1.0).glVertex();
      Pnt3d(1.0, +1.0, -1.0).glVertex();
      Pnt3d(1.0, +1.0, +1.0).glVertex();
      Pnt3d(1.0, -1.0, +1.0).glVertex();
      Vec3d(0.0, -1.0, 0.0).glNormal();
      Pnt3d(-1.0, -1.0, -1.0).glVertex();
      Pnt3d(+1.0, -1.0, -1.0).glVertex();
      Pnt3d(+1.0, -1.0, +1.0).glVertex();
      Pnt3d(-1.0, -1.0, +1.0).glVertex();
      Vec3d(0.0, 1.0, 0.0).glNormal();
      Pnt3d(-1.0, 1.0, -1.0).glVertex();
      Pnt3d(-1.0, 1.0, +1.0).glVertex();
      Pnt3d(+1.0, 1.0, +1.0).glVertex();
      Pnt3d(+1.0, 1.0, -1.0).glVertex();
      Vec3d(0.0, 0.0, -1.0).glNormal();
      Pnt3d(-1.0, -1.0, -1.0).glVertex();
      Pnt3d(-1.0, +1.0, -1.0).glVertex();
      Pnt3d(+1.0, +1.0, -1.0).glVertex();
      Pnt3d(+1.0, -1.0, -1.0).glVertex();
      Vec3d(0.0, 0.0, 1.0).glNormal();
      Pnt3d(-1.0, -1.0, 1.0).glVertex();
      Pnt3d(+1.0, -1.0, 1.0).glVertex();
      Pnt3d(+1.0, +1.0, 1.0).glVertex();
      Pnt3d(-1.0, +1.0, 1.0).glVertex();
      glEnd();
      glPopMatrix();//GL_MODELVIEW
      glPopAttrib();//GL_LIGHTING_BIT 
      glPopAttrib();//GL_TRANSFORM_BIT
      glEndList();
    }
}

// ************************************************************************* //
void	NGizmo::renderTranslation(DagNode &tr)
{
  const Mat4d	&m = tr.readAsMat4d("matChild");
  const Mat4d	&model(readAsMat4d("model"));
  const double	size(readAsDouble("size"));
  const int	axis(readAsInt("axis"));
  const bool	isEasy(readAsBool("isEasy"));

  Vec3d	m_x(m.getx().unit()), m_y(m.gety().unit());
  Vec3d	m_z(m.getz().unit());
  Pnt3d	m_o(m.geto());
  Vec3d	x(model.getx()), y(model.gety()), z(model.getz());

  x.eqUnit();
  y.eqUnit();
  z.eqUnit();

  Vec3d	px(x(0), y(0), z(0)), py(x(1), y(1), z(1));

  if (!isEasy)
    {
      glPushAttrib(GL_TRANSFORM_BIT);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      Mat4d(m_x, m_y, m_z, m_o).glMultMatrix();
      glScaled(size, size, size);
      (axis == 3)? glColor3d(1.0, 1.0, 0.0):
	glColor3d(0.0, 0.0, 1.0);

      glPushName(3);
      glCallList(_s_list);
      glPopName();
      glRotated(-90.0, 1.0, 0.0, 0.0);
      (axis == 2)? glColor3d(1.0, 1.0, 0.0):
	glColor3d(0.0, 1.0, 0.0);
      glPushName(2);
      glCallList(_s_list);
      glPopName();
      glRotated(90.0, 0.0, 1.0, 0.0);
      (axis == 1)? glColor3d(1.0, 1.0, 0.0):
	glColor3d(1.0, 0.0, 0.0);
      glPushName(1);
      glCallList(_s_list);
      glPopName();
      glPopMatrix();//GL_MODELVIEW
      glPopAttrib();//GL_TRANSFORM_BIT
    }

  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);
  px *= size / 6.0;
  py *= size / 6.0;
  (axis == 4)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.0, 1.0, 1.0);
  glPushName(4);
  glBegin(GL_QUADS);//TODO: QUAD
  (m.geto() + px - py).glVertex();
  (m.geto() + px + py).glVertex();
  (m.geto() - px + py).glVertex();
  (m.geto() - px - py).glVertex();
  glEnd();
  glPopName();
  glPopAttrib();//GL_LIGHTING_BIT

	    
  ///////////////////////////////////////////////////////
  glPushAttrib(GL_DEPTH_BUFFER_BIT);
  glDepthFunc(GL_GREATER);

  if (!isEasy)
    {
      glPushAttrib(GL_TRANSFORM_BIT);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      Mat4d(m_x, m_y, m_z, m_o).glMultMatrix();
      glScaled(size, size, size);
      (axis == 3)? glColor3d(0.7, 0.7, 0.6):
	glColor3d(0.6, 0.6, 0.7);
      glPushName(3);
      glCallList(_s_list);
      glPopName();
      glRotated(-90.0, 1.0, 0.0, 0.0);
      (axis == 2)? glColor3d(0.7, 0.7, 0.6):
	glColor3d(0.6, 0.7, 0.6);
      glPushName(2);
      glCallList(_s_list);
      glPopName();
      glRotated(90.0, 0.0, 1.0, 0.0);
      (axis == 1)? glColor3d(0.7, 0.7, 0.6):
	glColor3d(0.7, 0.6, 0.6);
      glPushName(1);
      glCallList(_s_list);
      glPopName();
      glPopMatrix();//GL_MODELVIEW
      glPopAttrib();//GL_TRANSFORM_BIT
    }

  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);
  (axis == 4)? glColor3d(0.7, 0.7, 0.6): glColor3d(0.6, 0.7, 0.7);
  glPushName(4);
  glBegin(GL_QUADS);
  (m.geto() + px - py).glVertex();
  (m.geto() + px + py).glVertex();
  (m.geto() - px + py).glVertex();
  (m.geto() - px - py).glVertex();
  glEnd();
  glPopName();
  glPopAttrib();//GL_LIGHTING_BIT
  glPopAttrib();//GL_DEPTH_BUFFER_BIT
}

/* ************************************************************************* */
void	renderCircle(const Vec3d	&x,
		     const Vec3d	&y,
		     const Vec3d	&v,
		     const double	size)
{
  unsigned	i;
  double	theta;
  Vec3d	px[32];

  for (theta = 0.0, i = 0; i < 32; theta += M_2PI / 32.0, ++i)
    px[i] = (x * cos(theta) + y * sin(theta)) * size;
  glBegin(GL_LINES);
  for (i = 0; i < 32; ++i)
    if (px[i] * v > 0.0)
      px[i].glVertex(), px[(i + 1) % 32].glVertex();
  glEnd();
  glPushAttrib(GL_LINE_BIT);
  glLineStipple(1, 0xAAAA);
  glEnable(GL_LINE_STIPPLE);
  glBegin(GL_LINES);
  for (i = 0; i < 32; ++i)
    if (px[i] * v < 0.0)
      px[i].glVertex(), px[(i + 1) % 32].glVertex();
  glEnd();
  glDisable(GL_LINE_STIPPLE);
  glPopAttrib();//GL_LINE_BIT
}

// ************************************************************************* //
void	NGizmo::renderRotation(DagNode	&tr)
{
  const Mat4d	&m = tr.readAsMat4d("matChild");
  const Mat4d	&model(readAsMat4d("model"));

  //DagNode	*camera = DagNode::getNodeByPath(":Camera0");
  std::list<DagNode*>	cameras;

  DagNode::root().DagNode::getNodesByType("Camera", cameras);

  DagNode	*camera = cameras.front();

  const double	size(readAsDouble("size"));
  const int	axis(readAsInt("axis"));
  const bool	isEasy(readAsBool("isEasy"));

  const Vec3d	m_x(m.getx().unit());
  const Vec3d	m_y(m.gety().unit());
  const Vec3d	m_z(m.getz().unit());
  const Pnt3d	m_o(m.geto());
  const Vec3d	vv(m.inverse() * unproject(*camera, Pnt3d(0.0)));
  const Vec3d	x(model.getx().unit());
  const Vec3d	y(model.gety().unit());
  const Vec3d	z(model.getz().unit());

  glPushAttrib(GL_TRANSFORM_BIT);
  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();

  Mat4d(m_x, m_y, m_z, m_o).glMultMatrix();
  if (!isEasy)
    {
      (axis == 11)? glColor3d(1.0, 1.0, 0.0): glColor3d(1.0, 0.0, 0.0);
      glPushName(11);
      renderCircle(Vec3d(0.0, 1.0, 0.0), Vec3d(0.0, 0.0, 1.0), vv, size);
      glPopName();

      (axis == 12)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.0, 1.0, 0.0);
      glPushName(12);
      renderCircle(Vec3d(0.0, 0.0, 1.0), Vec3d(1.0, 0.0, 0.0), vv, size);
      glPopName();

      (axis == 13)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.0, 0.0, 1.0);
      glPushName(13);
      renderCircle(Vec3d(1.0, 0.0, 0.0), Vec3d(0.0, 1.0, 0.0), vv, size);
      glPopName();
    }
  glPopMatrix();//GL_MODELVIEW

  Vec3d	px(x(0), y(0), z(0)), py(x(1), y(1), z(1));

  px *= size * 1.2;
  py *= size * 1.2;
  (axis == 14)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.0, 1.0, 1.0);
  glPushName(14);
  glBegin(GL_LINE_LOOP);
  for (double theta(0.0); theta < M_2PI; theta += M_2PI / 32.0)
    (m.geto() + cos(theta) * px + sin(theta) * py).glVertex();
  glEnd();
  glPopName();

  ///////////////////////////////////////////////////////
  glPushAttrib(GL_DEPTH_BUFFER_BIT);
  glDepthFunc(GL_GREATER);

  if (!isEasy)
    {
      glPushMatrix();
      Mat4d(m_x, m_y, m_z, m_o).glMultMatrix();
      (axis == 11)? glColor3d(0.7, 0.7, 0.6): glColor3d(0.7, 0.6, 0.6);
      glPushName(11);
      renderCircle(Vec3d(0.0, 1.0, 0.0), Vec3d(0.0, 0.0, 1.0), vv, size);
      glPopName();
      (axis == 12)? glColor3d(0.7, 0.7, 0.6): glColor3d(0.6, 0.7, 0.6);
      glPushName(12);
      renderCircle(Vec3d(0.0, 0.0, 1.0), Vec3d(1.0, 0.0, 0.0), vv, size);
      glPopName();
      (axis == 13)? glColor3d(0.7, 0.7, 0.6): glColor3d(0.6, 0.6, 0.7);
      glPushName(13);
      renderCircle(Vec3d(1.0, 0.0, 0.0), Vec3d(0.0, 1.0, 0.0), vv, size);
      glPopName();
      glPopMatrix();//GL_MODELVIEW
    }

  (axis == 14)? glColor3d(0.7, 0.7, 0.6): glColor3d(0.6, 0.7, 0.7);
  glPushName(14);
  glBegin(GL_LINE_LOOP);
  for (double theta(0.0); theta < M_2PI; theta += M_2PI / 32.0)
    (m.geto() + cos(theta) * px + sin(theta) * py).glVertex();
  glEnd();
  glPopName();
  glPopAttrib();//GL_DEPTH_BUFFER_BIT

  glPopAttrib();//GL_LIGHTING_BIT
  glPopAttrib();//GL_TRANSFORM_BIT
}

// ************************************************************************* //
void	NGizmo::renderScale(DagNode	&tr)
{
  const Mat4d	&m = tr.readAsMat4d("matChild");
  const Mat4d	&model(readAsMat4d("model"));
  const double	size(readAsDouble("size"));
  const int	axis(readAsInt("axis"));
  const bool	isEasy(readAsBool("isEasy"));

  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);

  Vec3d	x(model.getx());
  Vec3d	y(model.gety());
  Vec3d	z(model.getz());
  Vec3d	m_x(m.getx().unit());
  Vec3d	m_y(m.gety().unit());
  Vec3d	m_z(m.getz().unit());
  Pnt3d	m_o(m.geto());

  x.eqUnit();
  y.eqUnit();
  z.eqUnit();

  Vec3d	px(x(0), y(0), z(0)), py(x(1), y(1), z(1));
  px *= size * 1.2;
  py *= size * 1.2;
  glColor3d(0.0, 1.0, 1.0);

  if (!readAsBool("rotate"))
    {
      glPushAttrib(GL_LINE_BIT);
      glLineStipple(1, 0xAAAA);
      glEnable(GL_LINE_STIPPLE);
      glBegin(GL_LINE_LOOP);
      for (double theta(0.0); theta < M_2PI; theta += M_2PI / 32.0)
	(m.geto() + cos(theta) * px + sin(theta) * py).glVertex();
      glEnd();
      glDisable(GL_LINE_STIPPLE);
      glPopAttrib();//GL_LINE_BIT
    }

  const Pnt3d	tmp0(cos(M_PI / 4.0) * px + sin(M_PI / 4.0) * py);
  const Pnt3d	tmp1(m.geto() + tmp0 * 1.2);

  glBegin(GL_LINES);
  (m.geto() + tmp0).glVertex();
  tmp1.glVertex();
  glEnd();

  (axis == 24)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.0, 1.0, 1.0);
  glPushName(24);
  glBegin(GL_QUADS);
  (tmp1 + 0.1 * (px - py)).glVertex();
  (tmp1 + 0.1 * (3.0 * px + py)).glVertex();
  (tmp1 + 0.1 * (px + 3.0 * py)).glVertex();
  (tmp1 + 0.1 * (py - px)).glVertex();
  glEnd();
  glPopName();

  glEnable(GL_LIGHTING);

  if (!isEasy)
    {
      glPushAttrib(GL_TRANSFORM_BIT);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      Mat4d(m_x, m_y, m_z, m_o).glMultMatrix();
      glScaled(size, size, size);
      (axis == 23)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.0, 0.0, 1.0);
      glPushName(23);
      glCallList(_s_list + 1);
      glPopName();
      glRotated(-90.0, 1.0, 0.0, 0.0);
      (axis == 22)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.0, 1.0, 0.0);
      glPushName(22);
      glCallList(_s_list + 1);
      glPopName();
      glRotated(90.0, 0.0, 1.0, 0.0);
      (axis == 21)? glColor3d(1.0, 1.0, 0.0): glColor3d(1.0, 0.0, 0.0);
      glPushName(21);
      glCallList(_s_list + 1);
      glPopName();
      glPopMatrix();//GL_MODELVIEW
      glPopAttrib();//GL_TRANSFORM_BIT
    }

  ///////////////////////////////////////////////////////
  glPushAttrib(GL_DEPTH_BUFFER_BIT);
  glDepthFunc(GL_GREATER);
  glDisable(GL_LIGHTING);
  glColor3d(0.0, 0.7, 0.7);
  if (!readAsBool("rotate"))
    {
      glPushAttrib(GL_LINE_BIT);
      glLineStipple(1, 0xAAAA);
      glEnable(GL_LINE_STIPPLE);
      glBegin(GL_LINE_LOOP);
      for (double theta(0.0); theta < M_2PI; theta += M_2PI / 32.0)
	(m.geto() + cos(theta) * px + sin(theta) * py).glVertex();
      glEnd();
      glDisable(GL_LINE_STIPPLE);
      glPopAttrib();//GL_LINE_BIT
    }
  glBegin(GL_LINES);
  (m.geto() + tmp0).glVertex();
  tmp1.glVertex();
  glEnd();
  (axis == 24)? glColor3d(0.7, 0.7, 0.0): glColor3d(0.0, 0.7, 0.7);
  glPushName(24);
  glBegin(GL_QUADS);
  (tmp1 + 0.1 * (px - py)).glVertex();
  (tmp1 + 0.1 * (3.0 * px + py)).glVertex();
  (tmp1 + 0.1 * (px + 3.0 * py)).glVertex();
  (tmp1 + 0.1 * (py - px)).glVertex();
  glEnd();
  glPopName();

  glEnable(GL_LIGHTING);

  if (!isEasy)
    {
      glPushAttrib(GL_TRANSFORM_BIT);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      Mat4d(m_x, m_y, m_z, m_o).glMultMatrix();
      glScaled(size, size, size);
      (axis == 23)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.6, 0.6, 0.7);
      glPushName(23);
      glCallList(_s_list + 1);
      glPopName();
      glRotated(-90.0, 1.0, 0.0, 0.0);
      (axis == 22)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.6, 0.7, 0.6);
      glPushName(22);
      glCallList(_s_list + 1);
      glPopName();
      glRotated(90.0, 0.0, 1.0, 0.0);
      (axis == 21)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.7, 0.6, 0.6);
      glPushName(21);
      glCallList(_s_list + 1);
      glPopName();
      glPopMatrix();//GL_MODELVIEW
      glPopAttrib();//GL_TRANSFORM_BIT
    }
  glPopAttrib();//GL_DEPTH_BUFFER_BIT
  glPopAttrib();//GL_LIGHTING_BIT
}

// ************************************************************************* //
void	NGizmo::glRender(Plug &plugOut)
{
  std::vector<DagNode*>	transforms;

  glInitLists();
  readAsSignal("renderIn");
  getPlug("transform")->getOutputNodes(transforms);
  if (readAsBool("visible") && !transforms.empty())
    {
      pushName();

      const Mat4d	&m = transforms.back()->readAsMat4d("matChild");
      Vec3d	viewdir;
      Pnt3d	viewpoint;

      viewpoint.glGetViewpoint();
      viewdir.glGetViewdir();

      const double	size((m.geto() - viewpoint) * viewdir.unit() / 10.0);

      writeAsDouble("size", size);

      if (readAsBool("translate"))
	renderTranslation(*transforms.back());
      if (readAsBool("rotate"))
	renderRotation(*transforms.back());
      if (readAsBool("scale"))
	renderScale(*transforms.back());
      popName();
    }
  plugOut.writeAsSignal();
}

// ************************************************************************* //
/// DESCRIPTION: Returns the active axis constraining the transformation.
// ************************************************************************* //
Vec3d	NGizmo::getConstrainAxis()
{ 
  std::vector<DagNode*>	transforms;

  getPlug("transform")->getOutputNodes(transforms);
  if (transforms.empty())
    return 0.0;

  const Mat4d	&m = transforms.back()->readAsMat4d("matChild");

  switch (readAsInt("axis"))
    {
    case 01:
    case 11:
    case 21:
      return m.getx().unit();
    case 02:
    case 12:
    case 22:
      return m.gety().unit();
    case 03:
    case 13:
    case 23:
      return m.getz().unit();
    default:
      return Vec3d(0.0);
    }
}

// ************************************************************************* //
void	NGizmo::update(Plug &plugOut)
{
  const Pnt3d	&p0(readAsPnt3d("p0"));
  const Pnt3d	&p1(readAsPnt3d("p1"));
  std::vector<DagNode*>	transforms;
  std::vector<DagNode*>::iterator	tit;

  plugOut.writeAsSignal();
  getPlug("transform")->getOutputNodes(transforms);
  if (transforms.empty())
    return;

  DagNode	*tr(transforms.back());
  const Mat4d	&m = tr->readAsMat4d("matChild");
  const int	axis(readAsInt("axis"));
  //DagNode	*camera = DagNode::getNodeByPath(":Camera0");
  std::list<DagNode*>	cameras;

  DagNode::root().DagNode::getNodesByType("Camera", cameras);

  DagNode	*camera = cameras.front();

  switch (axis)
    {
    case 01:
    case 02:
    case 03:
    case 04:
      {
	Vec3d	n;

	if (axis == 04)
	  n = p1 - p0;
	else
	  {
	    const Pnt3d	o = m.geto();
	    const Vec3d	x = getConstrainAxis();
	    const Pnt2d	op = Pnt2d(project(*camera, o));
	    const Pnt2d	oxp = Pnt2d(project(*camera, o + x));
	    Pnt2d	w0p = Pnt2d(project(*camera, p0));
	    Pnt2d	w1p = Pnt2d(project(*camera, p1));

	    w0p = w0p.proj2line(op, oxp);
	    w1p = w1p.proj2line(op, oxp);
	    if ((oxp - op).norm2() < 1.0e-12)
	      n = Vec3d(0.0);
	    else
	      n = x * (((w1p - w0p) * (oxp - op)) / (oxp - op).norm2());
	  }
	for (tit = transforms.begin(); tit != transforms.end(); ++tit)
	  {
	    (*tit)->writeAsDouble("tx", (*tit)->readAsDouble("tx") + n.x);
	    (*tit)->writeAsDouble("ty", (*tit)->readAsDouble("ty") + n.y);
	    (*tit)->writeAsDouble("tz", (*tit)->readAsDouble("tz") + n.z);
	  }
	break;
      }
    case 11:
    case 12:
    case 13:
    case 14:
      {
	Quaternion	q;
	const Pnt3d	o = m.geto();
	const Pnt2d	op = Pnt2d(project(*camera, o));
	const Pnt2d	w0p = Pnt2d(project(*camera, p0));
	const Pnt2d	w1p = Pnt2d(project(*camera, p1));
	const Pnt3d	eye(camera->readAsPnt3d("eye"));

	if (axis == 14)
	  {
	    double	angle(myacos((w0p - op).unit() * (w1p - op).unit()));
	    const Vec3d	dir(camera->readAsVec3d("dir"));
  
	    if (((w1p - op)(0) * (w0p - op)(1) - (w1p - op)(1) * (w0p - op)(0))
		< 0.0)
	      q = Quaternion::creator(-angle, dir);
	    else
	      q = Quaternion::creator(angle, dir);
	  }
	else
	  {
	    const Vec3d	x = getConstrainAxis();
	    Vec3d n0 = (p0.proj2planeAlong(o, x, Vec3d(p0 - eye)) - o).unit();
	    Vec3d n1 = (p1.proj2planeAlong(o, x, Vec3d(p1 - eye)) - o).unit();

	    if (((w1p - op)(0) * (w0p - op)(1) - (w1p - op)(1) * (w0p - op)(0))
		* (x * (eye - o)) < 0.0)
	      q = Quaternion::creator(myacos(n1 * n0), x);
	    else
	      q = Quaternion::creator(-myacos(n1 * n0), x);
	  }
	for (tit = transforms.begin(); tit != transforms.end(); ++tit)
	  {
	    const Mat4d		&r = (*tit)->readAsMat4d("rotate");
	    const Mat4d		a = Mat4d(q) * r;
	    const double	rz = atan2(-a(1, 0), a(0, 0));
	    const double	ry = atan2(a(2, 0) , a(0, 0) / cos(rz));
	    const double	rx = atan2(-a(2, 1), a(2, 2));

	    (*tit)->writeAsDouble("rx", RAD2DEG(rx));
	    (*tit)->writeAsDouble("ry", RAD2DEG(ry));
	    (*tit)->writeAsDouble("rz", RAD2DEG(rz));
	  }
	break;
      }
    case 21:
    case 22:
    case 23:
    case 24:
      {
	const Pnt3d	o = m.geto();

	if (axis == 24)
	  {
	    const Pnt2d	w0p = Pnt2d(project(*camera, p0));
	    const Pnt2d	w1p = Pnt2d(project(*camera, p1));
	    const Vec2d	v(w1p - w0p);
	    const Vec2d	dir(1.0, 1.0);
	    const Mat2i	&view = camera->readAsMat2i("view");
	    double	s;
	    
	    if (0 <= v * dir)
	      s = 1.0 + v.norm() / max(view(2), view(3));
	    else
	      s = 1.0 - v.norm() / max(view(2), view(3));

	    for (tit = transforms.begin(); tit != transforms.end(); ++tit)
	      {
		(*tit)->writeAsDouble("sx", (*tit)->readAsDouble("sx") * s);
		(*tit)->writeAsDouble("sy", (*tit)->readAsDouble("sy") * s);
		(*tit)->writeAsDouble("sz", (*tit)->readAsDouble("sz") * s);
	      }
	  }
	else
	  {
	    const Vec3d	x = getConstrainAxis();
	    Pnt2d	w0p = Pnt2d(project(*camera, p0));
	    Pnt2d	w1p = Pnt2d(project(*camera, p1));
	    const Pnt2d	op = Pnt2d(project(*camera, o));
	    const Pnt2d	oxp = Pnt2d(project(*camera, o + x));
	    double	s;

	    w0p = w0p.proj2line(op, oxp);
	    w1p = w1p.proj2line(op, oxp);
	    s = 1.0 + (w1p - w0p) * (oxp - op) / (oxp - op).norm2();
	    if (s < 1.0e-6)
	      s = 1.0e-6;
	    for (tit = transforms.begin(); tit != transforms.end(); ++tit)
	      {
		if (axis == 21)
		  (*tit)->writeAsDouble("sx", (*tit)->readAsDouble("sx") * s);
		else if (axis == 22)
		  (*tit)->writeAsDouble("sy", (*tit)->readAsDouble("sy") * s);
		else if (axis == 23)
		  (*tit)->writeAsDouble("sz", (*tit)->readAsDouble("sz") * s);
	      }
    	  }
	break;
      }
    }
}

// ************************************************************************* //
void	NGizmo::compute(Plug &plugOut)
{
  if (plugOut.shortName() == "renderOut")
    glRender(plugOut);
  else if (plugOut.shortName() == "update")
    update(plugOut);
}

