// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:29 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include "nodeimagelayer.h"
#include "gbb.h"
#include "datablock.h"
#include "plug.h"
#include <QImage>

// ************************************************************************* //
int	NImageLayer::typeNbr(86);
String	NImageLayer::typeStr("ImageLayer");

// ************************************************************************* //
void	NImageLayer::initialize()
{
  DataBlock	&db(getDataBlock());

  db.addInt(0, 0, MAXINT, "renderIn", false);
  db.addSignal("renderOut", false);
  db.addString("", "matteFg");
  db.addSignal("camera", false);
  db.addInt(0, 0, MAXINT, "texnameLores", false, false);
  db.addInt(0, 0, MAXINT, "texnameHires", false, false);
  db.addDouble(0.0, 0.0, 1.0, "depth");
  db.addBool(true, "enabled");

  db.affect("renderIn", "renderOut");
  db.affect("matteFg", "renderOut");
  db.affect("camera", "renderOut");

  db.affect("matteFg", "texnameLores");
  db.affect("depth", "texnameLores");
  db.affect("enabled", "texnameLores");

  db.affect("matteFg", "texnameHires");
  db.affect("depth", "texnameHires");
  db.affect("enabled", "texnameHires");

  buildDagNodes();
}

// ************************************************************************* //
void	NImageLayer::buildDagNodes()
{
  std::list<DagNode*>	cameras;

  DagNode::root().DagNode::getNodesByType("Camera", cameras);

  DagNode	*camera = cameras.front();

  DagNode::connect(*camera, "signal", *this, "camera");
  DagNode::connect(*_m_parent, "renderIn", *this, "renderIn");
}

// ************************************************************************* //
void	NImageLayer::glRender(Plug &plugOut)
{
  GLuint	texname(0);
  const int	renderIn = readAsInt("renderIn");

  if (renderIn == Gbb::Rendering)
    texname = readAsInt("texnameHires");
  else if (renderIn == Gbb::Modeling)
    texname = readAsInt("texnameLores");
  if (texname)
    {
      const double	depth = -readAsDouble("depth");

      glPushAttrib(GL_LIGHTING_BIT | GL_TRANSFORM_BIT);
      glDisable(GL_LIGHTING);
      glMatrixMode(GL_PROJECTION);
      glPushMatrix();
      glLoadIdentity();
      glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      glLoadIdentity(); 
	  
      glEnable(GL_TEXTURE_2D);
      glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
      glBindTexture(GL_TEXTURE_2D, texname);
      glColor3d(1.0, 1.0, 1.0);
      glBegin(GL_QUADS);

      glTexCoord2d(0.0, 1.0);
      glVertex3d(-1.0, -1.0, depth);
      glTexCoord2d(1.0, 1.0);
      glVertex3d(1.0, -1.0, depth);
      glTexCoord2d(1.0, 0.0);
      glVertex3d(1.0, 1.0, depth);
      glTexCoord2d(0.0, 0.0);
      glVertex3d(-1.0, 1.0, depth);

      glEnd();//GL_QUADS
      glDisable(GL_TEXTURE_2D);
  
      glPopMatrix();
      glMatrixMode(GL_PROJECTION);
      glPopMatrix();
      glPopAttrib();// GL_LIGHTING_BIT | GL_TRANSFORM_BIT
    }
  plugOut.writeAsSignal();
}

// ************************************************************************* //
void	NImageLayer::compute(Plug &plugOut)
{
  if (plugOut.shortName() == "texnameHires")
    {
      QImage	qimage(readAsString("matteFg").data());
      GLuint	texnameNear = readAsInt("texnameHires");

      if (texnameNear)
	glDeleteTextures(1, &texnameNear);
      if (qimage.isNull())
	plugOut.writeAsInt(0);
      else
	if (readAsBool("enabled"))
	  {
	    glGenTextures(1, &texnameNear);
	    glBindTexture(GL_TEXTURE_2D, texnameNear);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
#ifndef _WIN32
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 
			 qimage.width(), qimage.height(), 
			 0, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV,
			 qimage.bits());
#else//_WIN32
	    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 
			 qimage.width(), qimage.height(), 
			 0, GL_RGBA, GL_UNSIGNED_INT,
			 qimage.bits());
#endif//_WIN32
	    plugOut.writeAsInt(texnameNear);
	  }
	else
	  plugOut.writeAsInt(0);
    }
  else if (plugOut.shortName() == "texnameLores")
    {
      QImage	qimage(readAsString("matteFg").data());
      GLuint	texnameNear = readAsInt("texnameLores");

      if (texnameNear)
	glDeleteTextures(1, &texnameNear);
      if (qimage.isNull())
	plugOut.writeAsInt(0);
      else
	if (readAsBool("enabled"))
	  {
	    qimage = qimage.scaled(256, 256, Qt::IgnoreAspectRatio,
				   Qt::SmoothTransformation);
	    glGenTextures(1, &texnameNear);
	    glBindTexture(GL_TEXTURE_2D, texnameNear);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
#ifndef _WIN32
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 
			 qimage.width(), qimage.height(), 
			 0, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV,
			 qimage.bits());
#else//_WIN32
	    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 
			 qimage.width(), qimage.height(), 
			 0, GL_RGBA, GL_UNSIGNED_INT,
			 qimage.bits());
#endif//_WIN32
		plugOut.writeAsInt(texnameNear);
	  }
	else
	  plugOut.writeAsInt(0);
    }
  else if (plugOut.name() == "renderOut")
    glRender(plugOut);
}
