// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:29 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include "nodelocator.h"
#include "datablock.h"
#include "plug.h"
#include "command.h"
#include "gbb.h"

// ************************************************************************* //
int	NLocator::typeNbr(79);
String	NLocator::typeStr("Locator");

// ************************************************************************* //
void	NLocator::initialize()
{
  DataBlock	&db(getDataBlock());

  NShape::initialize();
  db.addMat4d(Mat4d::identity(), "matParent", false);//position
  db.addMat4d(Mat4d::identity(), "imatParent", false);//position inverse
  db.addMat4d(Mat4d::identity(), "matChild", false);//position
  db.addDouble(0.5, 0.0, MAXDOUBLE, "size");
  db.addPnt3d(0.0, "pos");

  db.affect("matParent", "imatParent");
  db.affect("matParent", "pos");
  db.affect("matParent", "matChild");

  buildDagNodes();
}

// ************************************************************************* //
void	NLocator::buildDagNodes()
{
  DagNode	*oldParent(_m_parent);
  DagNode	*tr = Command::run(String("new -type Transform -path :"));

  if (oldParent)
    DagNode::parent(*oldParent, *tr);
  DagNode::parent(*tr, *this);
  DagNode::connect(*oldParent, "renderIn", *tr, "renderIn");
  DagNode::connect(*tr, "renderIn", *this, "renderIn");
  DagNode::connect(*tr, "matChild", *this, "matParent");
}

// ************************************************************************* //
void	NLocator::glRender(Plug &plugOut)
{
  if (readAsInt("renderIn") != Gbb::Rendering)
    {
      const double	size(readAsDouble("size"));
      const int	sel(readAsInt("sel"));

      pushName();
      glPushMatrix();
      readAsMat4d("matParent").glMultMatrix();
      glScaled(size, size, size);
      if (sel == 1)
	glColor4d(1.0, 0.0, 0.5, 1.0);
      else if (sel)
	glColor4d(1.0, 0.5, 0.0, 1.0);
      else
	glColor4d(0.0, 0.25, 1.0, 1.0);
      glPushAttrib(GL_LIGHTING_BIT);
      glDisable(GL_LIGHTING);
      glBegin(GL_LINES);
      glVertex3d(0.0, 0.0, 1.0);
      glVertex3d(0.0, 0.0, -1.0);
      glVertex3d(0.0, 1.0, 0.0);
      glVertex3d(0.0, -1.0, 0.0);
      glVertex3d(1.0, 0.0, 0.0);
      glVertex3d(-1.0, 0.0, 0.0);
      glEnd();//GL_LINES
      glPopAttrib();//GL_LIGHTING_BIT
      glPopMatrix();

      std::list<DagNode*>::const_iterator	cit;
      Plug	*data;
      
      for (cit = children().begin(); cit != children().end(); ++cit)
	if ((data = (*cit)->getPlug("renderOut")))
	  data->readAsSignal();

      popName();
    }
  plugOut.writeAsSignal();
}

// ************************************************************************* //
void	NLocator::compute(Plug &plugOut)
{
  if (plugOut.shortName() == "renderOut")
    glRender(plugOut);
  else if (plugOut.shortName() == "imatParent")
    plugOut.writeAsMat4d(readAsMat4d("matParent").inverse());
  else if (plugOut.shortName() == "pos")
    plugOut.writeAsPnt3d(readAsMat4d("matParent") * Pnt3d(0.0));
  else if (plugOut.shortName() == "matChild")
    plugOut.writeAsMat4d(readAsMat4d("matParent"));
}


