// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:29 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include <fstream>
#include <qprogressdialog.h>
#include <qapplication.h>
#include "datablock.h"
#include "nodeobstacle.h"
#include "command.h"
#include "ten3d.h"
#include "plugcustom.h"
#include "plugmesh.h"
#include "qtMainWidget.h"

// ************************************************************************* //
int	NObstacle::typeNbr(63);
String	NObstacle::typeStr("Obstacle");

// ************************************************************************* //
int	PlugMatMNd::typeNbr(71);
String	PlugMatMNd::typeStr("PlugMatMNd");
int	PlugOctreeVortex::typeNbr(74);
String	PlugOctreeVortex::typeStr("PlugOctreeVortex");
int	PlugObstacleVortex::typeNbr(75);
String	PlugObstacleVortex::typeStr("PlugObstacleVortex");

// ************************************************************************* //
double	rbfObstacle(const double	l2,
		    const double	r2)
{
  return (r2 < l2)? 0.0: sq(4.0 - 20.0 / (l2 / r2 + 4.0));
}

// ************************************************************************* //
/// TODO: remove ovortices.
/// TODO: Do second octree to blur out high freq. Put them back with ovortices.
// ************************************************************************* //
void	NObstacle::initialize()
{
  NShape::initialize();

  DataBlock	&db(getDataBlock());

  db.addInt(0, MININT, MAXINT, "f", false);
  db.addInt(0, MININT, MAXINT, "fPrev", false);
  db.addDouble(0, 0.0, MAXDOUBLE, "dt", false);
  db.addSignal("update", false);
  db.addMat4d(Mat4d::identity(), "matParent", false);//position
  db.addMat4d(Mat4d::identity(), "imatParent", false);//position
  db.addDouble(1.0, 0.0, MAXDOUBLE, "radius");//TODO: control face density!!!
  db.addDouble(0.0001, 0.0, MAXDOUBLE, "centerDepth");
  db.addInt(4, 1, MAXINT, "centerResolution");
  db.addInt(6, 1, MAXINT, "sampleResolution");
  db.addString("./obstacle.obs", "meshfile");
  db.addString("./obstacle1.0_4x6.mat", "matrixfile");
  db.addBool(false, "displayData");
  db.addBool(true, "isFlatShaded");

  db.addPnt3d(Pnt3d(0.0), "p", false);
  db.addTwistLog(TwistLog(0.0), "rlog", false);

  db.addPnt3d(Pnt3d(0.0), "pobstacle", false);
  db.addTwistLog(TwistLog(0.0), "rlogObstacle", false);

  db.addPnt3d(Pnt3d(0.0), "povortices", false);
  db.addTwistLog(TwistLog(0.0), "rlogOvortices", false);

  db.addCustom(*new PlugMatMNd(db), "v2vComputed", false, false);
  db.addCustom(*new PlugMatMNd(db), "v2vLoaded", false, false);
  db.addCustom(*new PlugMatMNd(db), "v2vTreated", false, false);
  db.addCustom(*new PlugMesh(db), "mesh", false);
  db.addCustom(*new PlugBBox3d(db), "bbox", false);
  db.addCustom(*new PlugOctreeVortex(db), "grid", false);

  db.addCustom(*new PlugObstacleVortex(db), "ovortex", false);
  db.addDouble(1.0, 0.0, MAXDOUBLE, "stickiness");
  db.addDouble(0.1, 0.0, MAXDOUBLE, "scale");

  db.addDouble(0.01, 0.0, MAXDOUBLE, "safeZone");// No-slip inside
  db.addDouble(0.01, 0.0, MAXDOUBLE, "smoothZone");// No-slip inside

  db.addVec3dArray(0, Vec3d(0.0), "cnormals", false);
  db.addPnt3dArray(0, Pnt3d(0.0), "ssamples", false);
  db.addPnt3dArray(0, Pnt3d(0.0), "ccenters", false);
  db.addVec3dArray(0, Vec3d(0.0), "cnormalsWorld", false);
  db.addPnt3dArray(0, Pnt3d(0.0), "ssamplesWorld", false);
  db.addPnt3dArray(0, Pnt3d(0.0), "ccentersWorld", false, false);
  db.addVec3dArray(0, Vec3d(0.0), "svorticesWorld", false, false);//that's it..
  db.addTwistLogArray(0, TwistLog(0.0), "slogWorld", false, false);

  db.addDouble(20000.0, 0.0, MAXDOUBLE, "detachAbove");//DEBUG
  db.addDouble(2.0, 0.0, MAXDOUBLE, "maxStretch");//DEBUG

  db.affect("ccentersWorld", "bbox");
  db.affect("radius", "bbox");

  db.affect("bbox", "grid");
  db.affect("radius", "grid");
  db.affect("ccentersWorld", "grid");

  db.affect("mesh", "renderOut");
  db.affect("f", "update");
  db.affect("meshfile", "mesh");
  db.affect("mesh", "ssamples");
  db.affect("sampleResolution", "ssamples");

  db.affect("mesh", "cnormals");
  db.affect("sampleResolution", "cnormals");

  db.affect("mesh", "ccenters");
  db.affect("centerDepth", "ccenters");
  db.affect("centerResolution", "ccenters");

  db.affect("matrixfile", "v2vLoaded");

  db.affect("v2vLoaded", "v2vTreated");
  db.affect("matParent", "v2vTreated");

  db.affect("matParent", "imatParent");

  db.affect("ssamples", "v2vComputed");
  db.affect("ccenters", "v2vComputed");
  db.affect("radius", "v2vComputed");

  db.affect("ssamples", "ssamplesWorld");
  db.affect("matParent", "ssamplesWorld");

  db.affect("cnormals", "cnormalsWorld");
  db.affect("matParent", "cnormalsWorld");

  db.affect("ccenters", "ccentersWorld");
  db.affect("matParent", "ccentersWorld");

  db.affect("svorticesWorld", "slogWorld");
  db.affect("ccentersWorld", "slogWorld");

  db.affect("ssamples", "rlog");
  db.affect("p", "rlog");

  db.affect("ssamples", "rlogOvortices");
  db.affect("slogWorld", "rlogObstacle");
  db.affect("pobstacle", "rlogObstacle");

  db.affect("povortices", "rlogOvortices");

  //db.affect("ssamples", "rlogObstacle");
  //db.affect("slogWorld", "rlogObstacle");

  buildDagNodes();
}

// ************************************************************************* //
void	NObstacle::buildDagNodes()
{
  DagNode	*oldParent(_m_parent);
  DagNode	*tr = Command::run(String("new -type Transform -path :"));
  DagNode	*clock = DagNode::getNodeByPath(":Clock0");

  if (oldParent)
    DagNode::parent(*oldParent, *tr);
  DagNode::parent(*tr, *this);
  DagNode::connect(*oldParent, "renderIn", *tr, "renderIn");
  DagNode::connect(*tr, "renderIn", *this, "renderIn");
  DagNode::connect(*tr, "matChild", *this, "matParent");
  DagNode::connect(*clock, "f", *this, "f");
  DagNode::connect(*clock, "dt", *this, "dt");
}

// ************************************************************************* //
void	NObstacle::computev2vCompute(Plug &plugOut)
{
  const double	radius(readAsDouble("radius"));
  const std::vector<Pnt3d>	&ssamples(readAsPnt3dArray("ssamples"));
  const std::vector<Pnt3d>	&ccenters(readAsPnt3dArray("ccenters"));
  const unsigned	nbsamples(ssamples.size());
  const unsigned	nbcenters(ccenters.size());
  MatMNd		m(3 * nbsamples, 3 * nbcenters);

  for (unsigned i(0); i < nbsamples; ++i)
    for (unsigned j(0); j < nbcenters; ++j)
      {
	const Pnt3d	&si(ssamples[i]);
	const Pnt3d	&cj(ccenters[j]);
	const Vec3d	v(si - cj);
	const double	l2(v.norm2());
	const Vec3d	e = v * rbfObstacle(l2, sq(radius));
	const Mat3d	mij = Ten3d::permutation() * e;

	for (unsigned ki(0); ki < 3; ++ki)
	  for (unsigned kj(0); kj < 3; ++kj)
	    m(3 * i + ki, 3 * j + kj) = mij(ki, kj);
      }
  MatMNd	*v2v = static_cast<MatMNd *>(readAsCustom("v2vComputed"));

  std::cout << "Inverting " << m.rows() << "x" << m.cols()
	    << " matrix..." << std::flush;
  v2v->eqInverse(m);
  std::cout << "done." << std::endl;
  plugOut.writeAsSignal();
}

// ************************************************************************* //
void	NObstacle::computev2vLoad(Plug &plugOut)
{
  const String	&matrixfile = readAsString("matrixfile");

  if (!matrixfile.empty())
    {
      std::ifstream	ifs(matrixfile.data());

      if (ifs)
	{
	  static_cast<MatMNd *>(readAsCustom("v2vLoaded"))->loadBits(ifs);
	  plugOut.writeAsSignal();
	  return;
	}
    }
  plugOut.writeAsSignal();
}

// ************************************************************************* //
/// DESCRIPTION: update signal pulled by parent calls this method.
// ************************************************************************* //
void	NObstacle::update(Plug &plugOut)
{
  const int	fPrev = readAsInt("fPrev");
  const int	f = readAsInt("f");

  writeAsInt("fPrev", f);

  const double	dt = readAsDouble("dt");

  const std::vector<Pnt3d>	&ccentersW = readAsPnt3dArray("ccentersWorld");
  const std::vector<Pnt3d>	&ssamplesW = readAsPnt3dArray("ssamplesWorld");
  const std::vector<Vec3d>	&cnormalsW = readAsVec3dArray("cnormalsWorld");
  MatMNd	*v2vTreated = static_cast<MatMNd*>(readAsCustom("v2vTreated"));
  DagNode	*rs = DagNode::getNodeByPath(":VortexSet0");
  const Mat4d	&matParent = readAsMat4d("matParent");
  const Mat4d	&imatParent = readAsMat4d("imatParent");
  const double	radius(readAsDouble("radius"));

  // Updates ovortices ***************************************************** //
  {
    DagNode	*rs = DagNode::getNodeByPath(":VortexSet0");
    Plug	*plugp(rs->getPlug("ptotal"));//TODO :use octree!!!
    Plug	*plugTlog(rs->getPlug("rlogTotal"));
    Plug	*plugJac(rs->getPlug("jacTotal"));

    std::list<OVortex>	&ov = 
      static_cast<DataObstacleVortex *>(readAsCustom("ovortex"))->_m_ovortices;
    std::list<OVortex>::iterator	oit;

    for (oit = ov.begin(); oit != ov.end(); ++oit)
      {
	TwistLog	tlog;
	Mat3d		jac;

	plugp->writeAsPnt3d(oit->c);
	tlog = plugTlog->readAsTwistLog();
	jac = plugJac->readAsMat3d();
	(tlog * dt).exp(oit->c);
	oit->stretch2 *= (jac * oit->w).norm2() / oit->w.norm2();
	oit->w = jac * oit->w;
      }
  }

  // Generates vortices around object ************************************** //
  // Deletes vortices around object **************************************** //
  // TODO: Accumulate vorticity? ... i.e. sum up vorticity vectors
  {
    DataObstacleVortex	*ov = 
      static_cast<DataObstacleVortex *>(readAsCustom("ovortex"));
    const std::vector<Vec3d> &svorticesW = readAsVec3dArray("svorticesWorld");
    const double	stickiness = readAsDouble("stickiness");
    const double	scale = readAsDouble("scale");
    const double	centerDepth = readAsDouble("centerDepth");
    const double	detachAbove(readAsDouble("detachAbove"));
    const double	maxStretch(readAsDouble("maxStretch"));

    if (f == 0 && f < fPrev)
      ov->_m_ovortices.clear();
    else
      for (unsigned i(0), n(svorticesW.size()); i < n; ++i)
	if (sq(detachAbove) < svorticesW[i].norm2() * sq(dt))
	  ov->_m_ovortices.push_back
	    (OVortex(ccentersW[i] + cnormalsW[i] * (centerDepth * 2 + scale),
		     svorticesW[i] * dt * stickiness, 1.0, radius));
    
    std::list<OVortex>::iterator	oit;

    for (oit = ov->_m_ovortices.begin(); oit != ov->_m_ovortices.end(); )
      if (sqrt(maxStretch) < oit->stretch2)
	ov->_m_ovortices.erase(oit++);
      else
	++oit;
  }

  // Compute obstacle matrix
  if (v2vTreated->size() != 0 && v2vTreated->cols() == 3 * ssamplesW.size())
    {
      VecNd	v(3 * ssamplesW.size());
      Plug	*plugsvortices(getPlug("svorticesWorld"));
      Plug	*plugpoctree(rs->getPlug("poctree"));
      Plug	*plugrlogoctree(rs->getPlug("rlogOctree"));
      Plug	*plugpovortices(getPlug("povortices"));
      Plug	*plugrlogovortices(getPlug("rlogOvortices"));

      double	sum2(0.0);

      // TODO: use also ovortices

      for (unsigned i(0), n(ssamplesW.size()); i < n; ++i)
	{
	  const Pnt3d	&si(ssamplesW[i]);
	  Vec3d	vi;

	  plugpoctree->writeAsPnt3d(si);
	  plugpovortices->writeAsPnt3d(si);
	  vi = -((plugrlogoctree->readAsTwistLog() + 
		  plugrlogovortices->readAsTwistLog()) * si);

	  v(3 * i + 0) = vi.x;
	  v(3 * i + 1) = vi.y;
	  v(3 * i + 2) = vi.z;
	  sum2 += vi.norm2();
	}
      plugsvortices->resize(ccentersW.size());
      if (1.0e-8 < sum2)
	{
	  std::vector<Vec3d>	w(ccentersW.size());
	  
	  //TODO: compute only the vortices that are used!!!
	  //std::cout << "in update" << std::endl;
	  VecNd::eqMul(w, *v2vTreated, v);
	  //std::cout << "in update::done" << std::endl;
	  plugsvortices->writeAsVec3dArray(w);
	}
      else
	{
	  for (unsigned j(0), n(ccentersW.size()); j < n; ++j)
	    plugsvortices->writeAsVec3dArray(j, Vec3d(0.0));
	}
    }
  plugOut.writeAsSignal();
}

// ************************************************************************* //
void	NObstacle::glRender(Plug &plugOut)
{
  const int	sel(readAsInt("sel"));
  const Mat4d	&matParent(readAsMat4d("matParent"));
  const Mat4d	&imatParent(readAsMat4d("imatParent"));

  if (sel == 1)
    glColor3d(1.0, 0.0, 0.5);
  else if (sel)
    glColor3d(1.0, 0.5, 0.0);
  else
    glColor3d(0.0, 0.0, 1.0);

  pushName();
  glPushMatrix();
  readAsMat4d("matParent").glMultMatrix();
  
  const bool	isFlatShaded(readAsBool("isFlatShaded"));

  static_cast<Mesh *>(readAsCustom("mesh"))->glRender(isFlatShaded);

  glPushMatrix();
  imatParent.glMultMatrix();
  static_cast<DataObstacleVortex *>(readAsCustom("ovortex"))->glRender();
  glPopMatrix();

  if (readAsBool("displayData"))
    {
      const std::vector<Pnt3d>	&ssamples(readAsPnt3dArray("ssamples"));
      const std::vector<Pnt3d>	&ccenters(readAsPnt3dArray("ccenters"));

      glPushAttrib(GL_LIGHTING_BIT);
      glDisable(GL_LIGHTING);
      glBegin(GL_POINTS);
      glColor3d(0.0, 0.0, 1.0);
      for (unsigned i(0); i < ssamples.size(); ++i)
	ssamples[i].glVertex();
      glColor3d(1.0, 0.0, 0.0);
      for (unsigned i(0); i < ccenters.size(); ++i)
	ccenters[i].glVertex();
      glEnd();//GL_POINTS

      glColor3d(1.0, 1.0, 0.0);

      {
	const std::vector<Vec3d>&svorticesW=readAsVec3dArray("svorticesWorld");
	const std::vector<Pnt3d>&ccentersW=readAsPnt3dArray("ccentersWorld");
	const double	dt = readAsDouble("dt");

	glColor3d(0.0, 0.0, 0.0);
	glBegin(GL_LINES);
	for (unsigned i(0), n(svorticesW.size()); i < n; ++i)
	  {
	    const Pnt3d	si(ccentersW[i]);
	    const Vec3d	vi = svorticesW[i] * dt;

	    (imatParent * (si)).glVertex();
	    (imatParent * (si + vi)).glVertex();
	  }
	glEnd();//GL_LINES
      }

      glPopAttrib();//GL_LIGHTING_BIT

      const BBox3d	&bbox = *static_cast<BBox3d *>(readAsCustom("bbox"));
      glPushMatrix();
      imatParent.glMultMatrix();
      bbox.glRender();
      glPopMatrix();
    }

  glPopMatrix();
  popName();
  readAsSignal("renderIn");
  plugOut.writeAsSignal();
}

// ************************************************************************* //
void	NObstacle::computeRlogObstacle(Plug &plugOut)
{
  TwistLog	rlog(0.0);
  const Pnt3d	&p(readAsPnt3d("pobstacle"));
  const BBox3d	&bbox = *static_cast<BBox3d *>(readAsCustom("bbox"));

  const std::vector<TwistLog>	&slogWorld = readAsTwistLogArray("slogWorld");
  const std::vector<Pnt3d>	&ccentersW =readAsPnt3dArray("ccentersWorld");

  const double	radius(readAsDouble("radius"));
  const Octree<std::vector<unsigned> >	&grid = 
    *static_cast<Octree<std::vector<unsigned> >*>(readAsCustom("grid"));
  const OLeaf<std::vector<unsigned> >	*leaf;

  if (bbox.contains(p) && (leaf = grid.query(p)))
    {
      const std::vector<unsigned>	&l = leaf->elm;
      std::vector<unsigned>::const_iterator	lit;

      for (lit = l.begin(); lit != l.end(); ++lit)
	{
	  const unsigned j = *lit;
	  const Pnt3d	&cj(ccentersW[j]);
	  const double	l2((p - cj).norm2());
	  const double	r2(sq(radius));

	  if (l2 < r2)
	    rlog += slogWorld[j] * rbfObstacle(l2, r2);
	}
    }
  plugOut.writeAsTwistLog(rlog);
}

// ************************************************************************* //
void	NObstacle::computeRlogOvortices(Plug &plugOut)
{
  TwistLog	rlog(0.0);
  const Pnt3d	&p(readAsPnt3d("povortices"));

  std::list<OVortex>	ov = 
    static_cast<DataObstacleVortex *>(readAsCustom("ovortex"))->_m_ovortices;
  std::list<OVortex>::const_iterator	oit;

  for (oit = ov.begin(); oit != ov.end(); ++oit)
    rlog += oit->rlog(p);
  plugOut.writeAsTwistLog(rlog);
}

// ************************************************************************* //
void	NObstacle::computeRlog(Plug &plugOut)
{
  const Pnt3d	&p(readAsPnt3d("p"));

  writeAsPnt3d("pobstacle", p);
  writeAsPnt3d("povortices", p);
  plugOut.writeAsTwistLog(readAsTwistLog("rlogObstacle") +
			  readAsTwistLog("rlogOvortices"));
}

// ************************************************************************* //
void	NObstacle::computeBBox(Plug &plugOut)
{
  const double	radius(readAsDouble("radius"));
  const std::vector<Pnt3d>	&ccentersW(readAsPnt3dArray("ccentersWorld"));
  std::vector<Pnt3d>::const_iterator	cit;
  BBox3d	bbox;

  if (ccentersW.empty())
    bbox = BBox3d(Pnt3d(0.0));
  else
    {
      bbox = BBox3d(ccentersW.front());
      for (cit = ccentersW.begin(), ++cit; cit != ccentersW.end(); ++cit)
	bbox.fill(*cit);
      bbox.eqExpand(radius);
    }
  plugOut.writeAsCustom(&bbox);
}

// ************************************************************************* //
void	NObstacle::computeGrid(Plug &plugOut)
{
  Octree<std::vector<unsigned> > *octree =
    static_cast<Octree<std::vector<unsigned> >*>(readAsCustom("grid"));
  const BBox3d	&bbox = *static_cast<BBox3d *>(readAsCustom("bbox"));
  const std::vector<Pnt3d> &ccentersW =readAsPnt3dArray("ccentersWorld");
  const double	radius(readAsDouble("radius"));

  octree->resize(bbox);
  for (unsigned c(0); c < ccentersW.size(); ++c)
    {
      std::list<OLeaf<std::vector<unsigned> >*>	ret;
      std::list<OLeaf<std::vector<unsigned> >*>::iterator	rit;

      octree->createSphereNeighbors(ccentersW[c], radius);
      octree->querySphereNeighbors(ccentersW[c], radius, ret);
      for (rit = ret.begin(); rit != ret.end(); ++rit)
	(*rit)->elm.push_back(c);
    }
  plugOut.writeAsSignal();
}

// ************************************************************************* //
void	NObstacle::compute(Plug &plugOut)
{
  if (plugOut.shortName() == "rlog") //sum of logs
    computeRlog(plugOut);
  else if (plugOut.shortName() == "rlogObstacle")//just the obstacle
    computeRlogObstacle(plugOut);//TODO
  else if (plugOut.shortName() == "rlogOvortices")//just the ovortices
    computeRlogOvortices(plugOut);//TODO
  else if (plugOut.shortName() == "renderOut")
    glRender(plugOut);
  else if (plugOut.shortName() == "update")
    update(plugOut);
  else if (plugOut.shortName() == "grid")
    computeGrid(plugOut);
  else if (plugOut.shortName() == "mesh")
    {
      std::ifstream	ifs(readAsString("meshfile").data());
      Mesh *mesh = static_cast<Mesh *>(readAsCustom("mesh"));

      if (ifs)
	ifs >> *mesh;
      plugOut.writeAsSignal();
    }
  else if (plugOut.shortName() == "bbox")
    computeBBox(plugOut);
  else if (plugOut.shortName() == "ccenters")
    {
      Mesh	*mesh = static_cast<Mesh *>(readAsCustom("mesh"));
      std::vector<Pnt3d>	ccenters;

      mesh->computeCenters(ccenters, readAsDouble("centerDepth"),
			   readAsInt("centerResolution"));
      plugOut.writeAsPnt3dArray(ccenters);
    }
  else if (plugOut.shortName() == "cnormals")
    {
      Mesh	*mesh = static_cast<Mesh *>(readAsCustom("mesh"));
      std::vector<Vec3d>	cnormals;

      mesh->computeNormals(cnormals, readAsInt("centerResolution"));
      plugOut.writeAsVec3dArray(cnormals);
    }
  else if (plugOut.shortName() == "ssamples")
    {
      Mesh	*mesh = static_cast<Mesh *>(readAsCustom("mesh"));
      std::vector<Pnt3d>	ssamples;

      mesh->computeSamples(ssamples, readAsInt("sampleResolution"));
      plugOut.writeAsPnt3dArray(ssamples);
    }
  else if (plugOut.shortName() == "slogWorld")
    {
      const std::vector<Pnt3d> &ccentersW =readAsPnt3dArray("ccentersWorld");
      const std::vector<Vec3d> &svorticesW(readAsVec3dArray("svorticesWorld"));
      std::vector<TwistLog>    slogWorld(svorticesW.size());

      for (unsigned j(0); j < svorticesW.size(); ++j)
	{
	  const Pnt3d	&cj(ccentersW[j]);
	  const Vec3d	&wj(svorticesW[j]);
	      
	  slogWorld[j] = TwistLog(wj, Vec3d(cj) ^ wj);
	}
      writeAsTwistLogArray("slogWorld", slogWorld);
    }
  else if (plugOut.shortName() == "ccentersWorld")
    {
      const Mat4d	&matParent(readAsMat4d("matParent"));
      const std::vector<Pnt3d>	&ccenters(readAsPnt3dArray("ccenters"));
      std::vector<Pnt3d>	ccentersWorld(ccenters.size());
      
      for (unsigned i(0); i < ccenters.size(); ++i)
	ccentersWorld[i] = matParent * ccenters[i];
      plugOut.writeAsPnt3dArray(ccentersWorld);
    }
  else if (plugOut.shortName() == "cnormalsWorld")
    {
      const Mat4d	&matParent(readAsMat4d("matParent"));
      const std::vector<Vec3d>	&cnormals(readAsVec3dArray("cnormals"));
      std::vector<Vec3d>	cnormalsWorld(cnormals.size());
      
      for (unsigned i(0); i < cnormals.size(); ++i)
	cnormalsWorld[i] = matParent * cnormals[i];
      plugOut.writeAsVec3dArray(cnormalsWorld);
    }
  else if (plugOut.shortName() == "ssamplesWorld")
    {
      const Mat4d	&matParent(readAsMat4d("matParent"));
      const std::vector<Pnt3d>	&ssamples(readAsPnt3dArray("ssamples"));
      std::vector<Pnt3d>	ssamplesWorld(ssamples.size());
      
      for (unsigned i(0); i < ssamples.size(); ++i)
	ssamplesWorld[i] = matParent * ssamples[i];
      plugOut.writeAsPnt3dArray(ssamplesWorld);
    }
  else if (plugOut.shortName() == "imatParent")
    plugOut.writeAsMat4d(readAsMat4d("matParent").inverse());
  else if (plugOut.shortName() == "v2vLoaded")
    computev2vLoad(plugOut);
  else if (plugOut.shortName() == "v2vTreated")
    {
      MatMNd	*v2vLoaded = static_cast<MatMNd*>(readAsCustom("v2vLoaded"));
      const unsigned rows(v2vLoaded->rows());
      const unsigned cols(v2vLoaded->cols());
      MatMNd	v2vTemp(rows, cols);
      MatMNd	*v2vTreated = static_cast<MatMNd*>(readAsCustom("v2vTreated"));
      const Mat4d	&matParent(readAsMat4d("matParent"));
      const Mat4d	&imatParent(readAsMat4d("imatParent"));
      QProgressDialog	progress("Placing obtsacle matrix...", "Abort", 
				 rows * 2, 16, MainWindow::mainwindow());

      MainWindow::mainwindow()->freeze();
      for (unsigned i(0); i < rows; ++i)
	{
	  progress.setValue(i);
	  qApp->processEvents();
	  if (progress.wasCanceled())
	    {
	      plugOut.writeAsSignal();
	      return;
	    }
	  for (unsigned j(0); j < cols; ++j)
	    {
	      double	&cur = v2vTemp(i, j);

	      cur = 0.0;
	      for (unsigned k(i - i % 3); k < i - i % 3 + 3; ++k)
		cur += matParent(i % 3, k % 3) * (*v2vLoaded)(k, j);
	    }
	}
      v2vTreated->resize(rows, cols);
      for (unsigned i(0); i < rows; ++i)
	{
	  progress.setValue(rows + i);
	  qApp->processEvents();
	  if (progress.wasCanceled())
	    {
	      plugOut.writeAsSignal();
	      return;
	    }
	  for (unsigned j(0); j < cols; ++j)
	    {
	      double	&cur = (*v2vTreated)(i, j);

	      cur = 0.0;
	      for (unsigned k(j - j % 3); k < j - j % 3 + 3; ++k)
		cur += v2vTemp(i, k) * imatParent(k % 3, j % 3);
	    }
	}
      MainWindow::mainwindow()->warmup();
      plugOut.writeAsSignal();
    }
  else if (plugOut.shortName() == "v2vComputed")
    computev2vCompute(plugOut);
}
