// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:29 $
/// $Revision: 1.1 $
// ************************************************************************* //
#ifndef __NODEOBSTACLE_HH__
#define __NODEOBSTACLE_HH__

// ************************************************************************* //
#include "nodeshape.h"
#include "matnd.h"
#include "matmnd.h"
#include "plug.h"
#include "dataoctree.h"
#include "gfxtypes.h"

// ************************************************************************* //
double	rbfObstacle(const double l2, const double r2);

// ************************************************************************* //
class NObstacle: public NShape
{
public:
  static int	typeNbr;
  static String	typeStr;
public:
  NObstacle(DagNode *parent = 0): NShape(parent) {}
  virtual void		initialize();
  virtual int		getTypeNbr() const { return typeNbr; }
  virtual const String	&getTypeStr() const { return typeStr; }
  static DagNode	*creator() { return new NObstacle; }
  void			buildDagNodes();
  virtual void		compute(Plug &plugOut);
  void			computev2vCompute(Plug &plugOut);
  void			computev2vLoad(Plug &plugOut);
  void			computeRlogObstacle(Plug &plugOut);
  void			computeRlogOvortices(Plug &plugOut);
  void			computeRlog(Plug &plugOut);
  void			computeJac(Plug &plugOut);
  void			computeBBox(Plug &plugOut);
  void			computeGrid(Plug &plugOut);
  void			update(Plug &plugOut);
  void			glRender(Plug &plugOut);
};

// ************************************************************************* //
/// DESCRIPTION: Extends a data node. This is stored in the scene graph.
// ************************************************************************* //
class PlugMatMNd: public Plug
{
protected:
  MatMNd	*_m_data;
public:
  static int	typeNbr;
  static String	typeStr;
public:
  PlugMatMNd(DataBlock &db): Plug(db), _m_data(0) { newData(); }
  virtual ~PlugMatMNd() { freeData(); }
  static Plug	*creator(DataBlock &db) { return new PlugMatMNd(db); }
  virtual int	getTypeNbr() const { return typeNbr; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual void	freeData() { if (_m_data) delete _m_data; _m_data = 0; }
  virtual void	newData() { if (!_m_data) _m_data = new MatMNd; }
  virtual void	saveData(std::ostream &) {}
  virtual void	loadData(std::istream &) {}
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const MatMNd*>(v); }
  virtual void	**rawdata() { return (void**)(&_m_data); }
};

// ************************************************************************* //
class PlugOctreeVortex: public Plug
{
protected:
  Octree<std::vector<unsigned> >	*_m_data;
public:
  static int	typeNbr;
  static String	typeStr;
public:
  PlugOctreeVortex(DataBlock &db): Plug(db), _m_data(0) { newData(); }
  virtual ~PlugOctreeVortex() { freeData(); }
  static Plug	*creator(DataBlock &db) { return new PlugOctreeVortex(db); }
  virtual int	getTypeNbr() const { return typeNbr; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual void	freeData() { if (_m_data) delete _m_data; _m_data = 0; }
  virtual void	newData()
  { if (!_m_data) _m_data = new Octree<std::vector<unsigned> >; }
  virtual void	saveData(std::ostream &) {}
  virtual void	loadData(std::istream &) {}
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const Octree<std::vector<unsigned> >*>(v); }
  virtual void	**rawdata() { return (void**)(&_m_data); }
};

// ************************************************************************* //
class OVortex
{
public:
  Pnt3d		c;
  Vec3d		w; // n * strength
  double	stretch2;
  double	radius;
public:
  OVortex() {}
  OVortex(const Pnt3d &c_, const Vec3d &w_, const double stretch2_,
	  const double radius_):
    c(c_), w(w_), stretch2(stretch2_), radius(radius_) {}
  TwistLog	rlog(const Pnt3d &p) const
  {
    return TwistLog(w, w ^ Vec3d(c)) * rbfObstacle((p - c).norm2(),
						   sq(radius));
  }
  void	glRender() const
  {
    /*glBegin(GL_POINTS);
    p.glVertex();
    glEnd();//GL_POINTS*/
    if (sq(0.01) < w.norm2())
      {
	glBegin(GL_LINES);
	c.glVertex();
	(c + w.unit() * 0.1).glVertex();
	glEnd();//GL_LINES
      }
  }
};

// ************************************************************************* //
/// DESCRIPTION:
// ************************************************************************* //
class DataObstacleVortex
{
public:
  std::list<OVortex>	_m_ovortices;
public:
  DataObstacleVortex() {}
  void	glRender() const
  {
    std::list<OVortex>::const_iterator	oit;
  
    glPushAttrib(GL_LIGHTING_BIT);
    glDisable(GL_LIGHTING);
    for (oit = _m_ovortices.begin(); oit != _m_ovortices.end(); ++oit)
      oit->glRender();
    glPopAttrib();//GL_LIGHTING_BIT
  }
};

// ************************************************************************* //
/// TODO:
class PlugObstacleVortex: public Plug
{
protected:
  DataObstacleVortex	*_m_data;
public:
  static int	typeNbr;
  static String	typeStr;
public:
  PlugObstacleVortex(DataBlock &db): Plug(db), _m_data(0) { newData(); }
  virtual ~PlugObstacleVortex() { freeData(); }
  static Plug	*creator(DataBlock &db) { return new PlugObstacleVortex(db); }
  virtual int	getTypeNbr() const { return typeNbr; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual void	freeData() { if (_m_data) delete _m_data; _m_data = 0; }
  virtual void	newData()
  { if (!_m_data) _m_data = new DataObstacleVortex; }
  virtual void	saveData(std::ostream &) {}
  virtual void	loadData(std::istream &) {}
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const DataObstacleVortex*>(v); }
  virtual void	**rawdata() { return (void**)(&_m_data); }
};

// ************************************************************************* //
#endif//__NODEOBSTACLE_HH__
