// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:29 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include <GL/glu.h>
#include "datablock.h"
#include "dagnode.h"
#include "nodescene.h"
#include "nodeclock.h"
#include "nodecamera.h"
#include "plugcustom.h"
#include "gbb.h"

// ************************************************************************* //
int	NScene::typeNbr(43);
String	NScene::typeStr("Scene");
int	NWind::typeNbr(58);
String	NWind::typeStr("Wind");

// ************************************************************************* //
/// DESCRIPTION: The root of the scene graph.
///  renderMode	- 
// ************************************************************************* //
void	NScene::initialize()
{
  DataBlock	&db(getDataBlock());

  db.addInt(false, 0, MAXINT, "sel", false);
  db.addInt(0, 0, MAXINT, "renderIn", false);
  db.addSignal("renderOut", false);
  db.addBool(false, "renderAll");
  db.addInt(0, MININT, MAXINT, "f", false);
  db.addSignal("update", false);
  db.addDouble(0.0, 0.0, MAXDOUBLE, "shadingTime", false);

  db.affect("f", "update");
  db.affect("renderIn", "renderOut");
  setRoot(*this);
  buildDagNodes();
}

// ************************************************************************* //
void	NScene::buildDagNodes()
{
  DagNode *clock = Command::run("new -type Clock -path :");
  DagNode *vortexset = Command::run("new -type VortexSet -path :");
  DagNode *smokeset = Command::run("new -type SmokeParticles -path :");
  DagNode *smokeGPUset = Command::run("new -type SmokeGPUParticles -path :");
  DagNode *camera = Command::run("new -type Camera -path :");
  DagNode *gizmo = Command::run("new -type Gizmo -path :");

  Command::run("new -type Sky -path :");

  DagNode *frontLayer = Command::run("new -type ImageLayer -path :");//front
  DagNode *backLayer = Command::run("new -type ImageLayer -path :");//back

  backLayer->writeAsDouble("depth", 0.9999);
  frontLayer->writeAsDouble("depth", 0.0001);

  DagNode::connect(*clock, "f", *this, "f");
  DagNode::connect(*camera, "eye", *smokeset, "eye");
  DagNode::connect(*camera, "eye", *smokeGPUset, "eye");
  DagNode::connect(*camera, "model", *gizmo, "model");
  //DagNode::connect(*smokeset, "bbox", *vortexset, "bboxSmoke");
  DagNode::connect(*smokeset, "sp", *vortexset, "sp");
  DagNode::connect(*smokeGPUset, "sp", *vortexset, "spGPU");

  camera->writeAsBool("rendering", true);
}

// ************************************************************************* //
void	NScene::compute(Plug &hndOut)
{
  if (hndOut.shortName() == "renderOut")
    glRender(hndOut);
  else if (hndOut.shortName() == "update")
    update(hndOut);
}

// ************************************************************************* //
/// DESCRIPTION: Updates the scene, i.e. the pysical simulation.
///  Pulls the signal "update" from every node that requires it.
/// CONTEXT: Compute.
/// Also, applies the wind to the particles (maybe another node
///  should wrap that).
// ************************************************************************* //
void	NScene::update(Plug &hndOut)
{
  readAsInt("f");

  std::list<DagNode*>::iterator	nit;
  std::list<DagNode *>	nodes;
      
  getNodesByType("RingSource", nodes);//To launch new rings.
  getNodesByType("CurrentGenerator", nodes);//To launch new rings.
  getNodesByType("SmokeGenerator", nodes);// To spawn smoke parcels.
  getNodesByType("SmokeGPUGenerator", nodes);// To spawn smoke parcels.
  getNodesByType("VortexSet", nodes);// To animate existing rings.
  getNodesByType("SmokeParticles", nodes);// To spawn smoke parcels.
  getNodesByType("SmokeGPUParticles", nodes);// To spawn smoke parcels.
  getNodesByType("Noise", nodes);//
  getNodesByType("Obstacle", nodes);//... TODO: hierarchy of update pluged in f
  getNodesByType("CurveConstraint", nodes);
  getNodesByType("Water", nodes);//... 
  getNodesByType("Turbulence", nodes);//... 
  //getNodesByType("NSmokeParticles", nodes);// To animate particles.
  //TODO: get nodes at depth 1!
  //      propagate update with Transform
  //     OR ... when update attributes are created, put them all in a list
  //            held by scene?... eek... can't control the order like that...
  for (nit = nodes.begin(); nit != nodes.end(); ++nit)
    (*nit)->readAsSignal("update");
  hndOut.writeAsSignal();
}

// ************************************************************************* //
/// DESCRIPTION: Renders the scene.
///   Pulls the signal renderOut from every child node that requires it.
// ************************************************************************* //
void	NScene::glRender(Plug &plug)
{
  readAsSignal("renderIn");

  std::list<DagNode*>::iterator	nit;
  std::list<DagNode*>	nodes;
  DagNode	*smokeparticles(0);
  DagNode	*smokeGPUparticles(0);
  DagNode	*sky(0);
  DagNode	*frontLayer(0);
  DagNode	*backLayer(0);
  DagNode	*obstacle(0);
  DagNode	*gizmo(0);
  Plug	*data;

  getNodesAtDepth(1, nodes);
  for (nit = nodes.begin(); nit != nodes.end(); )
    if ((*nit)->getTypeStr() == "SmokeParticles")
      {
	smokeparticles = *nit;
	nodes.erase(nit++);
      }
    else if ((*nit)->getTypeStr() == "SmokeGPUParticles")
      {
	smokeGPUparticles = *nit;
	nodes.erase(nit++);
      }
    else if ((*nit)->getTypeStr() == "Sky")
      {
	sky = *nit;
	nodes.erase(nit++);
      }
    else if ((*nit)->getInstStr() == "ImageLayer0")
      {
	frontLayer = *nit;
	nodes.erase(nit++);
      }
    else if ((*nit)->getInstStr() == "ImageLayer1")
      {
	backLayer = *nit;
	nodes.erase(nit++);
      }
    else if ((*nit)->getTypeStr() == "Gizmo")
      {
	gizmo = *nit;
	nodes.erase(nit++);
      }
    else if ((*nit)->getTypeStr() == "Transform")
      {
	std::list<DagNode*>::iterator	njt;
	std::list<DagNode*>	nodesRec;
	bool	isFound(false);

	(*nit)->getNodesAtDepth(1, nodesRec);
	for (njt = nodesRec.begin(); njt != nodesRec.end(); ++njt)
	  if ((*njt)->getTypeStr() == "Obstacle")
	    {
	      obstacle = *nit;
	      nodes.erase(nit++);
	      isFound = true;
	      break;
	    }
	if (!isFound)
	  ++nit;
      }
    else
      ++nit;
  if (obstacle)
    nodes.push_back(obstacle);
  nodes.push_back(smokeparticles);
  nodes.push_back(smokeGPUparticles);
  if (gizmo)
    nodes.push_back(gizmo);
  // *********************************************************************** //
  nodes.push_front(backLayer);
  nodes.push_front(sky);
  nodes.push_back(frontLayer);
  for (nit = nodes.begin(); nit != nodes.end(); ++nit)
    if ((data = (*nit)->getPlug("renderOut")))
      data->readAsSignal();
  plug.writeAsSignal();
}

// ************************************************************************* //
void	NWind::initialize()
{
  NShape::initialize();

  DataBlock	&db(getDataBlock());

  db.addPnt3d(Pnt3d(0.0), "p", false); //fake, field is constant!
  db.addPnt3d(Pnt3d(0.0), "src");
  db.addPnt3d(Pnt3d(0.0, 0.0, 1.0), "dst");
  db.addDouble(0.1, 0.0, MAXDOUBLE, "radius");
  db.addTwistLog(TwistLog(0.0), "rlog");

  db.addMat4d(Mat4d::identity(), "matParent", false);//position

  db.affect("src", "renderOut");
  db.affect("dst", "renderOut");
  db.affect("src", "rlog");
  db.affect("dst", "rlog");
  buildDagNodes();
}

// ************************************************************************* //
void	NWind::buildDagNodes()
{
  //attach wind parent
  DagNode	*oldParent(_m_parent);
  DagNode	*tr = Command::run(String("new -type Transform -path :"));
  //DagNode	*vortexset = DagNode::getNodeByPath(":VortexSet0");

  if (oldParent)
    DagNode::parent(*oldParent, *tr);

  //DagNode::parent(*vortexset, *tr);
  DagNode::parent(*tr, *this);
  DagNode::connect(*oldParent, "renderIn", *tr, "renderIn");
  DagNode::connect(*tr, "renderIn", *this, "renderIn");
  DagNode::connect(*tr, "matChild", *this, "matParent");
}

// ************************************************************************* //
void	NWind::glRender(Plug &plug)
{
  if (readAsInt("renderIn") != Gbb::Rendering)
    {
      Plug	*hndSrc(getPlug("src"));
      Plug	*hndDst(getPlug("dst"));
      const Pnt3d	&src(hndSrc->readAsPnt3d());
      const Pnt3d	&dst(hndDst->readAsPnt3d());
      const double	radius(readAsDouble("radius"));

      static GLUquadric	*quad = gluNewQuadric();
      const Mat4d	&pos(readAsMat4d("matParent"));
      const int	sel(readAsInt("sel"));
      const Vec3d	z((dst - src).unit());
      Vec3d		x, y, color;

      pushName();
      glPushMatrix();
      readAsMat4d("matParent").glMultMatrix();

      if (fabs(z.x) < fabs(z.y) && fabs(z.x) < fabs(z.z))
	y = Vec3d(1.0, 0.0, 0.0) - z * z.x;
      else if (fabs(z.y) < fabs(z.z))
	y = Vec3d(0.0, 1.0, 0.0) - z * z.y;
      else
	y = Vec3d(0.0, 0.0, 1.0) - z * z.z;
      y.eqUnit();
      x = y ^ z;
      if (sel == 1)
	color = Vec3d(1.0, 0.0, 0.5);
      else if (sel)
	color = Vec3d(1.0, 0.0, 0.0);
      else
	color = Vec3d(0.0, 0.0, 1.0);
      color.glColor();

      hndSrc->pushName();
      glPushAttrib(GL_CURRENT_BIT);
      if (hndSrc->isSelected())
	glColor3d(1.0, 1.0, 0.0);
      else
	color.glColor();
      glPushMatrix();
      Mat4d(x, y, z, src).glMultMatrix();
      glTranslated(0.0, 0.0, -radius);
      glPushMatrix();
      glRotated(180.0, 1.0, 0.0, 0.0);
      gluDisk(quad, 0, radius, 16, 1);
      glPopMatrix();
      glPushMatrix();
      glTranslated(0.0, 0.0, radius);
      gluDisk(quad, 0, radius, 16, 1);
      glPopMatrix();
      gluCylinder(quad, radius, radius, radius, 16, 1);
      glPopMatrix();
      glPopAttrib();//GL_CURRENT_BIT
      hndSrc->popName();

      hndDst->pushName();
      glPushAttrib(GL_CURRENT_BIT);
      if (hndDst->isSelected())
	glColor3d(1.0, 1.0, 0.0);
      else
	color.glColor();
      glPushMatrix();
      Mat4d(x, y, z, dst).glMultMatrix();
      gluCylinder(quad, radius, 0.0, 2.5 * radius, 16, 1);
      glPushMatrix();
      glRotated(180.0, 1.0, 0.0, 0.0);
      gluDisk(quad, 0, radius, 16, 1);
      glPopMatrix();
      glPopMatrix();
      glPopAttrib();//GL_CURRENT_BIT
      hndDst->popName();

      glBegin(GL_LINES);
      src.glVertex();
      dst.glVertex();
      glEnd();//GL_LINES

      glPopMatrix();
      popName();
    }
  plug.writeAsSignal();
}

// ************************************************************************* //
void	NWind::compute(Plug &plug)
{
  if (plug.shortName() == "rlog")
    {
      const Vec3d	rlog(readAsPnt3d("dst") - readAsPnt3d("src"));

      plug.writeAsTwistLog(TwistLog(Vec3d(0.0), rlog));
    }
  else if (plug.shortName() == "renderOut")
    glRender(plug);
}
