// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:29 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include <GL/glu.h>
//#include <GL/glxext.h>
#include <QImage>
#include "datablock.h"
#include "dagnode.h"
#include "nodesky.h"
#include "nodeclock.h"
#include "nodecamera.h"
#include "plugcustom.h"
#include "command.h"
#include "gbb.h"
//#include "glext.h"
#include <fstream>
#include <string.h>

// ************************************************************************* //
int	NSky::typeNbr(54);
String	NSky::typeStr("Sky");

// ************************************************************************* //
/// DESCRIPTION:
/// matte	- background image.
/// up		- up vector, used for color and camera differential scrolling.
/// CONTEXT: Assumes one camera exists, and that it's connected.
// ************************************************************************* //
void	NSky::initialize()
{
  DataBlock	&db(getDataBlock());
  
  db.addBool(false, "noSky");
  db.addInt(0, 0, MAXINT, "renderIn", false);
  db.addSignal("renderOut", false);
  db.addVec3d(Vec3d(0.0, 0.0, 1.0), "up");

  db.addRgb3d(Rgb3d(0.78, 0.87, 0.92), "colorSkyUp");
  db.addRgb3d(Rgb3d(0.56, 0.74, 0.84), "colorHorizon");
  db.addRgb3d(Rgb3d(0.78, 0.87, 0.92), "colorSkyDown");
  db.addDouble(2.0, 0.0, MAXDOUBLE, "sizeHorizon");
  db.addInt(8, 8, MAXINT, "resolution");
  db.addCustom(*new PlugSky(db), "skyLores", false);
  db.addCustom(*new PlugSky(db), "skyHires", false);
  db.addInt(0, MININT, MAXINT, "listLores", false);
  db.addInt(0, MININT, MAXINT, "listHires", false);
  db.addString("", "matteBg");
  db.addInt(0, 0, MAXINT, "texnameLores", false, false);
  db.addInt(0, 0, MAXINT, "texnameHires", false, false);
  db.addBool(false, "scroll");
  db.addDouble(0.5, 0.0, MAXDOUBLE, "scalex");
  db.addDouble(0.5, 0.0, MAXDOUBLE, "scaley");
  db.addSignal("camera", false);

  db.affect("renderIn", "renderOut");

  db.affect("camera", "texnameLores");
  db.affect("matteBg", "texnameLores");
  db.affect("texnameLores", "listLores");

  db.affect("camera", "texnameHires");
  db.affect("matteBg", "texnameHires");
  db.affect("texnameHires", "listHires");

  //db.affect("up", "skyLores");
  //db.affect("matteBg", "skyLores");
  //db.affect("camera", "skyLores");

  //db.affect("up", "skyHires");
  //db.affect("matteBg", "skyHires");
  //db.affect("camera", "skyHires");

  buildDagNodes();
}

// ************************************************************************* //
void	NSky::buildDagNodes()
{
  std::list<DagNode*>	cameras;

  DagNode::root().DagNode::getNodesByType("Camera", cameras);

  DagNode	*camera = cameras.front();

  //DagNode::connect(*camera, "eye", *this, "eye");
  //DagNode::connect(*camera, "frustum", *this, "frustum");
  DagNode::connect(*camera, "signal", *this, "camera");
  DagNode::connect(*_m_parent, "renderIn", *this, "renderIn");
}

// ************************************************************************* //
const Pnt3d	&NSky::getEye()
{
  return getPlug("camera")->getInputNode()->readAsPnt3d("eye");
}

// ************************************************************************* //
const Frustum	&NSky::getFrustum()
{
  const void *ft = getPlug("camera")->getInputNode()->readAsCustom("frustum");

  return *static_cast<const Frustum*>(ft);
}

// ************************************************************************* //
Pnt3d	NSky::getAimAt()
{
  DagNode *camera = getPlug("camera")->getInputNode();
  DagNode *constraintAim = camera->getPlug("constraintAim")->getInputNode();

  if (!constraintAim)
    {
      return Pnt3d(0.0);
    }
  else
    return constraintAim->readAsPnt3d("target");
}

// ************************************************************************* //
/*
static void (*my_glGenProgramsARB)(GLuint, GLuint *) = NULL;
static void (*my_glBindProgramARB)(GLuint, GLuint) = NULL;
static void (*my_glProgramStringARB)(GLuint, GLuint, GLint,
				     const GLbyte *) = NULL;

// ************************************************************************* //
/// DESCRIPTION: this function returns a string containing
/// the contents of the specified shader file
// ************************************************************************* //
static char *load_program_string(const char *filename)
{
  static char	program_string[16384];
  FILE		*fp;
  unsigned int	len;

  fp = fopen(filename, "r");
  if(!fp)
    return NULL;

  len = fread(program_string, 1, 16384, fp);
  program_string[len] = '\0';
  fclose(fp);

  return program_string;
}

// ************************************************************************* //
/// DESCRIPTION: loads the shader from file and returns shader number 
// ************************************************************************* //
static unsigned int	loadShader(GLuint	type,
				   const char	*filename)
{
  unsigned int	shader_num;
  char		*str;

  str = load_program_string(filename);
  glEnable(type);

  my_glGenProgramsARB = 
    (void (*)(GLuint, GLuint *))getProcAddress("glGenProgramsARB");
  my_glBindProgramARB = 
    (void (*)(GLuint, GLuint))getProcAddress("glBindProgramARB");
  my_glProgramStringARB = (void (*)(GLuint, GLuint, GLint, const GLbyte *))
    getProcAddress("glProgramStringARB");

  my_glGenProgramsARB(1, &shader_num);
  my_glBindProgramARB(type, shader_num);
  my_glProgramStringARB(type, GL_PROGRAM_FORMAT_ASCII_ARB, 
			strlen(str), (const GLbyte *)str);
  glDisable(type);
  return shader_num;
}
*/

// ************************************************************************* //
void	NSky::glRender(Plug &plugOut)
{
  const int renderIn = readAsInt("renderIn");

  if (!readAsBool("noSky"))
    {
      if (renderIn == Gbb::Modeling)
	readAsSignal("skyLores");
      else if (renderIn == Gbb::Rendering)
	readAsSignal("skyHires");

      glPushAttrib(GL_LIGHTING_BIT | GL_TRANSFORM_BIT);
      glDisable(GL_LIGHTING);
      glMatrixMode(GL_PROJECTION);
      glPushMatrix();
      glLoadIdentity();
      glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      glLoadIdentity();
      glDepthMask(GL_FALSE);

      if (readAsInt("renderIn") == Gbb::Modeling)
	glCallList(readAsInt("listLores"));
      else if (readAsInt("renderIn") == Gbb::Rendering)
	glCallList(readAsInt("listHires"));
      //TODO: dont write depth if alpha is transparent.

      glDepthMask(GL_TRUE);
      glPopMatrix();
      glMatrixMode(GL_PROJECTION);
      glPopMatrix();
      glPopAttrib();// GL_LIGHTING_BIT | GL_TRANSFORM_BIT
    }
  /*
  else
    {
      glPushAttrib(GL_LIGHTING_BIT | GL_TRANSFORM_BIT);
      glDisable(GL_LIGHTING);
      glMatrixMode(GL_PROJECTION);
      glPushMatrix();
      glLoadIdentity();
      glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      glLoadIdentity();
      glDepthMask(GL_FALSE);

      //Draw large transparent quad at back for depth???
      glBegin(GL_QUADS);
      glColor4d(1, 0, 0, 1);
      glVertex3d(1, 1, -0.99);
      glVertex3d(-1, 1, -0.99);
      glVertex3d(-1, -1, -0.99);
      glVertex3d(1, -1, -0.99);
      glEnd();//GL_QUADS

      glEnable(GL_BLEND);
      //glBlendFuncSeparate(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA,
      //		  GL_ONE, GL_ONE);
      //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

      glBegin(GL_QUADS);
      glColor4d(1, 1, 1, 0.5);
      glVertex3d(0.5, 0.5, 0.5);
      glVertex3d(-0.5, 0.5, 0.5);
      glVertex3d(-0.5, -0.5, 0.5);
      glVertex3d(0.5, -0.5, 0.5);
      glColor4d(1, 1, 1, 0.5);
      glVertex3d(0.25, 0.25, 0.0);
      glVertex3d(-0.25, 0.25, 0.0);
      glVertex3d(-0.25, -0.25, 0.0);
      glVertex3d(0.25, -0.25, 0.0);
      glEnd();//GL_QUADS

      //glDisable(GL_FRAGMENT_PROGRAM_ARB);

      glDepthMask(GL_TRUE);
      glPopMatrix();
      glMatrixMode(GL_PROJECTION);
      glPopMatrix();
      glPopAttrib();// GL_LIGHTING_BIT | GL_TRANSFORM_BIT
    }
  */
  plugOut.writeAsSignal();
}

// ************************************************************************* //
void	NSky::computeListHires(Plug &plugOut)
{
  const Frustum	&frustum = getFrustum();
  const Vec3d	up = readAsVec3d("up").unit();
  DTable<PlugSky::DataSky> &sp =
    *static_cast<DTable<PlugSky::DataSky>*>(readAsCustom("skyHires"));
  Pnt3d			farCorners[4] = { 0.0, 0.0, 0.0, 0.0 };
  const double	sizeHorizon(readAsDouble("sizeHorizon"));
  const int		resolution(readAsInt("resolution"));

  frustum.farCorners(farCorners);

  const Vec3d	x(farCorners[2] - farCorners[0]);
  const Vec3d	y(farCorners[1] - farCorners[0]);

  sp.resize(resolution, resolution);
  for (int i(0); i < resolution; ++i)
    for (int j(0); j < resolution; ++j)
      {
	const double	di(i / (resolution - 1.0));
	const double	dj(j / (resolution - 1.0));
	const Vec3d	vFar = Vec3d(farCorners[0] + x * di + y * dj).unit();
	const Rgb3d	&col0(readAsRgb3d("colorSkyUp"));
	const Rgb3d	&col1(readAsRgb3d("colorHorizon"));
	const Rgb3d	&col2(readAsRgb3d("colorSkyDown"));
	const double	u((1.0 - pow(fabs(vFar * up), sizeHorizon)));

	sp(i, j).pFar = Pnt3d(2.0 * di - 1.0, 2.0 * dj - 1.0, -0.9999);
	if (0.0 < vFar * up)
	  sp(i, j).p1 = col0 + (col1 - col0) * u;
	else
	  sp(i, j).p1 = col2 + (col1 - col2) * u;
      }

  GLuint	dlist, texname(0);

  if ((dlist = plugOut.readAsInt()))
    glDeleteLists(dlist, 1);
  dlist = glGenLists(1);
  texname = readAsInt("texnameHires");
  glNewList(dlist, GL_COMPILE);
  if (texname)
    {
      const Pnt3d	aimAt = getAimAt();
      const Vec3d	up = readAsVec3d("up");
      const Pnt3d	eye = getEye();
      const Vec3d	e = (eye - aimAt).unit();
      const Vec3d	scrollVec = (e - up * (e * up)).unit();
      const bool	isScroll(readAsBool("scroll"));
      const double	ax = isScroll?
	atan2(scrollVec.y, scrollVec.x) / (2.0 * M_PI): 0.0;
      const double	ay = isScroll? (1.0 - acos(e * up) / M_PI): 0.5;
      const double	sx = isScroll? readAsDouble("scalex"): 1.0;
      const double	sy = isScroll? readAsDouble("scaley"): 1.0;

      glEnable(GL_TEXTURE_2D);
      glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
      glBindTexture(GL_TEXTURE_2D, texname);
      glColor3d(1.0, 1.0, 1.0);
      for (int i(0); i < resolution - 1; ++i)
	{
	  glBegin(GL_QUAD_STRIP);
	  for (int j(0); j < resolution; ++j)
	    {
	      glTexCoord2d(-ax + ((i) / double(resolution - 1.0)) * sx,
			   0.5 + (ay - j / double(resolution - 1.0)) * sy);
	      sp(i, j).pFar.glVertex();
	      glTexCoord2d(-ax + (i + 1.0) / double(resolution - 1.0) * sx,
			   0.5 + (ay - j / double(resolution - 1.0)) * sy);
	      sp(i + 1, j).pFar.glVertex();
	    }
	  glEnd();//GL_QUAD_STRIP
	}
      glDisable(GL_TEXTURE_2D);
    }
  else
    for (int i(0); i < resolution - 1; ++i)
      {
	glBegin(GL_QUAD_STRIP);
	for (int j(0); j < resolution; ++j)
	  {
	    sp(i, j).p1.glColor();
	    sp(i, j).pFar.glVertex();
	    sp(i + 1, j).p1.glColor();
	    sp(i + 1, j).pFar.glVertex();
	  }
	glEnd();//GL_QUAD_STRIP
      }
  glEndList();
  plugOut.writeAsInt(dlist);
}

// ************************************************************************* //
void	NSky::computeListLores(Plug &plugOut)
{
  const Frustum	&frustum = getFrustum();
  const Vec3d	up = readAsVec3d("up").unit();
  DTable<PlugSky::DataSky> &sp =
    *static_cast<DTable<PlugSky::DataSky>*>(readAsCustom("skyLores"));
  Pnt3d			farCorners[4] = { 0.0, 0.0, 0.0, 0.0 };
  const double	sizeHorizon(readAsDouble("sizeHorizon"));
  const int		resolution(readAsInt("resolution"));

  frustum.farCorners(farCorners);

  const Vec3d	x(farCorners[2] - farCorners[0]);
  const Vec3d	y(farCorners[1] - farCorners[0]);

  sp.resize(resolution, resolution);
  for (int i(0); i < resolution; ++i)
    for (int j(0); j < resolution; ++j)
      {
	const double	di(i / (resolution - 1.0));
	const double	dj(j / (resolution - 1.0));
	const Vec3d	vFar = Vec3d(farCorners[0] + x * di + y * dj).unit();
	const Rgb3d	&col0(readAsRgb3d("colorSkyUp"));
	const Rgb3d	&col1(readAsRgb3d("colorHorizon"));
	const Rgb3d	&col2(readAsRgb3d("colorSkyDown"));
	const double	u((1.0 - pow(fabs(vFar * up), sizeHorizon)));

	sp(i, j).pFar = Pnt3d(2.0 * di - 1.0, 2.0 * dj - 1.0, -0.9999);
	if (0.0 < vFar * up)
	  sp(i, j).p1 = col0 + (col1 - col0) * u;
	else
	  sp(i, j).p1 = col2 + (col1 - col2) * u;
      }

  GLuint	dlist, texname(0);

  if ((dlist = plugOut.readAsInt()))
    glDeleteLists(dlist, 1);
  dlist = glGenLists(1);
  texname = readAsInt("texnameLores");
  glNewList(dlist, GL_COMPILE);
  if (texname)
    {
      const Pnt3d	aimAt = getAimAt();
      const Vec3d	up = readAsVec3d("up");
      const Pnt3d	eye = getEye();
      const Vec3d	e = (eye - aimAt).unit();
      const Vec3d	scrollVec = (e - up * (e * up)).unit();
      const bool	isScroll(readAsBool("scroll"));
      const double	ax = isScroll?
	atan2(scrollVec.y, scrollVec.x) / (2.0 * M_PI): 0.0;
      const double	ay = isScroll? (1.0 - acos(e * up) / M_PI): 0.5;
      const double	sx = isScroll? readAsDouble("scalex"): 1.0;
      const double	sy = isScroll? readAsDouble("scaley"): 1.0;

      glEnable(GL_TEXTURE_2D);
      glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
      glBindTexture(GL_TEXTURE_2D, texname);
      glColor3d(1.0, 1.0, 1.0);
      for (int i(0); i < resolution - 1; ++i)
	{
	  glBegin(GL_QUAD_STRIP);
	  for (int j(0); j < resolution; ++j)
	    {
	      glTexCoord2d(-ax + ((i) / double(resolution - 1.0)) * sx,
			   0.5 + (ay - j / double(resolution - 1.0)) * sy);
	      sp(i, j).pFar.glVertex();
	      glTexCoord2d(-ax + (i + 1.0) / double(resolution - 1.0) * sx,
			   0.5 + (ay - j / double(resolution - 1.0)) * sy);
	      sp(i + 1, j).pFar.glVertex();
	    }
	  glEnd();//GL_QUAD_STRIP
	}
      glDisable(GL_TEXTURE_2D);
    }
  else
    for (int i(0); i < resolution - 1; ++i)
      {
	glBegin(GL_QUAD_STRIP);
	for (int j(0); j < resolution; ++j)
	  {
	    sp(i, j).p1.glColor();
	    sp(i, j).pFar.glVertex();
	    sp(i + 1, j).p1.glColor();
	    sp(i + 1, j).pFar.glVertex();
	  }
	glEnd();//GL_QUAD_STRIP
      }
  glEndList();
  plugOut.writeAsInt(dlist);
}

// ************************************************************************* //
void	NSky::compute(Plug &plugOut)
{
  if (plugOut.shortName() == "listLores")
    computeListLores(plugOut);
  else if (plugOut.shortName() == "listHires")
    computeListHires(plugOut);
  else if (plugOut.shortName() == "texnameHires")
    {
      QImage	qimage(readAsString("matteBg").data());
      GLuint	texname = readAsInt("texnameHires");
      GLint		paramWrap = readAsBool("scroll")? GL_REPEAT: GL_CLAMP;

      if (texname)
	glDeleteTextures(1, &texname);
      if (qimage.isNull())
	plugOut.writeAsInt(0);
      else
	{
	  glGenTextures(1, &texname);
	  glBindTexture(GL_TEXTURE_2D, texname);
	  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, paramWrap);
	  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, paramWrap);
	  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 
		       qimage.width(), qimage.height(), 
		       0, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, qimage.bits());
	  plugOut.writeAsInt(texname);
	}
    }
  else if (plugOut.shortName() == "texnameLores")
    {
      QImage	qimage(readAsString("matteBg").data());
      GLuint	texname = readAsInt("texnameLores");
      GLint		paramWrap = readAsBool("scroll")? GL_REPEAT: GL_CLAMP;

      if (texname)
	glDeleteTextures(1, &texname);
      if (qimage.isNull())
	plugOut.writeAsInt(0);
      else
	{
	  qimage = qimage.scaled(256, 256, Qt::IgnoreAspectRatio,
				 Qt::SmoothTransformation);
	  glGenTextures(1, &texname);
	  glBindTexture(GL_TEXTURE_2D, texname);
	  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, paramWrap);
	  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, paramWrap);
	  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 
		       qimage.width(), qimage.height(), 
		       0, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, qimage.bits());
	  plugOut.writeAsInt(texname);
	}
    }
  else if (plugOut.shortName() == "renderOut")
    glRender(plugOut);
}
