// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:30 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include "nodetransform.h"
#include "plug.h"
#include "datablock.h"

// ************************************************************************* //
int	NTransform::typeNbr(52);
String	NTransform::typeStr("Transform");

// ************************************************************************* //
/// DESCRIPTION: Initializes a transform node.
// ************************************************************************* //
void	NTransform::initialize()
{
  NShape::initialize();

  DataBlock	&db(getDataBlock());

  db.addSignal("signal", false);

  db.addDouble(0.0, MINDOUBLE, MAXDOUBLE, "tx");
  db.addDouble(0.0, MINDOUBLE, MAXDOUBLE, "ty");
  db.addDouble(0.0, MINDOUBLE, MAXDOUBLE, "tz");
  db.addPnt3d(0.0, "t", false);

  db.addDouble(1.0, MINDOUBLE, MAXDOUBLE, "sx");
  db.addDouble(1.0, MINDOUBLE, MAXDOUBLE, "sy");
  db.addDouble(1.0, MINDOUBLE, MAXDOUBLE, "sz");

  db.addDouble(0.0, MINDOUBLE, MAXDOUBLE, "shx");
  db.addDouble(0.0, MINDOUBLE, MAXDOUBLE, "shy");
  db.addDouble(0.0, MINDOUBLE, MAXDOUBLE, "shz");

  db.addDouble(0.0, MINDOUBLE, MAXDOUBLE, "spx");
  db.addDouble(0.0, MINDOUBLE, MAXDOUBLE, "spy");
  db.addDouble(0.0, MINDOUBLE, MAXDOUBLE, "spz");

  db.addDouble(0.0, MINDOUBLE, MAXDOUBLE, "rpx");
  db.addDouble(0.0, MINDOUBLE, MAXDOUBLE, "rpy");
  db.addDouble(0.0, MINDOUBLE, MAXDOUBLE, "rpz");

  db.addDouble(0.0, MINDOUBLE, MAXDOUBLE, "rox");
  db.addDouble(0.0, MINDOUBLE, MAXDOUBLE, "roy");
  db.addDouble(0.0, MINDOUBLE, MAXDOUBLE, "roz");

  db.addDouble(0.0, MINDOUBLE, MAXDOUBLE, "rx");
  db.addDouble(0.0, MINDOUBLE, MAXDOUBLE, "ry");
  db.addDouble(0.0, MINDOUBLE, MAXDOUBLE, "rz");

  db.addMat4d(Mat4d::identity(), "translate", false);
  db.addMat4d(Mat4d::identity(), "rotate", false);
  db.addMat4d(Mat4d::identity(), "scale", false);

  db.addMat4d(Mat4d::identity(), "matParent", false);
  db.addMat4d(Mat4d::identity(), "mat", false);
  db.addMat4d(Mat4d::identity(), "matChild", false);

  db.affect("mat", "renderOut");

  db.affect("matParent", "matChild");
  db.affect("mat", "matChild");

  db.affect("tx", "translate");
  db.affect("ty", "translate");
  db.affect("tz", "translate");

  db.affect("tx", "t");
  db.affect("ty", "t");
  db.affect("tz", "t");

  db.affect("spx", "scale");
  db.affect("spy", "scale");
  db.affect("spz", "scale");
  db.affect("sx", "scale");
  db.affect("sy", "scale");
  db.affect("sz", "scale");

  db.affect("rpx", "rotate");// center of rotation
  db.affect("rpy", "rotate");// center of rotation
  db.affect("rpz", "rotate");// center of rotation
  db.affect("rox", "rotate");// angle
  db.affect("roy", "rotate");
  db.affect("roz", "rotate");
  db.affect("rx", "rotate");// angle
  db.affect("ry", "rotate");
  db.affect("rz", "rotate");

  db.affect("translate", "mat");
  db.affect("rotate", "mat");
  db.affect("scale", "mat");
}

// ************************************************************************* //
/// DESCRIPTION: Have to render this for selection in DAG.
// ************************************************************************* //
void	NTransform::glRender(Plug &hndOut)
{
  std::list<DagNode*>::iterator	nit;

  readAsSignal("renderIn");
  hndOut.writeAsSignal();
  pushName();
  for (nit = _m_children.begin(); nit != _m_children.end(); ++nit)
    if ((*nit)->getPlug("renderOut"))
      (*nit)->readAsSignal("renderOut");
  popName();
}

// ************************************************************************* //
void	NTransform::scale(Plug &plugOut)
{
  const double	spxold = getAsDouble("spx");
  const double	spyold = getAsDouble("spy");
  const double	spzold = getAsDouble("spz");
  const double	spx = readAsDouble("spx");
  const double	spy = readAsDouble("spy");
  const double	spz = readAsDouble("spz");
  const double	stx = spxold - spx;
  const double	sty = spyold - spy;
  const double	stz = spzold - spz;
  const double	sx = readAsDouble("sx");
  const double	sy = readAsDouble("sy");
  const double	sz = readAsDouble("sz");
  const double	shx = readAsDouble("shx");
  const double	shy = readAsDouble("shy");
  const double	shz = readAsDouble("shz");
  const Mat4d	st(1.0, 0.0, 0.0, stx,
		   0.0, 1.0, 0.0, sty,
		   0.0, 0.0, 1.0, stz,
		   0.0, 0.0, 0.0, 1.0);
  const Mat4d	sp(1.0, 0.0, 0.0, spx,
		   0.0, 1.0, 0.0, spy,
		   0.0, 0.0, 1.0, spz,
		   0.0, 0.0, 0.0, 1.0);
  const Mat4d	s(sx, 0.0, 0.0, 0.0,
		  0.0, sy, 0.0, 0.0,
		  0.0, 0.0, sz, 0.0,
		  0.0, 0.0, 0.0, 1.0);
  const Mat4d	sh(1.0, shz, shy, 0.0,
		   0.0, 1.0, shx, 0.0,
		   0.0, 0.0, 1.0, 0.0,
		   0.0, 0.0, 0.0, 1.0);
  const Mat4d	spi(1.0, 0.0, 0.0, -spx,
		    0.0, 1.0, 0.0, -spy,
		    0.0, 0.0, 1.0, -spz,
		    0.0, 0.0, 0.0, 1.0);

  plugOut.writeAsMat4d(st * sp * s * spi);
}

// ************************************************************************* //
void	NTransform::rotate(Plug &plugOut)
{
  const double	roxv = DEG2RAD(readAsDouble("rox"));
  const double	royv = DEG2RAD(readAsDouble("roy"));
  const double	rozv = DEG2RAD(readAsDouble("roz"));
  const double	cosrox(cos(roxv)), sinrox(sin(roxv));
  const Mat4d	rox(1.0, 0.0, 0.0, 0.0,
		    0.0, cosrox, sinrox, 0.0,
		    0.0, -sinrox, cosrox, 0.0,
		    0.0, 0.0, 0.0, 1.0);
  const double	cosroy(cos(royv)), sinroy(sin(royv));
  const Mat4d	roy(cosroy, 0.0, -sinroy, 0.0,
		    0.0, 1.0, 0.0, 0.0,
		    sinroy, 0.0, cosroy, 0.0,
		    0.0, 0.0, 0.0, 1.0);
  const double	cosroz(cos(rozv)), sinroz(sin(rozv));
  const Mat4d	roz(cosroz, sinroz, 0.0, 0.0,
		    -sinroz, cosroz, 0.0, 0.0,
		    0.0, 0.0, 1.0, 0.0,
		    0.0, 0.0, 0.0, 1.0);
  const Mat4d	ro(roz * roy * rox);
  const double	rxv = DEG2RAD(readAsDouble("rx"));
  const double	ryv = DEG2RAD(readAsDouble("ry"));
  const double	rzv = DEG2RAD(readAsDouble("rz"));
  const double	cosrx(cos(rxv)), sinrx(sin(rxv));
  const Mat4d	rx(1.0, 0.0, 0.0, 0.0,
		   0.0, cosrx, sinrx, 0.0,
		   0.0, -sinrx, cosrx, 0.0,
		   0.0, 0.0, 0.0, 1.0);
  const double	cosry(cos(ryv)), sinry(sin(ryv));
  const Mat4d	ry(cosry, 0.0, -sinry, 0.0,
		   0.0, 1.0, 0.0, 0.0,
		   sinry, 0.0, cosry, 0.0,
		   0.0, 0.0, 0.0, 1.0);
  const double	cosrz(cos(rzv)), sinrz(sin(rzv));
  const Mat4d	rz(cosrz, sinrz, 0.0, 0.0,
		   -sinrz, cosrz, 0.0, 0.0,
		   0.0, 0.0, 1.0, 0.0,
		   0.0, 0.0, 0.0, 1.0);
  const Mat4d	r(rz * ry * rx);
  const double	rpxold = getAsDouble("rpx");
  const double	rpyold = getAsDouble("rpy");
  const double	rpzold = getAsDouble("rpz");
  const double	rpx = readAsDouble("rpx");
  const double	rpy = readAsDouble("rpy");
  const double	rpz = readAsDouble("rpz");
  const double	rtx = rpxold - rpx;
  const double	rty = rpyold - rpy;
  const double	rtz = rpzold - rpz;

  const Mat4d	rp(1.0, 0.0, 0.0, rpx,
		   0.0, 1.0, 0.0, rpy,
		   0.0, 0.0, 1.0, rpz,
		   0.0, 0.0, 0.0, 1.0);
  const Mat4d	rpi(1.0, 0.0, 0.0, -rpx,
		    0.0, 1.0, 0.0, -rpy,
		    0.0, 0.0, 1.0, -rpz,
		    0.0, 0.0, 0.0, 1.0);
  const Mat4d	rt(1.0, 0.0, 0.0, rtx,
		   0.0, 1.0, 0.0, rty,
		   0.0, 0.0, 1.0, rtz,
		   0.0, 0.0, 0.0, 1.0);

  plugOut.writeAsMat4d(rt * rp * r * ro * rpi);
}

// ************************************************************************* //
void	NTransform::translate(Plug &plugOut)
{
  const double	tx = readAsDouble("tx");
  const double	ty = readAsDouble("ty");
  const double	tz = readAsDouble("tz");
  const Mat4d	t(1.0, 0.0, 0.0, tx,
		  0.0, 1.0, 0.0, ty,
		  0.0, 0.0, 1.0, tz,
		  0.0, 0.0, 0.0, 1.0);

  plugOut.writeAsMat4d(t);
}

// ************************************************************************* //
void	NTransform::compute(Plug &plugOut)
{
  if (plugOut.shortName() == "renderOut")
    glRender(plugOut);
  else if (plugOut.shortName() == "scale")
    scale(plugOut);
  else if (plugOut.shortName() == "rotate")
    rotate(plugOut);
  else if (plugOut.shortName() == "translate")
    translate(plugOut);
  else if (plugOut.shortName() == "t")
    plugOut.writeAsPnt3d(Pnt3d(readAsDouble("tx"),
			       readAsDouble("ty"),
			       readAsDouble("tz")));
  else if (plugOut.shortName() == "mat")
    plugOut.writeAsMat4d(readAsMat4d("translate") *
			 readAsMat4d("rotate") *
			 readAsMat4d("scale"));
  else if (plugOut.shortName() == "matChild")
    plugOut.writeAsMat4d(readAsMat4d("matParent") * readAsMat4d("mat"));
  NShape::compute(plugOut);
}
