// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:30 $
/// $Revision: 1.1 $
// ************************************************************************* //
#ifndef __NODERINGSET_HH__
#define __NODERINGSET_HH__

// ************************************************************************* //
#include "plug.h"
#include "dagnode.h"

// ************************************************************************* //
#ifndef EPSILON_JAC
#define EPSILON_JAC	1.0e-8
#endif//EPSILON_JAC

// ************************************************************************* //
/// DESCRIPTION: Collection of all the wind primitives. 
///   Also spawns the ring at the right time
// ************************************************************************* //
class NVortexSet: public DagNode
{
public:
  static int	typeNbr;
  static String	typeStr;
public:
  NVortexSet(DagNode *parent = 0): DagNode(parent) {}
  virtual void		initialize();
  virtual int		getTypeNbr() const { return typeNbr; }
  virtual const String	&getTypeStr() const { return typeStr; }
  static DagNode	*creator() { return new NVortexSet(); }
  void			buildDagNodes();
  virtual void	compute(Plug &plugOut);
  void		computeOctree(Plug &plugOut);
  void		computeBBoxSmoke(Plug &plugOut);
  void		computeBBoxSmokeGPU(Plug &plugOut);
  void		computeBBoxRings(Plug &plugOut);
  void		computeBBoxNoise(Plug &plugOut);
  void		computeBBoxOctree(Plug &plugOut);
  void		computeRlog(Plug &plugOut);
  void		computeRlogOctree(Plug &plugOut);
  void		computeRlogTotal(Plug &plugOut);
  void		computeRlogNoisy(Plug &plugOut);
  void		computeJac(Plug &plugOut, Plug &plugp, Plug &plugRlog);
  void		glRender(Plug &plugOut);
  void		update(Plug &plugOut);
};

// ************************************************************************* //
/// DESCRIPTION: Extends a data node. This is stored in the scene graph.
// ************************************************************************* //
class PlugNodeList: public Plug
{
protected:
  std::list<DagNode *>	*_m_data;
public:
  static int	typeNbr;
  static String	typeStr;
public:
  PlugNodeList(DataBlock &db): Plug(db), _m_data(0) { newData(); }
  virtual ~PlugNodeList() { freeData(); }
  static Plug	*creator(DataBlock &db) { return new PlugNodeList(db); }
  virtual int	getTypeNbr() const { return typeNbr; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual void	freeData() { if (_m_data) delete _m_data; _m_data = 0; }
  virtual void	newData()
  { if (!_m_data) _m_data = new std::list<DagNode *>; }
  virtual void	saveData(std::ostream &) {}
  virtual void	loadData(std::istream &) {}
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const std::list<DagNode *>*>(v); }
  virtual void	**rawdata() { return (void**)(&_m_data); }
};

// ************************************************************************* //
#endif//__NODERINGSET_HH__
