// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:30 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include "plug.h"
#include "datablock.h"
#include "dagnode.h"
#include <fstream>

// ************************************************************************* //
String	PlugSignal::typeStr("Signal");
String	PlugBool::typeStr("Bool");
String	PlugInt::typeStr("Int");
String	PlugDouble::typeStr("Double");
String	PlugString::typeStr("String");
String	PlugPnt2d::typeStr("Pnt2d");
String	PlugMat2i::typeStr("Mat2i");
String	PlugPnt3d::typeStr("Pnt3d");
String	PlugVec3d::typeStr("Vec3d");
String	PlugRgb3d::typeStr("Rgb3d");
String	PlugMat3d::typeStr("Mat3d");
String	PlugTwistLog::typeStr("TwistLog");
String	PlugMat4d::typeStr("Mat4d");
String	PlugIntArray::typeStr("IntArray");
String	PlugDoubleArray::typeStr("DoubleArray");
String	PlugPnt2dArray::typeStr("Pnt2dArray");
String	PlugPnt3dArray::typeStr("Pnt3dArray");
String	PlugVec3dArray::typeStr("Vec3dArray");
String	PlugTwistLogArray::typeStr("TwistLogArray");
String	PlugPnt2dList::typeStr("Pnt2dList");

// ************************************************************************* //
Plug::Plug(DataBlock	&db,
	   const String	&shortName,
	   const String	&,//longName
	   const bool	isVisible,
	   const bool	isSaved,
	   TPlug	*parent):
  _m_dataBlock(db), _m_parent(parent), _m_shortNamei(shortName),
  _m_longNamei(shortName),
  _m_bits((isVisible? BIT_ISVISIBLE: 0) | (isSaved? BIT_ISSAVED: 0)),
  _m_id(db._m_idCounter++),
  _m_input(0)
{
}

// ************************************************************************* //
Plug::Plug(DataBlock &db):
  _m_dataBlock(db), _m_parent(0), _m_bits(0), _m_id(db._m_idCounter++),
  _m_input(0)
{
}

// ************************************************************************* //
/// DESCRIPTION: disconnects all input/outputs
// ************************************************************************* //
Plug::~Plug()
{
  disconnectAll();
  _m_affects.clear();
}

// ************************************************************************* //
/// DESCRIPTION:
// ************************************************************************* //
void	Plug::registerPlug(const String		&typeStr,
			   Plug			*(*creator)(DataBlock &))
{
  creators()[typeStr] = creator;
}

// ************************************************************************* //
void	Plug::linkdata(Plug &d)
{
  *rawdata() = *d.rawdata();
}

// ************************************************************************* //
/// DESCRIPTION: Every Plug can be read as a Signal.
// ************************************************************************* //
void	Plug::readAsSignal()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	_m_input->readAsSignal();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }
  //std::cerr << "Error: Plug cannot be read as a Signal." << std::endl;
}

// ************************************************************************* //
/// DESCRIPTION: Every Plug can be written as a Signal.
// ************************************************************************* //
void	Plug::writeAsSignal()
{
  propagateDirtyBit();
}

// ************************************************************************* //
bool	Plug::readAsBool()
{
  std::cerr << "Error: Plug cannot be read as a Bool." << std::endl;
  return 0.0;
}

// ************************************************************************* //
void	Plug::writeAsBool(const bool)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be written as a Bool." << std::endl;
}

// ************************************************************************* //
int	Plug::readAsInt()
{
  std::cerr << "Error: Plug " << shortName()
	    << " cannot be read as a Int." << std::endl;
  return 0;
}

// ************************************************************************* //
void	Plug::writeAsInt(const int)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug " << shortName()
	    << " cannot be written as a Int." << std::endl;
}

// ************************************************************************* //
double	Plug::readAsDouble()
{
  std::cerr << "Error: Plug " << shortName()
	    << " cannot be read as a Double." << std::endl;
  return 0.0;
}

// ************************************************************************* //
void	Plug::writeAsDouble(const double)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be written as a Double." << std::endl;
}

// ************************************************************************* //
const String	&Plug::readAsString()
{
  static const String	value("");

  std::cerr << "Error: Plug " << shortName()
	    << " cannot be read as a String." << std::endl;
  return value;
}

// ************************************************************************* //
void	Plug::writeAsString(const String &)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be written as a String." << std::endl;
}

// ************************************************************************* //
const Pnt2d	&Plug::readAsPnt2d()
{
  static const Pnt2d	value(0.0);

  std::cerr << "Error: Plug cannot be read as a Pnt2d. "
	    << getTypeNbr() << " " << std::endl;
  return value;
}

// ************************************************************************* //
void	Plug::writeAsPnt2d(const Pnt2d &)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be written as a Pnt2d." << std::endl;
}

// ************************************************************************* //
const Mat2i	&Plug::readAsMat2i()
{
  static const Mat2i	value(0);

  std::cerr << "Error: Plug cannot be read as a Mat2i." << std::endl;
  return value;
}

// ************************************************************************* //
void	Plug::writeAsMat2i(const Mat2i &)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be written as a Mat2i." << std::endl;
}

// ************************************************************************* //
const Pnt3d	&Plug::readAsPnt3d()
{
  static const Pnt3d	value(0.0);

  std::cerr << "Error: Plug cannot be read as a Pnt3d." << std::endl;
  return value;
}

// ************************************************************************* //
void	Plug::writeAsPnt3d(const Pnt3d &)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be written as a Pnt3d." << std::endl;
}

// ************************************************************************* //
const Vec3d	&Plug::readAsVec3d()
{
  static const Vec3d	value(0.0);

  std::cerr << "Error: Plug cannot be read as a Vec3d." << std::endl;
  return value;
}

// ************************************************************************* //
void	Plug::writeAsVec3d(const Vec3d &)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be written as a Vec3d." << std::endl;
}

// ************************************************************************* //
const Rgb3d	&Plug::readAsRgb3d()
{
  static const Rgb3d	value(0.0);

  std::cerr << "Error: Plug cannot be read as a Rgb3d." << std::endl;
  return value;
}

// ************************************************************************* //
void	Plug::writeAsRgb3d(const Rgb3d &)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be written as a Rgb3d." << std::endl;
}

// ************************************************************************* //
const Mat3d	&Plug::readAsMat3d()
{
  static const Mat3d	value(0.0);

  std::cerr << "Error: Plug cannot be read as a Mat3d." << std::endl;
  return value;
}

// ************************************************************************* //
void	Plug::writeAsMat3d(const Mat3d &)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be written as a Mat3d." << std::endl;
}

// ************************************************************************* //
const TwistLog	&Plug::readAsTwistLog()
{
  static const TwistLog	value(0.0);

  std::cerr << "Error: Plug cannot be read as a TwistLog."
	    << std::endl;
  return value;
}

// ************************************************************************* //
void	Plug::writeAsTwistLog(const TwistLog &)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be written as a TwistLog."
	    << std::endl;
}

// ************************************************************************* //
const Mat4d	&Plug::readAsMat4d()
{
  static const Mat4d	value(0.0);

  std::cerr << "Error: Plug cannot be read as a Mat4d." << std::endl;
  return value;
}

// ************************************************************************* //
void	Plug::writeAsMat4d(const Mat4d &)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be written as a Mat4d." << std::endl;
}

// ************************************************************************* //
const std::vector<int>	&Plug::readAsIntArray()
{
  static const std::vector<int>	value;

  std::cerr << "Error: Plug cannot be read as a IntArray."
	    << std::endl;
  return value;
}

// ************************************************************************* //
void	Plug::writeAsIntArray(const std::vector<int> &)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be written as a IntArray."
	    << std::endl;
}

// ************************************************************************* //
const std::vector<double>	&Plug::readAsDoubleArray()
{
  static const std::vector<double>	value;

  std::cerr << "Error: Plug " << shortName() 
	    << " cannot be read as a DoubleArray."
	    << std::endl;
  return value;
}

// ************************************************************************* //
void	Plug::writeAsDoubleArray(const std::vector<double> &)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug " << shortName() 
	    << " cannot be written as a DoubleArray."
	    << std::endl;
}

// ************************************************************************* //
const std::vector<Pnt2d>	&Plug::readAsPnt2dArray()
{
  static const std::vector<Pnt2d>	value;

  std::cerr << "Error: Plug cannot be read as a Pnt2dArray."
	    << std::endl;
  return value;
}

// ************************************************************************* //
const Pnt2d	&Plug::readAsPnt2dArray(const int)
{
  static const Pnt2d	value;

  std::cerr << "Error: Plug cannot be written as a Pnt2dArray."
	    << std::endl;
  return value;
}

// ************************************************************************* //
void	Plug::writeAsPnt2dArray(const std::vector<Pnt2d> &)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be written as a Pnt2dArray."
	    << std::endl;
}

// ************************************************************************* //
void	Plug::writeAsPnt2dArray(const int	,//i
				const Pnt2d	&)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be read as a Pnt2dArray."
	    << std::endl;
}

// ************************************************************************* //
const std::vector<Pnt3d>	&Plug::readAsPnt3dArray()
{
  static const std::vector<Pnt3d>	value;

  std::cerr << "Error: Plug cannot be read as a Pnt3dArray."
	    << std::endl;
  return value;
}

// ************************************************************************* //
const Pnt3d	&Plug::readAsPnt3dArray(const int)
{
  static const Pnt3d	value;

  std::cerr << "Error: Plug cannot be written as a Pnt3dArray."
	    << std::endl;
  return value;
}

// ************************************************************************* //
void	Plug::writeAsPnt3dArray(const std::vector<Pnt3d> &)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be written as a Pnt3dArray."
	    << std::endl;
}

// ************************************************************************* //
void	Plug::writeAsPnt3dArray(const int	,//i
				const Pnt3d	&)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be read as a Pnt3dArray."
	    << std::endl;
}

// ************************************************************************* //
const std::vector<Vec3d>	&Plug::readAsVec3dArray()
{
  static const std::vector<Vec3d>	value;

  std::cerr << "Error: Plug cannot be read as a Vec3dArray."
	    << std::endl;
  return value;
}

// ************************************************************************* //
const Vec3d	&Plug::readAsVec3dArray(const int)
{
  static const Vec3d	value;

  std::cerr << "Error: Plug cannot be written as a Vec3dArray."
	    << std::endl;
  return value;
}

// ************************************************************************* //
void	Plug::writeAsVec3dArray(const std::vector<Vec3d> &)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be written as a Vec3dArray."
	    << std::endl;
}

// ************************************************************************* //
void	Plug::writeAsVec3dArray(const int	,//i
				const Vec3d	&)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be read as a Vec3dArray."
	    << std::endl;
}

// ************************************************************************* //
const std::vector<TwistLog>	&Plug::readAsTwistLogArray()
{
  static const std::vector<TwistLog>	value;

  std::cerr << "Error: Plug cannot be read as a TwistLogArray."
	    << std::endl;
  return value;
}

// ************************************************************************* //
const TwistLog	&Plug::readAsTwistLogArray(const int)
{
  static const TwistLog	value;

  std::cerr << "Error: Plug cannot be written as a TwistLogArray."
	    << std::endl;
  return value;
}

// ************************************************************************* //
void	Plug::writeAsTwistLogArray(const std::vector<TwistLog> &)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be written as a TwistLogArray."
	    << std::endl;
}

// ************************************************************************* //
void	Plug::writeAsTwistLogArray(const int	,//i
				   const TwistLog	&)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be read as a TwistLogArray."
	    << std::endl;
}

// ************************************************************************* //
const std::list<Pnt2d>	&Plug::readAsPnt2dList()
{
  static const std::list<Pnt2d>	value;

  std::cerr << "Error: Plug cannot be read as a Pnt2dList."
	    << std::endl;
  return value;
}

// ************************************************************************* //
void	Plug::writeAsPnt2dList(const std::list<Pnt2d>	&)
{
  propagateDirtyBit();
  std::cerr << "Error: Plug cannot be written as a Pnt2dList."
	    << std::endl;
}

// ************************************************************************* //
/// DESCRIPTION:
// ************************************************************************* //
void	*Plug::readAsCustom()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	_m_input->readAsCustom();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }
  return *rawdata();
}

// ************************************************************************* //
void	Plug::writeAsCustom(const void *value)
{
  copy(value);
  propagateDirtyBit();
}

// ************************************************************************* //
void	Plug::resize(const unsigned	,
		     const void		*)
{
  std::cerr << "Error: plug cannot be resized." << std::endl;
}

// ************************************************************************* //
void	Plug::getPlugByType(const int		type,
			    std::list<Plug *>	&hnd)
{
  if (getTypeNbr() == type)
    hnd.push_back(this);
}

// ************************************************************************* //
/// DESCRIPTION: Automatic registration of basic types.
// ************************************************************************* //
std::map<String, Plug *(*)(DataBlock &)>	&Plug::creators()
{
  static std::map<String, Plug *(*)(DataBlock &)>	screators;
  static bool	isInitialized(false);

  if (!isInitialized)
    {
      isInitialized = true;
      Plug::registerPlug(PlugSignal::typeStr, PlugSignal::creator);
      Plug::registerPlug(PlugBool::typeStr, PlugBool::creator);
      Plug::registerPlug(PlugInt::typeStr, PlugInt::creator);
      Plug::registerPlug(PlugDouble::typeStr, PlugDouble::creator);
      Plug::registerPlug(PlugString::typeStr, PlugString::creator);
      Plug::registerPlug(PlugPnt2d::typeStr, PlugPnt2d::creator);
      Plug::registerPlug(PlugMat2i::typeStr, PlugMat2i::creator);
      Plug::registerPlug(PlugPnt3d::typeStr, PlugPnt3d::creator);
      Plug::registerPlug(PlugVec3d::typeStr, PlugVec3d::creator);
      Plug::registerPlug(PlugRgb3d::typeStr, PlugRgb3d::creator);
      Plug::registerPlug(PlugMat3d::typeStr, PlugMat3d::creator);
      Plug::registerPlug(PlugTwistLog::typeStr, PlugTwistLog::creator);
      Plug::registerPlug(PlugMat4d::typeStr, PlugMat4d::creator);
      Plug::registerPlug(PlugIntArray::typeStr, PlugIntArray::creator);
      Plug::registerPlug(PlugDoubleArray::typeStr, PlugDoubleArray::creator);
      Plug::registerPlug(PlugPnt2dArray::typeStr, PlugPnt2dArray::creator);
      Plug::registerPlug(PlugPnt3dArray::typeStr, PlugPnt3dArray::creator);
      Plug::registerPlug(PlugVec3dArray::typeStr, PlugVec3dArray::creator);
      Plug::registerPlug(PlugTwistLogArray::typeStr,
			 PlugTwistLogArray::creator);
      Plug::registerPlug(PlugPnt2dList::typeStr, PlugPnt2dList::creator);
    }
  return screators;
}

// ************************************************************************* //
/// DESCRIPTION: Returns the node of the data.
// ************************************************************************* //
DagNode	&Plug::getNode()
{
  return _m_dataBlock._m_node;
}

// ************************************************************************* //
/// DESCRIPTION: Returns the node of input to the data.
///  If the data has no input, returns 0.
// ************************************************************************* //
DagNode	*Plug::getInputNode()
{
  if (!_m_input)
    return 0;
  return &_m_input->getNode();
}

// ************************************************************************* //
/// DESCRIPTION: Gets the nodes that are affected by this datafield.
///  If the data is the input two or more data field of another node,
///  multiple references are removed.
// ************************************************************************* //
void	Plug::getOutputNodes(std::vector<DagNode*> &onodes)
{
  std::vector<DagNode*>::iterator	oit;
  std::list<DagNode*>::iterator	nit;
  std::list<Plug*>::iterator	dit;
  std::list<DagNode*>	tmp;

  tmp.resize(_m_outputs.size());
  for (nit = tmp.begin(), dit = _m_outputs.begin(); nit != tmp.end();
       ++nit, ++dit)
    *nit = &(*dit)->getNode();
  tmp.unique();
  onodes.resize(tmp.size());
  for (nit = tmp.begin(), oit = onodes.begin(); nit != tmp.end();
       ++nit, ++oit)
    *oit = *nit;
}

// ************************************************************************* //
void	Plug::saveInfo(std::ostream	&os)
{
  os << ' ' << shortName() << ' '
     << longName() << ' '
     << _m_bits << ' '
     << _m_id << ' ';
}

// ************************************************************************* //
void	Plug::saveConnections(std::ostream	&os)
{
  std::list<Plug*>::const_iterator ait, oit;

  os << ' ' << int(*rawdata()) << ' ';
  os << '[';
  if (_m_input)
    os << int(_m_input);
  else
    os << '0';
  os << ']'
     << '[';
  ait = _m_affects.begin();
  if (ait != _m_affects.end())
    for (; ; )
      {
	os << int(*ait);
	if (++ait != _m_affects.end())
	  os << ',';
	else
	  break;
      }
  else
    os << '0';
  os << ']';
  os << '[';
  oit = _m_outputs.begin();
  if (oit != _m_outputs.end())
    for (; ; )
      {
	os << int(*oit);
	if (++oit != _m_outputs.end())
	  os << ',';
	else
	  break;
      }
  else
    os << '0';
  os << ']';
}

// ************************************************************************* //
void	Plug::loadIO(std::istream			&is,
		     std::map<int, std::list<void **> >	&ref2v)
{
  String	skip;
  int		input, affect, output;

  is >> skip;
  is >> input;
  if (input)
    {
      _m_input = this;
      ref2v[input].push_back((void**)(&_m_input));
    }
  else
    _m_input = 0;
  is >> skip;
  for (is >> skip; skip != "]"; is >> skip)
    {
      is >> affect;
      if (affect != 0)
	{
	  _m_affects.push_back(0);
	  ref2v[affect].push_back((void**)(&_m_affects.back()));
	}
      }
  for (is >> skip; skip != "]"; is >> skip)
    {
      is >> output;
      if (output != 0)
	{
	  _m_outputs.push_back(0);
	  ref2v[output].push_back((void**)(&_m_outputs.back()));
	}
    }
}

// ************************************************************************* //
/// DESCRIPTION: Converts the name of string (the name of the data) into an
///  index.
// ************************************************************************* //
unsigned	item2index(const String &item)
{
  if (item == "x" || item == "r")
    return 0;
  else if (item == "y" || item == "g")
    return 1;
  else if (item == "z" || item == "b")
    return 2;
  else if (item == "w" || item == "a")
    return 3;
  else
    return item.asInt();
}

// ************************************************************************* //
/// DESCRIPTION:
///  When loaded, plugs are not selected.
///  If plug is not saved, it is marked dirty, and the propagation of the 
///  dirty bit is performed once the graph is entirely loaded.
// ************************************************************************* //
void	Plug::loadInfo(std::istream				&is,
		       std::map<int, void *>			&,//v2real
		       std::map<int, std::list<void **> >	&)//ref2v
{
  is >> _m_shortNamei
     >> _m_longNamei
     >> _m_bits
     >> _m_id;
  _m_bits &= ~BIT_ISSELECTED;
  if (!isSaved())
    _m_bits |= BIT_ISDIRTY;
}

// ************************************************************************* //
/// DESCRIPTION:
// ************************************************************************* //
void	Plug::loadConnections(std::istream			&is,
			      std::map<int, void *>		&v2real,
			      std::map<int, std::list<void **> >	&ref2v)
{
  String	skip;
  int		vadress;

  is >> vadress;
  loadIO(is, ref2v);
  if (vadress)
    if (_m_parent)
      {
	freeData();
	if (hasTopParentInput())
	  {
	    ref2v[vadress].push_back(rawdata());
	  }
	else
	  {
	    const unsigned	index = item2index(shortName());

	    *rawdata() = _m_parent->subrawdata(index);
	    v2real[vadress] = *rawdata();
	  }
      }
    else if (_m_input)
      {
	freeData();
	ref2v[vadress].push_back(rawdata());
      }
    else
      v2real[vadress] = *rawdata();
}

// ************************************************************************* //
void	Plug::saveAll(std::ostream &os)
{
  os << getTypeStr() << ' ' << int(this) << ' ';
  saveInfo(os);
  if (isSaved())
    saveData(os);
  saveConnections(os);
}

// ************************************************************************* //
/// DESCRIPTION:
// ************************************************************************* //
Plug	*Plug::loadAll(std::istream				&is,
		       std::map<int, void *>			&v2real,
		       std::map<int, std::list<void **> >	&ref2v)
{
  loadInfo(is, v2real, ref2v);
  if (isSaved())
    loadData(is);
  loadConnections(is, v2real, ref2v);
  return 0;
}

// ************************************************************************* //
/// DESCRIPTION: Propagates the dirty bits to affected attributes.
// ************************************************************************* //
void	Plug::propagateDirtyBit()
{
  std::list<Plug*>::iterator	hit;

  for (hit = _m_affects.begin(); hit != _m_affects.end(); ++hit)
    if (!((*hit)->_m_bits & BIT_ISDIRTY))
      {
	(*hit)->_m_bits |= BIT_ISDIRTY;
	(*hit)->propagateDirtyBit();
      }
  for (hit = _m_outputs.begin(); hit != _m_outputs.end(); ++hit)
    if (!((*hit)->_m_bits & BIT_ISDIRTY))
      {
	(*hit)->_m_bits |= BIT_ISDIRTY;
	(*hit)->propagateDirtyBit();
      }
  if (_m_parent)
    _m_parent->propagateDirtyBit();
}

// ************************************************************************* //
void	Plug::disconnectAll()
{
  if (_m_input)
    {
      DataBlock::disconnect(*_m_input, *this);
      _m_input = 0;
    }
  while (!_m_outputs.empty())
    DataBlock::disconnect(*this, *_m_outputs.front());
}


// ************************************************************************* //
/// TODO:
/// DESCRIPTION: When computing a plug in the "compute" method,
///  if you've lost tracks of the plugs to pull, then call this 
///  method. In most cases, pulling manually the plugs is more efficient,
///  and the clean method should not be used... Yhis method requires
///  the storage of the plug's input.
// ************************************************************************* //
void	Plug::clean()
{
  /*
  if (_m_input)
    _m_input->readAsSignal();
  else
    {
      std::list<Plug *>	&plugs(_m_dataBlock._m_plugs);
      std::list<Plug*>::iterator	pit;

      for (pit = plugs.begin(); pit != plugs.end(); ++pit)
	{
	  std::list<Plug *>	&affects((*pit)->_m_affects);
	  std::list<Plug*>::iterator	ait;

	  for (ait = affects.begin(); ait != affects.end(); ++ait)
	    if (*ait == this)
	      (*ait)->clean();
	}
    }
  writeAsSignal();
  */
  _m_bits &= ~BIT_ISDIRTY;
}

// ************************************************************************* //
void	Plug::select(const bool on)
{
  on? (_m_bits |= BIT_ISSELECTED): (_m_bits &= ~BIT_ISSELECTED);
}

// ************************************************************************* //
void	Plug::setVisible(const bool on)
{
  on? (_m_bits |= BIT_ISVISIBLE): (_m_bits &= ~BIT_ISVISIBLE);
}

// ************************************************************************* //
void	Plug::pushName() const
{
  glPushName(1);//DataArray
  glPushName(getTypeNbr());//Type
  glPushName(getId());//Id
}

// ************************************************************************* //
void	Plug::popName() const
{
  glPopName();//Instance
  glPopName();//Type
  glPopName();//Node
}

// ************************************************************************* //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
PlugSignal::~PlugSignal()
{
  disconnectAll();
}

// ************************************************************************* //
void	PlugSignal::readAsSignal()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	_m_input->readAsSignal();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }
}

// ************************************************************************* //
void	PlugSignal::writeAsSignal()
{
  propagateDirtyBit();
}

// ************************************************************************* //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
PlugBool::~PlugBool()
{
  disconnectAll();
  if (!_m_parent && _m_data)
    {
      delete _m_data;
      _m_data = 0;
    }
}

// ************************************************************************* //
bool	PlugBool::getAsBool()
{
  return *_m_data;
}

// ************************************************************************* //
bool	PlugBool::readAsBool()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	if (!hasTopParentInput())
	  *_m_data = _m_input->readAsBool();
	else
	  _m_input->readAsBool();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }
  return *_m_data;
}

// ************************************************************************* //
void	PlugBool::writeAsBool(const bool value)
{
  if (value != *_m_data)
    {
      *_m_data = value;
      propagateDirtyBit();
    }
}

// ************************************************************************* //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
PlugInt::~PlugInt()
{
  disconnectAll();
  /*
  if (!_m_parent && _m_data)
    {
      delete _m_data;
      _m_data = 0;
      }*/
}

// ************************************************************************* //
int	PlugInt::getAsInt()
{
  return *_m_data;
}

// ************************************************************************* //
int	PlugInt::readAsInt()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	if (!hasTopParentInput())
	  *_m_data = _m_input->readAsInt();
	else
	  _m_input->readAsInt();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }
  return *_m_data;
}

// ************************************************************************* //
void	PlugInt::writeAsInt(const int value)
{
  if (value != *_m_data)
    {
      *_m_data = value;
      propagateDirtyBit();
    }
}

// ************************************************************************* //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
PlugDouble::~PlugDouble()
{
  disconnectAll();
  /*
  if (!_m_parent && _m_data)
    {
      delete _m_data;
      _m_data = 0;
      }*/
}

// ************************************************************************* //
/// DESCRIPTION: Returns current, non-updated value.
// ************************************************************************* //
double	PlugDouble::getAsDouble()
{
  return *_m_data;
}

// ************************************************************************* //
/// DESCRIPTION: Updates the value of the plug, and returns new value.
// ************************************************************************* //
double	PlugDouble::readAsDouble()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	if (!hasTopParentInput())
	  *_m_data = _m_input->readAsDouble();
	else
	  _m_input->readAsDouble();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }
  return *_m_data;
}

// ************************************************************************* //
void	PlugDouble::writeAsDouble(const double value)
{
  if (value != *_m_data)
    {
      *_m_data = value;
      propagateDirtyBit();
    }
}

// ************************************************************************* //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
PlugString::~PlugString()
{
  disconnectAll();
  /*
  if (!_m_parent && _m_data)
    {
      delete _m_data;
      _m_data = 0;
      }*/
}

// ************************************************************************* //
const String	&PlugString::readAsString()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	if (!hasTopParentInput())
	  *_m_data = _m_input->readAsString();
	else
	  _m_input->readAsString();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }
  return *_m_data;
}

// ************************************************************************* //
void	PlugString::writeAsString(const String &value)
{
  if (value != *_m_data)
    {
      *_m_data = value;
      propagateDirtyBit();
    }
}

// ************************************************************************* //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
PlugPnt2d::~PlugPnt2d()
{
  disconnectAll();
  /*
  if (!_m_parent && _m_data)
    {
      delete _m_data;
      _m_data = 0;
      }*/
}

// ************************************************************************* //
const Pnt2d	&PlugPnt2d::readAsPnt2d()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	if (!hasTopParentInput())
	  *_m_data = _m_input->readAsPnt2d();
	else
	  _m_input->readAsPnt2d();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }

  std::map<unsigned, Plug*>::iterator cit;

  for (cit = _m_children.begin(); cit != _m_children.end(); ++cit)
    if (cit->second)
      cit->second->readAsDouble();
  return *_m_data;
}

// ************************************************************************* //
void	PlugPnt2d::writeAsPnt2d(const Pnt2d &value)
{
  if (value != *_m_data)
    {
      *_m_data = value;
      propagateDirtyBit();
    }
}

// ************************************************************************* //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
PlugMat2i::~PlugMat2i()
{
  disconnectAll();
  /*
  if (!_m_parent && _m_data)
    {
      delete _m_data;
      _m_data = 0;
      }*/
}

// ************************************************************************* //
const Mat2i	&PlugMat2i::readAsMat2i()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	if (!hasTopParentInput())
	  *_m_data = _m_input->readAsMat2i();
	else
	  _m_input->readAsMat2i();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }

  std::map<unsigned, Plug*>::iterator cit;

  for (cit = _m_children.begin(); cit != _m_children.end(); ++cit)
    if (cit->second)
      cit->second->readAsDouble();
  return *_m_data;
}

// ************************************************************************* //
void	PlugMat2i::writeAsMat2i(const Mat2i &value)
{
  if (value != *_m_data)
    {
      *_m_data = value;
      propagateDirtyBit();
    }
}

// ************************************************************************* //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
PlugPnt3d::~PlugPnt3d()
{
  disconnectAll();
  /*
  if (!_m_parent && _m_data)
    {
      delete _m_data;
      _m_data = 0;
      }*/
}

// ************************************************************************* //
const Pnt3d	&PlugPnt3d::readAsPnt3d()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	if (!hasTopParentInput())
	  *_m_data = _m_input->readAsPnt3d();
	else
	  _m_input->readAsPnt3d();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }

  std::map<unsigned, Plug*>::iterator cit;

  for (cit = _m_children.begin(); cit != _m_children.end(); ++cit)
    if (cit->second)
      cit->second->readAsDouble();
  return *_m_data;
}

// ************************************************************************* //
void	PlugPnt3d::writeAsPnt3d(const Pnt3d &value)
{
  if (value != *_m_data)
    {
      *_m_data = value;
      propagateDirtyBit();
    }
}

// ************************************************************************* //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
PlugVec3d::~PlugVec3d()
{
  disconnectAll();
  /*
  if (!_m_parent && _m_data)
    {
      delete _m_data;
      _m_data = 0;
      }*/
}

// ************************************************************************* //
const Vec3d	&PlugVec3d::readAsVec3d()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	if (!hasTopParentInput())
	  *_m_data = _m_input->readAsVec3d();
	else
	  _m_input->readAsVec3d();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }

  std::map<unsigned, Plug*>::iterator cit;

  for (cit = _m_children.begin(); cit != _m_children.end(); ++cit)
    if (cit->second)
      cit->second->readAsDouble();
  return *_m_data;
}

// ************************************************************************* //
void	PlugVec3d::writeAsVec3d(const Vec3d &value)
{
  if (value != *_m_data)
    {
      *_m_data = value;
      propagateDirtyBit();
    }
}

// ************************************************************************* //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
PlugRgb3d::~PlugRgb3d()
{
  disconnectAll();
  /*
  if (!_m_parent && _m_data)
    {
      delete _m_data;
      _m_data = 0;
      }*/
}

// ************************************************************************* //
const Rgb3d	&PlugRgb3d::readAsRgb3d()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	if (!hasTopParentInput())
	  *_m_data = _m_input->readAsRgb3d();
	else
	  _m_input->readAsRgb3d();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }

  std::map<unsigned, Plug*>::iterator cit;

  for (cit = _m_children.begin(); cit != _m_children.end(); ++cit)
    if (cit->second)
      cit->second->readAsDouble();
  return *_m_data;
}

// ************************************************************************* //
void	PlugRgb3d::writeAsRgb3d(const Rgb3d &value)
{
  if (value != *_m_data)
    {
      *_m_data = value;
      propagateDirtyBit();
    }
}

// ************************************************************************* //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
PlugMat3d::~PlugMat3d()
{
  disconnectAll();
  /*
  if (!_m_parent && _m_data)
    {
      delete _m_data;
      _m_data = 0;
      }*/
}

// ************************************************************************* //
const Mat3d	&PlugMat3d::readAsMat3d()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	if (!hasTopParentInput())
	  *_m_data = _m_input->readAsMat3d();
	else
	  _m_input->readAsMat3d();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }

  std::map<unsigned, Plug*>::iterator cit;

  for (cit = _m_children.begin(); cit != _m_children.end(); ++cit)
    if (cit->second)
      cit->second->readAsDouble();
  return *_m_data;
}

// ************************************************************************* //
void	PlugMat3d::writeAsMat3d(const Mat3d &value)
{
  if (value != *_m_data)
    {
      *_m_data = value;
      propagateDirtyBit();
    }
}

// ************************************************************************* //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
PlugTwistLog::~PlugTwistLog()
{
  disconnectAll();
  /*
  if (!_m_parent && _m_data)
    {
      delete _m_data;
      _m_data = 0;
      }*/
}

// ************************************************************************* //
const TwistLog	&PlugTwistLog::readAsTwistLog()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	if (!hasTopParentInput())
	  *_m_data = _m_input->readAsTwistLog();
	else
	  _m_input->readAsTwistLog();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }

  std::map<unsigned, Plug*>::iterator cit;

  for (cit = _m_children.begin(); cit != _m_children.end(); ++cit)
    if (cit->second)
      cit->second->readAsDouble();
  return *_m_data;
}

// ************************************************************************* //
void	PlugTwistLog::writeAsTwistLog(const TwistLog &value)
{
  if (value != *_m_data)
    {
      *_m_data = value;
      propagateDirtyBit();
    }
}

// ************************************************************************* //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
PlugMat4d::~PlugMat4d()
{
  disconnectAll();
  /*
  if (!_m_parent && _m_data)
    {
      delete _m_data;
      _m_data = 0;
      }*/
}

// ************************************************************************* //
const Mat4d	&PlugMat4d::readAsMat4d()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	if (!hasTopParentInput())
	  *_m_data = _m_input->readAsMat4d();
	else
	  _m_input->readAsMat4d();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }

  std::map<unsigned, Plug*>::iterator cit;

  for (cit = _m_children.begin(); cit != _m_children.end(); ++cit)
    if (cit->second)
      cit->second->readAsDouble();
  return *_m_data;
}

// ************************************************************************* //
void	PlugMat4d::writeAsMat4d(const Mat4d &value)
{
  if (value != *_m_data)
    {
      *_m_data = value;
      propagateDirtyBit();
    }
}

// ************************************************************************* //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
PlugIntArray::~PlugIntArray()
{
  disconnectAll();
  /*
  if (!_m_parent && _m_data)
    {
      delete _m_data;
      _m_data = 0;
      }*/
}

// ************************************************************************* //
void	PlugIntArray::freeData()
{
  if (!_m_parent && _m_data)
    delete _m_data;
  _m_childrensize = 0;
  _m_data = 0;
  TPlug::freeData();
}

// ************************************************************************* //
void	PlugIntArray::newData()
{
  if (!_m_parent && !_m_data)
    {
      _m_data = new std::vector<int>;
      _m_childrensize = 0;
    }
  TPlug::newData();
}

// ************************************************************************* //
void	PlugIntArray::saveData(std::ostream &os)
{
  std::vector<int>::iterator	dit;

  os << _m_data->size() << ' ';
  for (dit = _m_data->begin(); dit != _m_data->end(); ++dit)
    os << *dit << ' ';
}

// ************************************************************************* //
void	PlugIntArray::loadData(std::istream &is)
{
  _m_data->clear();
  int		size;
  int	p;
	
  is >> size;
  _m_childrensize = size;
  for (int i(0); i < size; ++i)
    {
      is >> p;
      _m_data->push_back(p);
    }
}

// ************************************************************************* //
const std::vector<int>	&PlugIntArray::readAsIntArray()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	if (!hasTopParentInput())
	  *_m_data = _m_input->readAsIntArray();
	else
	  _m_input->readAsIntArray();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }

  std::map<unsigned, Plug*>::iterator cit;

  for (cit = _m_children.begin(); cit != _m_children.end(); ++cit)
    if (cit->second)
      cit->second->readAsInt();
  return *_m_data;
}

// ************************************************************************* //
void	PlugIntArray::writeAsIntArray(const std::vector<int> &value)
{
  *_m_data = value;
  _m_childrensize = value.size();
  propagateDirtyBit();
}

// ************************************************************************* //
void	PlugIntArray::resize(const unsigned	n,
				const void	*defaultValue)
{
  _m_data->resize(n);
  _m_childrensize = n;
  for (unsigned i(0); i < n; ++i)
    if (_m_children[i])
      *_m_children[i]->rawdata() = subrawdata(i);
  if (defaultValue)
    {
      const int	dv(*static_cast<const int*>(defaultValue));

      for (unsigned i(0); i < n; ++i)
	(*_m_data)[i] = dv;
    }
}

// ************************************************************************* //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
PlugDoubleArray::~PlugDoubleArray()
{
  disconnectAll();
  /*
  if (!_m_parent && _m_data)
    {
      delete _m_data;
      _m_data = 0;
      }*/
}

// ************************************************************************* //
void	PlugDoubleArray::freeData()
{
  if (!_m_parent && _m_data)
    delete _m_data;
  _m_childrensize = 0;
  _m_data = 0;
  TPlug::freeData();
}

// ************************************************************************* //
void	PlugDoubleArray::newData()
{
  if (!_m_parent && !_m_data)
    {
      _m_data = new std::vector<double>;
      _m_childrensize = 0;
    }
  TPlug::newData();
}

// ************************************************************************* //
void	PlugDoubleArray::saveData(std::ostream &os)
{
  std::vector<double>::iterator	dit;

  os << _m_data->size() << ' ';
  for (dit = _m_data->begin(); dit != _m_data->end(); ++dit)
    os << *dit << ' ';
}

// ************************************************************************* //
void	PlugDoubleArray::loadData(std::istream &is)
{
  _m_data->clear();
  int		size;
  double	p;
	
  is >> size;
  _m_childrensize = size;
  for (int i(0); i < size; ++i)
    {
      is >> p;
      _m_data->push_back(p);
    }
}

// ************************************************************************* //
const std::vector<double>	&PlugDoubleArray::readAsDoubleArray()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	if (!hasTopParentInput())
	  *_m_data = _m_input->readAsDoubleArray();
	else
	  _m_input->readAsDoubleArray();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }

  std::map<unsigned, Plug*>::iterator cit;

  for (cit = _m_children.begin(); cit != _m_children.end(); ++cit)
    if (cit->second)
      cit->second->readAsDouble();
  return *_m_data;
}

// ************************************************************************* //
void	PlugDoubleArray::writeAsDoubleArray(const std::vector<double> &value)
{
  *_m_data = value;
  _m_childrensize = value.size();
  propagateDirtyBit();
}

// ************************************************************************* //
void	PlugDoubleArray::resize(const unsigned	n,
				const void	*defaultValue)
{
  _m_data->resize(n);
  _m_childrensize = n;
  for (unsigned i(0); i < n; ++i)
    if (_m_children[i])
      *_m_children[i]->rawdata() = subrawdata(i);
  if (defaultValue)
    {
      const double	dv(*static_cast<const double*>(defaultValue));

      for (unsigned i(0); i < n; ++i)
	(*_m_data)[i] = dv;
    }
}

// ************************************************************************* //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
PlugPnt2dArray::~PlugPnt2dArray()
{
  disconnectAll();
  /*
  if (!_m_parent && _m_data)
    {
      delete _m_data;
      _m_data = 0;
      }*/
}

// ************************************************************************* //
void	PlugPnt2dArray::freeData()
{
  if (!_m_parent && _m_data)
    delete _m_data;
  _m_childrensize = 0;
  _m_data = 0;
  TPlug::freeData();
}

// ************************************************************************* //
void	PlugPnt2dArray::newData()
{
  if (!_m_parent && !_m_data)
    {
      _m_data = new std::vector<Pnt2d>;
      _m_childrensize = 0;
    }
  TPlug::newData();
}

// ************************************************************************* //
void	PlugPnt2dArray::saveData(std::ostream &os)
{
  std::vector<Pnt2d>::iterator	dit;

  os << _m_data->size() << ' ';
  for (dit = _m_data->begin(); dit != _m_data->end(); ++dit)
    os << *dit << ' ';
}

// ************************************************************************* //
void	PlugPnt2dArray::loadData(std::istream &is)
{
  _m_data->clear();
  int		size;
  Pnt2d		p;

  is >> size;
  _m_childrensize = size;
  for (int i(0); i < size; ++i)
    {
      is >> p;
      _m_data->push_back(p);
    }
}

// ************************************************************************* //
const std::vector<Pnt2d>	&PlugPnt2dArray::readAsPnt2dArray()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	if (!hasTopParentInput())
	  *_m_data = _m_input->readAsPnt2dArray();
	else
	  _m_input->readAsPnt2dArray();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }

  std::map<unsigned, Plug*>::iterator cit;

  for (cit = _m_children.begin(); cit != _m_children.end(); ++cit)
    if (cit->second)
      cit->second->readAsPnt2d();
  return *_m_data;
}

// ************************************************************************* //
const Pnt2d	&PlugPnt2dArray::readAsPnt2dArray(const int i)
{
  if (!isArray())
    {
      std::cerr << "Error: Plug cannot be read as an array." << std::endl;
      std::exit(-1);
    }
  if (!_m_children[i])
    addPnt2d(i);
  return _m_children[i]->readAsPnt2d();
}

// ************************************************************************* //
void	PlugPnt2dArray::writeAsPnt2dArray(const std::vector<Pnt2d> &value)
{
  *_m_data = value;
  _m_childrensize = value.size();
  propagateDirtyBit();
}

// ************************************************************************* //
void	PlugPnt2dArray::writeAsPnt2dArray(const int	i,
					  const Pnt2d	&value)
{
  if (!_m_children[i])
    addPnt2d(i);
  _m_children[i]->writeAsPnt2d(value);
  propagateDirtyBit();
}

// ************************************************************************* //
void	PlugPnt2dArray::resize(const unsigned	n,
			       const void	*defaultValue)
{
  _m_data->resize(n);
  _m_childrensize = n;
  for (unsigned i(0); i < n; ++i)
    if (_m_children[i])
      *_m_children[i]->rawdata() = subrawdata(i);
  if (defaultValue)
    {
      const Pnt2d	dv(*static_cast<const Pnt2d*>(defaultValue));

      for (unsigned i(0); i < n; ++i)
	(*_m_data)[i] = dv;
    }
}

// ************************************************************************* //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
PlugPnt3dArray::~PlugPnt3dArray()
{
  disconnectAll();
  /*
  if (!_m_parent && _m_data)
    {
      delete _m_data;
      _m_data = 0;
      }*/
}

// ************************************************************************* //
void	PlugPnt3dArray::freeData()
{
  if (!_m_parent && _m_data)
    delete _m_data;
  _m_childrensize = 0;
  _m_data = 0;
  TPlug::freeData();
}

// ************************************************************************* //
void	PlugPnt3dArray::newData()
{
  if (!_m_parent && !_m_data)
    {
      _m_data = new std::vector<Pnt3d>;
      _m_childrensize = 0;
    }
  TPlug::newData();
}

// ************************************************************************* //
void	PlugPnt3dArray::saveData(std::ostream &os)
{
  std::vector<Pnt3d>::iterator	dit;

  os << _m_data->size() << ' ';
  for (dit = _m_data->begin(); dit != _m_data->end(); ++dit)
    os << *dit << ' ';
}

// ************************************************************************* //
void	PlugPnt3dArray::loadData(std::istream &is)
{
  _m_data->clear();
  int		size;
  Pnt3d		p;

  is >> size;
  _m_childrensize = size;
  for (int i(0); i < size; ++i)
    {
      is >> p;
      _m_data->push_back(p);
    }
}

// ************************************************************************* //
const std::vector<Pnt3d>	&PlugPnt3dArray::readAsPnt3dArray()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	if (!hasTopParentInput())
	  *_m_data = _m_input->readAsPnt3dArray();
	else
	  _m_input->readAsPnt3dArray();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }

  std::map<unsigned, Plug*>::iterator cit;

  for (cit = _m_children.begin(); cit != _m_children.end(); ++cit)
    if (cit->second)
      cit->second->readAsPnt3d();
  return *_m_data;
}

// ************************************************************************* //
const Pnt3d	&PlugPnt3dArray::readAsPnt3dArray(const int i)
{
  if (!isArray())
    {
      std::cerr << "Error: Plug cannot be read as an array." << std::endl;
      std::exit(-1);
    }
  if (!_m_children[i])
    addPnt3d(i);
  return _m_children[i]->readAsPnt3d();
}

// ************************************************************************* //
void	PlugPnt3dArray::writeAsPnt3dArray(const std::vector<Pnt3d> &value)
{
  *_m_data = value;
  _m_childrensize = value.size();
  propagateDirtyBit();
}

// ************************************************************************* //
void	PlugPnt3dArray::writeAsPnt3dArray(const int	i,
					  const Pnt3d	&value)
{
  if (!_m_children[i])
    addPnt3d(i);
  _m_children[i]->writeAsPnt3d(value);
  propagateDirtyBit();
}

// ************************************************************************* //
void	PlugPnt3dArray::resize(const unsigned	n,
			       const void	*defaultValue)
{
  _m_data->resize(n);
  _m_childrensize = n;
  for (unsigned i(0); i < n; ++i)
    if (_m_children[i])
      *_m_children[i]->rawdata() = subrawdata(i);
  if (defaultValue)
    {
      const Pnt3d	dv(*static_cast<const Pnt3d*>(defaultValue));

      for (unsigned i(0); i < n; ++i)
	(*_m_data)[i] = dv;
    }
}

// ************************************************************************* //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
PlugVec3dArray::~PlugVec3dArray()
{
  disconnectAll();
  /*
  if (!_m_parent && _m_data)
    {
      delete _m_data;
      _m_data = 0;
      }*/
}

// ************************************************************************* //
void	PlugVec3dArray::freeData()
{
  if (!_m_parent && _m_data)
    delete _m_data;
  _m_childrensize = 0;
  _m_data = 0;
  TPlug::freeData();
}

// ************************************************************************* //
void	PlugVec3dArray::newData()
{
  if (!_m_parent && !_m_data)
    {
      _m_data = new std::vector<Vec3d>;
      _m_childrensize = 0;
    }
  TPlug::newData();
}

// ************************************************************************* //
void	PlugVec3dArray::saveData(std::ostream &os)
{
  std::vector<Vec3d>::iterator	dit;

  os << _m_data->size() << ' ';
  for (dit = _m_data->begin(); dit != _m_data->end(); ++dit)
    os << *dit << ' ';
}

// ************************************************************************* //
void	PlugVec3dArray::loadData(std::istream &is)
{
  _m_data->clear();
  int		size;
  Vec3d		p;

  is >> size;
  _m_childrensize = size;
  for (int i(0); i < size; ++i)
    {
      is >> p;
      _m_data->push_back(p);
    }
}

// ************************************************************************* //
const std::vector<Vec3d>	&PlugVec3dArray::readAsVec3dArray()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	if (!hasTopParentInput())
	  *_m_data = _m_input->readAsVec3dArray();
	else
	  _m_input->readAsVec3dArray();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }
  std::map<unsigned, Plug*>::iterator cit;

  for (cit = _m_children.begin(); cit != _m_children.end(); ++cit)
    if (cit->second)
      cit->second->readAsVec3d();
  return *_m_data;
}

// ************************************************************************* //
const Vec3d	&PlugVec3dArray::readAsVec3dArray(const int i)
{
  if (!isArray())
    {
      std::cerr << "Error: Plug cannot be read as an array." << std::endl;
      std::exit(-1);
    }
  if (!_m_children[i])
    addVec3d(i);// TODO: add pointer to data!
  if (!*_m_children[i]->rawdata())
    *_m_children[i]->rawdata() = subrawdata(i);//TEST: TODO: put it in others
  return _m_children[i]->readAsVec3d();
}

// ************************************************************************* //
void	PlugVec3dArray::writeAsVec3dArray(const std::vector<Vec3d> &value)
{
  *_m_data = value;
  _m_childrensize = value.size();
  propagateDirtyBit();
}

// ************************************************************************* //
void	PlugVec3dArray::writeAsVec3dArray(const int	i,
					  const Vec3d	&value)
{
  if (!_m_children[i])
    addVec3d(i);
  _m_children[i]->writeAsVec3d(value);
  propagateDirtyBit();
}

// ************************************************************************* //
void	PlugVec3dArray::resize(const unsigned	n,
			       const void	*defaultValue)
{
  _m_data->resize(n);
  _m_childrensize = n;
  for (unsigned i(0); i < n; ++i)
    if (_m_children[i])
      *_m_children[i]->rawdata() = subrawdata(i);
  if (defaultValue)
    {
      const Vec3d	dv(*static_cast<const Vec3d*>(defaultValue));

      for (unsigned i(0); i < n; ++i)
	(*_m_data)[i] = dv;
    }
}

// ************************************************************************* //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
PlugTwistLogArray::~PlugTwistLogArray()
{
  disconnectAll();
  /*
  if (!_m_parent && _m_data)
    {
      delete _m_data;
      _m_data = 0;
      }*/
}

// ************************************************************************* //
void	PlugTwistLogArray::freeData()
{
  if (!_m_parent && _m_data)
    delete _m_data;
  _m_childrensize = 0;
  _m_data = 0;
  TPlug::freeData();
}

// ************************************************************************* //
void	PlugTwistLogArray::newData()
{
  if (!_m_parent && !_m_data)
    {
      _m_data = new std::vector<TwistLog>;
      _m_childrensize = 0;
    }
  TPlug::newData();
}

// ************************************************************************* //
void	PlugTwistLogArray::saveData(std::ostream &os)
{
  std::vector<TwistLog>::iterator	dit;

  os << _m_data->size() << ' ';
  for (dit = _m_data->begin(); dit != _m_data->end(); ++dit)
    os << *dit << ' ';
}

// ************************************************************************* //
void	PlugTwistLogArray::loadData(std::istream &is)
{
  _m_data->clear();
  int		size;
  TwistLog		p;

  is >> size;
  _m_childrensize = size;
  for (int i(0); i < size; ++i)
    {
      is >> p;
      _m_data->push_back(p);
    }
}

// ************************************************************************* //
const std::vector<TwistLog>	&PlugTwistLogArray::readAsTwistLogArray()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	if (!hasTopParentInput())
	  *_m_data = _m_input->readAsTwistLogArray();
	else
	  _m_input->readAsTwistLogArray();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }

  std::map<unsigned, Plug*>::iterator cit;

  for (cit = _m_children.begin(); cit != _m_children.end(); ++cit)
    if (cit->second)
      cit->second->readAsTwistLog();
  return *_m_data;
}

// ************************************************************************* //
const TwistLog	&PlugTwistLogArray::readAsTwistLogArray(const int i)
{
  if (!isArray())
    {
      std::cerr << "Error: Plug cannot be read as an array." << std::endl;
      std::exit(-1);
    }
  if (!_m_children[i])
    addTwistLog(i);
  return _m_children[i]->readAsTwistLog();
}

// ************************************************************************* //
void PlugTwistLogArray::writeAsTwistLogArray(const std::vector<TwistLog>&value)
{
  *_m_data = value;
  _m_childrensize = value.size();
  propagateDirtyBit();
}

// ************************************************************************* //
void	PlugTwistLogArray::writeAsTwistLogArray(const int	i,
					  const TwistLog	&value)
{
  if (!_m_children[i])
    addTwistLog(i);
  _m_children[i]->writeAsTwistLog(value);
  propagateDirtyBit();
}

// ************************************************************************* //
void	PlugTwistLogArray::resize(const unsigned	n,
				  const void	*defaultValue)
{
  _m_data->resize(n);
  _m_childrensize = n;
  for (unsigned i(0); i < n; ++i)
    if (_m_children[i])
      *_m_children[i]->rawdata() = subrawdata(i);
  if (defaultValue)
    {
      const TwistLog	dv(*static_cast<const TwistLog*>(defaultValue));

      for (unsigned i(0); i < n; ++i)
	(*_m_data)[i] = dv;
    }
}

// ************************************************************************* //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
PlugPnt2dList::~PlugPnt2dList()
{
  disconnectAll();
  /*
  if (!_m_parent && _m_data)
    {
      delete _m_data;
      _m_data = 0;
      }*/
}

// ************************************************************************* //
void	PlugPnt2dList::freeData()
{
  if (!_m_parent && _m_data)
    delete _m_data;
  _m_childrensize = 0;;
  _m_data = 0;
  TPlug::freeData();
}

// ************************************************************************* //
void	PlugPnt2dList::newData()
{
  if (!_m_parent && !_m_data)
    {
      _m_data = new std::list<Pnt2d>;
      _m_childrensize = 0;
    }
  for (unsigned i(0); i < _m_childrensize; ++i)
    if (_m_children[i])
      {
	*_m_children[i]->rawdata() = subrawdata(i);
	_m_children[i]->newData();
      }
  TPlug::newData();
}

// ************************************************************************* //
void	PlugPnt2dList::saveData(std::ostream &os)
{
  std::list<Pnt2d>::iterator	dit;

  os << _m_data->size() << ' ';
  for (dit = _m_data->begin(); dit != _m_data->end(); ++dit)
    os << *dit << ' ';
}

// ************************************************************************* //
void	PlugPnt2dList::loadData(std::istream &is)
{
  _m_data->clear();
  int	size;
  Pnt2d	p;

  is >> size;
  _m_childrensize = size;
  for (int i(0); i < size; ++i)
    {
      is >> p;
      _m_data->push_back(p);
    }
}

// ************************************************************************* //
const std::list<Pnt2d>	&PlugPnt2dList::readAsPnt2dList()
{
  if (!(_m_bits & BIT_ISCOMPUTING) && (_m_bits & BIT_ISDIRTY))
    {
      if (_m_input)
	if (!hasTopParentInput())
	  *_m_data = _m_input->readAsPnt2dList();
	else
	  _m_input->readAsPnt2dList();
      else
	{
	  _m_bits |= BIT_ISCOMPUTING;
	  _m_dataBlock.node().compute(*this);
	  _m_bits &= ~BIT_ISCOMPUTING;
	}
      clean();
    }

  std::map<unsigned, Plug*>::iterator cit;

  for (cit = _m_children.begin(); cit != _m_children.end(); ++cit)
    if (cit->second)
      cit->second->readAsPnt2d();
  return *_m_data;
}

// ************************************************************************* //
void	PlugPnt2dList::writeAsPnt2dList(const std::list<Pnt2d> &value)
{
  *_m_data = value;
  _m_childrensize = value.size();
  propagateDirtyBit();
}

// ************************************************************************* //
void	PlugPnt2dList::resize(const unsigned	n,
			      const void	*)//defaultvalue
{
  _m_childrensize = n;// TODO: try to put more elms in the list
}

// ************************************************************************* //
void	*PlugPnt2dList::subrawdata(unsigned i)
{
  std::list<Pnt2d>::iterator	pit;
  std::list<Pnt2d> *data = static_cast<std::list<Pnt2d>*>(_m_data);
  unsigned	c(0);

  for (pit = data->begin(); pit != data->end(); ++pit, ++c)
    if (c == i)
      break;
  if (pit == data->end())
    std::cerr << "ERROR: " << " index overflows data." << std::endl;
  return &(*pit);
}

// ************************************************************************* //
/// DESCRIPTION: 
///  Finds the top parent. Returns true if the top parent has input.
// ************************************************************************* //
bool	Plug::hasTopParentInput() const
{
  const Plug *p0, *p1;

  for (p0 = p1 = this; p0; p1 = p0, p0 = p0->_m_parent)
    ;
  return p1 && p1->_m_input;
}

// ************************************************************************* //
/// DESCRIPTION:
///  Memory of the data tree is handled by the root.
///  If data is accessed from child, then parent is assumed not connected.
/// WARNING: Cannot connect data and children. only one level connections!!
// ************************************************************************* //
TPlug::TPlug(DataBlock		&db,
	     const String	&name, 
	     const bool		isVisible,
	     const bool		isSaved,
	     TPlug		*parent):
  Plug(db, name, name, isVisible, isSaved, parent)
{
  _m_bits &= ~BIT_ISSELECTED;
}

// ************************************************************************* //
/// DESCRIPTION: disconnects this as well as children.
// ************************************************************************* //
void	TPlug::disconnectAll()
{
  std::map<unsigned, Plug*>::iterator	cit;

  Plug::disconnectAll();
  for (cit = _m_children.begin(); cit != _m_children.end(); ++cit)
    if (cit->second)
      cit->second->disconnectAll();
}

// ************************************************************************* //
/// DESCRIPTION: Add attribute child.
// ************************************************************************* //
void	TPlug::addSignal(const int	index)
{
  if (!_m_children[index])
    _m_children[index] = new PlugSignal(_m_dataBlock, index,
					_m_bits & BIT_ISVISIBLE,
					_m_bits & BIT_ISSAVED,
					this);
}

// ************************************************************************* //
/// DESCRIPTION: Add attribute child.
// ************************************************************************* //
void	TPlug::addBool(const int	index)
{
  if (!_m_children[index])
    _m_children[index] = new PlugBool(_m_dataBlock, index,
				      _m_bits & BIT_ISVISIBLE,
				      _m_bits & BIT_ISSAVED, this,
				      static_cast<bool*>(subrawdata(index)));
}

// ************************************************************************* //
/// DESCRIPTION: Parent assumed not connected, thus isADuplicate is always
///   true.
///  Adds a double.
// ************************************************************************* //
void	TPlug::addInt(const int	index)
{
  if (!_m_children[index])
    _m_children[index] = new PlugInt(_m_dataBlock, index, 
				     _m_bits & BIT_ISVISIBLE,
				     _m_bits & BIT_ISSAVED, this,
				     static_cast<int*>(subrawdata(index)));
}

// ************************************************************************* //
/// DESCRIPTION: Parent assumed not connected, thus isADuplicate is always
///   true.
///  Adds a double.
// ************************************************************************* //
void	TPlug::addDouble(const int	index)
{
  if (!_m_children[index])
    _m_children[index] =
      new PlugDouble(_m_dataBlock, index,
		     _m_bits & BIT_ISVISIBLE,
		     _m_bits & BIT_ISSAVED, this,
		     static_cast<double*>(subrawdata(index)));
}

// ************************************************************************* //
/// DESCRIPTION: Parent assumed not connected, thus isADuplicate is always
///   true.
///  Adds a double.
// ************************************************************************* //
void	TPlug::addString(const int	index)
{
  if (!_m_children[index])
    _m_children[index] =
      new PlugString(_m_dataBlock, index,
		     _m_bits & BIT_ISVISIBLE,
		     _m_bits & BIT_ISSAVED, this,
		     static_cast<String*>(subrawdata(index)));
}

// ************************************************************************* //
/// DESCRIPTION: Parent assumed not connected, thus isADuplicate is always
///   true.
///  Adds a double.
// ************************************************************************* //
void	TPlug::addPnt2d(const int	index)
{
  if (!_m_children[index])
    _m_children[index] = new PlugPnt2d(_m_dataBlock, index,
				       _m_bits & BIT_ISVISIBLE,
				       _m_bits & BIT_ISSAVED, this,
				       static_cast<Pnt2d*>(subrawdata(index)));
}

// ************************************************************************* //
/// DESCRIPTION: Parent assumed not connected, thus isADuplicate is always
///   true.
///  Adds a double.
// ************************************************************************* //
void	TPlug::addPnt3d(const int	index)
{
  if (!_m_children[index])
    _m_children[index] = new PlugPnt3d(_m_dataBlock, index,
				       _m_bits & BIT_ISVISIBLE,
				       _m_bits & BIT_ISSAVED, this,
				       static_cast<Pnt3d*>(subrawdata(index)));
}

// ************************************************************************* //
/// DESCRIPTION: Parent assumed not connected, thus isADuplicate is always
///   true.
///  Adds a double.
// ************************************************************************* //
void	TPlug::addVec3d(const int	index)
{
  if (!_m_children[index])
    _m_children[index] = new PlugVec3d(_m_dataBlock, index,
				       _m_bits & BIT_ISVISIBLE,
				       _m_bits & BIT_ISSAVED, this,
				       static_cast<Vec3d*>(subrawdata(index)));
}

// ************************************************************************* //
/// DESCRIPTION: Parent assumed not connected, thus isADuplicate is always
///   true.
///  Adds a double.
// ************************************************************************* //
void	TPlug::addTwistLog(const int	index)
{
  if (!_m_children[index])
    _m_children[index] =
      new PlugTwistLog(_m_dataBlock, index,
		       _m_bits & BIT_ISVISIBLE,
		       _m_bits & BIT_ISSAVED, this,
		       static_cast<TwistLog*>(subrawdata(index)));
}

// ************************************************************************* //
/// DESCRIPTION:
// ************************************************************************* //
void	TPlug::add(const int index)
{
  if (getChildTypeStr() == "Signal")
    addSignal(index);
  else if (getChildTypeStr() == "Bool")
    addBool(index);
  else if (getChildTypeStr() == "Int")
    addInt(index);
  else if (getChildTypeStr() == "Double")
    addDouble(index);
  else if (getChildTypeStr() == "Pnt2d")
    addPnt2d(index);
  else if (getChildTypeStr() == "Pnt3d")
    addPnt3d(index);
  else
    std::cerr << "Error: cannot add subdata to data." << std::endl;
}

// ************************************************************************* //
/// DESCRIPTION: Allocates a new data.
// ************************************************************************* //
void	TPlug::newData()
{
  for (unsigned i(0); i < size(); ++i)
    if (_m_children[i])
      {
	*_m_children[i]->rawdata() = subrawdata(i);
	_m_children[i]->newData();
      }
}

// ************************************************************************* //
/// DESCRIPTION: Keeps only parent.
// ************************************************************************* //
void	TPlug::freeData()
{
  for (unsigned i(0); i < size(); ++i)
    if (_m_children[i])
      {
	_m_children[i]->freeData();
	delete _m_children[i];
      }
  _m_children.clear();
}

// ************************************************************************* //
/// DESCRIPTION: Attaches the data of this to the data of that.
///  Keeps only a pointer, so that data between node is never copied, but 
///  is held by the source node ("this" in this context).
// ************************************************************************* //
void	TPlug::linkdata(Plug &that)
{
  TPlug	&tthat(static_cast<TPlug&>(that));

  Plug::linkdata(that);
  for (unsigned i(0); i < size(); ++i)
    {
      if (!_m_children[i])
	add(i);
      if (!tthat._m_children[i])
	tthat.add(i);
      _m_children[i]->linkdata(*tthat._m_children[i]);
    }
}

// ************************************************************************* //
/// DESCRIPTION: Deselects the whole hierarchy.
// ************************************************************************* //
void	TPlug::deselect()
{
  _m_bits &= ~BIT_ISSELECTED;

  std::map<unsigned, Plug*>::iterator	cit;

  for (cit = _m_children.begin(); cit != _m_children.end(); ++cit)
    if (cit->second)
      cit->second->deselect();
}

// ************************************************************************* //
/// DESCRIPTION:
/// ref2v	- usage: ref2v["virtual" int data].push_back(void **v);
///			 ask to replace rawdata v with data, virtual pointer.
/// v2real	- maps virtual pointers to real pointers.
// ************************************************************************* //
void	TPlug::loadConnections(std::istream			&is,
			       std::map<int, void *>		&v2real,
			       std::map<int, std::list<void **> >	&ref2v)
{
  String	skip;
  int		vadress;

  is >> vadress;
  loadIO(is, ref2v);
  if (vadress)
    if (_m_parent)
      {
	freeData();
	if (hasTopParentInput())
	  ref2v[vadress].push_back(rawdata());
	else
	  {
	    const unsigned	index = item2index(shortName());
      
	    *rawdata() = _m_parent->subrawdata(index);
	    v2real[vadress] = *rawdata();
	  }
      }
    else if (_m_input)
      {
	freeData();
	ref2v[vadress].push_back(rawdata());
	//the data pointer must reference some other data
      }
    else
      {
	v2real[vadress] = *rawdata();//the data is real. Keep it ans pass it on
	//for (unsigned i(0); i < _m_childrensize; )
	//v2real[data] = subrawdata(i);
      }
}

// ************************************************************************* //
/// DESCRIPTION: Saves tree... recursive 0[ 1.1[] 1.2[] 1.3[] ]
/// store size... links
/// If unconnected children, saves "0".
// ************************************************************************* //
void	TPlug::saveAll(std::ostream &os)
{
  os << getTypeStr() << ' ' << int(this) << ' ';
  saveInfo(os);
  if (isSaved())
    saveData(os);
  //propagateDirtyBit();
  saveConnections(os);

  // if not saved, do not save children info!
  if (isSaved())
    {
      os << '{';
      for (unsigned i(0); i < size(); )
	{
	  if (_m_children[i])
	    _m_children[i]->saveAll(os);
	  else
	    os << "0";
	  if (++i != size())
	    os << ',' << std::endl;
	}
      os << '}';
    }
}

// ************************************************************************* //
/// DESCRIPTION:
/// ref2v	- usage: ref2v["virtual" int data].push_back(void **v);
///			 ask to replace rawdata v with data, virtual pointer.
/// v2real	- maps virtual pointers to real pointers.
// ************************************************************************* //
Plug	*TPlug::loadAll(std::istream						&is,
						std::map<int, void *>				&v2real,
						std::map<int, std::list<void **> >	&ref2v)
{
  String	skip;

  loadInfo(is, v2real, ref2v);
  if (isSaved())
    loadData(is);
  loadConnections(is, v2real, ref2v);
  if (isSaved())
    {
		is >> skip;
		while (skip != "}")
		{
			String	typeStr;
			int	adress;

			is >> typeStr;
			if (typeStr == "}")
				break;
			if (typeStr != "0")
			{
				Plug	*new_data = (*Plug::creators()[typeStr])(_m_dataBlock);

				is >> adress;
				v2real[adress] = new_data;
				new_data->_m_parent = this;
				new_data->loadAll(is, v2real, ref2v);

				const String	item(new_data->shortName());
				const unsigned	index = item2index(item);

				//_m_childrensize = max(_m_childrensize, index + 1);
				resize(max(size(), index + 1));
				_m_children[index] = new_data;
				// if not saved ... how can children be saved???
			}
			is >> skip;// EEEK!! /... should read the next!
		}
	}
	return 0;
}

// ************************************************************************* //
/// DESCRIPTION: Returns true if the ith child is selected.
// ************************************************************************* //
bool	TPlug::isSelected(const int i)
{
  if (!isArray())
    {
      std::cerr << "Error: Plug cannot be treated as an array." << std::endl;
      return false;
    }
  if (!_m_children[i])
    add(i);
  return _m_children[i]->isSelected();
}

// ************************************************************************* //
void	TPlug::getPlugByType(const int		type,
			     std::list<Plug *>	&hnd)
{
  if (getTypeNbr() == type)
    hnd.push_back(this);
  for (unsigned i(0); i < size(); ++i)
    {
      if ((type == ID_DOUBLE) && (getTypeStr() == "Pnt2d"))
	add(i);
      if ((type == ID_DOUBLE) && (getTypeStr() == "Pnt3d"))
	add(i);
      if ((type == ID_PNT2D) && (getTypeStr() == "Pnt2dArray"))
	add(i);
      if ((type == ID_PNT3D) && (getTypeStr() == "Pnt3dArray"))
	add(i);
      if ((type == ID_VEC3D) && (getTypeStr() == "Vec3dArray"))
	add(i);
      if ((type == ID_TWISTLOG) && (getTypeStr() == "TwistLogArray"))
	add(i);
      if ((type == ID_PNT2D) && (getTypeStr() == "Pnt2dList"))
	add(i);
      if (_m_children[i])
	_m_children[i]->getPlugByType(type, hnd);
    }
}

// ************************************************************************* //
void	TPlug::push_back(const void *x)
{
  std::list<Pnt2d>	&pl(*static_cast<std::list<Pnt2d>*>(*rawdata()));
  const Pnt2d		d(*static_cast<const Pnt2d*>(x));
  const unsigned	insertindex(size());

  pl.push_back(Pnt2d(0.0));
  _m_children[insertindex] = 0;
  add(insertindex);
  //resize(_m_childrensize + 1);
  _m_children[insertindex]->writeAsPnt2d(d);
  ++static_cast<PlugPnt2dList*>(this)->_m_childrensize;
  //++_m_childrensize;
}

// ************************************************************************* //
/// DESCRIPTION:
// ************************************************************************* //
void	TPlug::push_front(const void *x)
{
  std::list<Pnt2d>	&pl(*static_cast<std::list<Pnt2d>*>(*rawdata()));
  const Pnt2d		d(*static_cast<const Pnt2d*>(x));
  const unsigned	insertindex(0);
  const unsigned	childrensize(size());

  pl.push_front(d);
  for (unsigned i(childrensize); i > insertindex; --i)
    {
      _m_children[i] = _m_children[i - 1];
      _m_children[i]->_m_shortNamei = i;
      _m_children[i]->_m_longNamei = i;
   }
  _m_children[insertindex] = 0;
  add(insertindex);
  //resize(_m_childrensize + 1);
  _m_children[insertindex]->writeAsPnt2d(d);
  ++static_cast<PlugPnt2dList*>(this)->_m_childrensize;
  //++_m_childrensize;
}

// ************************************************************************* //
/// DESCRIPTION: Horror! Shifts the whole thing ... 
///   Inserts x before pos.
// ************************************************************************* //
void	TPlug::insert(const iterator	&pos,
		      const void	*x)
{
  std::list<Pnt2d>	&pl(*static_cast<std::list<Pnt2d>*>(*rawdata()));
  const Pnt2d		d(*static_cast<const Pnt2d*>(x));
  const unsigned	insertindex(pos._m_index);

  pl.insert(pos._m_pit, Pnt2d(0.0));
  for (unsigned i(size()); i > insertindex; --i)
    {
      _m_children[i] = _m_children[i - 1];
      _m_children[i]->_m_shortNamei = i;
      _m_children[i]->_m_longNamei = i;
    }
  _m_children[insertindex] = 0;
  add(insertindex);
  _m_children[insertindex]->writeAsPnt2d(d);
  ++static_cast<PlugPnt2dList*>(this)->_m_childrensize;
  //++_m_childrensize;
}

// ************************************************************************* //
/// DESCRIPTION:
// ************************************************************************* //
void	TPlug::erase(const iterator &pos)
{
  if (size() == 0)
    return;

  std::list<Pnt2d>	&pl(*static_cast<std::list<Pnt2d>*>(*rawdata()));
  const unsigned	eraseindex(pos._m_index);

  pl.erase(pos._m_pit);
  delete _m_children[eraseindex];
  for (unsigned i(eraseindex); i < size() - 1; ++i)
    {
      _m_children[i] = _m_children[i + 1];
      _m_children[i]->_m_shortNamei = i;
      _m_children[i]->_m_longNamei = i;
    }
  _m_children.erase(size() - 1);
  --static_cast<PlugPnt2dList*>(this)->_m_childrensize;
}

// ************************************************************************* //
TPlug::iterator	TPlug::begin()
{
  std::list<Pnt2d>	&pl(*static_cast<std::list<Pnt2d>*>(*rawdata()));

  return iterator(_m_children.begin(), pl.begin(), 0);
}

// ************************************************************************* //
TPlug::iterator	TPlug::end()
{
  std::list<Pnt2d>	&pl(*static_cast<std::list<Pnt2d>*>(*rawdata()));

  return iterator(_m_children.end(), pl.end(), size());
}

// ************************************************************************* //
Plug	*&TPlug::front()
{
  return _m_children[0];
}

// ************************************************************************* //
Plug	*&TPlug::back()
{
  return _m_children[size() - 1];
}
