// ************************************************************************* //
/// AUTHOR Alexis Angelidis
/// STARTED Sat Sep 10 09:40:24 NZST 2005
// ************************************************************************* //
#ifndef __QTRENDEREDITOR_HH__
#define __QTRENDEREDITOR_HH__

// ************************************************************************* //
#include <QGLWidget>
#include <QImage>
#include <qmainwindow.h>

// ************************************************************************* //
class RenderWidget;
class RenderPreferences;
class DagNode;

// ************************************************************************* //
class RenderEditor: public QMainWindow
{
  Q_OBJECT
  RenderWidget	*_m_renderwidget;
  static DagNode	*_m_renderCamera;
public:
  RenderEditor(const QGLContext *shareContext, QWidget *parent = 0);
  static DagNode	&getRenderCamera();
  virtual void		keyPressEvent(QKeyEvent *e);
};

// ************************************************************************* //
class RenderWidget: public QGLWidget
{
  Q_OBJECT
  QImage	_m_image;//image buffer
  bool		_m_renderFrame;
  QGLContext	*_m_context;
  RenderPreferences	*_m_renderpreferences;
  bool		_m_wasCanceled;
protected:
  void	paintEvent(QPaintEvent *e);
public:
  RenderWidget(const QGLContext *shareContext, QWidget *parent = 0);
  //void	makeCurrent() {}
  //void	makeOverlayCurrent() {}
  void	resizeGL(int w, int h);
  void	initializeGL();
  void	paintGL();
  void	makeCurrent();
public slots:
  void	saveImage();
  void	editPreferences();
  void	renderFrame();
  void	renderAnimation();
  void	cancel() { _m_wasCanceled = true; }
  bool	wasCanceled() const { return _m_wasCanceled; }
};

// ************************************************************************* //
class RenderPreferences: public QMainWindow
{
  Q_OBJECT
public:
  std::list<QWidget*>	children;
public:
  RenderPreferences(QWidget *parent = 0);
};

// ************************************************************************* //
class CameraSelector: public QObject
{
  Q_OBJECT
  DagNode	*_m_camera;
public:
  CameraSelector(DagNode &camera);
public slots:
  void	select(bool checked);
};

// ************************************************************************* //
#endif//__QTRENDEREDITOR_HH__
