// ************************************************************************* //
/// AUTHOR: Alexis Angelidis.
/// STARTED:Wed Sep 14 08:42:13 NZST 2005
// ************************************************************************* //
#include "gfxtypes.h"

// ************************************************************************* //
class Ten3d
{
  double	_m_m[27];
public:
  Ten3d() {}
  Ten3d(const double v[27]) { for (unsigned i(0); i < 27; ++i) _m_m[i] = v[i];}
  double	operator()(unsigned i, unsigned j, unsigned k) const;
  double	&operator()(unsigned i, unsigned j, unsigned k);
  static const Ten3d	&permutation();
  Mat3d		operator*(const Vec3d &v) const;
  Ten3d		operator-() const;
};

// ******************************************************************* ***** //
inline Ten3d	Ten3d::operator-() const
{
  Ten3d ret; 

  for (unsigned i(0); i < 27; ++i)
    ret._m_m[i] = -_m_m[i];
  return ret;
}

// ******************************************************************* ***** //
inline double	Ten3d::operator()(unsigned i,
				  unsigned j,
				  unsigned k) const
{
  return _m_m[9 * k + 3 * j + i];
}

// ******************************************************************* ***** //
inline double	&Ten3d::operator()(unsigned i,
				   unsigned j,
				   unsigned k)
{
  return _m_m[9 * k + 3 * j + i];
}

// ******************************************************************* ***** //
inline const Ten3d	&Ten3d::permutation()
{
  static double v[27] = {0.0, 0.0, 0.0, 0.0, 0.0, -1.0, 0.0, 1.0, 0.0,
			 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0,
			 0.0, -1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
  static const Ten3d ret(v);

  return ret;
}

// *************************************** created by Angelidis Alexis ***** //
inline Mat3d	Ten3d::operator*(const Vec3d &v) const
{
  Mat3d	ret(0.0);

  for (int i(0); i < 3; ++i)
    for (int j(0); j < 3; ++j)
      for (int k(0); k < 3; ++k)
	ret(i, j) += operator()(i, j, k) * v(k);
  return ret;
}
