// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:29 $
/// $Revision: 1.1 $
///
/// DESCRIPTION:
/// This file contains all the very basic 3d types:
///  Pnt2i, Pnt2d, Pnt2i, Vec2d, Pnt3i, Vec2i, Vec2d, Vec3i, Pnt3d, Vec3d,
///  Rgb3d, Plane, Frustum, BBox3d, Quaternion, Mat2i, Mat3d, Mat4d, 
///  Translation, Rotation, Scale, ScaleT.
/// As a guideline (not a rule), methods that begin with "eq" assign a new 
/// value to or modify the value of this.
// ************************************************************************* //
#ifndef	__GFXTYPES_HH__
#define	__GFXTYPES_HH__

// ***************************************** created by Alexis Angelidis *** //
extern "C"
{
#include <math.h>
#include <stdlib.h>
#include <GL/gl.h>
}

// ***************************************** created by Alexis Angelidis *** //
#include <iostream>

// *************************************** created by Alexis Angelidis ***** //
#ifndef RAD2DEG
#define RAD2DEG(a) ((a) * 57.2957795130823208768)
#endif//RAD2DEG
#ifndef DEG2RAD
#define DEG2RAD(a) ((a) * 0.01745329251994329576)
#endif//DEG2RAD
#ifndef MAXFLOAT
#define MAXFLOAT 3.40282347e+38F
#endif//MINFLOAT
#ifndef MINFLOAT
#define MINFLOAT -3.40282347e+38F
#endif//MAXFLOAT
#ifndef MINDOUBLE
#define MINDOUBLE -1.7976931e308
#endif//MINDOUBLE
#ifndef MAXDOUBLE
#define MAXDOUBLE 1.7976931e308
#endif//MAXDOUBLE
#ifndef MININT//(1 << (8 * sizeof(int) - 1)) - 1
#define MININT -2147483647
#endif//MININT
#ifndef MAXINT//(1 << (8 * sizeof(int) - 1)) - 1
#define MAXINT 2147483647
#endif//MAXINT

// ************************************************************************* //
#ifndef M_PI
#define M_PI   3.14159265358979323846  /* pi */
#endif//M_PI
#ifndef M_2PI
#define M_2PI	6.28318530717958647692  /* 2 * pi */
#endif//M_2PI

// ************************************************************************* //
#ifndef rand_r
#define RAND_R_MAX 32767

// ************************************************************************* //
//// RAND_MAX assumed to be 32767
// ************************************************************************* //
inline int		rand_r(unsigned int *seed)
{
  *seed = *seed * 1103515245 + 12345;

  return (unsigned)(*seed / 65536) % 32768;
}

// ************************************************************************* //
#else//rand_r
#define RAND_R_MAX RAND_MAX
#endif//rand_r

// ******************************************* coded by Angelidis Alexis *** //
namespace
{

// *************************************** created by Alexis Angelidis ***** //
inline double	myacos(const double costheta)
{
  return (1.0 <= costheta)? 0.0: ((costheta <= -1.0)? M_PI: acos(costheta));
}

// ***************************************** created by Angelidis Alexis *** //
template <class T> inline
T	sq(const T t)
{
  return t * t;
}

// ******************************************* coded by Alexis Angelidis *** //
template <class T> inline
const int	argmax(const T		a,
		       const int	n)
{
  int	ret(0);

  for (int i(1); i < n; ++i)
    if (a[ret] < a[i])
      ret = i;
  return ret;
}

// *************************************** created by Angelidis Alexis ***** //
template <class T> inline
const int	argfabmax(const T	a,
			  const int	n)
{
  int	ret(0);

  for (int i(1); i < n; ++i)
    if (fabs(a[ret]) < fabs(a[i]))
      ret = i;
  return ret;
}

// ***************************************** created by Alexis Angelidis *** //
template <class T> inline
const int	argfabsgnmax(const T	a,
			     const int	n)
{
  int	ret(0);

  for (int i(1); i < n; ++i)
    if (fabs(a[ret]) < fabs(a[i]))
      ret = i;
  return (a[ret] < 0)? -ret: ret;
}

// ***************************************** created by Angelidis Alexis *** //
template <class T> inline
T	min(T a,
	    T b)
{
  return (a < b) ? a: b;
}

// ***************************************** coded by Alexis Angelidis ***** //
template <class T> inline
T	min(T a,
	    T b,
	    T c)
{
  return a < b ? ((a < c)? a: c): ((b < c)? b: c);
}

// ***************************************** created by Angelidis Alexis *** //
template <class T> inline
T	max(T a,
	    T b)
{
  return (a < b) ? b: a;
}

// ******************************************* coded by Alexis Angelidis *** //
template <class T> inline
T	max(T a,
	    T b,
	    T c)
{
  return (a < b) ? ((b < c)? c: b): ((a < c)? c: a);
}

// *************************************** created by Angelidis Alexis ***** //
/// DESCRIPTION: Clamps x in the interval [a, b].
// ***************************************** created by Alexis Angelidis *** //
template <class T> inline
T	bound(const T x,
	      const T a,
	      const T b)
{
  return (x <= a) ? a: ((b <= x)? b: x);
}

// ***************************************** created by Alexis Angelidis *** //
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Classes prototypes.
// *************************************** created by Alexis Angelidis ***** //
struct Vec2i;
struct Vec2d;
struct Vec2ld;
struct Pnt2i;
struct Pnt2d;
struct Pnt2ld;
struct Vec3i;
struct Pnt3i;
struct Vec3d;
struct Vec3ld;
struct Pnt3d;
struct Pnt3ld;
struct Rgb3d;
struct BBox3d;
class Quaternion;
struct Mat2i;
struct Mat3d;
struct Mat4d;
class Transformation;
class Translation;
class Rotation;
class UnitRotation;
class Scale;
class ScaleT;
class Plane;
class Frustum;

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Basic Vec2d class.
// ***************************************** created by Alexis Angelidis *** //
struct Vec2d
{
  double	x, y;
public:
  Vec2d() {}
  Vec2d(const double x_, const double y_): x(x_), y(y_) {}
  explicit Vec2d(const Pnt2d &p);
  Vec2d	operator*(const double d) const
  { return Vec2d(d * x, d * y); }
  Vec2d	operator/(const double d) const
  { return Vec2d(x / d, y / d); }
  double	operator*(const Vec2d &v) const 
  { return x * v.x + y * v.y; }
  Vec2d	operator-(const Vec2d &v) const 
  { return Vec2d(x - v.x, y - v.y); }
  Vec2d	operator+(const Vec2d &v) const
  { return Vec2d(x + v.x, y + v.y); }
  double	operator^(const Vec2d &v) const
  { return x * v.y - y * v.x; }
  Vec2d	&operator*=(const double d) { x *= d; y *= d; return *this; }
  Vec2d	&operator/=(const double d) { x /= d; y /= d; return *this; }
  Vec2d	&operator-=(const Vec2d &v)
  { x -= v.x; y -= v.y; return *this; }
  Vec2d	&operator+=(const Vec2d &v)
  { x += v.x; y += v.y; return *this; }
  Vec2d	operator-() const { return Vec2d(-x, -y); }
  double	norm2() const { return x * x + y * y; }
  double	norm() const { return sqrt(norm2()); }
  double	operator()(const int i) const { return i? y: x; }
  double	&operator()(const int i) { return i? y: x; }
  double	operator[](const int i) const { return i? y: x; }
  double	&operator[](const int i) { return i? y: x; }
  void		eqUnit() { operator*=(1.0 / norm()); }
  Vec2d	unit() const { return operator*(1.0 / norm()); }
  friend std::ostream	&operator<<(std::ostream &os, const Vec2d &v)
  { return os << v.x << " " << v.y; }
  Vec2d	cw90() const { return Vec2d(-y, x); }
  Vec2d	ccw90() const { return Vec2d(y, -x); }
  Vec2d	cw180() const { return Vec2d(-x, -y); }
  Vec2d	ccw180() const { return Vec2d(-x, -y); }
};

// *************************************** created by Angelidis Alexis ***** //
struct Vec2ld
{
  long double	x, y;
public:
  Vec2ld() {}
  Vec2ld(const long double x_, const long double y_): x(x_), y(y_) {}
  Vec2ld	operator-(const Vec2ld &v) const 
  { return Vec2ld(x - v.x, y - v.y); }
  Vec2ld	operator+(const Vec2ld &v) const 
  { return Vec2ld(x + v.x, y + v.y); }
  Vec2ld	operator*(const long double d) const
  { return Vec2ld(x * d, y * d); }
  long double	operator^(const Vec2ld &v) const
  { return x * v.y - y * v.x; }
};

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Basic Pnt2i class.
// ***************************************** coded by Angelidis Alexis ***** //
struct Vec2i
{
  int	x, y;
public:
  Vec2i() {}
  Vec2i(const int d): x(d), y(d) {}
  Vec2i(const int x_, const int y_): x(x_), y(y_) {}
  explicit Vec2i(const Vec2d &v): x(int(v.x)), y(int(v.y)) {}
  Vec2i	operator*(const int d) const
  { return Vec2i(x * d, y * d); }
  Vec2d	operator*(const double d) const
  { return Vec2d(x * d, y * d); }
};

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Basic Pnt2i class.
// ***************************************** created by Angelidis Alexis *** //
struct Pnt2i
{
  int	x, y;
public:
  Pnt2i() {}
  Pnt2i(const int d): x(d), y(d) {}
  Pnt2i(const int x_, const int y_): x(x_), y(y_) {}
  void		glVertex() const { glVertex2i(x, y); }
  void		glTexCoord() const { glTexCoord2i(x, y); }
  int	operator()(const int i) const { return i? y: x; }
  int	&operator()(const int i) { return i? y: x; }
  Vec2i	operator-(const Pnt2i &p) const
  { return Vec2i(x - p.x, y - p.y); }
  Pnt2i	operator-(const Vec2i &v) const
  { return Pnt2i(x - v.x, y - v.y); }
  Pnt2i	operator+(const Vec2i &v) const
  { return Pnt2i(x + v.x, y + v.y); }
  friend std::istream	&operator>>(std::istream &is, Pnt2i &p)
  { return is >> p.x >> p.y; }
  friend std::ostream	&operator<<(std::ostream &os, const Pnt2i &p)
  { return os << p.x << " " << p.y; }
  int	norm2() const { return x * x + y * y; }
};

// ***************************************** coded by Alexis Angelidis ***** //
/// DESCRIPTION: Basic Pnt2d class.
// ***************************************** created by Angelidis Alexis *** //
struct Pnt2d
{
  double	x, y;
public:
  Pnt2d() {}
  Pnt2d(const double d): x(d), y(d) {}
  Pnt2d(const double x_, const double y_): x(x_), y(y_) {}
  explicit Pnt2d(const Pnt2i &p): x(p.x), y(p.y) {}
  explicit Pnt2d(const Pnt3d &p);
  void		glVertex() const { glVertex2d(x, y); }
  void		glTexCoord() const { glTexCoord2d(x, y); }
  Vec2d	operator-(const Pnt2d &p) const { return Vec2d(x - p.x, y - p.y); }
  Pnt2d	operator+(const Vec2d &v) const { return Pnt2d(x + v.x, y + v.y); }
  Pnt2d	operator+(const Pnt2d &p) const { return Pnt2d(x + p.x, y + p.y); }
  Pnt2d	operator*(const double d) const { return Pnt2d(x * d, y * d); }
  Pnt2d	operator-(const Vec2d &v) const { return Pnt2d(x - v.x, y - v.y); }
  Pnt2d	&operator-=(const Vec2d &v) { x -= v.x; y -= v.y; return *this; }
  Pnt2d	&operator+=(const Vec2d &v) { x += v.x; y += v.y; return *this; }
  Pnt2d	&operator*=(const double d) { x *= d; y *= d; return *this; }
  Pnt2d	&operator/=(const double d) { x /= d; y /= d; return *this; }
  bool	operator==(const Pnt2d &p) const { return x == p.x && y == p.y; }
  bool	operator!=(const Pnt2d &p) const { return x != p.x || y != p.y; }
  double	operator()(const int i) const { return i? y: x; }
  double	&operator()(const int i) { return i? y: x; }
  friend std::istream	&operator>>(std::istream &is, Pnt2d &p)
  { return is >> p.x >> p.y; }
  friend std::ostream	&operator<<(std::ostream &os, const Pnt2d &p)
  { return os << p.x << " " << p.y; }
  double	norm() const { return sqrt(x * x + y * y); }
  double	norm2() const { return x * x + y * y; }
  double	angle() const { return atan2(y, x); }
  Pnt2d	proj2line(const Pnt2d &a, const Pnt2d &b) const;
};

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Pnt2ld class, long double.
// ***************************************** coded by Angelidis Alexis ***** //
struct Pnt2ld
{
  long double	x, y;
public:
  Pnt2ld() {}
  Pnt2ld(const long double d): x(d), y(d) {}
  Pnt2ld(const long double x_, const long double y_): x(x_), y(y_) {}
  operator Pnt2d() const { return Pnt2d(x, y); }
  Vec2ld	operator-(const Pnt2ld &p) const
  { return Vec2ld(x - p.x, y - p.y); }
  Pnt2ld	operator+(const Pnt2ld &p) const
  { return Pnt2ld(x + p.x, y + p.y); }
  Pnt2ld	operator+(const Vec2ld &v) const
  { return Pnt2ld(x + v.x, y + v.y); }
  Pnt2ld	operator*(const long double d) const
  { return Pnt2ld(x * d, y * d); }
  friend std::ostream	&operator<<(std::ostream &os, const Pnt2ld &p)
  { return os << p.x << " " << p.y; }
};

// ***************************************** created by Alexis Angelidis *** //
inline Vec2d::Vec2d(const Pnt2d &p): x(p.x), y(p.y)
{
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Projects point on line (a, b).
// *************************************** created by Alexis Angelidis ***** //
inline Pnt2d	Pnt2d::proj2line(const Pnt2d	&a,
				 const Pnt2d	&b) const
{
  const Vec2d	m(b - a);

  return a + m * (((*this - a) * m) / (a - b).norm2());
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Basic Vec3d class.
// ***************************************** created by Alexis Angelidis *** //
struct Vec3d
{
  double	x, y, z;
public:
  Vec3d() {}
  Vec3d(const double d): x(d), y(d), z(d) {}
  Vec3d(const double x_, const double y_, const double z_):
    x(x_), y(y_), z(z_) {}
  explicit Vec3d(const Pnt3d &p);
  Vec3d	&operator+=(const Vec3d &v)
  { x += v.x; y += v.y; z += v.z; return *this; }
  Vec3d	&operator-=(const Vec3d &v)
  { x -= v.x; y -= v.y; z -= v.z; return *this; }
  Vec3d	&operator*=(const double d)
  { x *= d; y *= d; z *= d; return *this; }
  Vec3d	&operator/=(const double d)
  { x /= d; y /= d; z /= d; return *this; }
  friend Vec3d	operator*(const double d, const Vec3d &v)
  { return Vec3d(d * v.x, d * v.y, d * v.z); }
  Vec3d	operator+(const Vec3d &v) const
  { return Vec3d(x + v.x, y + v.y, z + v.z); }
  Vec3d	operator-(const Vec3d &v) const
  { return Vec3d(x - v.x, y - v.y, z - v.z); }
  double	operator*(const Vec3d &v) const
  { return x * v.x + y * v.y + z * v.z; }
  Vec3d	operator/(const double d) const
  { return Vec3d(x / d, y / d, z / d); }
  Vec3d	operator*(const double d) const
  { return Vec3d(x * d, y * d, z * d); }
  double &operator()(const unsigned i) { return i? ((i == 2)? z: y): x; }
  double operator()(const unsigned i) const { return i? ((i == 2)? z: y): x; }
  double &operator[](const unsigned i) { return i? ((i == 2)? z: y): x; }
  double operator[](const unsigned i) const { return i? ((i == 2)? z: y): x; }
  friend std::istream	&operator>> (std::istream &is, Vec3d &v);
  friend std::ostream	&operator<<(std::ostream &os, const Vec3d &v);
  void		eqJitter(const double max);
  Vec3d	jitter(const double max) const;
  Vec3d	jitter(const double max, unsigned *seedp) const;
  Vec3d	operator-() const { return Vec3d(-x, -y, -z); }
  Vec3d	&operator*=(const Mat4d &m);
  Vec3d	&operator*=(const Mat3d &m);
  bool	operator!=(const Vec3d &v) const
  { return v.x != x || v.y != y || v.z != z; }
  double	norm() const { return sqrt(x * x + y * y + z * z); }
  double	norm2() const { return x * x + y * y + z * z; }
  void		eqUnit();
  Vec3d	unit() const;
  Vec3d	operator^(const Vec3d &v) const
  { return Vec3d(y * v.z - z * v.y, z * v.x - x * v.z, x * v.y - y * v.x); }
  void		glVertex() const { glVertex3d(x, y, z); }
  void		glNormal() const { glNormal3d(x, y, z); }
  void		glColor() const { glColor3d(x, y, z); }
  void		glTexCoord(int u, int v) const
  { glTexCoord2d(operator()(u), operator()(v)); }
  void		eqSymx() { x = -x; }
  void		eqSymy() { y = -y; }
  void		eqSymz() { z = -z; }
  void		eqSym(const unsigned sym);
  Vec3d	symx() const { return Vec3d(-x, y, z); }
  Vec3d	symy() const { return Vec3d(x, -y, z); }
  Vec3d  	symz() const { return Vec3d(x, y, -z); }
  Vec3d	sym(const unsigned sym) const;
  void	getOrthoPair(Vec3d &u, Vec3d &v) const;
  void	glGetUp();
  void	glGetViewdir();
};

// *************************************** created by Angelidis Alexis ***** //
inline void	Vec3d::eqUnit()
{
  const double n2 = norm2();

  if (n2 != 0.0)
    operator*=(1.0 / sqrt(n2));
}

// *************************************** created by Angelidis Alexis ***** //
inline Vec3d	Vec3d::unit() const
{
  const double n2 = norm2();

  if (n2 != 0.0)
    return operator*(1.0 / sqrt(n2));
  return *this;
}

// *************************************** created by Angelidis Alexis ***** //
struct Vec3ld
{
  long double	x, y, z;
public:
  Vec3ld() {}
  Vec3ld(const long double d): x(d), y(d), z(d) {}
  Vec3ld(const long double x_, const long double y_, const long double z_):
    x(x_), y(y_), z(z_) {}
  explicit Vec3ld(const Vec3d &v): x(v.x), y(v.y), z(v.z) {}
  operator Vec3d() const { return Vec3d(x, y, z); }
  Vec3ld	operator+(const Vec3ld &v) const
  { return Vec3ld(x + v.x, y + v.y, z + v.z); }
  Vec3ld	operator-(const Vec3ld &v) const
  { return Vec3ld(x - v.x, y - v.y, z - v.z); }
  Vec3ld	operator/(const long double d) const
  { return Vec3ld(x / d, y / d, z / d); }
  Vec3ld	operator*(const long double d) const
  { return Vec3ld(x * d, y * d, z * d); }
  long double	operator*(const Vec3ld &v) const
  { return x * v.x + y * v.y + z * v.z; }
  Vec3ld	&operator*=(const long double d)
  { x *= d; y *= d; z *= d; return *this; }
  Vec3ld	operator^(const Vec3ld &v) const
  { return Vec3ld(y * v.z - z * v.y, z * v.x - x * v.z,
		     x * v.y - y * v.x); }
  long double	norm() const { return sqrtl(x * x + y * y + z * z); }
  long double	norm2() const { return x * x + y * y + z * z; }
  void		eqUnit() { operator*=(1.0l / norm()); }
  Vec3ld	unit() const { return operator/(norm()); }
  long double	operator()(const unsigned i) const
  { return i? ((i == 2)? z: y): x; }
  long double	&operator()(const unsigned i)
  { return i? ((i == 2)? z: y): x; }
  friend std::ostream	&operator<<(std::ostream &os, const Vec3ld &p);
};

// ******************************************* coded by Alexis Angelidis *** //
inline std::ostream	&operator<<(std::ostream	&os,
				    const Vec3ld	&v)
{
  return os << v.x << " "  << v.y << " "  << v.z;
  /*os.write(reinterpret_cast<const char *>(&v.x), 8);
  os.write(reinterpret_cast<const char *>(&v.y), 8);
  os.write(reinterpret_cast<const char *>(&v.z), 8);
  return os;*/
}

// ***************************************** created by Angelidis Alexis *** //
/// DESCRIPTION: Creates a pair of orthogonal vectors.
// ***************************************** created by Alexis Angelidis *** //
inline void	Vec3d::getOrthoPair(Vec3d &u,
				    Vec3d &v) const
{
  if (fabs(x) < fabs(y) && fabs(x) < fabs(z))
    u = Vec3d(1.0, 0.0, 0.0);
  else if (fabs(y) < fabs(z))
    u = Vec3d(0.0, 1.0, 0.0);
  else
    u = Vec3d(0.0, 0.0, 1.0);
  
  u = (u - ((u * *this) * *this) / norm2()).unit();
  v = operator^(u).unit();
}

// ***************************************** created by Angelidis Alexis *** //
/// DESCRIPTION: assign to the point one of its symmetries.
// ***************************************** coded by Alexis Angelidis ***** //
inline void	Vec3d::eqSym(const unsigned sym)
{
  if (sym == 1)
    eqSymx();
  else if (sym == 2)
    eqSymy();
  else if (sym == 3)
    eqSymz();
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Returns symmetries of the point.
// ***************************************** created by Angelidis Alexis *** //
inline Vec3d	Vec3d::sym(const unsigned sym) const
{
  if (sym == 1)
    return symx();
  else if (sym == 2)
    return symy();
  else if (sym == 3)
    return symz();
  return *this;
}

// ***************************************** coded by Alexis Angelidis ***** //
inline std::istream	&operator>>(std::istream	&is,
				    Vec3d		&v)
{
  return is >> v.x >> v.y >> v.z;
  /* is.read(reinterpret_cast<char *>(&v.x), 8);
  is.read(reinterpret_cast<char *>(&v.y), 8);
  is.read(reinterpret_cast<char *>(&v.z), 8);
  return is;*/
}

// ***************************************** created by Angelidis Alexis *** //
inline std::ostream	&operator<<(std::ostream	&os,
				    const Vec3d		&v)
{
  return os << v.x << " "  << v.y << " "  << v.z;
  /*os.write(reinterpret_cast<const char *>(&v.x), 8);
  os.write(reinterpret_cast<const char *>(&v.y), 8);
  os.write(reinterpret_cast<const char *>(&v.z), 8);
  return os;*/
}

// *************************************** created by Alexis Angelidis ***** //
inline void	Vec3d::eqJitter(const double max)
{
  const double	scale(max / RAND_MAX);
  const double	trans(max * 0.5);

  x += (scale * rand()) - trans;
  y += (scale * rand()) - trans;
  z += (scale * rand()) - trans;
}

// ******************************************* coded by Angelidis Alexis *** //
inline Vec3d	Vec3d::jitter(const double max) const
{
  const double	scale(max / RAND_MAX);
  const double	trans(max * 0.5);

  return Vec3d(x + (scale * rand()) - trans,
	       y + (scale * rand()) - trans,
	       z + (scale * rand()) - trans);
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION:
///  The method jitter() is not reentrant or thread-safe, since it uses hid-
///  den state that is modified on each call. This might just  be  the  seed
///  value to be used by the next call, or it might be something more elabo-
///  rate. In order to get reproducible behaviour in a threaded application,
///  this  state  must  be  made explicit. The function jitter() can be used
///  with a pointer to an unsigned int, to be used as state.
// *************************************** created by Alexis Angelidis ***** //
inline Vec3d	Vec3d::jitter(const double	max,
			      unsigned		*seedp) const
{
  const double	scale(max / RAND_R_MAX);
  const double	trans(max * 0.5);

  return Vec3d(x + (scale * rand_r(seedp)) - trans,
			   y + (scale * rand_r(seedp)) - trans,
			   z + (scale * rand_r(seedp)) - trans);
}

// *************************************** created by Alexis Angelidis ***** //
/// DESCRIPTION: Basic Pnt3d class.
// ***************************************** created by Angelidis Alexis *** //
struct Pnt3d
{
  double	x, y, z;
public:
  Pnt3d() {}
  Pnt3d(const double d): x(d), y(d), z(d) {}
  Pnt3d(const double x_, const double y_, const double z_):
    x(x_), y(y_), z(z_) {}
  explicit Pnt3d(const Vec3d &v): x(v.x), y(v.y), z(v.z) {}
  explicit Pnt3d(const Pnt2d &p): x(p.x), y(p.y), z(0.0) {}
  explicit Pnt3d(const Pnt3i &p);
  void		glVertex() const { glVertex3d(x, y, z); }
  void		glColor() const { glColor3d(x, y, z); }
  double	&operator()(const unsigned i)
  { return i? ((i == 2)? z: y): x; }
  double	operator()(const unsigned i) const
  { return i? ((i == 2)? z: y): x; }
  double	&operator[](const unsigned i)
  { return i? ((i == 2)? z: y): x; }
  double	operator[](const unsigned i) const
  { return i? ((i == 2)? z: y): x; }
  Vec3d	operator-(const Pnt3d &p) const
  { return Vec3d(x - p.x, y - p.y, z - p.z); }
  Pnt3d	operator+(const Pnt3d &p) const
  { return Pnt3d(x + p.x, y + p.y, z + p.z); }
  Pnt3d	operator+(const Vec3d &v) const
  { return Pnt3d(x + v.x, y + v.y, z + v.z); }
  Pnt3d	operator-(const Vec3d &v) const
  { return Pnt3d(x - v.x, y - v.y, z - v.z); }
  Pnt3d	operator*(const double d) const
  { return Pnt3d(x * d, y * d, z * d); }
  Pnt3d	operator/(const double d) const
  { return Pnt3d(x / d, y / d, z / d); }
  Pnt3d	&operator-=(const Vec3d &v)
  { x -= v.x; y -= v.y; z -= v.z; return *this; }
  Pnt3d	&operator+=(const Vec3d &v)
  { x += v.x; y += v.y; z += v.z; return *this; }
  Pnt3d	&operator+=(const Pnt3d &p)
  { x += p.x; y += p.y; z += p.z; return *this; }
  Pnt3d	&operator*=(const double d)
  { x *= d; y *= d; z *= d; return *this; }
  Pnt3d	&operator/=(const double d)
  { x /= d; y /= d; z /= d; return *this; }
  Pnt3d	&operator*=(const Mat4d &m);
  bool	operator!=(const Pnt3d &p) const
  { return p.x != x || p.y != y || p.z != z; }
  bool	operator==(const Pnt3d &p) const
  { return x == p.x && y == p.y && p.z == z; }
  Pnt3d	operator-() const { return Pnt3d(-x, -y, -z); }
  friend std::istream	&operator>>(std::istream &is, Pnt3d &p);
  friend std::ostream	&operator<<(std::ostream &os, const Pnt3d &p);
  void		eqJitter(const double max);
  Pnt3d	jitter(const double max) const;
  Pnt3d	jitter(const double max, unsigned *seedp) const;
  Pnt3d	sphJitter(const double max) const;
  void		glTranslate() const { glTranslated(x, y, z); }
  void		glGetViewpoint();
  void		eqSphereIntersection(const Pnt3d &s, const Vec3d &d,
				     const Pnt3d &c, const double r);
  Pnt3d	proj2planeAlong(const Pnt3d &o, const Vec3d &n,
			const Vec3d &dir) const;//Deprecated
  Pnt3d	proj2line(const Pnt3d &a, const Pnt3d &b) const;
  void		eqProjection(const Plane &p);
  void		eqProjection(const Plane &p, const Vec3d &dir);
  void		eqSymx() { x = -x; }
  void		eqSymy() { y = -y; }
  void		eqSymz() { z = -z; }
  void		eqSym(const unsigned sym);
  void		eqSym(const Plane &p);
  Pnt3d	symx() const { return Pnt3d(-x, y, z); }
  Pnt3d	symy() const { return Pnt3d(x, -y, z); }
  Pnt3d  	symz() const { return Pnt3d(x, y, -z); }
  Pnt3d	sym(const unsigned sym) const;
  double	dist2edge(const Pnt3d &p0, const Pnt3d &p1) const;
};

// ***************************************** coded by Alexis Angelidis ***** //
struct Pnt3ld
{
  long double	x, y, z;
public:
  Pnt3ld() {}
  Pnt3ld(const long double d): x(d), y(d), z(d) {}
  Pnt3ld(const long double x_, const long double y_, const long double z_):
    x(x_), y(y_), z(z_) {}
  Pnt3ld(const Vec3ld &v): x(v.x), y(v.y), z(v.z) {}
  explicit Pnt3ld(const Pnt3d &p): x(p.x), y(p.y), z(p.z) {}
  operator Pnt3d() const { return Pnt3d(x, y, z); }
  operator Vec3ld() const { return Vec3ld(x, y, z); }
  Pnt3ld	operator+(const Pnt3ld &p) const
  { return Pnt3ld(x + p.x, y + p.y, z + p.z); }
  Pnt3ld	operator+(const Vec3ld &v) const
  { return Pnt3ld(x + v.x, y + v.y, z + v.z); }
  Vec3ld	operator-(const Pnt3ld &p) const
  { return Vec3ld(x - p.x, y - p.y, z - p.z); }
  Pnt3ld	operator*(const long double d) const
  { return Pnt3ld(x * d, y * d, z * d); }
};

// ***************************************** created by Angelidis Alexis *** //
/// DESCRIPTION: Squared distance from point to the edge (p0,p1)
// ***************************************** created by Alexis Angelidis *** //
inline double	Pnt3d::dist2edge(const Pnt3d	&p0,
				 const Pnt3d	&p1) const
{
  const Pnt3d	&p(*this);
  const Vec3d	n(p1 - p0);
  const double	n2(n.norm2());

  if (fabs(n2) < 1.0e-12)
    return (p - p0).norm2();

  const Pnt3d	h(p - n * (((p - p0) * n) / n2));
  const double	tn2 = (h - p0) * n;

  if (tn2 <= 0.0)
    return (p - p0).norm2();
  if (n2 <= tn2)
    return (p - p1).norm2();
  return (p - h).norm2();
}

// ***************************************** created by Angelidis Alexis *** //
/// DESCRIPTION: assign to the point one of its symmetries.
// ***************************************** created by Alexis Angelidis *** //
inline void	Pnt3d::eqSym(const unsigned sym)
{
  if (sym == 1)
    eqSymx();
  else if (sym == 2)
    eqSymy();
  else if (sym == 3)
    eqSymz();
}

// ***************************************** coded by Angelidis Alexis ***** //
/// DESCRIPTION: Returns symmetries of the point.
// ***************************************** created by Alexis Angelidis *** //
inline Pnt3d	Pnt3d::sym(const unsigned sym) const
{
  if (sym == 1)
    return symx();
  else if (sym == 2)
    return symy();
  else if (sym == 3)
    return symz();
  return *this;
}

// ******************************************* coded by Angelidis Alexis *** //
inline std::istream	&operator>>(std::istream	&is,
				    Pnt3d		&p)
{ 
  return is >> p.x >> p.y >> p.z;
  /*is.read(reinterpret_cast<char *>(&p.x), 8);
  is.read(reinterpret_cast<char *>(&p.y), 8);
  is.read(reinterpret_cast<char *>(&p.z), 8);
  return is;*/
}

// *************************************** created by Alexis Angelidis ***** //
inline std::ostream	&operator<<(std::ostream	&os,
				    const Pnt3d		&p)
{
  return os << p.x << " "  << p.y << " "  << p.z;
  /*os.write(reinterpret_cast<const char *>(&p.x), 8);
  os.write(reinterpret_cast<const char *>(&p.y), 8);
  os.write(reinterpret_cast<const char *>(&p.z), 8);
  return os;*/
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Projects point on plane defined by point o and normal n,
///   along vector dir.
/// Deprecated.
// ***************************************** created by Alexis Angelidis *** //
inline Pnt3d	Pnt3d::proj2planeAlong(const Pnt3d	&o,
				       const Vec3d	&n,
				       const Vec3d	&dir) const
{
  return *this - dir * (((*this - o) * n) / (dir * n));
}

// *************************************** created by Angelidis Alexis ***** //
inline Pnt3d	Pnt3d::proj2line(const Pnt3d &a,
				 const Pnt3d &b) const
{
  const Vec3d	m(b - a);

  return a + m * (((*this - a) * m) / (a - b).norm2());
}

// *************************************** created by Angelidis Alexis ***** //
inline Pnt2d::Pnt2d(const Pnt3d &p): x(p.x), y(p.y)
{
}

// ******************************************* coded by Alexis Angelidis *** //
inline void	Pnt3d::eqSphereIntersection(const Pnt3d		&s,
					    const Vec3d		&d,
					    const Pnt3d		&center,
					    const double	r)
{
  Vec3d	s2(s - center);
  const double	a(sq(d(0)) + sq(d(1)) + sq(d(2)));
  const double	b(2.0 * (s2(0) * d(0) + s2(1) * d(1) + s2(2) * d(2)));
  const double	c(sq(s2(0)) + sq(s2(1)) + sq(s2(2)) - sq(r));
  const double	delta(sq(b) - 4.0 * a * c);
  const double	t1((-b + sqrt(delta)) / (2.0 * a));
  const double	t2((-b - sqrt(delta)) / (2.0 * a));
  const double	t(min(t1, t2));

  *this = s + t * d;
}

// ***************************************** created by Angelidis Alexis *** //
inline void	Pnt3d::eqJitter(const double max)
{
  const double	scale(max / RAND_MAX);
  const double	trans(max * 0.5);

  x += (scale * rand()) - trans;
  y += (scale * rand()) - trans;
  z += (scale * rand()) - trans;
}

// *************************************** created by Alexis Angelidis ***** //
/// DESCRIPTION: Jitters the point in a neighboorhood of diameter max.
// ******************************************* coded by Angelidis Alexis *** //
inline Pnt3d	Pnt3d::jitter(const double max) const
{
  const double	scale(max / RAND_MAX);
  const double	trans(max * 0.5);

  return Pnt3d(x + (scale * rand()) - trans,
		 y + (scale * rand()) - trans,
		 z + (scale * rand()) - trans);
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION:
///  The method jitter() is not reentrant or thread-safe, since it uses hid-
///  den state that is modified on each call. This might just  be  the  seed
///  value to be used by the next call, or it might be something more elabo-
///  rate. In order to get reproducible behaviour in a threaded application,
///  this  state  must  be  made explicit. The function jitter() can be used
///  with a pointer to an unsigned int, to be used as state.
// *************************************** created by Alexis Angelidis ***** //
inline Pnt3d	Pnt3d::jitter(const double	max,
			      unsigned		*seedp) const
{
  const double	scale(max / RAND_R_MAX);
  const double	trans(max * 0.5);

  return Pnt3d(x + (scale * rand_r(seedp)) - trans,
			   y + (scale * rand_r(seedp)) - trans,
			   z + (scale * rand_r(seedp)) - trans);
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Jitters the point in a sph neighboorhood of diameter max.
// ******************************************* coded by Angelidis Alexis *** //
inline Pnt3d	Pnt3d::sphJitter(const double max) const
{
  const double	scale(max / RAND_MAX);
  const double	trans(max * 0.5);

  while (1)
    {
      const Vec3d ret(scale * rand() - trans,
		      scale * rand() - trans,
		      scale * rand() - trans);
      if (ret.norm2() < max * max * 0.25)
	return operator+(ret);
    }
}

// ***************************************** created by Alexis Angelidis *** //
inline Vec3d::Vec3d(const Pnt3d &p): x(p.x), y(p.y), z(p.z)
{
}

// ***************************************** created by Angelidis Alexis *** //
/// DESCRIPTION: Rgb color, in [0,1]^3
// ***************************************** created by Angelidis Alexis *** //
struct Rgb3d
{
  double	r, g, b;
public:
  Rgb3d(const double r_, const double g_, const double b_):
    r(r_), g(g_), b(b_) {}
  Rgb3d(const double c): r(c), g(c), b(c) {}
  Rgb3d() {}
  double	&operator()(const unsigned i)
  { return i? ((i == 2)? b: g): r; }
  double	operator()(const unsigned i) const
  { return i? ((i == 2)? b: g): r; }
  double	&operator[](const unsigned i)
  { return i? ((i == 2)? b: g): r; }
  double	operator[](const unsigned i) const
  { return i? ((i == 2)? b: g): r; }
  bool	operator!=(const Rgb3d &p) const
  { return p.r != r || p.g != g || p.b != b; }
  bool	operator==(const Rgb3d &p) const
  { return r == p.r && g == p.g && p.b == b; }
  Rgb3d	operator+(const Rgb3d &p) const
  { return Rgb3d(r + p.r, g + p.g, b + p.b); }
  Rgb3d	operator-(const Rgb3d &p) const
  { return Rgb3d(r - p.r, g - p.g, b - p.b); }
  Rgb3d	operator*(const long double d) const
  { return Rgb3d(r * d, g * d, b * d); }
  void	glColor() const { glColor3d(r, g, b); }
  friend std::istream	&operator>>(std::istream &is, Rgb3d &p);
  friend std::ostream	&operator<<(std::ostream &os, const Rgb3d &p);
};

// ******************************************* coded by Angelidis Alexis *** //
inline std::istream	&operator>>(std::istream	&is,
				    Rgb3d		&p)
{ 
  return is >> p.r >> p.g >> p.b;
  /*is.read(reinterpret_cast<char *>(&p.x), 8);
  is.read(reinterpret_cast<char *>(&p.y), 8);
  is.read(reinterpret_cast<char *>(&p.z), 8);
  return is;*/
}

// *************************************** created by Alexis Angelidis ***** //
inline std::ostream	&operator<<(std::ostream	&os,
				    const Rgb3d		&p)
{
  return os << p.r << " "  << p.g << " "  << p.b;
  /*os.write(reinterpret_cast<const char *>(&p.x), 8);
  os.write(reinterpret_cast<const char *>(&p.y), 8);
  os.write(reinterpret_cast<const char *>(&p.z), 8);
  return os;*/
}

// ***************************************** created by Angelidis Alexis *** //
/// DESCRIPTION: 3D plane. Equation is v * p + t = 0.
///  The inside is the half space v * p + t > 0 i.e.
///  Attribute _m_v points inside, and is unit.
// ******************************************* coded by Alexis Angelidis *** //
class Plane
{
  friend class Mat4d;
  friend class Pnt3d;
  Vec3d		_m_v;
  double	_m_t;
public:
  Plane() {}
  Plane(const Vec3d &v, const double t): _m_v(v), _m_t(t) {}
  Plane(const Vec3d &v, const Pnt3d &c);
  static bool	intersect(const Plane &p0, const Plane &p1,
			  const Plane &p2, Pnt3d &ret);
  const Vec3d	&getv() const { return _m_v; }
  double		gett() const { return _m_t; }
  Pnt3d		getc() const { return Pnt3d(_m_v * (-_m_t)); }
  static const Plane	&x();
  static const Plane	&y();
  static const Plane	&z();
  bool	contains(const Pnt3d &p, const double epsilon = 1.0e-8) const;
  void	projectAlong(const Vec3d &v, Pnt3d &p) const;
  void	glRender() const;
  double	dist(const Pnt3d &p) const;
  friend std::ostream	&operator<<(std::ostream &os, const Plane &p)
  { return os << p._m_v << ' ' << p._m_t; }
  friend std::istream	&operator>>(std::istream &is, Plane &p)
  { return is >> p._m_v >> p._m_t; }
};

// *************************************** created by Angelidis Alexis ***** //
/// DESCRIPTION: Distance from point to plane.
/// TODO: turn in a function with 2 params.
// ******************************************* coded by Alexis Angelidis *** //
inline double	Plane::dist(const Pnt3d &p) const
{
  const double	d(_m_v * Vec3d(p) + _m_t);

  return 0.0 <= d? 0.0: -d;
}

// *************************************** created by Angelidis Alexis ***** //
/// DESCRIPTION: Projects point on plane defined by point o and normal n,
///   along vector dir.
// ***************************************** created by Alexis Angelidis *** //
inline void	Pnt3d::eqProjection(const Plane	&p,
				    const Vec3d	&dir)
{
  operator-=(dir * (Vec3d(*this) * p._m_v + p._m_t) / (p._m_v * dir));
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Projects point on plane defined by point o and normal n.
// *************************************** created by Alexis Angelidis ***** //
inline void	Pnt3d::eqProjection(const Plane	&p)
{
  operator-=(p._m_v * (Vec3d(*this) * p._m_v + p._m_t));
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: assign to the point its symmetry about the plane p.
// *************************************** created by Alexis Angelidis ***** //
inline void	Pnt3d::eqSym(const Plane &p)
{
  operator-=(p._m_v * (2.0 * (Vec3d(*this) * p._m_v + p._m_t)));
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Sets a plane, with v pointing outside, and point c on
///  the plane.
// *************************************** created by Alexis Angelidis ***** //
inline Plane::Plane(const Vec3d	&v,
		    const Pnt3d	&c): _m_v(-v), _m_t(Vec3d(c) * v)
{
}

// ***************************************** created by Angelidis Alexis *** //
/// DESCRIPTION:
// *************************************** created by Alexis Angelidis ***** //
inline void	Plane::glRender() const
{
  Vec3d	x, y;

  if (fabs(_m_v.x) < fabs(_m_v.y) && fabs(_m_v.x) < fabs(_m_v.z))
    y = Vec3d(1.0, 0.0, 0.0);
  else if (fabs(_m_v.y) < fabs(_m_v.z))
    y = Vec3d(0.0, 1.0, 0.0);
  else
    y = Vec3d(0.0, 0.0, 1.0);
  y = (y - _m_v * (y * _m_v)).unit();
  x = y ^ _m_v;
  glBegin(GL_QUADS);
  (x + y - _m_v * _m_t).glVertex();
  (x - y - _m_v * _m_t).glVertex();
  (-x - y - _m_v * _m_t).glVertex();
  (y - x - _m_v * _m_t).glVertex();
  glEnd();//GL_QUADS
}

// ******************************************* coded by Angelidis Alexis *** //
inline const Plane	&Plane::x()
{
  static const Plane px(Vec3d(1.0, 0.0, 0.0), 0.0);

  return px;
}

// ***************************************** created by Alexis Angelidis *** //
inline const Plane	&Plane::y()
{
  static const Plane py(Vec3d(0.0, 1.0, 0.0), 0.0);

  return py;
}

// ***************************************** coded by Angelidis Alexis ***** //
inline const Plane	&Plane::z()
{
  static const Plane pz(Vec3d(0.0, 0.0, 1.0), 0.0);

  return pz;
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Returns true if the plane contains p, within tolerance 
///   epsilon.
/// NOTES: This method works if the planes are not normalized.
// ***************************************** coded by Angelidis Alexis ***** //
inline bool	Plane::contains(const Pnt3d	&p,
				const double	epsilon) const
{
  return -epsilon < Vec3d(p) * _m_v + _m_t;
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Projects p on plane following v.
// ******************************************* coded by Angelidis Alexis *** //
inline void	Plane::projectAlong(const Vec3d	&n,
				    Pnt3d		&p) const
{
  p -= n * (Vec3d(p) * _m_v + _m_t) / (_m_v * n);
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Clipping planes class.
// *************************************** created by Angelidis Alexis ***** //
class Frustum
{
  Plane	_m_clipPlanes[6];
public:
  Frustum() {}
  void	getWorldFrustum();
  void	getCameraFrustum();
  const Plane	&getLeftPlane() const { return _m_clipPlanes[0]; }
  const Plane	&getRightPlane() const { return _m_clipPlanes[1]; }
  const Plane	&getBottomPlane() const { return _m_clipPlanes[2]; }
  const Plane	&getTopPlane() const { return _m_clipPlanes[3]; }
  const Plane	&getNearPlane() const { return _m_clipPlanes[4]; }
  const Plane	&getFarPlane() const { return _m_clipPlanes[5]; }
  bool	contains(const Pnt3d &p, const double epsilon = 1.0e-8) const;
  bool	intersects(const BBox3d &bbox) const;
  void	computez(Pnt3d  planez[4]) const;
  int	intersect(const Plane &p, Pnt3d  planez[6]) const;
  void	corners(Pnt3d c[8]) const;
  void	nearCorners(Pnt3d c[4]) const;
  void	farCorners(Pnt3d c[4]) const;
  Pnt3d	eye() const;
  int	getHorizon(Pnt3d h[2]) const;
  friend std::ostream	&operator<<(std::ostream &os, const Frustum &fs)
  { for (int i(0); i < 6; ++i) os << fs._m_clipPlanes[i] << ' '; return os; }
  friend std::istream	&operator>>(std::istream &is, Frustum &fs)
  { for (int i(0); i < 6; ++i) is >> fs._m_clipPlanes[i]; return is; }
};

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Returns true if the clip planes contain p.
///  Epsilon can be provided as a safety threshold.
// ***************************************** coded by Angelidis Alexis ***** //
inline bool	Frustum::contains(const Pnt3d	&p,
				  const double	epsilon) const
{
  for (int ip(0); ip < 6; ++ip)
    if (!_m_clipPlanes[ip].contains(p, epsilon))
      return false;
  return true;
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Returns the intersection points between the cliping
///  planes and the plane z = 0:  1----3
///                               |    |
///                               0----2
/// TODO: Marching cube-type case detection.
// ***************************************** coded by Angelidis Alexis ***** //
inline void	Frustum::computez(Pnt3d	planez[4]) const
{
  Plane::intersect(getLeftPlane(), getBottomPlane(), Plane::z(), planez[0]);
  Plane::intersect(getLeftPlane(), getFarPlane(), Plane::z(), planez[1]);
  if (!contains(planez[1]))
    Plane::intersect(getLeftPlane(), getTopPlane(), Plane::z(), planez[1]);
  Plane::intersect(getRightPlane(), getBottomPlane(), Plane::z(), planez[2]);
  Plane::intersect(getRightPlane(), getFarPlane(), Plane::z(), planez[3]);
  if (!contains(planez[3]))
    Plane::intersect(getRightPlane(), getTopPlane(), Plane::z(), planez[3]);
}

// ************************************************************************* //
/// DESCRIPTION: Converts the vertices position from one of the 19 canonic
///   configuration into the i'th configuration (i of 256).
/// The 19 basic configurations, coded on 8 bits:
/// Type 0     Type 1    Type 2    Type 3    Type 4
///  0         00000001  10000001  00001001  00010001
///  0x00      0x01      0x81      0x09      0x11
///    6----7    .----.    .----*    .----.    .----.
///   /|   /|   /|   /|   /|   /|   /|   /|   /|   /|
///  4----5 |  .----. |  .----. |  .----. |  *----. |
///  | 2--|-3  | .--|-.  | .--|-.  | .--|-*  | .--|-.
///  |/   |/   |/   |/   |/   |/   |/   |/   |/   |/ 
///  0----1    *----.    *----.    *----.    *----.  
/// Type 5     Type 6    Type 7    Type 8    Type 9
///  00011001  00000111  00010110  10011001  00001111
///  0x19      0x07      0x16      0x99      0x0F
///    .----.    .----.    .----.    .----*    .----.
///   /|   /|   /|   /|   /|   /|   /|   /|   /|   /|
///  *----. |  .----. |  *----. |  *----. |  .----. |
///  | .--|-*  | *--|-.  | *--|-.  | .--|-*  | *--|-*
///  |/   |/   |/   |/   |/   |/   |/   |/   |/   |/ 
///  *----.    *----*    .----*    *----.    *----*  
/// Type 10    Type 11   Type 12   Type 13   Type 14
///  00010111  00100111  01000111  10000111  01101001
///  0x17      0x27      0x47      0x87      0x69
///    .----.    .----.    *----.    .----*    *----.
///   /|   /|   /|   /|   /|   /|   /|   /|   /|   /|
///  *----. |  .----* |  .----. |  .----. |  .----* |
///  | *--|-.  | *--|-.  | *--|-.  | *--|-.  | .--|-*
///  |/   |/   |/   |/   |/   |/   |/   |/   |/   |/ 
///  *----*    *----*    *----*    *----*    *----.  
/// Type 0     Type 1    Type 2    Type 15   Type 4
///  11111111  11111110  01111110  11110110  11101110 
///  0xFF      0xFE      0x7E      0xF6      0xEE
///    *----*    *----*    *----.    *----*    *----*
///   /|   /|   /|   /|   /|   /|   /|   /|   /|   /|
///  *----* |  *----* |  *----* |  *----* |  .----* |
///  | *--|-*  | *--|-*  | *--|-*  | *--|-.  | *--|-*
///  |/   |/   |/   |/   |/   |/   |/   |/   |/   |/ 
///  *----*    .----*    .----*    .----*    .----*  
/// Type 16    Type 6    Type 17   Type 8    Type 9
///  11100110  11111000  11101001  01100110  11110000
///  0xE6      0xF8      0xE9      0x66      0xF0
///    *----*    *----*    *----*    *----.    *----*
///   /|   /|   /|   /|   /|   /|   /|   /|   /|   /|
///  .----* |  *----* |  .----* |  .----* |  *----* |
///  | *--|-.  | .--|-*  | .--|-*  | *--|-.  | .--|-.
///  |/   |/   |/   |/   |/   |/   |/   |/   |/   |/ 
///  .----*    .----.    *----.    .----*    .----.  
/// Type 10    Type 11   Type 12   Type 18   Type 14
///  11101000  11011000  10111000  01111000  10010110
///  0xE8      0xD8      0xB8      0x78      0x96
///    *----*    *----*    .----*    *----.    .----*
///   /|   /|   /|   /|   /|   /|   /|   /|   /|   /|
///  .----* |  *----. |  *----* |  *----* |  *----. |
///  | .--|-*  | .--|-*  | .--|-*  | .--|-*  | *--|-.
///  |/   |/   |/   |/   |/   |/   |/   |/   |/   |/ 
///  .----.    .----.    .----.    .----.    .----*  
// ************************************************************************* //
const int map[256][8] = 
{ {0, 1, 2, 3, 4, 5, 6, 7}, {0, 1, 2, 3, 4, 5, 6, 7}, {1, 0, 5, 4, 3, 2, 7, 6},
  {0, 2, 4, 6, 1, 3, 5, 7}, {2, 0, 3, 1, 6, 4, 7, 5}, {2, 3, 6, 7, 0, 1, 4, 5},
  {2, 0, 3, 1, 6, 4, 7, 5}, {0, 1, 2, 3, 4, 5, 6, 7}, {3, 7, 2, 6, 1, 5, 0, 4},
  {0, 1, 2, 3, 4, 5, 6, 7}, {1, 0, 5, 4, 3, 2, 7, 6}, {1, 3, 0, 2, 5, 7, 4, 6},
  {2, 6, 0, 4, 3, 7, 1, 5}, {2, 0, 3, 1, 6, 4, 7, 5}, {3, 2, 1, 0, 7, 6, 5, 4},
  {0, 1, 2, 3, 4, 5, 6, 7}, {4, 0, 6, 2, 5, 1, 7, 3}, {0, 1, 2, 3, 4, 5, 6, 7},
  {4, 5, 0, 1, 6, 7, 2, 3}, {0, 4, 1, 5, 2, 6, 3, 7}, {2, 6, 0, 4, 3, 7, 1, 5},
  {0, 2, 4, 6, 1, 3, 5, 7}, {0, 1, 2, 3, 4, 5, 6, 7}, {0, 1, 2, 3, 4, 5, 6, 7},
  {3, 1, 7, 5, 2, 0, 6, 4}, {0, 1, 2, 3, 4, 5, 6, 7}, {1, 0, 5, 4, 3, 2, 7, 6},
  {0, 4, 1, 5, 2, 6, 3, 7}, {2, 6, 0, 4, 3, 7, 1, 5}, {0, 2, 4, 6, 1, 3, 5, 7},
  {3, 2, 1, 0, 7, 6, 5, 4}, {7, 5, 6, 4, 3, 1, 2, 0}, {5, 7, 1, 3, 4, 6, 0, 2},
  {0, 4, 1, 5, 2, 6, 3, 7}, {5, 4, 7, 6, 1, 0, 3, 2}, {1, 0, 5, 4, 3, 2, 7, 6},
  {2, 6, 0, 4, 3, 7, 1, 5}, {0, 4, 1, 5, 2, 6, 3, 7}, {1, 3, 0, 2, 5, 7, 4, 6},
  {0, 1, 2, 3, 4, 5, 6, 7}, {5, 7, 1, 3, 4, 6, 0, 2}, {1, 5, 3, 7, 0, 4, 2, 6},
  {1, 5, 3, 7, 0, 4, 2, 6}, {1, 5, 3, 7, 0, 4, 2, 6}, {3, 1, 7, 5, 2, 0, 6, 4},
  {6, 7, 4, 5, 2, 3, 0, 1}, {1, 5, 3, 7, 0, 4, 2, 6}, {6, 7, 4, 5, 2, 3, 0, 1},
  {5, 7, 1, 3, 4, 6, 0, 2}, {4, 5, 0, 1, 6, 7, 2, 3}, {5, 1, 4, 0, 7, 3, 6, 2},
  {0, 4, 1, 5, 2, 6, 3, 7}, {4, 0, 6, 2, 5, 1, 7, 3}, {0, 2, 4, 6, 1, 3, 5, 7},
  {5, 1, 4, 0, 7, 3, 6, 2}, {7, 6, 3, 2, 5, 4, 1, 0}, {5, 7, 1, 3, 4, 6, 0, 2},
  {4, 5, 0, 1, 6, 7, 2, 3}, {1, 5, 3, 7, 0, 4, 2, 6}, {6, 2, 7, 3, 4, 0, 5, 1},
  {6, 4, 2, 0, 7, 5, 3, 1}, {7, 3, 5, 1, 6, 2, 4, 0}, {6, 4, 2, 0, 7, 5, 3, 1},
  {6, 4, 2, 0, 7, 5, 3, 1}, {6, 2, 7, 3, 4, 0, 5, 1}, {6, 4, 2, 0, 7, 5, 3, 1},
  {1, 5, 3, 7, 0, 4, 2, 6}, {0, 2, 4, 6, 1, 3, 5, 7}, {6, 7, 4, 5, 2, 3, 0, 1},
  {2, 6, 0, 4, 3, 7, 1, 5}, {2, 0, 3, 1, 6, 4, 7, 5}, {0, 1, 2, 3, 4, 5, 6, 7},
  {3, 7, 2, 6, 1, 5, 0, 4}, {2, 0, 3, 1, 6, 4, 7, 5}, {3, 7, 2, 6, 1, 5, 0, 4},
  {1, 3, 0, 2, 5, 7, 4, 6}, {2, 3, 6, 7, 0, 1, 4, 5}, {2, 0, 3, 1, 6, 4, 7, 5},
  {5, 4, 7, 6, 1, 0, 3, 2}, {5, 4, 7, 6, 1, 0, 3, 2}, {4, 5, 0, 1, 6, 7, 2, 3},
  {4, 0, 6, 2, 5, 1, 7, 3}, {4, 5, 0, 1, 6, 7, 2, 3}, {0, 4, 1, 5, 2, 6, 3, 7},
  {6, 4, 2, 0, 7, 5, 3, 1}, {3, 1, 7, 5, 2, 0, 6, 4}, {6, 4, 2, 0, 7, 5, 3, 1},
  {7, 3, 5, 1, 6, 2, 4, 0}, {6, 2, 7, 3, 4, 0, 5, 1}, {5, 7, 1, 3, 4, 6, 0, 2},
  {4, 5, 0, 1, 6, 7, 2, 3}, {7, 6, 3, 2, 5, 4, 1, 0}, {5, 7, 1, 3, 4, 6, 0, 2},
  {5, 7, 1, 3, 4, 6, 0, 2}, {5, 1, 4, 0, 7, 3, 6, 2}, {7, 6, 3, 2, 5, 4, 1, 0},
  {6, 7, 4, 5, 2, 3, 0, 1}, {4, 5, 0, 1, 6, 7, 2, 3}, {5, 4, 7, 6, 1, 0, 3, 2},
  {1, 0, 5, 4, 3, 2, 7, 6}, {6, 7, 4, 5, 2, 3, 0, 1}, {3, 1, 7, 5, 2, 0, 6, 4},
  {0, 1, 2, 3, 4, 5, 6, 7}, {7, 5, 6, 4, 3, 1, 2, 0}, {7, 5, 6, 4, 3, 1, 2, 0},
  {0, 1, 2, 3, 4, 5, 6, 7}, {1, 5, 3, 7, 0, 4, 2, 6}, {6, 7, 4, 5, 2, 3, 0, 1},
  {2, 3, 6, 7, 0, 1, 4, 5}, {5, 7, 1, 3, 4, 6, 0, 2}, {4, 6, 5, 7, 0, 2, 1, 3},
  {7, 5, 6, 4, 3, 1, 2, 0}, {4, 6, 5, 7, 0, 2, 1, 3}, {4, 5, 0, 1, 6, 7, 2, 3},
  {4, 6, 5, 7, 0, 2, 1, 3}, {3, 7, 2, 6, 1, 5, 0, 4}, {1, 3, 0, 2, 5, 7, 4, 6},
  {3, 1, 7, 5, 2, 0, 6, 4}, {3, 2, 1, 0, 7, 6, 5, 4}, {7, 5, 6, 4, 3, 1, 2, 0},
  {0, 1, 2, 3, 4, 5, 6, 7}, {3, 2, 1, 0, 7, 6, 5, 4}, {2, 3, 6, 7, 0, 1, 4, 5},
  {7, 6, 3, 2, 5, 4, 1, 0}, {1, 5, 3, 7, 0, 4, 2, 6}, {7, 3, 5, 1, 6, 2, 4, 0},
  {0, 1, 2, 3, 4, 5, 6, 7}, {7, 5, 6, 4, 3, 1, 2, 0}, {7, 6, 3, 2, 5, 4, 1, 0},
  {0, 1, 2, 3, 4, 5, 6, 7}, {7, 3, 5, 1, 6, 2, 4, 0}, {1, 5, 3, 7, 0, 4, 2, 6},
  {2, 3, 6, 7, 0, 1, 4, 5}, {2, 3, 6, 7, 0, 1, 4, 5}, {3, 7, 2, 6, 1, 5, 0, 4},
  {0, 1, 2, 3, 4, 5, 6, 7}, {3, 2, 1, 0, 7, 6, 5, 4}, {3, 2, 1, 0, 7, 6, 5, 4},
  {3, 1, 7, 5, 2, 0, 6, 4}, {3, 1, 7, 5, 2, 0, 6, 4}, {3, 7, 2, 6, 1, 5, 0, 4},
  {2, 0, 3, 1, 6, 4, 7, 5}, {4, 5, 0, 1, 6, 7, 2, 3}, {4, 6, 5, 7, 0, 2, 1, 3},
  {7, 5, 6, 4, 3, 1, 2, 0}, {4, 6, 5, 7, 0, 2, 1, 3}, {5, 7, 1, 3, 4, 6, 0, 2},
  {0, 4, 1, 5, 2, 6, 3, 7}, {6, 2, 7, 3, 4, 0, 5, 1}, {0, 2, 4, 6, 1, 3, 5, 7},
  {0, 1, 2, 3, 4, 5, 6, 7}, {7, 3, 5, 1, 6, 2, 4, 0}, {7, 5, 6, 4, 3, 1, 2, 0},
  {0, 1, 2, 3, 4, 5, 6, 7}, {3, 1, 7, 5, 2, 0, 6, 4}, {6, 7, 4, 5, 2, 3, 0, 1},
  {3, 7, 2, 6, 1, 5, 0, 4}, {5, 4, 7, 6, 1, 0, 3, 2}, {4, 5, 0, 1, 6, 7, 2, 3},
  {5, 4, 7, 6, 1, 0, 3, 2}, {7, 6, 3, 2, 5, 4, 1, 0}, {5, 1, 4, 0, 7, 3, 6, 2},
  {5, 7, 1, 3, 4, 6, 0, 2}, {2, 3, 6, 7, 0, 1, 4, 5}, {7, 6, 3, 2, 5, 4, 1, 0},
  {1, 0, 5, 4, 3, 2, 7, 6}, {4, 0, 6, 2, 5, 1, 7, 3}, {6, 2, 7, 3, 4, 0, 5, 1},
  {7, 3, 5, 1, 6, 2, 4, 0}, {6, 4, 2, 0, 7, 5, 3, 1}, {2, 6, 0, 4, 3, 7, 1, 5},
  {6, 4, 2, 0, 7, 5, 3, 1}, {0, 4, 1, 5, 2, 6, 3, 7}, {4, 5, 0, 1, 6, 7, 2, 3},
  {4, 0, 6, 2, 5, 1, 7, 3}, {4, 5, 0, 1, 6, 7, 2, 3}, {5, 4, 7, 6, 1, 0, 3, 2},
  {4, 5, 0, 1, 6, 7, 2, 3}, {2, 6, 0, 4, 3, 7, 1, 5}, {2, 3, 6, 7, 0, 1, 4, 5},
  {5, 4, 7, 6, 1, 0, 3, 2}, {3, 7, 2, 6, 1, 5, 0, 4}, {2, 3, 6, 7, 0, 1, 4, 5},
  {6, 2, 7, 3, 4, 0, 5, 1}, {0, 1, 2, 3, 4, 5, 6, 7}, {2, 0, 3, 1, 6, 4, 7, 5},
  {2, 6, 0, 4, 3, 7, 1, 5}, {2, 0, 3, 1, 6, 4, 7, 5}, {0, 2, 4, 6, 1, 3, 5, 7},
  {1, 5, 3, 7, 0, 4, 2, 6}, {6, 4, 2, 0, 7, 5, 3, 1}, {6, 4, 2, 0, 7, 5, 3, 1},
  {6, 4, 2, 0, 7, 5, 3, 1}, {6, 4, 2, 0, 7, 5, 3, 1}, {7, 3, 5, 1, 6, 2, 4, 0},
  {1, 5, 3, 7, 0, 4, 2, 6}, {6, 2, 7, 3, 4, 0, 5, 1}, {2, 6, 0, 4, 3, 7, 1, 5},
  {4, 5, 0, 1, 6, 7, 2, 3}, {5, 7, 1, 3, 4, 6, 0, 2}, {7, 6, 3, 2, 5, 4, 1, 0},
  {5, 1, 4, 0, 7, 3, 6, 2}, {4, 5, 0, 1, 6, 7, 2, 3}, {4, 0, 6, 2, 5, 1, 7, 3},
  {0, 4, 1, 5, 2, 6, 3, 7}, {5, 1, 4, 0, 7, 3, 6, 2}, {4, 5, 0, 1, 6, 7, 2, 3},
  {4, 0, 6, 2, 5, 1, 7, 3}, {6, 7, 4, 5, 2, 3, 0, 1}, {3, 2, 1, 0, 7, 6, 5, 4},
  {2, 0, 3, 1, 6, 4, 7, 5}, {3, 1, 7, 5, 2, 0, 6, 4}, {1, 0, 5, 4, 3, 2, 7, 6},
  {1, 5, 3, 7, 0, 4, 2, 6}, {1, 3, 0, 2, 5, 7, 4, 6}, {5, 7, 1, 3, 4, 6, 0, 2},
  {0, 1, 2, 3, 4, 5, 6, 7}, {1, 3, 0, 2, 5, 7, 4, 6}, {0, 4, 1, 5, 2, 6, 3, 7},
  {2, 6, 0, 4, 3, 7, 1, 5}, {1, 0, 5, 4, 3, 2, 7, 6}, {1, 3, 0, 2, 5, 7, 4, 6},
  {0, 4, 1, 5, 2, 6, 3, 7}, {5, 7, 1, 3, 4, 6, 0, 2}, {7, 5, 6, 4, 3, 1, 2, 0},
  {7, 5, 6, 4, 3, 1, 2, 0}, {7, 5, 6, 4, 3, 1, 2, 0}, {2, 6, 0, 4, 3, 7, 1, 5},
  {1, 3, 0, 2, 5, 7, 4, 6}, {1, 0, 5, 4, 3, 2, 7, 6}, {0, 1, 2, 3, 4, 5, 6, 7},
  {4, 6, 5, 7, 0, 2, 1, 3}, {0, 1, 2, 3, 4, 5, 6, 7}, {0, 1, 2, 3, 4, 5, 6, 7},
  {0, 2, 4, 6, 1, 3, 5, 7}, {2, 6, 0, 4, 3, 7, 1, 5}, {0, 4, 1, 5, 2, 6, 3, 7},
  {1, 0, 5, 4, 3, 2, 7, 6}, {0, 1, 2, 3, 4, 5, 6, 7}, {4, 5, 0, 1, 6, 7, 2, 3},
  {0, 1, 2, 3, 4, 5, 6, 7}, {3, 2, 1, 0, 7, 6, 5, 4}, {2, 0, 3, 1, 6, 4, 7, 5},
  {3, 1, 7, 5, 2, 0, 6, 4}, {1, 3, 0, 2, 5, 7, 4, 6}, {3, 7, 2, 6, 1, 5, 0, 4},
  {0, 1, 2, 3, 4, 5, 6, 7}, {3, 7, 2, 6, 1, 5, 0, 4}, {0, 1, 2, 3, 4, 5, 6, 7},
  {1, 3, 0, 2, 5, 7, 4, 6}, {0, 4, 1, 5, 2, 6, 3, 7}, {2, 6, 0, 4, 3, 7, 1, 5},
  {1, 5, 3, 7, 0, 4, 2, 6}, {1, 3, 0, 2, 5, 7, 4, 6}, {0, 1, 2, 3, 4, 5, 6, 7},
  {0, 1, 2, 3, 4, 5, 6, 7} };

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Number of vertices in canonic configuration i.
///                           0  1  2  3  4  5  6  7  8  9 10 11 12 13 14
///                           15 16 17 18
// ***************************************** coded by Angelidis Alexis ***** //
const int	numVert[19] = {0, 3, 0, 0, 4, 0, 5, 0, 0, 4, 6, 6, 6, 0, 0,
			       0, 0, 0, 0};

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Polygons, defined on the edges of the cube.
// ***************************************** coded by Angelidis Alexis ***** //
const int	vindex[19][54] = 
  { {},					    //Type 0 0
    { 0, 1, 0, 2, 0, 4 },		    //Type 1 3
    {},					    //Type 2 0
    {},					    //Type 3 0
    { 0, 2, 4, 6, 4, 5, 0, 1 },             //Type 4 4
    {},					    //Type 5 0
    { 0, 4, 1, 5, 1, 3, 2, 3, 2, 6 },       //Type 6 5
    { },				    //Type 7 0
    { },				    //Type 8 0
    { 0, 4, 1, 5, 3, 7, 2, 6 },		    //Type 9 4
    { 1, 5, 1, 3, 2, 3, 2, 6, 4, 6, 4, 5 }, //Type 10 6
    { 5, 7, 1, 3, 2, 3, 2, 6, 0, 4, 5, 4 }, //Type 11 6
    { 0, 4, 1, 5, 1, 3, 2, 3, 6, 7, 6, 4 }, //Type 12 6
    { },				    //Type 13 0
    { },				    //Type 14 0
    { },				    //Type 15 0
    { },				    //Type 16 0
    { },				    //Type 17 0
    { }					    //Type 18 0
  };

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Converts a configuration into a known type (one of the 19
///   canonic types).
///   The sign tell whether the type is the complement configuration
///   (i.e. swap inside/outside).
// ***************************************** coded by Angelidis Alexis ***** //
const int confToType[256] = 
  {0, 1, 1, 4, 1, 4, 3, 6, 1, 3, 4, 6, 4, 6, 6, 9, 1, 4, 3, 6, 3, 6, 7, 10, 2,
   5, 5, 12, 5, 11, 13, -6, 1, 3, 4, 6, 2, 5, 5, 11, 3, 7, 6, 10, 5, 18, 12,
   -6, 4, 6, 6, 9, 5, 12, 13, -6, 5, 13, 11, -6, -8, 16, 16, -4, 1, 3, 2, 5, 4,
   6, 5, 12, 3, 7, 5, 13, 6, 10, -11, -6, 4, 6, 5, 11, 6, -9, 13, -6, 5, 18, 8,
   16, -12, -6, 16, -4, 3, 7, 5, 13, 5, 18, -8, 16, 7, 14, 13, 17, 13, 17, 16,
   15, 6, 10, 12, -6, -11, -6, 16, -4, 18, 17, 16, 15, 16, 15, -2, -1, 1, 2, 3,
   5, 3, 5, 7, 13, 4, 5, 6, 11, 6, 12, -10, -6, 3, 5, 7, 13, 7, 13, -14, 17, 5,
   8, 13, 16, 13, 16, 17, 15, 4, 5, 6, -12, 5, -8, 18, 16, 6, 18, -9, -6, -11,
   16, -6, -4, 6, 11, -10, -6, 13, 16, 17, 15, -12, 16, -6, -4, 16, -2, 15, -1,
   4, 5, 5, 8, 6, 11, 18, 16, 6, 18, -12, 16, -9, -6, -6, -4, 6, -12, 18, 16,
   -10, -6, 17, 15, -11, 16, 16, -2, -6, -4, 15, -1, 6, 13, 11, 16, -12, 16,
   16, -2, -10, 17, -6, 15, -6, 15, -4, -1, -9, -6, -6, -4, -6, -4, 15, -1, -6,
   15, -4, -1, -4, -1, -1, 0 };

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Computes the intersection vertices of the intersection
///  polygon between the plane p and this frustum.
///  The polygon is returned in planez, and the number of vertices of the 
///  polygon is the returned value.
// ***************************************** coded by Angelidis Alexis ***** //
inline int	Frustum::intersect(const Plane	&p,
				   Pnt3d	planez[6]) const
{
  int		conf(0), absType;
  Pnt3d	c[8];

  corners(c);
  for (int i(0); i < 8; ++i)
    if (p.contains(c[i]))
      conf |= (1 << i);
  if (!(absType = abs(confToType[conf])))
    return 0;
  if (0 < confToType[conf])
    for (int i(0), nVert(numVert[absType]); i < nVert; ++i)
      {
	const Pnt3d	&c0(c[map[conf][vindex[absType][2 * i]]]);
	const Pnt3d	&c1(c[map[conf][vindex[absType][2 * i + 1]]]);

	planez[i] = c0;
	p.projectAlong(c1 - c0, planez[i]);
       }
  else
    for (int j(0), i(numVert[absType] - 1); i >= 0; --i, ++j)
      {
	const Pnt3d	&c0(c[map[conf][vindex[absType][2 * i]]]);
	const Pnt3d	&c1(c[map[conf][vindex[absType][2 * i + 1]]]);

	planez[j] = c0;
	p.projectAlong(c1 - c0, planez[j]);
       }
  return numVert[absType];
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Computes the 8 corners.
///
///      1----3...5....7
/// View |    |   :    |
///      0----2...4----6
// ***************************************** coded by Angelidis Alexis ***** //
inline void	Frustum::corners(Pnt3d c[8]) const
{
  nearCorners(c);
  farCorners(c + 4);
}

// ***************************************** created by Alexis Angelidis *** //
inline void	Frustum::nearCorners(Pnt3d c[4]) const
{
  Plane::intersect(getNearPlane(), getLeftPlane(), getBottomPlane(), c[0]);
  Plane::intersect(getNearPlane(), getLeftPlane(), getTopPlane(), c[1]);
  Plane::intersect(getNearPlane(), getRightPlane(), getBottomPlane(), c[2]);
  Plane::intersect(getNearPlane(), getRightPlane(), getTopPlane(), c[3]);
}

// *************************************** created by Angelidis Alexis ***** //
inline void	Frustum::farCorners(Pnt3d c[4]) const
{
  Plane::intersect(getFarPlane(), getLeftPlane(), getBottomPlane(), c[0]);
  Plane::intersect(getFarPlane(), getLeftPlane(), getTopPlane(), c[1]);
  Plane::intersect(getFarPlane(), getRightPlane(), getBottomPlane(), c[2]);
  Plane::intersect(getFarPlane(), getRightPlane(), getTopPlane(), c[3]);
}

// ***************************************** created by Alexis Angelidis *** //
inline Pnt3d	Frustum::eye() const
{
  Pnt3d	ret;

  Plane::intersect(getLeftPlane(), getBottomPlane(), getRightPlane(), ret);
  return ret;
}

// ***************************************** coded by Angelidis Alexis ***** //
/// DESCRIPTION: Computes the intersection between the horizon (z = 0) 
///  and far plane of frustum
/// RETURN: 0 if horsion is in view, -1 if horizon is above and 1 if bellow.
// ***************************************** created by Alexis Angelidis *** //
inline int	Frustum::getHorizon(Pnt3d h[2]) const
{
  Pnt3d	fc[4];

  farCorners(fc);
  
  const Vec3d	dLeft(fc[1] - fc[0]), dRight(fc[3] - fc[2]);
  const double		tLeft(-fc[0].z / dLeft.z), tRight(-fc[2].z / dLeft.z);

  if (1.0 <= tLeft && 1.0 <= tRight)
    return -1;
  if (tLeft <= 0.0 && tRight <= 0.0)
    return 1;
  h[0] = fc[0] + dLeft * tLeft;
  h[1] = fc[2] + dRight * tRight;
  return 0;
}

// ***************************************** coded by Angelidis Alexis ***** //
/// DESCRIPTION: Basic Vec3i class.
// ***************************************** created by Alexis Angelidis *** //
struct Vec3i
{
  int	x, y, z;
public:
  Vec3i() {}
  Vec3i(const int d): x(d), y(d), z(d) {}
  Vec3i(const int x_, const int y_, const int z_):
    x(x_), y(y_), z(z_) {}
  explicit Vec3i(const Pnt3i &p);
  Vec3i	&operator+=(const Vec3i &v)
  { x += v.x; y += v.y; z += v.z; return *this; }
  Vec3i	&operator*=(const int d)
  { x *= d; y *= d; z *= d; return *this; }
  Vec3i	&operator/=(const int d)
  { x /= d; y /= d; z /= d; return *this; }
  friend Vec3i	operator*(const int d, const Vec3i &v)
  { return Vec3i(d * v.x, d * v.y, d * v.z); }
  Vec3i	operator+(const Vec3i &v) const
  { return Vec3i(x + v.x, y + v.y, z + v.z); }
  Vec3i	operator-(const Vec3i &v) const
  { return Vec3i(x - v.x, y - v.y, z - v.z); }
  int	operator*(const Vec3i &v) const
  { return x * v.x + y * v.y + z * v.z; }
  Vec3i	operator/(const int d) const
  { return Vec3i(x / d, y / d, z / d); }
  Vec3i	operator*(const int d) const
  { return Vec3i(x * d, y * d, z * d); }
  int &operator()(const unsigned i) { return i? ((i == 2)? z: y): x; }
  int operator()(const unsigned i) const { return i? ((i == 2)? z: y): x; }
  int &operator[](const unsigned i) { return i? ((i == 2)? z: y): x; }
  int operator[](const unsigned i) const { return i? ((i == 2)? z: y): x; }
  friend std::ostream	&operator<<(std::ostream &os, const Vec3i &v)
  { return os << v.x << " " << v.y << " " << v.z; }
  Vec3i	operator-() const { return Vec3i(-x, -y, -z); }
  Vec3i	operator^(const Vec3i &v) const
  { return Vec3i(y * v.z - z * v.y, z * v.x - x * v.z, x * v.y - y * v.x); }
  void		glVertex() const { glVertex3i(x, y, z); }
  void		glNormal() const { glNormal3i(x, y, z); }
};

// ***************************************** coded by Angelidis Alexis ***** //
/// DESCRIPTION: Basic Pnt3i class.
// ***************************************** created by Alexis Angelidis *** //
struct Pnt3i
{
  int	x, y, z;
public:
  Pnt3i() {}
  Pnt3i(const int d): x(d), y(d), z(d) {}
  Pnt3i(const int x_, const int y_, const int z_):
    x(x_), y(y_), z(z_) {}
  explicit Pnt3i(const Pnt3d &p): x(int(p.x)), y(int(p.y)), z(int(p.z)) {}
  void		glVertex() const { glVertex3i(x, y, z); }
  Vec3i	operator-(const Pnt3i &p) const
  { return Vec3i(x - p.x, y - p.y, z - p.z); }
  Pnt3i	operator+(const Vec3i &v) const
  { return Pnt3i(x + v.x, y + v.y, z + v.z); }
  friend std::istream	&operator>>(std::istream &is, Pnt3i &p)
  { return is >> p.x >> p.y >> p.z; }
  friend std::ostream	&operator<<(std::ostream &os, const Pnt3i &p)
  { return os << p.x << " " << p.y << " " << p.z; }
};

// ******************************************* coded by Alexis Angelidis *** //
inline Vec3i::Vec3i(const Pnt3i &p): x(p.x), y(p.y), z(p.z)
{
}

// ******************************************* coded by Alexis Angelidis *** //
inline Pnt3d::Pnt3d(const Pnt3i &p): x(p.x), y(p.y), z(p.z)
{
}

// *************************************** created by Angelidis Alexis ***** //
/// DESCRIPTION: Standard non-oriented bounded box.
// *************************************** created by Angelidis Alexis ***** //
struct BBox3d
{
  Pnt3d	pmin, pmax;
public:
  BBox3d() {}
  BBox3d(const Pnt3d &p0, const Pnt3d &p1);
  BBox3d(const Pnt3d &p0, const Pnt3d &p1, const Pnt3d &p2);
  BBox3d(const Pnt3d &q): pmin(q), pmax(q) {}
  BBox3d	&operator*=(const Mat4d &m);
  BBox3d	operator+(const BBox3d &b) const;
  BBox3d	&operator+=(const Pnt3d &p);
  BBox3d	&operator+=(const BBox3d &b);
  void		fill(const Pnt3d &p);
  void		fill(const BBox3d &b);
  void		eqExpand(const double d);
  void		eqExpand(const BBox3d &b, const double d);
  void		glRender() const;
  void		glRenderSmooth() const;
  Pnt3d		corner(const char i) const;
  void		corners(Pnt3d c[8]) const;
  bool		contains(const Pnt3d &p) const;
  friend std::ostream	&operator<<(std::ostream &os, const BBox3d &b)
  { return os << b.pmin << " " << b.pmax; }
  friend std::istream	&operator>>(std::istream &is, BBox3d &b)
  { return is >> b.pmin >> b.pmax; }
  void		eqSymx();
  void		eqSymy();
  void		eqSymz();
  void		eqSym(const unsigned sym);
  BBox3d	symx() const;
  BBox3d	symy() const;
  BBox3d  	symz() const;
  BBox3d	sym(const unsigned sym) const;
  bool		intersects(const BBox3d &b) const;
  double	dist2(const Pnt3d &p) const;
  double	dist2min(const Pnt3d &p) const { return dist2(p); }
  double	dist2max(const Pnt3d &p) const;
  Pnt3d		closest(const Pnt3d &p) const;
  void		eqCubify();
};

// ************************************************************************* //
/// DESCRIPTION: Turns the box into a cube that includes the box
///  i.e. all edges will have the same length as the longest edge of the box.
///  The extra volume is added evenly around the box.
// ************************************************************************* //
inline void	BBox3d::eqCubify()
{
  const Vec3d diag(pmax - pmin);

  if (diag.y <= diag.x && diag.z <= diag.x)
    {
      const double dy((diag.x - diag.y) * 0.5);
      const double dz((diag.x - diag.z) * 0.5);

      pmin.y -= dy;
      pmax.y += dy;
      pmin.z -= dz;
      pmax.z += dz;
    }
  else if (diag.z <= diag.y)
    {
      const double dz((diag.y - diag.z) * 0.5);
      const double dx((diag.y - diag.x) * 0.5);

      pmin.z -= dz;
      pmax.z += dz;
      pmin.x -= dx;
      pmax.x += dx;
    }
  else
    {
      const double dx((diag.z - diag.x) * 0.5);
      const double dy((diag.z - diag.y) * 0.5);

      pmin.x -= dx;
      pmax.x += dx;
      pmin.y -= dy;
      pmax.y += dy;
    }
}

// ******************************************* coded by Alexis Angelidis *** //
/// DESCRIPTION: Finds the closest point on the box.
// ******************************************* coded by Alexis Angelidis *** //
inline Pnt3d	BBox3d::closest(const Pnt3d &p) const
{
  if (p.x < pmin.x)
    if (p.y < pmin.y)
      if (p.z < pmin.z)
	return pmin;
      else if (pmax.z < p.z)
	return Pnt3d(pmin.x, pmin.y, pmax.z);
      else
	return Pnt3d(pmin.x, pmin.y, p.z);
    else if (pmax.y < p.y)
      if (pmax.z < p.z)
	return Pnt3d(pmin.x, pmax.y, pmax.z);
      else if (p.z < pmin.z)
	return Pnt3d(pmin.x, pmax.y, pmin.z);
      else
	return Pnt3d(pmin.x, pmax.y, p.z);
    else if (pmax.z < p.z)
      return Pnt3d(pmin.x, p.y, pmax.z);
    else if (p.z < pmin.z)
      return Pnt3d(pmin.x, p.y, pmin.z);
    else
      return Pnt3d(pmin.x, p.y, p.z);
  else if (pmax.x < p.x)
    if (p.y < pmin.y)
      if (p.z < pmin.z)
	return Pnt3d(pmax.x, pmin.y, pmin.z);
      else if (pmax.z < p.z)
	return Pnt3d(pmax.x, pmin.y, pmax.z);
      else
	return Pnt3d(pmax.x, pmin.y, p.z);
    else if (pmax.y < p.y)
      if (p.z < pmin.z)
	return Pnt3d(pmax.x, pmax.y, pmin.z);
      else if (pmax.z < p.z)
	return pmax;
      else
	return Pnt3d(pmax.x, pmax.y, p.z);
    else if (pmax.z < p.z)
      return Pnt3d(pmax.x, p.y, pmax.z);
    else if (p.z < pmin.z)
      return Pnt3d(pmax.x, p.y, pmin.z);
    else
      return Pnt3d(pmax.x, p.y, p.z);
  else if (p.y < pmin.y)
    if (p.z < pmin.z)
      return Pnt3d(p.x, pmin.y, pmin.z);
    else if (pmax.z < p.z)
      return Pnt3d(p.x, pmin.y, pmax.z);
    else
      return Pnt3d(p.x, pmin.y, p.z);
  else if (pmax.y < p.y)
    if (p.z < pmin.z)
      return Pnt3d(p.x, pmax.y, pmin.z);
    else if (pmax.z < p.z)
      return Pnt3d(p.x, pmax.y, pmax.z);
    else
      return Pnt3d(p.x, pmax.y, p.z);
  else
    if (p.z < pmin.z)
      return Pnt3d(p.x, p.y, pmin.z);
    else if (pmax.z < p.z)
      return Pnt3d(p.x, p.y, pmax.z);
    else
      return p;
}

// ******************************************* coded by Alexis Angelidis *** //
/// DESCRIPTION: Returns the distance from the point to the box.
///  The box is full, i.e. the distance is 0 for points inside it.
// ******************************************* coded by Alexis Angelidis *** //
inline double	BBox3d::dist2(const Pnt3d &p) const
{
  if (p.x < pmin.x)
    if (p.y < pmin.y)
      if (p.z < pmin.z)
	return (p - Pnt3d(pmin.x, pmin.y, pmin.z)).norm2();
      else if (pmax.z < p.z)
	return (p - Pnt3d(pmin.x, pmin.y, pmax.z)).norm2();
      else
	return sq(p.x - pmin.x) + sq(p.y - pmin.y);
    else if (pmax.y < p.y)
      if (pmax.z < p.z)
	return (p - Pnt3d(pmin.x, pmax.y, pmax.z)).norm2();
      else if (p.z < pmin.z)
	return (p - Pnt3d(pmin.x, pmax.y, pmin.z)).norm2();
      else
	return sq(p.x - pmin.x) + sq(p.y - pmax.y);
    else if (pmax.z < p.z)
      return sq(p.z - pmax.z) + sq(p.x - pmin.x);
    else if (p.z < pmin.z)
      return sq(p.z - pmin.z) + sq(p.x - pmin.x);
    else
      return sq(p.x - pmin.x);
  else if (pmax.x < p.x)
    if (p.y < pmin.y)
      if (p.z < pmin.z)
	return (p - Pnt3d(pmax.x, pmin.y, pmin.z)).norm2();
      else if (pmax.z < p.z)
	return (p - Pnt3d(pmax.x, pmin.y, pmax.z)).norm2();
      else
	return sq(p.x - pmax.x) + sq(p.y - pmin.y);
    else if (pmax.y < p.y)
      if (p.z < pmin.z)
	return (p - Pnt3d(pmax.x, pmax.y, pmin.z)).norm2();
      else if (pmax.z < p.z)
	return (p - pmax).norm2();
      else
	return sq(p.x - pmax.x) + sq(p.y - pmax.y);
    else if (pmax.z < p.z)
      return sq(p.z - pmax.z) + sq(p.x - pmax.x);
    else if (p.z < pmin.z)
      return sq(p.z - pmin.z) + sq(p.x - pmax.x);
    else
      return sq(p.x - pmax.x);
  else if (p.y < pmin.y)
    if (p.z < pmin.z)
      return sq(p.y - pmin.y) + sq(p.z - pmin.z);
    else if (pmax.z < p.z)
      return sq(p.y - pmin.y) + sq(p.z - pmax.z);
    else
      return sq(p.y - pmin.y);
  else if (pmax.y < p.y)
    if (p.z < pmin.z)
      return sq(p.y - pmax.y) + sq(p.z - pmin.z);
    else if (pmax.z < p.z)
      return sq(p.y - pmax.y) + sq(p.z - pmax.z);
    else
      return sq(p.y - pmax.y);
  else
    if (p.z < pmin.z)
      return sq(p.z - pmin.z);
    else if (pmax.z < p.z)
      return sq(p.z - pmax.z);
    else
      return 0.0;
}

// ******************************************* coded by Alexis Angelidis *** //
/// DESCRIPTION: Finds a point which is the farthest point from p, in the box.
///  Returns the squared distance to that point.
/// NOTE: could be better (i.e. depending on p's region, fewer distances
///  could be computed: todo later!
// ******************************************* coded by Alexis Angelidis *** //
inline double	BBox3d::dist2max(const Pnt3d &p) const
{
  double	d2max(0.0), d2tmp;

  for (unsigned i(0); i < 8; ++i)
    {
      const Pnt3d	ci((i & 0x1)? pmin.x: pmax.x,
			   (i & 0x2)? pmin.y: pmax.y,
			   (i & 0x4)? pmin.z: pmax.z);

      if (d2max < (d2tmp = (p - ci).norm2()))
	d2max = d2tmp;
    }
  return d2max;
}

// ******************************************* coded by Alexis Angelidis *** //
inline BBox3d::BBox3d(const Pnt3d &p0,
		      const Pnt3d &p1)
{
  pmin.x = min(p0.x, p1.x);
  pmin.y = min(p0.y, p1.y);
  pmin.z = min(p0.z, p1.z);
  pmax.x = max(p0.x, p1.x);
  pmax.y = max(p0.y, p1.y);
  pmax.z = max(p0.z, p1.z);
}

// ***************************************** coded by Angelidis Alexis ***** //
inline BBox3d::BBox3d(const Pnt3d &p0,
		      const Pnt3d &p1,
		      const Pnt3d &p2)
{
  pmin.x = min(min(p0.x, p1.x), p2.x);
  pmin.y = min(min(p0.y, p1.y), p2.y);
  pmin.z = min(min(p0.z, p1.z), p2.z);
  pmax.x = max(max(p0.x, p1.x), p2.x);
  pmax.y = max(max(p0.y, p1.y), p2.y);
  pmax.z = max(max(p0.z, p1.z), p2.z);
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Returns true if frustum intersects bbox.
// *************************************** created by Angelidis Alexis ***** //
inline bool	Frustum::intersects(const BBox3d &bbox) const
{
  Pnt3d	corners[8];
  bool	isInside(true);

  bbox.corners(corners);
  for (int ip(0); ip < 6 && isInside; ++ip)
    {
      int	iCount(0);
      for (int ic(0); ic < 8; ++ic)
	if (!_m_clipPlanes[ip].contains(corners[ic]))
	  ++iCount;
      if (iCount == 8)
	isInside = false;
    }
  return isInside;
}

// ***************************************** created by Alexis Angelidis *** //
inline void	BBox3d::eqSym(const unsigned sym)
{
  if (sym == 1)
    eqSymx();
  else if (sym == 2)
    eqSymy();
  else if (sym == 3)
    eqSymz();
}

// ***************************************** coded by Angelidis Alexis ***** //
inline BBox3d	BBox3d::sym(const unsigned sym) const
{
  if (sym == 1)
    return symx();
  else if (sym == 2)
    return symy();
  else if (sym == 3)
    return symz();
  return *this;
}

// ******************************************* coded by Alexis Angelidis *** //
inline bool	BBox3d::intersects(const BBox3d &b) const
{
  return b.pmin.x <= pmax.x && pmin.x <= b.pmax.x && 
    b.pmin.y <= pmax.y && pmin.y <= b.pmax.y &&
    b.pmin.z <= pmax.z && pmin.z <= b.pmax.z;
}

// ***************************************** coded by Angelidis Alexis ***** //
/// DESCRIPTION: Turns this into a bbox symetric relative to plane x = 0.
// ***************************************** created by Alexis Angelidis *** //
inline void	BBox3d::eqSymx()
{
  const double	tmp(pmin(0));

  pmin(0) = -pmax(0);
  pmax(0) = -tmp;
}

// *************************************** created by Angelidis Alexis ***** //
/// DESCRIPTION: Turns this into a bbox symetric relatively to plane y = 0.
// ******************************************* coded by Alexis Angelidis *** //
inline void	BBox3d::eqSymy()
{
  const double	tmp(pmin(1));

  pmin(1) = -pmax(1);
  pmax(1) = -tmp;
}

// *************************************** created by Angelidis Alexis ***** //
/// DESCRIPTION: Turns this into a bbox symetric relatively to plane z = 0.
// ******************************************* coded by Alexis Angelidis *** //
inline void	BBox3d::eqSymz()
{
  const double	tmp(pmin(2));

  pmin(2) = -pmax(2);
  pmax(2) = -tmp;
}

// *************************************** created by Angelidis Alexis ***** //
/// DESCRIPTION: Turns this into a bbox symetric relative to plane x = 0.
// ******************************************* coded by Alexis Angelidis *** //
inline BBox3d	BBox3d::symx() const
{
  return BBox3d(Pnt3d(-pmax(0), pmin(1), pmin(2)),
		Pnt3d(-pmin(0), pmax(1), pmax(2)));
}

// ***************************************** created by Angelidis Alexis *** //
/// DESCRIPTION: Turns this into a bbox symetric relative to plane x = 0.
// ***************************************** coded by Alexis Angelidis ***** //
inline BBox3d	BBox3d::symy() const
{
  return BBox3d(Pnt3d(pmin(0), -pmax(1), pmin(2)),
		Pnt3d(pmax(0), -pmin(1), pmax(2)));
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Turns this into a bbox symetric relative to plane x = 0.
// ***************************************** coded by Angelidis Alexis ***** //
inline BBox3d	BBox3d::symz() const
{
  return BBox3d(Pnt3d(pmin(0), pmin(1), -pmax(2)),
		Pnt3d(pmax(0), pmax(1), -pmin(2)));
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Returns a box that contains both this box and box b.
// ******************************************* coded by Alexis Angelidis *** //
inline BBox3d	BBox3d::operator+(const BBox3d &b) const
{
  return BBox3d(Pnt3d(min(pmin.x, b.pmin.x),
		      min(pmin.y, b.pmin.y),
		      min(pmin.z, b.pmin.z)),
		Pnt3d(max(pmax.x, b.pmax.x),
		      max(pmax.y, b.pmax.y),
		      max(pmax.z, b.pmax.z)));
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Enlarges this box to include point p as well.
// ******************************************* coded by Alexis Angelidis *** //
inline BBox3d	&BBox3d::operator+=(const Pnt3d &p)
{
  pmin.x = min(p.x, pmin.x);
  pmin.y = min(p.y, pmin.y);
  pmin.z = min(p.z, pmin.z);
  pmax.x = max(p.x, pmax.x);
  pmax.y = max(p.y, pmax.y);
  pmax.z = max(p.z, pmax.z);
  return *this;
}

// *************************************** created by Angelidis Alexis ***** //
/// DESCRIPTION: Enlarges this box to include box b as well.
// ******************************************* coded by Alexis Angelidis *** //
inline BBox3d	&BBox3d::operator+=(const BBox3d &b)
{
  fill(b.pmin);
  fill(b.pmax);
  return *this;
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Enlarges this box to include point p as well.
// ******************************************* coded by Alexis Angelidis *** //
inline void	BBox3d::fill(const Pnt3d &p)
{
  pmin.x = min(p.x, pmin.x);
  pmin.y = min(p.y, pmin.y);
  pmin.z = min(p.z, pmin.z);
  pmax.x = max(p.x, pmax.x);
  pmax.y = max(p.y, pmax.y);
  pmax.z = max(p.z, pmax.z);
}

// *************************************** created by Angelidis Alexis ***** //
/// DESCRIPTION: Enlarges this box to include box b as well.
// ******************************************* coded by Alexis Angelidis *** //
inline void	BBox3d::fill(const BBox3d &b)
{
  fill(b.pmin);
  fill(b.pmax);
}

// ******************************************* coded by Alexis Angelidis *** //
/// DESCRIPTION: Enlarges this box by d in every direction.
// ******************************************* coded by Alexis Angelidis *** //
inline void	BBox3d::eqExpand(const double d)
{
  pmin.x -= d;
  pmin.y -= d;
  pmin.z -= d;
  pmax.x += d;
  pmax.y += d;
  pmax.z += d;
}

// ******************************************* coded by Alexis Angelidis *** //
/// DESCRIPTION: Enlarges the box b by d in every direction.
// ******************************************* coded by Alexis Angelidis *** //
inline void	BBox3d::eqExpand(const BBox3d	&b,
				 const double	d)
{
  pmin.x = b.pmin.x - d;
  pmin.y = b.pmin.y - d;
  pmin.z = b.pmin.z - d;
  pmax.x = b.pmax.x + d;
  pmax.y = b.pmax.y + d;
  pmax.z = b.pmax.z + d;
}

// *************************************** created by Angelidis Alexis ***** //
/// DESCRIPTION: Returns the ith corner of the box (out of 8).
// *************************************** created by Angelidis Alexis ***** //
inline Pnt3d	BBox3d::corner(const char i) const
{
  return Pnt3d((i & 0x1)? pmin.x: pmax.x,
	       (i & 0x2)? pmin.y: pmax.y,
	       (i & 0x4)? pmin.z: pmax.z);
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Computes all 8 corners of the box.
// ***************************************** created by Alexis Angelidis *** //
inline void	BBox3d::corners(Pnt3d c[8]) const
{
  for (int i(0); i < 8; ++i)
    c[i] = corner(i);
}

// ***************************************** coded by Angelidis Alexis ***** //
/// DESCRIPTION: Returns true if this box contains p.
// ***************************************** coded by Angelidis Alexis ***** //
inline bool	BBox3d::contains(const Pnt3d &p) const
{
  return pmin(0) <= p(0) && pmin(1) <= p(1) && pmin(2) <= p(2) &&
    p(0) < pmax(0) && p(1) < pmax(1) && p(2) < pmax(2);
}

// ***************************************** created by Alexis Angelidis *** //
// DESCRIPTION: Renders the box in wireframe.
// ***************************************** created by Alexis Angelidis *** //
inline void	BBox3d::glRender() const
{
  glBegin(GL_LINE_STRIP);
  pmax.glVertex();
  Pnt3d(pmax.x, pmin.y, pmax.z).glVertex();
  Pnt3d(pmin.x, pmin.y, pmax.z).glVertex();
  pmin.glVertex();
  Pnt3d(pmax.x, pmin.y, pmin.z).glVertex();
  Pnt3d(pmax.x, pmax.y, pmin.z).glVertex();
  Pnt3d(pmin.x, pmax.y, pmin.z).glVertex();
  Pnt3d(pmin.x, pmax.y, pmax.z).glVertex();
  pmax.glVertex();
  Pnt3d(pmax.x, pmax.y, pmin.z).glVertex();
  glEnd();//GL_LINE_STRIP
  glBegin(GL_LINES);
  Pnt3d(pmax.x, pmin.y, pmax.z).glVertex();
  Pnt3d(pmax.x, pmin.y, pmin.z).glVertex();
  Pnt3d(pmin.x, pmax.y, pmax.z).glVertex();
  Pnt3d(pmin.x, pmin.y, pmax.z).glVertex();
  Pnt3d(pmin.x, pmax.y, pmin.z).glVertex();
  pmin.glVertex();
  glEnd();//GL_LINES
}

// ***************************************** coded by Angelidis Alexis ***** //
inline void	BBox3d::glRenderSmooth() const
{
  glBegin(GL_QUADS);

  glNormal3d(0.0, 0.0, 1.0);
  corner(0).glVertex();
  corner(1).glVertex();
  corner(3).glVertex();
  corner(2).glVertex();

  glNormal3d(0.0, 0.0, -1.0);
  corner(6).glVertex();
  corner(7).glVertex();
  corner(5).glVertex();
  corner(4).glVertex();

  glNormal3d(0.0, 1.0, 0.0);
  corner(4).glVertex();
  corner(5).glVertex();
  corner(1).glVertex();
  corner(0).glVertex();

  glNormal3d(0.0, -1.0, 0.0);
  corner(2).glVertex();
  corner(3).glVertex();
  corner(7).glVertex();
  corner(6).glVertex();

  glNormal3d(1.0, 0.0, 0.0);
  corner(0).glVertex();
  corner(2).glVertex();
  corner(6).glVertex();
  corner(4).glVertex();

  glNormal3d(-1.0, 0.0, 0.0);
  corner(5).glVertex();
  corner(7).glVertex();
  corner(3).glVertex();
  corner(1).glVertex();

  glEnd();//GL_QUADS
}

// ***************************************** coded by Angelidis Alexis ***** //
/// DESCRIPTION: Virtual transformation class.
///  Can transform points and normals (when dealing with surfaces).
///  Transformation can be alibi (relative, i.e. cartesian tensor) or alias 
///  (absolute).
// ***************************************** created by Alexis Angelidis *** //
class Transformation
{
public:
  virtual ~Transformation() {}
  virtual Transformation	*copy() const = 0;
  virtual Mat4d		asMatrix() const = 0;
  virtual Mat4d		operator*(const Mat4d &m) const;
  virtual Mat4d		operator*(const Transformation &t) const;
  virtual Transformation	&operator^=(const double t) = 0;
  virtual void		operator()(Pnt3d &p, Vec3d &v) const = 0;
  virtual Pnt3d		&operator()(Pnt3d &p) const = 0;
  virtual Pnt3d		&operator()(Pnt3d &p, const double t) const = 0;
  virtual void		cast(const Transformation &t) = 0;
  virtual unsigned	nbSteps(const BBox3d &b, const double k) const = 0;
  virtual void		eqSymx() = 0;
  virtual void		eqSymy() = 0;
  virtual void		eqSymz() = 0;
  virtual void		eqSym(const unsigned sym);
  virtual void		eqInverse() = 0;
  virtual double	det() const = 0;
  virtual Mat4d		mlog() const = 0;
};

// ******************************************* coded by Angelidis Alexis *** //
inline void	Transformation::eqSym(const unsigned sym)
{
  if (sym == 1)
    eqSymx();
  else if (sym == 2)
    eqSymy();
  else if (sym == 3)
    eqSymz();
}

// *************************************** created by Alexis Angelidis ***** //
/// DESCRIPTION: Rotation around the origin.
// ******************************************* coded by Angelidis Alexis *** //
class Quaternion: public Transformation
{
  friend class	Rotation;
  double	_m_s;
  Vec3d	_m_v;
public:
  Quaternion() {}
  virtual ~Quaternion() {}
  Quaternion(const double d): _m_s(d), _m_v(d) {}
  Quaternion(const double s, const Vec3d &v): _m_s(s), _m_v(v) {}
  Quaternion(const double s, const double x, const double y, const double z):
    _m_s(s), _m_v(x, y, z) {}
  Quaternion(const Vec3d &v): _m_s(0.0), _m_v(v) {}
  Transformation	*copy() const { return new Quaternion(*this); }
  Quaternion	&operator+=(const Quaternion &q)
  { _m_s += q._m_s; _m_v += q._m_v; return *this; }
  Quaternion	&operator*=(const double d) 
  { _m_s *= d; _m_v *= d; return *this; }
  Quaternion	&operator/=(const double d)
  { _m_s /= d; _m_v /= d; return *this; }
  Quaternion	operator-() const { return Quaternion(-_m_s, -_m_v); }
  Quaternion	operator^(const double d) const;
  Quaternion	&operator^=(const double d);
  Quaternion	operator*(const double d) const 
  { return Quaternion(_m_s * d, _m_v * d); }
  Mat4d		asMatrix() const;
  friend Quaternion	operator*(const double d, const Quaternion &q)
  { return q * d; }
  Quaternion	operator/(const double d) const 
  { return Quaternion(_m_s / d, _m_v / d); }
  Quaternion	operator*(const Quaternion &q) const 
  { return Quaternion(_m_s * q._m_s - _m_v * q._m_v,
		      _m_s * q._m_v + q._m_s * _m_v + (_m_v ^ q._m_v)); }
  Quaternion	operator*(const Vec3d &v) const 
  { return Quaternion(-_m_v * v, _m_s * v + (_m_v ^ v)); }
  Quaternion	operator*(const Pnt3d &v) const 
  { return Quaternion(-_m_v * Vec3d(v),
		      Vec3d(v) * _m_s + (_m_v ^ Vec3d(v))); }
  Quaternion	operator+(const Quaternion &q) const 
  { return Quaternion(_m_s + q._m_s, _m_v + q._m_v); }
  Quaternion	operator-(const Quaternion &q) const 
  { return Quaternion(_m_s - q._m_s, _m_v - q._m_v); }
  double	norm2() const { return _m_s * _m_s + _m_v * _m_v; }
  double	norm() const { return sqrt(norm2()); }
  void		eqUnit() { operator*=(1.0 / norm()); }
  Quaternion	unit() const { return operator*(1.0 / norm()); }
  void		eqLog();
  Quaternion	log() const;
  Mat4d		mlog() const;
  void		eqExp();
  Quaternion	exp() const;
  Quaternion	conjugate() const { return Quaternion(_m_s, -_m_v); }
  void		eqConjugate() { _m_v = -_m_v; }
  Quaternion	inverse() const { return conjugate() / norm2(); }
  void		eqSymx() { _m_v(1) = -_m_v(1); _m_v(2) = -_m_v(2); }
  void		eqSymy() { _m_v(2) = -_m_v(2); _m_v(0) = -_m_v(0); }
  void		eqSymz() { _m_v(0) = -_m_v(0); _m_v(1) = -_m_v(1); }
  void		eqInverse() { eqConjugate(); operator/(norm2()); }
  friend std::ostream &operator<<(std::ostream &os, const Quaternion &q)
  { return os << q._m_v << q._m_s; }
  friend std::istream &operator>>(std::istream &is, Quaternion &q)
  { return is >> q._m_v >> q._m_s; }
  double	&operator()(const unsigned i) { return i? _m_v(i - 1): _m_s; }
  double	operator()(const unsigned i) const
  { return i? _m_v(i - 1): _m_s; }
  void		glRotate() const;
  static Quaternion	creator(const double angle, const Vec3d &n);
  Pnt3d	&operator()(Pnt3d &p) const;
  Vec3d	&operator()(Vec3d &v) const;
  Pnt3d	&operator()(Pnt3d &p, const double t) const;
  void		operator()(Pnt3d &p, Vec3d &v) const;
  void		cast(const Transformation &t);
  unsigned	nbSteps(const BBox3d &b, const double k) const;
  double	det() const { return 1.0; }
};

// *************************************** created by Alexis Angelidis ***** //
inline unsigned	Quaternion::nbSteps(const BBox3d	&b,
				    const double	k) const
{
  const Vec3d	v(_m_v.unit());
  Vec3d	p;
  double	l2tmp, lmax(0.0);

  for (int i(0); i < 8; ++i)
    {
      p = Vec3d(b.corner(i));
      l2tmp = (p - (p * v) * v).norm2();
      if (lmax < l2tmp)
	lmax = l2tmp;
    }
  lmax = sqrt(lmax);

  const double s = ceil(1.875 / k * 2.0 * acos(_m_s) * lmax);

  if (s <= 1.0)
    return 1;
  return unsigned(s);
}

// ******************************************* coded by Angelidis Alexis *** //
inline void	Quaternion::cast(const Transformation &t)
{
  const Quaternion *q(dynamic_cast<const Quaternion *>(&t));

  if (!q)
    {
      std::cerr << "A Quaternion can only cast a Quaternion." << std::endl;
      std::exit(1);
    }
  *this = *q;
}

// *************************************** created by Alexis Angelidis ***** //
inline Pnt3d	&Quaternion::operator()(Pnt3d &p) const
{
  p = Pnt3d((_m_s * _m_s - _m_v * _m_v) * Vec3d(p)
	      + (2.0 * (_m_v * Vec3d(p))) * _m_v
	      + (2.0 * _m_s) * (_m_v ^ Vec3d(p)));
  return p;
}


// ******************************************* coded by Angelidis Alexis *** //
inline Vec3d	&Quaternion::operator()(Vec3d &v) const
{
  v = v * (_m_s * _m_s - _m_v * _m_v)
    + _m_v * (2.0 * (_m_v * v))
    + (_m_v ^ v) * (2.0 * _m_s);
  return v;
}

// *************************************** created by Alexis Angelidis ***** //
inline Pnt3d	&Quaternion::operator()(Pnt3d		&p,
					const double	t) const
{
  const Quaternion	qt(operator^(t));
  
  qt(p);
  return p;
}

// ******************************************* coded by Angelidis Alexis *** //
inline void	Quaternion::operator()(Pnt3d	&p,
				       Vec3d	&v) const
{
  p = Pnt3d(Vec3d(p) * (_m_s * _m_s - _m_v * _m_v)
	      + _m_v * (2.0 * (_m_v * Vec3d(p)))
	      + (_m_v ^ Vec3d(p)) * (2.0 * _m_s));
  v = v * (_m_s * _m_s - _m_v * _m_v)
    + _m_v * (2.0 * (_m_v * v))
    + (_m_v ^ v) * (2.0 * _m_s);
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Constructor, for easy construction.
// ******************************************* coded by Angelidis Alexis *** //
inline Quaternion	Quaternion::creator(const double	angle,
					    const Vec3d		&n)
{
  const double	angle2(angle * 0.5);

  return Quaternion(cos(angle2), n.unit() * sin(angle2));
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: this needs to be a unit quaternion.
// ***************************************** coded by Angelidis Alexis ***** //
inline void	Quaternion::eqLog()
{
  const double	sinTheta(_m_v.norm());

  if (sinTheta == 0.0)
    _m_v = 0.0;
  else
    _m_v *= atan2(sinTheta, _m_s) / sinTheta;
  _m_s = 0.0;
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: this needs to be a unit quaternion.
// ***************************************** coded by Angelidis Alexis ***** //
inline Quaternion	Quaternion::log() const
{
  const double	sinTheta(_m_v.norm());
    
  if (sinTheta == 0.0)
    return Quaternion(0.0);
  return Quaternion(0.0, (atan2(sinTheta, _m_s) / sinTheta) * _m_v);
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Produces a unit quaternion.
// ******************************************* coded by Alexis Angelidis *** //
inline void	Quaternion::eqExp()
{
  const double	theta(_m_v.norm());

  if (theta == 0.0)
    _m_s = 1.0, _m_v = 0.0;
  else
    _m_s = cos(theta), _m_v *= sin(theta) / theta;
}

// *************************************** created by Angelidis Alexis ***** //
/// DESCRIPTION: Produces a unit quaternion.
// ******************************************* coded by Alexis Angelidis *** //
inline Quaternion	Quaternion::exp() const
{
  const double	theta(_m_v.norm());

  if (theta == 0.0)
    return Quaternion(1.0, 0.0);
  return Quaternion(cos(theta), (sin(theta) / theta) * _m_v);
}

// ***************************************** created by Alexis Angelidis *** //
inline void	Quaternion::glRotate() const
{
  glRotated(RAD2DEG(2.0 * myacos(_m_s)), _m_v.x, _m_v.y, _m_v.z);
}

// ***************************************** coded by Angelidis Alexis ***** //
/// DESCRIPTION: this need to be a unit quaternion.
// ***************************************** created by Alexis Angelidis *** //
inline Quaternion	Quaternion::operator^(const double d) const
{
  const double	sinTheta(_m_v.norm());
  
  if (sinTheta == 0.0)
    return Quaternion(1.0, 0.0);

  const double	angle((atan2(sinTheta, _m_s) * d));

  return Quaternion(cos(angle), _m_v * (sin(angle) / sinTheta));
}

// ***************************************** coded by Alexis Angelidis ***** //
inline Quaternion	&Quaternion::operator^=(const double d)
{
  const double	sinTheta(_m_v.norm());
  
  if (sinTheta == 0.0)
    _m_s = 1.0, _m_v = 0.0;
  else
    {
      const double	angle((atan2(sinTheta, _m_s) * d));

      _m_s = cos(angle), _m_v *= (sin(angle) / sinTheta);
    }
  return *this;
}

// ***************************************** created by Angelidis Alexis *** //
/// DESCRIPTION: Basic Mat2i class. 2D and 1D indexes:
/// m = ( a00 a01 ) = ( m0  m2 )
//      ( a10 a11 )   ( m1  m3 )
// ***************************************** coded by Alexis Angelidis ***** //
struct Mat2i
{
  int	_m_m[4];
public:
  Mat2i() {}
  Mat2i(const int a00, const int a01, const int a10, const int a11)
  { _m_m[0] = a00; _m_m[2] = a01; _m_m[1] = a10; _m_m[3] = a11; }
  Mat2i(const GLenum pname) { glGetIntegerv(pname, _m_m); }
  int	&operator()(unsigned i, unsigned j) { return _m_m[2 * j + i]; }
  int	operator()(unsigned i, unsigned j) const {return _m_m[2 * j + i];}
  int	&operator()(unsigned i) { return _m_m[i]; }
  int	operator()(unsigned i) const { return _m_m[i]; }
  void	glGet(const GLenum pname) { glGetIntegerv(pname, _m_m); }
  void	glLoad() const;
  friend std::ostream	&operator<<(std::ostream &os, const Mat2i &m)
  { for (unsigned i(0); i < 4; ++i) os << m._m_m[i] << " "; return os; }
  friend std::istream	&operator>>(std::istream &is, Mat2i &m)
  { for (unsigned i(0); i < 4; ++i) is >> m._m_m[i]; return is; }
  bool	operator!=(const Mat2i &m) const;
};

// ******************************************* coded by Alexis Angelidis *** //
inline bool	Mat2i::operator!=(const Mat2i &m) const
{
  for (int i(0); i < 4; ++i)
    if (m._m_m[i] != _m_m[i])
      return true;
  return false;
}

// ***************************************** created by Angelidis Alexis *** //
/// DESCRIPTION: Loads this as a viewport (a 2x2 int matrix can't be loaded
///  anywhere else).
// ******************************************* coded by Alexis Angelidis *** //
inline void	Mat2i::glLoad() const
{
  glViewport(operator()(0), operator()(1), operator()(2), operator()(3));
}

// *************************************** created by Angelidis Alexis ***** //
/// DESCRIPTION:
/// Stored at following indexes:
/// M = ( m[0] m[2] )
///     ( m[1] m[3] )
// ******************************************* coded by Alexis Angelidis *** //
class Mat2d
{
  double	_m_m[4];
public:
  Mat2d() {}
  Mat2d(double a) {_m_m[0] = _m_m[2] = _m_m[1] = _m_m[3] = a; }
  Mat2d(double a00, double a01, double a10, double a11)
  { _m_m[0] = a00, _m_m[2] = a01; _m_m[1] = a10, _m_m[3] = a11; }
  Mat2d(const Vec2d &v)
  { for (int i(0); i < 2; ++i) for (int j(0); j < 2; ++j)
    operator()(i, j) = v(i) * v(j); }
  Mat2d(const Vec2d &p0, const Vec2d &p1)
  { _m_m[0] = p0.x, _m_m[1] = p0.y, _m_m[2] = p1.x, _m_m[3] = p1.y; }
  Mat2d(const Pnt2d &p0, const Pnt2d &p1)
  { _m_m[0] = p0.x, _m_m[1] = p0.y, _m_m[2] = p1.x, _m_m[3] = p1.y; }
  Mat2d	inverse() const;
  Mat2d	transpose() const;
  void		eqInverse();
  void		eqTranspose();
  Mat2d	operator*(const Mat2d &m) const;
  Mat2d	operator+(const Mat2d &m) const;
  Mat2d	&operator-=(const Mat2d &that);
  Mat2d	&operator+=(const Mat2d &that);
  Mat2d	operator-(const Mat2d &m) const;
  Mat2d	operator*(const double d) const;
  Vec2d	operator*(const Vec2d &v) const;
  double	det() const { return _m_m[0] * _m_m[3] - _m_m[1] * _m_m[2]; }
  double	tr() const { return _m_m[0] + _m_m[3]; }
  double	&operator()(unsigned i, unsigned j) { return _m_m[2 * j + i]; }
  const double	&operator()(unsigned i, unsigned j) const
  { return _m_m[2 * j + i]; }
  friend std::ostream	&operator<<(std::ostream &os, const Mat2d &m)
  { for (unsigned i(0); i < 4; ++i) os << m._m_m[i] << " "; return os; }
  static const Mat2d	&identity();

  int		eigenvalues(double ret[2], const double epsilon = 0.0) const;
  int		sortEigenvalues(double ret[2]) const;
  Vec2d		eigenvector(const double z) const;
  void		eigenvectors(const double evals[2], Vec2d ev[2]) const;
  int		eigenDecomposition(double eval[2], Vec2d evec[2]) const;
  bool	equals(const Mat2d &m, const double epsilon = 1.0e-12) const;
  /*
  int	eigenvalues(double ret[2]) const;
  int	sortEigenvalues(double ret[2]) const;
  void	eigenvector(const double evals[2], Vec2d ev[2]) const;
  Vec2d	eigenvector(const double z) const;
  //int	eigen(double evals[2], Vec2d ev[2]) const;
  bool	equals(const Mat2d &m, const double epsilon = 1.0e-12) const;
  */
};

// *************************************** created by Angelidis Alexis ***** //
inline Mat2d	Mat2d::inverse() const
{
  return Mat2d(_m_m[3], -_m_m[2], -_m_m[1], _m_m[0]) * (1.0 / det());
}

// ******************************************* coded by Alexis Angelidis *** //
inline void	Mat2d::eqInverse()
{
  *this =  Mat2d(_m_m[3], -_m_m[2], -_m_m[1], _m_m[0]) * (1.0 / det());
}

// *************************************** created by Angelidis Alexis ***** //
inline Mat2d	Mat2d::transpose() const
{
  return Mat2d(_m_m[0], _m_m[1],
		  _m_m[2], _m_m[3]);
}

// ******************************************* coded by Alexis Angelidis *** //
inline void	Mat2d::eqTranspose()
{
  std::swap(_m_m[1], _m_m[2]);
}

// *************************************** created by Angelidis Alexis ***** //
inline Mat2d	Mat2d::operator*(const Mat2d &m) const
{
  Mat2d	ret(0.0);

  for (int i(0); i < 2; ++i)
    for (int j(0); j < 2; ++j)
      ret(i, j) = operator()(i, 0) * m(0, j) + operator()(i, 1) * m(1, j);
  return ret;
}

// ******************************************* coded by Alexis Angelidis *** //
inline Mat2d	Mat2d::operator+(const Mat2d &m) const
{
  Mat2d	ret;

  for (int i(0); i < 4; ++i)
    ret._m_m[i] = _m_m[i] + m._m_m[i];
  return ret;
}

// *************************************** created by Angelidis Alexis ***** //
inline Mat2d	&Mat2d::operator-=(const Mat2d &that)
{
  for (unsigned i(0); i < 4; ++i)
    _m_m[i] -= that._m_m[i];
  return *this;
}

// ******************************************* coded by Alexis Angelidis *** //
inline Mat2d	&Mat2d::operator+=(const Mat2d &that)
{
  for (unsigned i(0); i < 4; ++i)
    _m_m[i] += that._m_m[i];
  return *this;
}

// ***************************************** created by Angelidis Alexis *** //
inline Mat2d	Mat2d::operator-(const Mat2d &m) const
{
  Mat2d	ret;
  
  for (int i(0); i < 4; ++i)
    ret._m_m[i] = _m_m[i] - m._m_m[i];
  return ret;
}

// ******************************************* coded by Alexis Angelidis *** //
inline Mat2d	Mat2d::operator*(const double d) const
{
  Mat2d	ret;
  
  for (int i(0); i < 4; ++i)
    ret._m_m[i] = _m_m[i] * d;
  return ret;
}

// *************************************** created by Angelidis Alexis ***** //
inline Vec2d	Mat2d::operator*(const Vec2d &v) const
{
  return Vec2d(operator()(0, 0) * v.x + operator()(0, 1) * v.y,
		  operator()(1, 0) * v.x + operator()(1, 1) * v.y);
}

// ******************************************* coded by Alexis Angelidis *** //
inline bool	Mat2d::equals(const Mat2d	&m,
			      const double	epsilon) const
{
  for (int i(0); i < 4; ++i)
    if (epsilon < fabs(_m_m[i] - m._m_m[i]))
      return false;
  return true;
}

// *************************************** created by Angelidis Alexis ***** //
inline const Mat2d	&Mat2d::identity()
{
  static const Mat2d	id(1.0, 0.0, 0.0, 1.0);

  return id;
}

// ******************************************* coded by Alexis Angelidis *** //
/// DESCRIPTION:
///  const double a2(1.0);
// *************************************** created by Angelidis Alexis ***** //
inline int	Mat2d::eigenvalues(double	ret[2],
				   const double	epsilon) const
{
  const double a0(det());
  const double a1(-tr());
  const double delta(a1 * a1 - 4.0 * a0);

  if (delta < -epsilon)
    return 0;
  if (fabs(delta) <= epsilon)
    {
      ret[0] = ret[1] = -a1 * 0.5;
      return 2;
    }
  const double	sqrtdelta(sqrt(delta));

  ret[0] = (sqrtdelta - a1) * 0.5;
  ret[1] = -(sqrtdelta + a1) * 0.5;
  return 2;
}

// ******************************************* coded by Alexis Angelidis *** //
/// DESCRIPTION: Sorts the eigenvalues in order of decreasing magnitude.
// *************************************** created by Angelidis Alexis ***** //
inline int	Mat2d::sortEigenvalues(double ret[2]) const
{
  const int	n(eigenvalues(ret));

  if (n == 2 && ret[0] < ret[1])
    std::swap(ret[0], ret[1]);
  return n;
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Returns an eigenvector associated with eigenvalue z.
// ***************************************** created by Alexis Angelidis *** //
inline Vec2d	Mat2d::eigenvector(const double z) const
{
  const double m00(_m_m[0] - z), m11(_m_m[3] - z);
  const double m01(_m_m[2]), m10(_m_m[1]);

  if (fabs(m00) > fabs(m10))
    if (fabs(m00) > fabs(m01))
      if (fabs(m00) > fabs(m11))
	return Vec2d(-m01 / m00, 1.0);
      else
	return Vec2d(1.0, -m10 / m11);
    else 
      if (fabs(m01) > fabs(m11))
	return Vec2d(1.0, -m00 / m01);
      else
	return Vec2d(1.0, -m10 / m11);
  else
    if (fabs(m10) > fabs(m01))
      if (fabs(m10) > fabs(m11))
	return Vec2d(-m11 / m10, 1.0);
      else
	return Vec2d(1.0, -m10 / m11);
    else
      if (fabs(m01) > fabs(m11))
	return Vec2d(1.0, -m00 / m01);
      else
	return Vec2d(1.0, -m10 / m11);
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: For two eigenvalues (implicit).
///  Gets one vector, while the second is implicitly orthogonal to the first
///  one.
// ***************************************** coded by Angelidis Alexis ***** //
inline void	Mat2d::eigenvectors(const double	eval[2],
				    Vec2d		ev[2]) const
{
  ev[0] = eigenvector(eval[0]);
  ev[1] = Vec2d(-ev[0].y, ev[0].x);
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Decomposes the matrix.
// ***************************************** coded by Angelidis Alexis ***** //
inline int	Mat2d::eigenDecomposition(double	eval[2],
					  Vec2d		evec[2]) const
{
  const int	n(sortEigenvalues(eval));

  eigenvectors(eval, evec);
  return n;
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Basic Mat3d class.
///  m[0] m[3] m[6]
///  m[1] m[4] m[7]
///  m[2] m[5] m[8]
// ******************************************* coded by Angelidis Alexis *** //
class Mat3d
{
  double	_m_m[9];
public:
  Mat3d() {}
  Mat3d(const double d) { for (int i(0); i < 9; ++i) _m_m[i] = d; }
  Mat3d(const double l0, const double l1, const double l2);
  Mat3d(double a00, double a01, double a02,
	double a10, double a11, double a12,
	double a20, double a21, double a22);
  Mat3d(const Vec3d &v0, const Vec3d &v1, const Vec3d &v2);
  Mat3d(const Vec3d &v);
  void		eqCommatrix();
  void		eqInverse();
  void		eqTranspose();
  void		eqSymmetric();
  Mat3d	commatrix() const;
  Mat3d	inverse() const;
  Mat3d	transpose() const;
  Mat2d		project(const Vec3d &x, const Vec3d &y) const;
  double	project(const Vec3d &x) const;
  Vec3d	operator*(const Vec3d &v) const;
  Pnt3d	operator*(const Pnt3d &p) const;
  Mat3d	operator*(const Mat3d &m) const;
  Mat3d	operator*=(const Mat3d &m);
  Mat3d	operator*(const double d) const;
  Mat3d	operator/(const double d) const;
  Mat3d	operator+(const Mat3d &m) const;
  Mat3d	operator-(const Mat3d &m) const;
  Mat3d	&operator+=(const Mat3d &m);
  Mat3d	&operator-=(const Mat3d &m);
  Mat3d	&operator*=(const double d);
  Mat3d	&operator/=(const double d);
  Mat3d	&operator^=(const unsigned k);
  bool	operator!=(const Mat3d &m) const;
  double	&operator()(unsigned i, unsigned j) { return _m_m[3 * j + i]; }
  double	operator()(unsigned i, unsigned j) const 
  { return _m_m[3 * j + i]; }
  double	&operator()(unsigned i) { return _m_m[i]; }
  double	operator()(unsigned i) const { return _m_m[i]; }
  void		operator()(Pnt3d &p, Vec3d &v) const;
  Pnt3d	&operator()(Pnt3d &p) const;
  Pnt3d	&operator()(Pnt3d &p, const double t) const;
  double	det() const;
  double	tr() const;
  double	third() const;
  double	norm2() const;

  double	maxEigenvalue() const;
  int		eigenvalues(double ret[3]) const;
  int		sortEigenvalues(double ret[3]) const;
  Vec3d		eigenvector(const double z) const;
  void		eigenvectors(const double eval[3], Vec3d evec[3]) const;
  int		eigenDecomposition(double eval[3], Vec3d evec[3]) const;
  /*
  int		eigenvalues(double ret[3]) const;
  int		sortEigenvalues(double ret[3]) const;
  double	maxEigenvalue() const;
  Vec3d	eigenvector(const double z) const;
  int		eigen(double eval[3], Vec3d evec[3]) const;
  */
  friend std::ostream	&operator<<(std::ostream &os, const Mat3d &m)
  { for (unsigned i(0); i < 9; ++i) os << m._m_m[i] << " "; return os; }
  friend std::istream	&operator>>(std::istream &is, Mat3d &m)
  { for (unsigned i(0); i < 9; ++i) is >> m._m_m[i]; return is; }
  static const Mat3d	&identity();
  bool	equals(const Mat3d &m, const double epsilon = 1.0e-12) const;
  bool	isIdentity() const;
  void	eqMultTransMult(const Mat3d &m);
  Vec3d	col0() const { return Vec3d(_m_m[0], _m_m[1], _m_m[2]); }
  Vec3d	col1() const { return Vec3d(_m_m[3], _m_m[4], _m_m[5]); }
  Vec3d	col2() const { return Vec3d(_m_m[6], _m_m[7], _m_m[8]); }
  Vec3d	row0() const { return Vec3d(_m_m[0], _m_m[3], _m_m[6]); }
  Vec3d	row1() const { return Vec3d(_m_m[1], _m_m[4], _m_m[7]); }
  Vec3d	row2() const { return Vec3d(_m_m[2], _m_m[5], _m_m[8]); }
  Mat3d	log() const;
  Mat3d	exp() const;
  void		eqSqrt();
  double	normInfty() const;
  int		size() const { return 9; }
};

// ******************************************* coded by Alexis Angelidis *** //
inline bool	Mat3d::operator!=(const Mat3d &m) const
{
  for (int i(0); i < 9; ++i)
    if (m._m_m[i] != _m_m[i])
      return true;
  return false;
}

// ***************************************** created by Alexis Angelidis *** //
inline Mat3d	&Mat3d::operator^=(const unsigned k)
{
  Mat3d	r(*this);

  *this = identity();
  for (unsigned i(1); i <= k; i *= 2, r *= r)
    if (k & i)
      operator*=(r);
  return *this;
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Matrix square root.
// ***************************************** created by Alexis Angelidis *** //
inline void	Mat3d::eqSqrt()
{
  Mat3d	&x(*this), ix, y(identity()), iy;
  unsigned	l(4096);

  while (1.0e-6 < (x * x - *this).normInfty() && --l)
    {
      ix = x.inverse();
      iy = y.inverse();
      x += iy;
      x *= 0.5;
      y += ix;
      y *= 0.5;
    }
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Matrix norm.
// ***************************************** created by Alexis Angelidis *** //
inline double	Mat3d::normInfty() const
{
  unsigned	i, j;
  double	norm(0.0), tmp;
  
  for (i = 0; i < 3; ++i)
    {
      for (tmp = 0.0, j = 0; j < 3; ++j)
	tmp += fabs(operator()(i, j));
      if (norm < tmp)
	norm = tmp;
    }	
  return norm;
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Matrix logarithm.
///  If closed-form does not work, uses a numerical approximation.
// ***************************************** created by Alexis Angelidis *** //
inline Mat3d	Mat3d::log() const
{
  const Mat3d	n(identity() - *this);
  double	eval[3];
  Vec3d	evec[3];
  int tmp;

  if ((tmp = n.eigenDecomposition(eval, evec)) == 3 && eval[0] < 1.0 && 
      eval[1] < 1.0 && eval[2] < 1.0)
    return Mat3d(evec[0] * ::log(1.0 - eval[0]),
		    evec[1] * ::log(1.0 - eval[1]),
		    evec[2] * ::log(1.0 - eval[2])) * 
      Mat3d(evec[0], evec[1], evec[2]).inverse();

  std::cout << eval[0] << " " << eval[1] << " " << eval[2] 
	    << " " << tmp << std::endl;

  double	i(1.0);
  Mat3d	z, x , a(*this);
  unsigned	l(512), k(0);
  
  while (0.5 < (a - identity()).normInfty())
    {
      a.eqSqrt();
      ++k;
    }
  z = x = a = Mat3d::identity() - a;
  while (1.0e-6 < z.normInfty() && --l)
    {
      z *= a;
      x += z * (1.0 / ++i);
    }
  x *= -double(1 << k);
  return x;
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Matrix exponential.
///  If closed-form does not work, uses a numerical approximation.
// ***************************************** created by Angelidis Alexis *** //
inline Mat3d	Mat3d::exp() const
{
  double	eval[3];
  Vec3d	evec[3];

  if (eigenDecomposition(eval, evec) == 3)
    return Mat3d(evec[0] * ::exp(eval[0]),
		 evec[1] * ::exp(eval[1]),
		 evec[2] * ::exp(eval[2])) * 
      Mat3d(evec[0], evec[1], evec[2]).inverse();

  Mat3d	d(identity()), n(identity());
  Mat3d	x(identity()), a(*this);
  unsigned	j(max(0, 1 + int(floor(::log(a.normInfty())))));
  unsigned	powj(1 << j);
  double	c(1.0), k, q(6.0);

  a *= 1.0 / double(powj);
  for (k = 1.0; k < q; ++k)
    {
      c *= (q - k + 1.0) / (k * (2.0 * q - k + 1.0));
      x *= a;
      n += x * c;
      d += x * ((int(k) % 2)? -c: c);
    }
  d.eqInverse();
  x = d * n;
  x ^= powj;
  return x;
}

// ******************************************* coded by Alexis Angelidis *** //
inline bool	Mat3d::equals(const Mat3d	&m,
			      const double	epsilon) const
{
  for (int i(0); i < 9; ++i)
    if (epsilon < fabs(_m_m[i] - m._m_m[i]))
      return false;
  return true;
}

// ***************************************** created by Angelidis Alexis *** //
/// DESCRIPTION: Returns true if the matrix is the identity
// ***************************************** created by Alexis Angelidis *** //
inline bool	Mat3d::isIdentity() const
{
  return 
    (fabs(_m_m[0]) == 1.0 && fabs(_m_m[3]) == 0.0 && fabs(_m_m[6]) == 0.0 && 
     fabs(_m_m[1]) == 0.0 && fabs(_m_m[4]) == 1.0 && fabs(_m_m[7]) == 0.0 &&
     fabs(_m_m[2]) == 0.0 && fabs(_m_m[5]) == 0.0 && fabs(_m_m[8]) == 1.0);
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: If there is one, computes the intersection of three planes
///  and returns true, else returns false.
/// NOTES: This method works if the planes are not normalized.
// ***************************************** created by Alexis Angelidis *** //
inline bool	Plane::intersect(const Plane	&p0,
				 const Plane	&p1,
				 const Plane	&p2,
				 Pnt3d		&ret)
{
  const Vec3d	&v0(p0._m_v), &v1(p1._m_v), &v2(p2._m_v);
  const double		det(Mat3d(v0, v1, v2).det());
  
  if (fabs(det) < 1.0e-8)
    return false;
  ret = Pnt3d(-((v1 ^ v2) * p0._m_t + (v2 ^ v0) * p1._m_t +
		  (v0 ^ v1) * p2._m_t) / det);
  return true;
}

// ***************************************** created by Angelidis Alexis *** //
inline Mat3d::Mat3d(double a00, double a01, double a02,
		    double a10, double a11, double a12,
		    double a20, double a21, double a22)
{
  _m_m[0] = a00, _m_m[3] = a01, _m_m[6] = a02; 
  _m_m[1] = a10, _m_m[4] = a11, _m_m[7] = a12; 
  _m_m[2] = a20, _m_m[5] = a21, _m_m[8] = a22;
}

// ******************************************* coded by Alexis Angelidis *** //
inline Mat3d	Mat3d::operator*(const Mat3d &m) const
{
  return Mat3d
    (_m_m[0] * m._m_m[0] + _m_m[3] * m._m_m[1] + _m_m[6] * m._m_m[2],
     _m_m[0] * m._m_m[3] + _m_m[3] * m._m_m[4] + _m_m[6] * m._m_m[5],
     _m_m[0] * m._m_m[6] + _m_m[3] * m._m_m[7] + _m_m[6] * m._m_m[8],
     _m_m[1] * m._m_m[0] + _m_m[4] * m._m_m[1] + _m_m[7] * m._m_m[2],
     _m_m[1] * m._m_m[3] + _m_m[4] * m._m_m[4] + _m_m[7] * m._m_m[5],
     _m_m[1] * m._m_m[6] + _m_m[4] * m._m_m[7] + _m_m[7] * m._m_m[8],
     _m_m[2] * m._m_m[0] + _m_m[5] * m._m_m[1] + _m_m[8] * m._m_m[2],
     _m_m[2] * m._m_m[3] + _m_m[5] * m._m_m[4] + _m_m[8] * m._m_m[5],
     _m_m[2] * m._m_m[6] + _m_m[5] * m._m_m[7] + _m_m[8] * m._m_m[8]);
}

// ***************************************** created by Angelidis Alexis *** //
inline Mat3d	Mat3d::operator*=(const Mat3d &m)
{
  return *this = (*this * m);
}

// ******************************************* coded by Alexis Angelidis *** //
/// DESCRIPTION: this = M * this * M^T
/// Since the matrix is symmetric, all coeffs do not need to be computed.
// ***************************************** created by Angelidis Alexis *** //
inline void	Mat3d::eqMultTransMult(const Mat3d &m)
{
  const Mat3d	ret(m * (*this));

  _m_m[0] = ret(0, 0) * m(0, 0) + ret(0, 1) * m(0, 1) + ret(0, 2) * m(0, 2);
  _m_m[4] = ret(1, 0) * m(1, 0) + ret(1, 1) * m(1, 1) + ret(1, 2) * m(1, 2);
  _m_m[8] = ret(2, 0) * m(2, 0) + ret(2, 1) * m(2, 1) + ret(2, 2) * m(2, 2);
  _m_m[2] = _m_m[6] = ret(0, 0) * m(2, 0) + ret(0, 1) * m(2, 1)
    + ret(0, 2) * m(2, 2);
  _m_m[1] = _m_m[3] = ret(1, 0) * m(0, 0) + ret(1, 1) * m(0, 1)
    + ret(1, 2) * m(0, 2);
  _m_m[5] = _m_m[7] = ret(2, 0) * m(1, 0) + ret(2, 1) * m(1, 1)
    + ret(2, 2) * m(1, 2);
}

// ******************************************* coded by Alexis Angelidis *** //
inline const Mat3d	&Mat3d::identity()
{
  static const Mat3d	id(1.0, 0.0, 0.0,
			   0.0, 1.0, 0.0,
			   0.0, 0.0, 1.0);

  return id;
}

// ***************************************** created by Angelidis Alexis *** //
/// DESCRIPTION: this = this ^ -1
// ******************************************* coded by Alexis Angelidis *** //
inline void	Mat3d::eqInverse()
{
  const double	_m_m0(_m_m[0]), _m_m1(_m_m[1]), _m_m2(_m_m[2]);
  const double	_m_m3(_m_m[3]), _m_m4(_m_m[4]), _m_m6(_m_m[6]);
  const double	invd(1.0 / (_m_m0 * (_m_m4 * _m_m[8] - _m_m[5] * _m_m[7])
			    + _m_m1 * (_m_m[5] * _m_m6 - _m_m3 * _m_m[8])
			    + _m_m2 * (_m_m3 * _m_m[7] - _m_m4 * _m_m6)));

  _m_m[0] = (_m_m4 * _m_m[8] - _m_m[5] * _m_m[7]) * invd;
  _m_m[1] = (_m_m[7] * _m_m2 - _m_m[8] * _m_m1) * invd;
  _m_m[2] = (_m_m1 * _m_m[5] - _m_m2 * _m_m4) * invd;
  _m_m[3] = (_m_m[5] * _m_m[6] - _m_m3 * _m_m[8]) * invd;
  _m_m[4] = (_m_m[8] * _m_m0 - _m_m[6] * _m_m2) * invd;
  _m_m[5] = (_m_m2 * _m_m3 - _m_m0 * _m_m[5]) * invd;
  _m_m[6] = (_m_m3 * _m_m[7] - _m_m4 * _m_m[6]) * invd;
  _m_m[7] = (_m_m6 * _m_m1 - _m_m[7] * _m_m0) * invd;
  _m_m[8] = (_m_m0 * _m_m4 - _m_m1 * _m_m3) * invd;
}

// ***************************************** created by Angelidis Alexis *** //
/// DESCRIPTION: this = this ^ -1
// ******************************************* coded by Alexis Angelidis *** //
inline void	Mat3d::eqTranspose()
{
  std::swap(_m_m[3], _m_m[1]);
  std::swap(_m_m[6], _m_m[2]);
  std::swap(_m_m[7], _m_m[5]);
}

// ***************************************** created by Angelidis Alexis *** //
inline void	Mat3d::eqSymmetric()
{
  _m_m[3] = _m_m[1] = (_m_m[3] + _m_m[1]) * 0.5;
  _m_m[6] = _m_m[2] = (_m_m[6] + _m_m[2]) * 0.5;
  _m_m[7] = _m_m[5] = (_m_m[7] + _m_m[5]) * 0.5;
}

// ******************************************* coded by Alexis Angelidis *** //
inline Mat3d	Mat3d::commatrix() const
{
  Mat3d	ret(*this);

  ret.eqCommatrix();
  return ret;
}

// ***************************************** created by Alexis Angelidis *** //
inline Mat3d	Mat3d::inverse() const
{
  const double invd(1.0 / (_m_m[0] * (_m_m[4] * _m_m[8] - _m_m[5] * _m_m[7]) +
			   _m_m[1] * (_m_m[5] * _m_m[6] - _m_m[3] * _m_m[8]) +
			   _m_m[2] * (_m_m[3] * _m_m[7] - _m_m[4] * _m_m[6])));

  return Mat3d((_m_m[4] * _m_m[8] - _m_m[5] * _m_m[7]) * invd,
	       (_m_m[5] * _m_m[6] - _m_m[3] * _m_m[8]) * invd,
	       (_m_m[3] * _m_m[7] - _m_m[4] * _m_m[6]) * invd,
	       (_m_m[7] * _m_m[2] - _m_m[8] * _m_m[1]) * invd,
	       (_m_m[8] * _m_m[0] - _m_m[6] * _m_m[2]) * invd,
	       (_m_m[6] * _m_m[1] - _m_m[7] * _m_m[0]) * invd,
	       (_m_m[1] * _m_m[5] - _m_m[2] * _m_m[4]) * invd,
	       (_m_m[2] * _m_m[3] - _m_m[0] * _m_m[5]) * invd,
	       (_m_m[0] * _m_m[4] - _m_m[1] * _m_m[3]) * invd);
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: If this is a covariance matrix,
///  then this method projects it on a space of dimension 2.
/// Example: If this is the matrix of an ellipsoid, the projection makes 
///  the matrix of an ellipse.
// ***************************************** created by Alexis Angelidis *** //
inline Mat2d	Mat3d::project(const Vec3d &x,
			       const Vec3d &y) const
{
  const Vec3d	l0(x * col0(), x * col1(), x * col2());
  const Vec3d	l1(y * col0(), y * col1(), y * col2());

  return Mat2d(l0 * x, l0 * y,
	       l1 * x, l1 * y);
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: If this is a covariance matrix,
///  then this method projects it on a space of dimension 1.
/// Example: If this is the matrix of an ellipsoid, the projection makes 
///  the matrix of an ellipse.
// ***************************************** created by Alexis Angelidis *** //
inline double	Mat3d::project(const Vec3d &x) const
{
  return Vec3d(x * col0(), x * col1(), x * col2()) * x;
}

// ******************************************* coded by Angelidis Alexis *** //
inline Mat3d	Mat3d::transpose() const
{
  return Mat3d(_m_m[0], _m_m[1], _m_m[2], 
	       _m_m[3], _m_m[4], _m_m[5], 
	       _m_m[6], _m_m[7], _m_m[8]);
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Matrix determinant.
// ******************************************* coded by Alexis Angelidis *** //
inline double	Mat3d::det() const
{
  return _m_m[0] * (_m_m[4] * _m_m[8] - _m_m[5] * _m_m[7])
    + _m_m[1] * (_m_m[5] * _m_m[6] - _m_m[3] * _m_m[8])
    + _m_m[2] * (_m_m[3] * _m_m[7] - _m_m[4] * _m_m[6]);
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Matrix determinant.
// ***************************************** created by Alexis Angelidis *** //
inline double	Mat3d::tr() const
{
  return _m_m[0] + _m_m[4] + _m_m[8];
}

// ******************************************* coded by Angelidis Alexis *** //
inline double	Mat3d::third() const
{
  return -(_m_m[4] * _m_m[0] + _m_m[0] * _m_m[8] + _m_m[8] * _m_m[4]
	   - _m_m[3] * _m_m[1] - _m_m[6] * _m_m[2] - _m_m[7] * _m_m[5]);
}

// ***************************************** created by Alexis Angelidis *** //
inline double	Mat3d::norm2() const
{
  double	sum(0.0);

  for (int i(0); i < 9; ++i)
    sum += sq(_m_m[i]);
  return sqrt(sum / 3.0);
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Sorts the values and vectors.
///  Swaps with a minimum of variables.
// ***************************************** created by Alexis Angelidis *** //
inline void	sorteigen(double	eval[3],
			  Vec3d		evec[3])
{
  Vec3d	evecSwap;
  double	evalSwap;

  if (eval[1] < eval[0])
    if (eval[2] < eval[0])
      if (eval[1] < eval[2]) // 1 < 2 < 0
	std::swap(eval[1], eval[2]), std::swap(evec[1], evec[2]);
      else // 2 < 1 < 0
	;
    else // 1 < 0 < 2
      {
	evalSwap = eval[0], eval[0] = eval[2], eval[2] = eval[1], 
	  eval[1] = evalSwap;
	evecSwap = evec[0], evec[0] = evec[2], evec[2] = evec[1],
	  evec[1] = evecSwap;
      }
  else 
    if (eval[2] < eval[1])
      if (eval[0] < eval[2]) // 0 < 2 < 1
	{
	  evalSwap = eval[0], eval[0] = eval[1], eval[1] = eval[2], 
	    eval[2] = evalSwap;
	  evecSwap = evec[0], evec[0] = evec[1], evec[1] = evec[2],
	    evec[2] = evecSwap;
	}
      else // 2 < 0 < 1
	std::swap(eval[0], eval[1]), std::swap(evec[0], evec[1]);
    else // 0 < 1 < 2
      std::swap(eval[0], eval[2]), std::swap(evec[0], evec[2]);
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Returns the eigenvector of eigenvalue z.
///  If none can be found, a constant vector is returned.
// ***************************************** created by Alexis Angelidis *** //
inline Vec3d	Mat3d::eigenvector(const double z) const
{
  const Mat3d	m(*this - Mat3d::identity() * z);
  const Mat2d	subm[9] = { Mat2d(m(1, 1), m(1, 2), m(2, 1), m(2, 2)),
			    Mat2d(m(1, 2), m(1, 0), m(2, 2), m(2, 0)),
			    Mat2d(m(1, 0), m(1, 1), m(2, 0), m(2, 1)),
			    Mat2d(m(2, 1), m(2, 2), m(0, 1), m(0, 2)),
			    Mat2d(m(2, 2), m(2, 0), m(0, 2), m(0, 0)),
			    Mat2d(m(2, 0), m(2, 1), m(0, 0), m(0, 1)),
			    Mat2d(m(0, 1), m(0, 2), m(1, 1), m(1, 2)),
			    Mat2d(m(0, 2), m(0, 0), m(1, 2), m(1, 0)),
			    Mat2d(m(0, 0), m(0, 1), m(1, 0), m(1, 1)) };
  double	maxdet(subm[0].det()), tmpdet;
  int		imax(0);

  for (int i(1); i < 9; ++i)
    if (fabs(maxdet) < fabs(tmpdet = subm[i].det()))
      imax = i, maxdet = tmpdet;
  if (1.0e-12 < fabs(maxdet))
    {
      const Vec2d r =  subm[imax].inverse() * 
	Vec2d(-m((imax % 3 + 1) % 3, imax / 3),
	      -m((imax % 3 + 2) % 3, imax / 3));
      if (imax % 3 == 0)
	return Vec3d(1.0, r.x, r.y);
      if (imax % 3 == 1)
	return Vec3d(r.y, 1.0, r.x);
      if (imax % 3 == 2)
	return Vec3d(r.x, r.y, 1.0);
    }

  bool	b0, b1, b2;

  b0 = (m(0, 0) != 0.0);
  b1 = (m(0, 1) != 0.0);
  b2 = (m(0, 2) != 0.0);
  if (b0 || b1 || b2)
    if (b0 && b1 && b2)
      //TODO: find the max/min of the m(ij)?
      return Vec3d(-(m(0, 1) + m(0, 2)) / m(0, 0), 1.0, 1.0);
    else
      return Vec3d(b0? 0.0: 1.0, b1? 0.0: 1.0, b2? 0.0: 1.0);
  b0 = (m(1, 0) != 0.0);
  b1 = (m(1, 1) != 0.0);
  b2 = (m(1, 2) != 0.0);
  if (b0 || b1 || b2)
    if (b0 && b1 && b2)
      return Vec3d(-(m(1, 1) + m(1, 2)) / m(1, 0), 1.0, 1.0);
    else
      return Vec3d(b0? 0.0: 1.0, b1? 0.0: 1.0, b2? 0.0: 1.0);
  b0 = (m(2, 0) != 0.0);
  b1 = (m(2, 1) != 0.0);
  b2 = (m(2, 2) != 0.0);
  if (b0 || b1 || b2)
    if (b0 && b1 && b2)
      return Vec3d(-(m(2, 1) + m(2, 2)) / m(2, 0), 1.0, 1.0);
    else
      return Vec3d(b0? 0.0: 1.0, b1? 0.0: 1.0, b2? 0.0: 1.0);
  return Vec3d(1.0, 0.0, 0.0);
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Finds three ORTHOGONAL eigenvectors out of 
///  three eigenvalues.
///  If eigenvectors are not nicely orthogonal within 
///  an epsilon, forces them.
// ******************************************* coded by Alexis Angelidis *** //
inline void	Mat3d::eigenvectors(const double	eval[3],
				    Vec3d		evec[3]) const
{
  evec[0] = eigenvector(eval[0]);
  evec[1] = eigenvector(eval[1]);
  
  const double e0norm2(evec[0].norm2());

  if (0.01 * e0norm2 * evec[1].norm2() < sq(evec[0] * evec[1]))
    if (fabs(evec[0].x) < fabs(evec[0].y) && fabs(evec[0].x) < fabs(evec[0].z))
      evec[1] = evec[0] * (-evec[0].x / e0norm2), evec[1].x += 1.0;
    else if (fabs(evec[0].y) < fabs(evec[0].z))
      evec[1] = evec[0] * (-evec[0].y / e0norm2), evec[1].y += 1.0;
    else
      evec[1] = evec[0] * (-evec[0].z / e0norm2), evec[1].z += 1.0;
  evec[2] = evec[0] ^ evec[1];
}

// ***************************************** created by Angelidis Alexis *** //
/// DESCRIPTION: 
///  Decomposes the matrix.
///  Returns the number of real eigenvalues.
///  If there's 1, it's in z[0],
///  Uses Cardano's Formula: z^3 a2 z^2 + a1 z + a0 = 0
// ******************************************* coded by Alexis Angelidis *** //
inline int	Mat3d::eigenDecomposition(double	eval[3],
					  Vec3d		evec[3]) const
{
  const double	a0(-det()), a2(-tr()), a1(-third());
  const double	a2d3(a2 / 3.0);
  const double	Q((a1 - a2 * a2d3) / 3.0);
  const double	R((a1 * a2d3 - a0 - a2 * a2 * a2 / 13.5) * 0.5);
  const double	D(Q * Q * Q + R * R);

  if (0.0 <= D && D <= 1.0e-6)//all roots are real and at least two are equal
    {
      if (a1 == 3.0 * sq(a2d3) && a0 == sq(a2d3) * a2d3)
	{
	  //All roots are real and equal.
	  eval[0] = eval[1] = eval[2] = -a2d3;
	  evec[0] = Vec3d(1.0, 0.0, 0.0);
	  evec[1] = Vec3d(0.0, 1.0, 0.0);
	  evec[2] = Vec3d(0.0, 0.0, 1.0);
	  return 3;
	}
      //all roots are real and two are equal
      const double sqrtTmp2(a2d3 * a2d3 - a1 / 3.0);
      const double sqrtTmp(sqrtTmp2 < 0.0? 0.0: sqrt(sqrtTmp2));
      const double z1a(-2.0 * sqrtTmp - a2d3), z2a(sqrtTmp - a2d3);
      const double z1b(2.0 * sqrtTmp - a2d3), z2b(-sqrtTmp - a2d3);

      if (fabs(a0 + z1a * z2a * z2a) < fabs(a0 + z1b * z2b * z2b))
	eval[0] = z1a, eval[1] = eval[2] = z2a;
      else
	eval[0] = z1b, eval[1] = eval[2] = z2b;
      eigenvectors(eval, evec);
      sorteigen(eval, evec);
      return 3;
    }
  if (0.0 < D)//one root is real and two are complex conjugates
    {
      const double sqrtD(sqrt(D)), s(R + sqrtD), t(R - sqrtD);
      
      eval[0] = ((0.0 < s)? pow(s, 1.0 / 3.0): -pow(-s, 1.0 / 3.0)) + 
	((0.0 < t)? pow(t, 1.0 / 3.0): -pow(-t, 1.0 / 3.0)) - a2d3;
      evec[0] = eigenvector(eval[0]);

      // ANY OTHER PAIR? DEBUG
      eval[1] = eval[2] = 0.0;
      if (fabs(evec[0].x) < fabs(evec[0].y) &&
	  fabs(evec[0].x) < fabs(evec[0].z))
	evec[1] = Vec3d(1.0, 0.0, 0.0);
      else if (fabs(evec[0].y) < fabs(evec[0].z))
	evec[1] = Vec3d(0.0, 1.0, 0.0);
      else
	evec[1] = Vec3d(0.0, 0.0, 1.0);
      evec[1] = 
	(evec[1] - evec[0] * (evec[1] * evec[0]) / evec[0].norm2()).unit();
      evec[2] = evec[1] ^ evec[0];
      sorteigen(eval, evec);
      return 3;//DEBUG

      return 1;
    }
  //all roots are real and unequal.
  if (0.0 <= Q)
    std::cerr << "Error: singular decomposition." << std::endl;

  const double	sqrtmQ(sqrt(-Q)), sqrtmQ2(2.0 * sqrtmQ);
  const double	theta(acos(bound(R / (sqrtmQ * sqrtmQ * sqrtmQ), -1.0, 1.0)));

  eval[0] = sqrtmQ2 * cos(theta / 3.0) - a2d3;
  eval[1] = sqrtmQ2 * cos((theta + 2.0 * M_PI) / 3.0) - a2d3;
  eval[2] = sqrtmQ2 * cos((theta + 4.0 * M_PI) / 3.0) - a2d3;
  eigenvectors(eval, evec);
  sorteigen(eval, evec);
  return 3;
}

// ***************************************** created by Angelidis Alexis *** //
/// DESCRIPTION:
/// Eigenvectors may not be othogonal (could even aligned be aligned)
// ******************************************* coded by Alexis Angelidis *** //
inline void	eigenorthogonize(Vec3d evec[3])
{
  if ((evec[0] ^ evec[1]).norm2() < 0.1)
    if ((evec[0] ^ evec[2]).norm2() < 0.1)
      evec[0] = Vec3d(1.0, 0.0, 0.0), evec[1] = Vec3d(0.0, 1.0, 0.0),
	evec[2] = Vec3d(0.0, 0.0, 1.0);
    else
      evec[0] = evec[1] ^ evec[2];
  else if ((evec[1] ^ evec[2]).norm2() < 0.1)
    if ((evec[1] ^ evec[0]).norm2() < 0.1)
      evec[0] = Vec3d(1.0, 0.0, 0.0), evec[1] = Vec3d(0.0, 1.0, 0.0),
	evec[2] = Vec3d(0.0, 0.0, 1.0);
    else
      evec[1] = evec[2] ^ evec[0];
  else if ((evec[2] ^ evec[0]).norm2() < 0.1)
    if ((evec[2] ^ evec[1]).norm2() < 0.1)
      evec[0] = Vec3d(1.0, 0.0, 0.0), evec[1] = Vec3d(0.0, 1.0, 0.0),
	evec[2] = Vec3d(0.0, 0.0, 1.0);
    else
      evec[2] = evec[0] ^ evec[1];
  evec[1] -= evec[0] * ((evec[1] * evec[0]) / evec[0].norm2());
  evec[2] = evec[0] ^ evec[1];
}

// ***************************************** created by Angelidis Alexis *** //
/// DESCRIPTION: Returns the number of real eigenvalues.
///  If there's 1, it's in z[0],
/// Uses Cardano's Formula: z^3 a2 z^2 + a1 z + a0 = 0
// ******************************************* coded by Angelidis Alexis *** //
inline int	Mat3d::eigenvalues(double ret[3]) const
{
  const double a0((_m_m[6] * _m_m[4] - _m_m[3] * _m_m[7]) * _m_m[2]
		  + (_m_m[0] * _m_m[7] - _m_m[6] * _m_m[1]) * _m_m[5]
		  + (_m_m[3] * _m_m[1] - _m_m[0] * _m_m[4]) * _m_m[8]);
  const double a1(_m_m[4] * _m_m[0] + _m_m[0] * _m_m[8] + _m_m[8] * _m_m[4]
		  - _m_m[3] * _m_m[1] - _m_m[6] * _m_m[2] - _m_m[7] * _m_m[5]);
  const double a2(-_m_m[0] - _m_m[4] - _m_m[8]);
  const double	a2d3(a2 / 3.0);
  const double	Q((a1 - a2 * a2d3) / 3.0);
  const double	R((a1 * a2d3 - a0 - a2 * a2 * a2 / 13.5) * 0.5);
  const double	D(Q * Q * Q + R * R);

  if (D == 0.0)//all roots are real and at least two are equal
    {
      if (a1 == 3.0 * sq(a2d3) && a0 == sq(a2d3) * a2d3)
	{
	  //All roots are real and equal.
	  ret[0] = ret[1] = ret[2] = -a2d3;
	  return 3;
	}
      //all roots are real and two are equal
      const double sqrtTmp2(a2d3 * a2d3 - a1 / 3.0);
      const double sqrtTmp(sqrtTmp2 < 0.0? 0.0: sqrt(sqrtTmp2));
      //const double sqrtTmp(sqrt(a2d3 * a2d3 - a1 / 3.0));
      const double z1a(-2.0 * sqrtTmp - a2d3), z2a(sqrtTmp - a2d3);
      const double z1b(2.0 * sqrtTmp - a2d3), z2b(-sqrtTmp - a2d3);
      
      if (fabs(a0 + z1a * z2a * z2a) < fabs(a0 + z1b * z2b * z2b))
	ret[0] = z1a, ret[1] = ret[2] = z2a;
      else
	ret[0] = z1b, ret[1] = ret[2] = z2b;
      return 3;
    }
  if (0.0 < D)//one root is real and two are complex conjugates
    {
      const double sqrtD(sqrt(D)), s(R + sqrtD), t(R - sqrtD);

      ret[0] = ((0.0 < s)? pow(s, 1.0 / 3.0): -pow(-s, 1.0 / 3.0)) + 
	((0.0 < t)? pow(t, 1.0 / 3.0): -pow(-t, 1.0 / 3.0)) - a2d3;
      return 1;
    }
  //all roots are real and unequal.
  if (0.0 <= Q)
    std::cerr << "Error: singular decomposition." << std::endl;

  const double	sqrtmQ(sqrt(-Q)), sqrtmQ2(2.0 * sqrtmQ);
  const double	theta(acos(R / (sqrtmQ * sqrtmQ * sqrtmQ)));

  ret[0] = sqrtmQ2 * cos(theta / 3.0) - a2d3;
  ret[1] = sqrtmQ2 * cos((theta + 2.0 * M_PI) / 3.0) - a2d3;
  ret[2] = sqrtmQ2 * cos((theta + 4.0 * M_PI) / 3.0) - a2d3;
  return 3;
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Sorts the eigenvalues in order of decreasing magnitude.
// ******************************************* coded by Alexis Angelidis *** //
inline int	Mat3d::sortEigenvalues(double ret[3]) const
{
  const int	n(eigenvalues(ret));

  if (n == 1)
    return n;
  if (ret[2] < ret[1])
    {
      if (ret[0] < ret[1])
	std::swap(ret[1], ret[0]);
    }
  else if (ret[0] < ret[2])
    std::swap(ret[2], ret[0]);
  if (ret[1] < ret[2])
    std::swap(ret[1], ret[2]);
  return n;
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Returns the biggest eigenvalue.
// ******************************************* coded by Angelidis Alexis *** //
inline double	Mat3d::maxEigenvalue() const
{
  const double a0((_m_m[6] * _m_m[4] - _m_m[3] * _m_m[7]) * _m_m[2]
		  + (_m_m[0] * _m_m[7] - _m_m[6] * _m_m[1]) * _m_m[5]
		  + (_m_m[3] * _m_m[1] - _m_m[0] * _m_m[4]) * _m_m[8]);
  const double a1(_m_m[4] * _m_m[0] + _m_m[0] * _m_m[8] + _m_m[8] * _m_m[4]
		  - _m_m[3] * _m_m[1] - _m_m[6] * _m_m[2] - _m_m[7] * _m_m[5]);
  const double a2(-_m_m[0] - _m_m[4] - _m_m[8]);
  const double	a2d3(a2 / 3.0);
  const double	Q((a1 - a2 * a2d3) / 3.0);
  const double	R((a1 * a2d3 - a0 - a2 * a2 * a2 / 13.5) * 0.5);
  const double	D(Q * Q * Q + R * R);

  if (D == 0.0)//all roots are real and at least two are equal
    {
      if (a1 == 3.0 * sq(a2d3) && a0 == sq(a2d3) * a2d3)
	return -a2d3; //All roots are real and equal.

      //all roots are real and two are equal
      const double sqrtTmp2(a2d3 * a2d3 - a1 / 3.0);
      const double sqrtTmp(sqrtTmp2 < 0.0? 0.0: sqrt(sqrtTmp2));
      const double z1a(-2.0 * sqrtTmp - a2d3), z2a(sqrtTmp - a2d3);
      const double z1b(2.0 * sqrtTmp - a2d3), z2b(-sqrtTmp - a2d3);
      
      if (fabs(a0 + z1a * z2a * z2a) < fabs(a0 + z1b * z2b * z2b))
	return (z1a < z2a)? z2a: z1a;
      else
	return (z1b < z2b)? z2b: z1b;
    }
  if (0.0 < D)//one root is real and two are complex conjugates
    {
      const double sqrtD(sqrt(D)), s(R + sqrtD), t(R - sqrtD);

      return ((0.0 < s)? pow(s, 1.0 / 3.0): -pow(-s, 1.0 / 3.0)) + 
	((0.0 < t)? pow(t, 1.0 / 3.0): -pow(-t, 1.0 / 3.0)) - a2d3;
    }
  //all roots are real and unequal.
  if (0.0 <= Q)
    std::cerr << "Error: singular decomposition." << std::endl;

  const double	sqrtmQ(sqrt(-Q)), sqrtmQ2(2.0 * sqrtmQ);
  const double	theta(acos(R / (sqrtmQ * sqrtmQ * sqrtmQ)));
  const double	alphas[3] = { theta, (2.0 * M_PI + theta),
			      (2.0 * M_PI - theta) };
  double alpha;
  
  if (0.0 < sqrtmQ2)
    if (alphas[0] < alphas[1])
      alpha = (alphas[2] < alphas[0])? alphas[2]: alphas[0];
    else
      alpha = (alphas[2] < alphas[1])? alphas[2]: alphas[1];
  else
    if (alphas[1] < alphas[0])
      alpha = (alphas[0] < alphas[2])? alphas[2]: alphas[0];
    else
      alpha = (alphas[1] < alphas[2])? alphas[2]: alphas[1];
  return sqrtmQ2 * cos(alpha / 3.0) - a2d3;
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Returns the biggest eigenvalue.
// ******************************************* coded by Angelidis Alexis *** //
/*
inline double	Mat3d::maxEigenvalue() const
{
  double	ret[3];
  const int	n(eigenvalues(ret));
  double	maxe(ret[0]);
 
  for (int i(1); i < n; ++i)
    if (maxe < ret[i])
      maxe = ret[i];
  return maxe;
}*/

// ***************************************** created by Alexis Angelidis *** //
inline void	Mat3d::operator()(Pnt3d	&p,
				  Vec3d	&v) const
{
  p = operator*(p);
  v = operator*(v);
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Applies the matrix to the point.
// ******************************************* coded by Alexis Angelidis *** //
inline Pnt3d	&Mat3d::operator()(Pnt3d	&p) const
{
  p = operator*(p);
  return p;
}

// ***************************************** created by Angelidis Alexis *** //
inline Mat3d::Mat3d(const double l0,
		    const double l1,
		    const double l2)
{
  *this = Mat3d(l0, 0.0, 0.0, 0.0, l1, 0.0, 0.0, 0.0, l2);
}

// ******************************************* coded by Alexis Angelidis *** //
/// DESCRIPTION: Stores v0, v1 and v2 as columns of the matrix.
// ***************************************** created by Alexis Angelidis *** //
inline Mat3d::Mat3d(const Vec3d &v0,
		    const Vec3d &v1,
		    const Vec3d &v2)
{
  _m_m[0] = v0(0), _m_m[3] = v1(0), _m_m[6] = v2(0);
  _m_m[1] = v0(1), _m_m[4] = v1(1), _m_m[7] = v2(1);
  _m_m[2] = v0(2), _m_m[5] = v1(2), _m_m[8] = v2(2);
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Creates a symmetric matrix.
// ***************************************** created by Alexis Angelidis *** //
inline Mat3d::Mat3d(const Vec3d &v)
{
  for (int i(0); i < 3; ++i)
    for (int j(0); j < 3; ++j)
      operator()(i, j) = v(i) * v(j);
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Matrix of the cofactors.
// ***************************************** created by Alexis Angelidis *** //
inline void	Mat3d::eqCommatrix()
{
  const double	_m_m0(_m_m[0]), _m_m3(_m_m[3]);
  const double	_m_m6(_m_m[6]);
  const double	_m_m1(_m_m[1]), _m_m4(_m_m[4]);
  const double	_m_m7(_m_m[7]);

  _m_m[0] = _m_m[4] * _m_m[8] - _m_m[5] * _m_m[7];
  _m_m[3] = _m_m[7] * _m_m[2] - _m_m[8] * _m_m[1];
  _m_m[6] = _m_m[1] * _m_m[5] - _m_m[2] * _m_m[4];
  _m_m[1] = _m_m[5] * _m_m6 - _m_m3 * _m_m[8];
  _m_m[4] = _m_m[8] * _m_m0 - _m_m6 * _m_m[2];
  _m_m[7] = _m_m[2] * _m_m3 - _m_m0 * _m_m[5];
  _m_m[2] = _m_m3 * _m_m7 - _m_m4 * _m_m6;
  _m_m[5] = _m_m6 * _m_m1 - _m_m7 * _m_m0;
  _m_m[8] = _m_m0 * _m_m4 - _m_m1 * _m_m3;
}

// ******************************************* coded by Alexis Angelidis *** //
inline Vec3d	Mat3d::operator*(const Vec3d &v) const
{
  return Vec3d(_m_m[0] * v(0) + _m_m[3] * v(1) + _m_m[6] * v(2),
		  _m_m[1] * v(0) + _m_m[4] * v(1) + _m_m[7] * v(2),
		  _m_m[2] * v(0) + _m_m[5] * v(1) + _m_m[8] * v(2));
}

// ***************************************** created by Angelidis Alexis *** //
inline Pnt3d	Mat3d::operator*(const Pnt3d &p) const
{
  return Pnt3d(_m_m[0] * p(0) + _m_m[3] * p(1) + _m_m[6] * p(2),
		 _m_m[1] * p(0) + _m_m[4] * p(1) + _m_m[7] * p(2),
		 _m_m[2] * p(0) + _m_m[5] * p(1) + _m_m[8] * p(2));
}

// ******************************************* coded by Alexis Angelidis *** //
inline Mat3d	Mat3d::operator*(const double d) const
{
  Mat3d	ret;

  for (int i(0); i < 9; ++i)
    ret._m_m[i] = _m_m[i] * d;
  return ret;
}

// ***************************************** created by Angelidis Alexis *** //
inline Mat3d	Mat3d::operator/(const double d) const
{
  Mat3d	ret;

  for (int i(0); i < 9; ++i)
    ret._m_m[i] = _m_m[i] / d;
  return ret;
}

// ******************************************* coded by Alexis Angelidis *** //
inline Mat3d	Mat3d::operator+(const Mat3d &m) const
{
  Mat3d	ret;

  for (int i(0); i < 9; ++i)
    ret._m_m[i] = _m_m[i] + m._m_m[i];
  return ret;
}

// ***************************************** created by Angelidis Alexis *** //
inline Mat3d	Mat3d::operator-(const Mat3d &m) const
{
  Mat3d	ret;

  for (int i(0); i < 9; ++i)
    ret._m_m[i] = _m_m[i] - m._m_m[i];
  return ret;
}

// ******************************************* coded by Alexis Angelidis *** //
inline Mat3d	&Mat3d::operator+=(const Mat3d &m)
{
  for (unsigned i(0); i < 9; ++i)
    _m_m[i] += m._m_m[i];
  return *this;
}

// ***************************************** created by Alexis Angelidis *** //
inline Mat3d	&Mat3d::operator*=(const double d)
{
  for (unsigned i(0); i < 9; ++i)
    _m_m[i] *= d;
  return *this;
}

// ******************************************* coded by Angelidis Alexis *** //
inline Mat3d	&Mat3d::operator/=(const double d)
{
  for (unsigned i(0); i < 9; ++i)
    _m_m[i] /= d;
  return *this;
}

// ***************************************** created by Alexis Angelidis *** //
inline Mat3d	&Mat3d::operator-=(const Mat3d &m)
{
  for (unsigned i(0); i < 9; ++i)
    _m_m[i] -= m._m_m[i];
  return *this;
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Basic Mat4d class, openGL compatible.
///  The most generic container for transformation. Not the 
///  most efficient. Use Quaternion Translation Rotation Scale and ScaleT for 
///  fast transformations.
/// m = ( a00 a01 a02 a03 ) = ( a0 a4 a8  a12 )
///     ( a10 a11 a12 a13 ) = ( a1 a5 a9  a13 )
///     ( a20 a21 a22 a23 ) = ( a2 a6 a10 a14 )
///     ( a30 a31 a32 a33 ) = ( a3 a7 a11 a15 )
// ***************************************** created by Alexis Angelidis *** //
struct Mat4d: public Transformation
{
  double	_m_m[16];
public:
  Mat4d() {}
  Mat4d(const Mat4d &m);
  Mat4d(const Vec3d &x, const Vec3d &y, const Vec3d &z, const Pnt3d &o);
  Mat4d(const double a);
  Mat4d(double a00, double a01, double a02, double a03, 
	double a10, double a11, double a12, double a13, 
	double a20, double a21, double a22, double a23, 
	double a30, double a31, double a32, double a33);
  virtual ~Mat4d() {}
  Transformation	*copy() const { return new Mat4d(*this); }
  explicit	Mat4d(const Quaternion &q);
  explicit	Mat4d(const Translation &t);
  explicit	Mat4d(const Rotation &r);
  explicit	Mat4d(const Scale &s);
  explicit	Mat4d(const ScaleT &s);
  Mat4d	asMatrix() const { return *this; }
  double	operator()(unsigned i, unsigned j) const 
  { return _m_m[4 * j + i]; }
  double	&operator()(unsigned i, unsigned j)
  { return _m_m[4 * j + i]; }
  double	operator()(unsigned i) const { return _m_m[i]; }
  double	&operator()(unsigned i) { return _m_m[i]; }
  friend std::ostream	&operator<<(std::ostream &os, const Mat4d &m)
  { for (unsigned i(0); i < 16; ++i) os << m._m_m[i] << " "; return os; }
  friend std::istream	&operator>>(std::istream &is, Mat4d &m)
  { for (unsigned i(0); i < 16; ++i) is >> m._m_m[i]; return is; }
  Mat4d	inverse() const;
  Mat4d	myInverse() const;
  void		eqInverse();
  void		eqSymx();
  void		eqSymy();
  void		eqSymz();
  Mat4d	log() const;
  Mat4d	mlog() const { return log(); }
  void		eqLog();
  Mat4d	exp() const;
  void		eqExp();
  Mat4d	sqrt() const;
  void		eqSqrt();
  double	normInfty() const;
  Mat4d	operator*(const Mat4d &m) const;
  Mat4d	operator*(const double d) const;
  friend Mat4d	operator*(const double d, const Mat4d &m)
  { return m * d; }
  Mat4d	&operator*=(const Mat4d &that);
  Mat4d	&operator*=(const double d);
  bool	operator!=(const Mat4d &m) const;
  Mat4d	&operator+=(const Mat4d &that);
  Mat4d	&operator-=(const Mat4d &that);
  Mat4d	operator+(const Mat4d &that) const;
  Mat4d	operator-(const Mat4d &that) const;
  Mat4d	operator^(const unsigned k) const;
  Mat4d	operator^(const double t) const;
  Mat4d	&operator^=(const double t);
  Mat4d	&operator^=(const unsigned k);
  Vec3d	operator*(const Vec3d &v) const;
  Pnt3d	operator*(const Pnt3d &p) const;
  BBox3d	operator*(const BBox3d &b) const;
  double	det() const;
  static const Mat4d	&identity();
  static const Mat4d	&symx();
  static const Mat4d	&symy();
  static const Mat4d	&symz();
  static const Mat4d	&sym(const unsigned sym);
  void	glGet(const GLenum pname) { glGetDoublev(pname, _m_m); }
  void	glLoad(const GLenum pname) const;
  void	glGetModelview() { glGetDoublev(GL_MODELVIEW_MATRIX, _m_m); }
  void	glGetProjection() { glGetDoublev(GL_PROJECTION_MATRIX, _m_m); }
  Vec3d	getx() const { return Vec3d(_m_m[0], _m_m[1], _m_m[2]); }
  Vec3d	gety() const { return Vec3d(_m_m[4], _m_m[5], _m_m[6]); }
  Vec3d	getz() const { return Vec3d(_m_m[8], _m_m[9], _m_m[10]); }
  Pnt3d	geto() const { return Pnt3d(_m_m[12], _m_m[13], _m_m[14]); }
  void		glMultMatrix() const { glMultMatrixd(_m_m); }
  Pnt3d	&operator()(Pnt3d &p) const
  { p = operator*(p); return p; }
  Pnt3d		&operator()(Pnt3d &p, const double t) const
  { p = operator^(t) * p; return p; }
  void			operator()(Pnt3d &p, Vec3d &v) const
  { p = operator*(p); v = operator*(v); }
  void		cast(const Transformation &t);
  unsigned	nbSteps(const BBox3d &b, const double k) const;
  void		getLeftPlane(Plane &p) const;
  void		getRightPlane(Plane &p) const;
  void		getBottomPlane(Plane &p) const;
  void		getTopPlane(Plane &p) const;
  void		getNearPlane(Plane &p) const;
  void		getFarPlane(Plane &p) const;
  void		getFrustum(Plane p[8]) const;
  int		size() const { return 16; }
};

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Computes the viewpoint, using the modelview matrix.
///  The projection matrix should NEVER contain scene transformation!!!
// ***************************************** created by Alexis Angelidis *** //
inline void	Pnt3d::glGetViewpoint()
{
  Mat4d	m;

  m.glGetModelview();
  m.eqInverse();
  *this = m * Pnt3d(0.0);
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Computes the view direction, using the modelview matrix.
///  The projection matrix should NEVER contain scene transformation!!!
// ***************************************** created by Alexis Angelidis *** //
inline void	Vec3d::glGetViewdir()
{
  Mat4d	m;

  m.glGetModelview();
  m.eqInverse();
  *this = m * Vec3d(0.0, 0.0, -1.0);
}

// ***************************************** created by Angelidis Alexis *** //
inline void	Vec3d::glGetUp()
{
  Mat4d	m;

  m.glGetModelview();
  m.eqInverse();
  *this = m * Vec3d(0.0, 1.0, 0.0);
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Computes the frustum in world space
// ***************************************** created by Alexis Angelidis *** //
inline void	Frustum::getWorldFrustum()
{
  Mat4d	m, p, r;

  m.glGetModelview();
  p.glGetProjection();
  r = p * m;
  r.getFrustum(_m_clipPlanes);
}

// ******************************************* coded by Alexis Angelidis *** //
/// DESCRIPTION: Computes the frustum in camera space
// ******************************************* coded by Angelidis Alexis *** //
inline void	Frustum::getCameraFrustum()
{
  Mat4d	p;

  p.glGetProjection();
  p.getFrustum(_m_clipPlanes);
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Extracts the left clipping plane.
///   To obtain the clipping planes in model space, the matrix is P * M.
///   The equation of the half-space is returned in the form  v * x - t > 0.
// ******************************************* coded by Angelidis Alexis *** //
inline void	Mat4d::getLeftPlane(Plane	&p) const
{
  p._m_v = Vec3d(operator()(3, 0) + operator()(0, 0),
		    operator()(3, 1) + operator()(0, 1),
		    operator()(3, 2) + operator()(0, 2));
  p._m_t = operator()(3, 3) + operator()(0, 3);
}

// ***************************************** created by Alexis Angelidis *** //
inline void	Mat4d::getRightPlane(Plane	&p) const
{
  p._m_v = Vec3d(operator()(3, 0) - operator()(0, 0),
		    operator()(3, 1) - operator()(0, 1),
		    operator()(3, 2) - operator()(0, 2));
  p._m_t = operator()(3, 3) - operator()(0, 3);
}

// ******************************************* coded by Angelidis Alexis *** //
inline void	Mat4d::getBottomPlane(Plane	&p) const
{
  p._m_v = Vec3d(operator()(3, 0) + operator()(1, 0),
		    operator()(3, 1) + operator()(1, 1),
		    operator()(3, 2) + operator()(1, 2));
  p._m_t = operator()(3, 3) + operator()(1, 3);
}

// ******************************************* coded by Alexis Angelidis *** //
inline void	Mat4d::getTopPlane(Plane	&p) const
{
  p._m_v = Vec3d(operator()(3, 0) - operator()(1, 0),
		    operator()(3, 1) - operator()(1, 1),
		    operator()(3, 2) - operator()(1, 2));
  p._m_t = operator()(3, 3) - operator()(1, 3);
}

// ******************************************* coded by Angelidis Alexis *** //
inline void	Mat4d::getNearPlane(Plane	&p) const
{
  p._m_v = Vec3d(operator()(3, 0) + operator()(2, 0),
		    operator()(3, 1) + operator()(2, 1),
		    operator()(3, 2) + operator()(2, 2));
  p._m_t = operator()(3, 3) + operator()(2, 3);
}

// ***************************************** created by Alexis Angelidis *** //
inline void	Mat4d::getFarPlane(Plane	&p) const
{
  p._m_v = Vec3d(operator()(3, 0) - operator()(2, 0),
		    operator()(3, 1) - operator()(2, 1),
		    operator()(3, 2) - operator()(2, 2));
  p._m_t = operator()(3, 3) - operator()(2, 3);
}

// ******************************************* coded by Alexis Angelidis *** //
inline void	Mat4d::getFrustum(Plane	p[8]) const
{
  getLeftPlane(p[0]);
  getRightPlane(p[1]);
  getBottomPlane(p[2]);
  getTopPlane(p[3]);
  getNearPlane(p[4]);
  getFarPlane(p[5]);
}

// ***************************************** created by Angelidis Alexis *** //
/// DESCRIPTION: Determinant.
// ******************************************* coded by Alexis Angelidis *** //
inline double	Mat4d::det() const
{
  Mat3d	sub;
  double	s, det = 0.0;

  for (unsigned i(0); i < 4; ++i)
    {
      s = (i % 2)? -1.0: 1.0;
      for (unsigned si(0); si < 3; ++si)
	for (unsigned sj(0); sj < 3; ++sj)
	  sub(si, sj) = s * operator()((i + 1 + si) % 4, (1 + sj) % 4);
      det += operator()(i, 0) * sub.det();
    }
  return det;
}

// ***************************************** created by Angelidis Alexis *** //
inline unsigned	Mat4d::nbSteps(const BBox3d	&b,
				  const double	k) const
{
  const Mat4d	logThis(log());
  double	l2tmp, lmax(0.0);

  for (int i(0); i < 8; ++i)
    {
      l2tmp = Vec3d((logThis * b.corner(i))).norm2();
      if (lmax < l2tmp)
	lmax = l2tmp;
    }
  lmax = ::sqrt(lmax);

  const double s = ceil(1.875 / k * 2.0 * lmax);

  if (s <= 1.0)
    return 1;
  return unsigned(s);
}

// ******************************************* coded by Alexis Angelidis *** //
inline void	Mat4d::cast(const Transformation &t)
{
  const Mat4d *m(dynamic_cast<const Mat4d *>(&t));

  if (!m)
    {
      std::cerr << "A Mat4d can only cast a Mat4d." << std::endl;
      std::exit(1);
    }
  *this = *m;
}

// ***************************************** created by Angelidis Alexis *** //
inline Mat4d	Transformation::operator*(const Mat4d &m) const
{
  return asMatrix() * m;
}

// ******************************************* coded by Alexis Angelidis *** //
inline Mat4d	Transformation::operator*(const Transformation &t) const
{
  return operator*(t.asMatrix());
}

// ***************************************** created by Angelidis Alexis *** //
inline Mat4d	Quaternion::asMatrix() const
{
  return Mat4d(*this);
}

// ******************************************* coded by Alexis Angelidis *** //
inline Mat4d	Quaternion::mlog() const
{
  const double	sinTheta(_m_v.norm());
    
  if (sinTheta == 0.0)
    return Mat4d(0.0);

  const Vec3d	n((2.0 * atan2(sinTheta, _m_s) / sinTheta) * _m_v);

  return Mat4d(0.0, -n.z, n.y, 0.0,
		  n.z, 0.0, -n.x, 0.0,
		  -n.y, n.x, 0.0, 0.0,
		  0.0, 0.0, 0.0, 0.0);
}

// ***************************************** created by Angelidis Alexis *** //
inline Vec3d	&Vec3d::operator*=(const Mat4d	&m)
{
  const double	tmpx(x), tmpy(y); 

  x = m(0, 0) * x + m(0, 1) * y + m(0, 2) * z;
  y = m(1, 0) * tmpx + m(1, 1) * y + m(1, 2) * z;
  z = m(2, 0) * tmpx + m(2, 1) * tmpy + m(2, 2) * z;
  return *this;
}

// ******************************************* coded by Alexis Angelidis *** //
inline Vec3d	&Vec3d::operator*=(const Mat3d	&m)
{
  const double	tmpx(x), tmpy(y); 

  x = m(0, 0) * x + m(0, 1) * y + m(0, 2) * z;
  y = m(1, 0) * tmpx + m(1, 1) * y + m(1, 2) * z;
  z = m(2, 0) * tmpx + m(2, 1) * tmpy + m(2, 2) * z;
  return *this;
}

// ***************************************** created by Angelidis Alexis *** //
inline Pnt3d	&Pnt3d::operator*=(const Mat4d	&m)
{
  const double	tmpx(x), tmpy(y); 
  
  x = m(0, 0) * x + m(0, 1) * y + m(0, 2) * z + m(0, 3);
  y = m(1, 0) * tmpx + m(1, 1) * y + m(1, 2) * z + m(1, 3);
  z = m(2, 0) * tmpx + m(2, 1) * tmpy + m(2, 2) * z + m(2, 3);
  return *this;
}

// ******************************************* coded by Alexis Angelidis *** //
/// DESCRIPTION: Value constructor
// ******************************************* coded by Alexis Angelidis *** //
inline	Mat4d::Mat4d(const double a)
{
  for (unsigned i(0); i < 16; ++i)
    _m_m[i] = a;
}

// ***************************************** created by Angelidis Alexis *** //
inline Mat4d::Mat4d(const Vec3d	&x,
			  const Vec3d	&y,
			  const Vec3d	&z,
			  const Pnt3d		&o)
{
  _m_m[0] = x(0); _m_m[4] = y(0); _m_m[8] = z(0); _m_m[12] = o(0);
  _m_m[1] = x(1); _m_m[5] = y(1); _m_m[9] = z(1); _m_m[13] = o(1);
  _m_m[2] = x(2); _m_m[6] = y(2); _m_m[10] = z(2); _m_m[14] = o(2);
  _m_m[3] = 0.0; _m_m[7] = 0.0; _m_m[11] = 0.0; _m_m[15] = 1.0; 
}

// ******************************************* coded by Alexis Angelidis *** //
inline Mat4d::Mat4d(const Mat4d &m): Transformation()
{
  for (unsigned i(0); i < 16; ++i)
    _m_m[i] = m._m_m[i];
}

// ***************************************** created by Angelidis Alexis *** //
/// DESCRIPTION: Value constructor.
// ***************************************** created by Alexis Angelidis *** //
inline	Mat4d::Mat4d(double a00, double a01, double a02, double a03, 
		     double a10, double a11, double a12, double a13, 
		     double a20, double a21, double a22, double a23, 
		     double a30, double a31, double a32, double a33)
{
  _m_m[0] = a00, _m_m[4] = a01, _m_m[8] = a02, _m_m[12] = a03;
  _m_m[1] = a10, _m_m[5] = a11, _m_m[9] = a12, _m_m[13] = a13;
  _m_m[2] = a20, _m_m[6] = a21, _m_m[10] = a22, _m_m[14] = a23;
  _m_m[3] = a30, _m_m[7] = a31, _m_m[11] = a32, _m_m[15] = a33;
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Matrix inverse which supports -O3 optimization.
// ***************************************** created by Alexis Angelidis *** //
inline Mat4d	Mat4d::inverse() const
{
  Mat4d	inv;
  Mat3d	sub0, sub1;
  double	s, det = 0.0;

  for (unsigned i(0); i < 4; ++i)
    {
      s = (i % 2)? -1.0: 1.0;
      for (unsigned si(0); si < 3; ++si)
	for (unsigned sj(0); sj < 3; ++sj)
	  sub0(si, sj) = s * operator()((i + 1 + si) % 4, (1 + sj) % 4),
	    sub1(si, sj) = s * operator()((i + 1 + si) % 4, (3 + sj) % 4);
      inv(0, i) = sub0.det();
      inv(2, i) = sub1.det();
      det += operator()(i, 0) * sub0.det();
      for (unsigned si(0); si < 3; ++si)
	for (unsigned sj(0); sj < 3; ++sj)
	  sub0(si, sj) = -s * operator()((i + 1 + si) % 4, (2 + sj) % 4),
	    sub1(si, sj) = -s * operator()((i + 1 + si) % 4, sj % 4);
      inv(1, i) = sub0.det();
      inv(3, i) = sub1.det();
    }
  if (fabs(det) < 1.0e-20)
    {
      std::cerr << "Singular matrix" << std::endl;
      std::exit(1);
    }
  return inv * (1.0 / det);
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Matrix inversion.
// ******************************************* coded by Alexis Angelidis *** //
inline void	Mat4d::eqInverse()
{
  Mat4d	inv;
  Mat3d	sub0, sub1;
  double	s, det = 0.0;

  for (unsigned i(0); i < 4; ++i)
    {
      s = (i % 2)? -1.0: 1.0;
      for (unsigned si(0); si < 3; ++si)
	for (unsigned sj(0); sj < 3; ++sj)
	  sub0(si, sj) = s * operator()((i + 1 + si) % 4, (1 + sj) % 4),
	    sub1(si, sj) = s * operator()((i + 1 + si) % 4, (3 + sj) % 4);
      inv(0, i) = sub0.det();
      inv(2, i) = sub1.det();
      det += operator()(i, 0) * sub0.det();
      for (unsigned si(0); si < 3; ++si)
	for (unsigned sj(0); sj < 3; ++sj)
	  sub0(si, sj) = -s * operator()((i + 1 + si) % 4, (2 + sj) % 4),
	    sub1(si, sj) = -s * operator()((i + 1 + si) % 4, sj % 4);
      inv(1, i) = sub0.det();
      inv(3, i) = sub1.det();
    }
  if (det < 1.0e-20)
    {
      std::cerr << "Singular matrix" << std::endl;
      std::exit(1);
    }
  *this = inv * (1.0 / det);
}

// ***************************************** created by Alexis Angelidis *** //
inline void	Mat4d::eqSymx()
{
  *this = Mat4d::symx() * *this;
}

// ******************************************* coded by Angelidis Alexis *** //
inline void	Mat4d::eqSymy()
{
  *this = Mat4d::symy() * *this;
}

// ***************************************** created by Alexis Angelidis *** //
inline void	Mat4d::eqSymz()
{
  *this = Mat4d::symz() * *this;
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Matrix logarithm.
// ***************************************** created by Alexis Angelidis *** //
inline Mat4d	Mat4d::log() const
{
  double	i(1.0);
  Mat4d	z, x , a(*this);
  unsigned	l(4096), k(0);
  
  while (0.5 < (a - identity()).normInfty())
    {
      a.eqSqrt();
      ++k;
    }
  z = x = a = Mat4d::identity() - a;
  while (1.0e-6 < z.normInfty() && --l)
    {
      z *= a;
      x += z * (1.0 / ++i);
    }
  x *= -double(1 << k);
  return x;
}

// ******************************************* coded by Alexis Angelidis *** //
/// DESCRIPTION: Matrix logarithm.
// ***************************************** created by Angelidis Alexis *** //
inline void	Mat4d::eqLog()
{
  double	i(1.0);
  Mat4d	z, &x(*this), a(*this);
  unsigned	l(4096), k(0);
  
  while (0.5 < (a - identity()).normInfty())
    {
      a.eqSqrt();
      ++k;
    }
  z = x = a = Mat4d::identity() - a;
  while (1.0e-6 < z.normInfty() && --l)
    {
      z *= a;
      x += z * (1.0 / ++i);
    }
  x *= -double(1 << k);
}

// ******************************************* coded by Alexis Angelidis *** //
/// DESCRIPTION: Matrix exponential.
// ******************************************* coded by Alexis Angelidis *** //
inline Mat4d	Mat4d::exp() const
{
  Mat4d	d(identity()), n(identity());
  Mat4d	x(identity()), a(*this);
  unsigned	j(max(0, 1 + int(floor(::log(a.normInfty())))));
  unsigned	powj(1 << j);

  a *= 1.0 / double(powj);
  x *= a;
  n += x * 0.5;
  d -= x * 0.5;
  x *= a;
  n += x * 0.113636363636;
  d += x * 0.113636363636;
  x *= a;
  n += x * 0.015151515151;
  d -= x * 0.015151515151;
  x *= a;
  n += x * 0.001262626262;
  d += x * 0.001262626262;
  x *= a;
  n += x * 0.000063131313;
  d -= x * 0.000063131313;

  d.eqInverse();
  x = d * n;

  for (unsigned i(1); i <= j; ++i)
    x *= x;
  return x;
  /*
  Mat4d	d(identity()), n(identity());
  Mat4d	x(identity()), a(*this);
  unsigned	j(max(0, 1 + int(floor(::log(a.normInfty())))));
  unsigned	powj(1 << j);
  double	c(1.0), k, q(6.0);

  a *= 1.0 / double(powj);
  for (k = 1.0; k < q; ++k)
    {
      c *= (q - k + 1.0) / (k * (2.0 * q - k + 1.0));
      x *= a;
      n += x * c;
      d += x * ((int(k) % 2)? -c: c);
    }
  d.eqInverse();
  x = d * n;
  x ^= powj;
  return x;
  */
}

// ***************************************** created by Angelidis Alexis *** //
/// DESCRIPTION: Matrix exponential.
// ******************************************* coded by Alexis Angelidis *** //
inline void	Mat4d::eqExp()
{
  Mat4d	d(identity()), n(identity());
  Mat4d	&x(*this), a(*this);
  unsigned	j(max(0, 1 + int(floor(::log(a.normInfty())))));
  unsigned	powj(1 << j);
  double	c(1.0), k, q(6.0);

  a *= 1.0 / double(powj);
  for (k = 1.0; k < q; ++k)
    {
      c *= (q - k + 1.0) / (k * (2.0 * q - k + 1.0));
      x *= a;
      n += x * c;
      d += x * ((int(k) % 2)? -c: c);
    }
  d.eqInverse();
  x = d * n;
  x ^= powj;
}

// ***************************************** created by Angelidis Alexis *** //
/// DESCRIPTION: Matrix square root.
// ******************************************* coded by Alexis Angelidis *** //
inline Mat4d	Mat4d::sqrt() const
{
  Mat4d	x(*this), ix, y(identity()), iy;
  unsigned	l(4096);

  while (1.0e-6 < (x * x - *this).normInfty() && --l)
    {
      ix = x.inverse();
      iy = y.inverse();
      x += iy;
      x *= 0.5;
      y += ix;
      y *= 0.5;
    }
  return x;
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Matrix square root.
// ******************************************* coded by Angelidis Alexis *** //
inline void	Mat4d::eqSqrt()
{
  Mat4d	&x(*this), ix, y(identity()), iy;
  unsigned	l(4096);

  while (1.0e-6 < (x * x - *this).normInfty() && --l)
    {
      ix = x.inverse();
      iy = y.inverse();
      x += iy;
      x *= 0.5;
      y += ix;
      y *= 0.5;
    }
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Matrix norm.
// ******************************************* coded by Angelidis Alexis *** //
inline double	Mat4d::normInfty() const
{
  unsigned	i, j;
  double	norm(0.0), tmp;
  
  for (i = 0; i < 4; ++i)
    {
      for (tmp = 0.0, j = 0; j < 4; ++j)
	tmp += fabs(operator()(i, j));
      if (norm < tmp)
	norm = tmp;
    }	
  return norm;
}

// ***************************************** created by Alexis Angelidis *** //
inline Mat4d	Mat4d::operator*(const Mat4d &that) const
{
  Mat4d	ret(0.0);

  for (unsigned i(0); i < 4; ++i)
    for (unsigned j(0); j < 4; ++j)
      for (unsigned k(0); k < 4; ++k)
	ret(i, j) += operator()(i, k) * that(k, j);
  return ret;
}

// ******************************************* coded by Alexis Angelidis *** //
/// DESCRIPTION: this *= that => this = this * that
// ***************************************** created by Angelidis Alexis *** //
inline Mat4d	&Mat4d::operator*=(const Mat4d &that)
{
  Mat4d	ret(0.0);

  for (unsigned i(0); i < 4; ++i)
    for (unsigned j(0); j < 4; ++j)
      for (unsigned k(0); k < 4; ++k)
	ret(i, j) += operator()(i, k) * that(k, j);
  *this = ret;
  return *this;
}

// ******************************************* coded by Alexis Angelidis *** //
inline bool	Mat4d::operator!=(const Mat4d &m) const
{
  for (int i(0); i < 16; ++i)
    if (m._m_m[i] != _m_m[i])
      return true;
  return false;
}

// ******************************************* coded by Alexis Angelidis *** //
inline Mat4d	&Mat4d::operator*=(const double d)
{
  for (unsigned i(0); i < 16; ++i)
    _m_m[i] *= d;
  return *this;
}

// ***************************************** created by Alexis Angelidis *** //
inline Mat4d	&Mat4d::operator+=(const Mat4d &that)
{
  for (unsigned i(0); i < 16; ++i)
    _m_m[i] += that._m_m[i];
  return *this;
}

// ******************************************* coded by Angelidis Alexis *** //
inline Mat4d	&Mat4d::operator-=(const Mat4d &that)
{
  for (unsigned i(0); i < 16; ++i)
    _m_m[i] -= that._m_m[i];
  return *this;
}

// ***************************************** created by Alexis Angelidis *** //
inline Mat4d	Mat4d::operator*(const double d) const
{
  return Mat4d
    (_m_m[0] * d, _m_m[4] * d, _m_m[8] * d, _m_m[12] * d,
     _m_m[1] * d, _m_m[5] * d, _m_m[9] * d, _m_m[13] * d,
     _m_m[2] * d, _m_m[6] * d, _m_m[10] * d, _m_m[14] * d,
     _m_m[3] * d, _m_m[7] * d, _m_m[11] * d, _m_m[15] * d);
}

// ******************************************* coded by Angelidis Alexis *** //
inline Mat4d	Mat4d::operator+(const Mat4d &that) const
{
  return Mat4d
    (_m_m[0] + that._m_m[0], _m_m[4] + that._m_m[4],
     _m_m[8] + that._m_m[8], _m_m[12] + that._m_m[12],
     _m_m[1] + that._m_m[1], _m_m[5] + that._m_m[5],
     _m_m[9] + that._m_m[9], _m_m[13] + that._m_m[13],
     _m_m[2] + that._m_m[2], _m_m[6] + that._m_m[6],
     _m_m[10] + that._m_m[10], _m_m[14] + that._m_m[14],
     _m_m[3] + that._m_m[3], _m_m[7] + that._m_m[7],
     _m_m[11] + that._m_m[11], _m_m[15] + that._m_m[15]);
}

// ***************************************** created by Alexis Angelidis *** //
inline Mat4d	Mat4d::operator-(const Mat4d &that) const
{
  return Mat4d
    (_m_m[0] - that._m_m[0], _m_m[4] - that._m_m[4],
     _m_m[8] - that._m_m[8], _m_m[12] - that._m_m[12],
     _m_m[1] - that._m_m[1], _m_m[5] - that._m_m[5],
     _m_m[9] - that._m_m[9], _m_m[13] - that._m_m[13],
     _m_m[2] - that._m_m[2], _m_m[6] - that._m_m[6],
     _m_m[10] - that._m_m[10], _m_m[14] - that._m_m[14],
     _m_m[13] - that._m_m[3], _m_m[7] - that._m_m[7],
     _m_m[11] - that._m_m[11], _m_m[15] - that._m_m[15]);
}

// ******************************************* coded by Alexis Angelidis *** //
/// DESCRIPTION: Returns the k^th integer power of d.
// ******************************************* coded by Angelidis Alexis *** //
inline Mat4d	Mat4d::operator^(const unsigned k) const
{
  Mat4d	ret(identity()), r(*this);

  for (unsigned i(1); i <= k; i *= 2, r *= r)
    if (k & i)
      ret *= r;
  return ret;
}

// ***************************************** created by Alexis Angelidis *** //
inline Mat4d	Mat4d::operator^(const double t) const
{
  return (log() * t).exp();
}

// ******************************************* coded by Alexis Angelidis *** //
inline Mat4d	&Mat4d::operator^=(const double t)
{
  eqLog();
  operator*=(t);
  eqExp();
  return *this;
}

// ***************************************** created by Angelidis Alexis *** //
inline Mat4d	&Mat4d::operator^=(const unsigned k)
{
  Mat4d	r(*this);

  *this = identity();
  for (unsigned i(1); i <= k; i *= 2, r *= r)
    if (k & i)
      operator*=(r);
  return *this;
}

// ******************************************* coded by Alexis Angelidis *** //
inline Vec3d	Mat4d::operator*(const Vec3d &v) const
{
  return Vec3d(_m_m[0] * v.x + _m_m[4] * v.y + _m_m[8] * v.z,
		  _m_m[1] * v.x + _m_m[5] * v.y + _m_m[9] * v.z,
		  _m_m[2] * v.x + _m_m[6] * v.y + _m_m[10] * v.z);
}

// ***************************************** created by Alexis Angelidis *** //
inline Pnt3d	Mat4d::operator*(const Pnt3d &p) const
{
  return Pnt3d
    (_m_m[0] * p.x + _m_m[4] * p.y + _m_m[8] * p.z + _m_m[12],
     _m_m[1] * p.x + _m_m[5] * p.y + _m_m[9] * p.z + _m_m[13],
     _m_m[2] * p.x + _m_m[6] * p.y + _m_m[10] * p.z + _m_m[14]);
}

// ******************************************* coded by Angelidis Alexis *** //
inline BBox3d	Mat4d::operator*(const BBox3d	&b) const
{
  BBox3d	ret(operator*(b.pmin));
  
  ret.fill(operator*(b.pmax));
  ret.fill(operator*(Pnt3d(b.pmax.x, b.pmin.y, b.pmin.z)));
  ret.fill(operator*(Pnt3d(b.pmin.x, b.pmax.y, b.pmin.z)));
  ret.fill(operator*(Pnt3d(b.pmin.x, b.pmin.y, b.pmax.z)));
  ret.fill(operator*(Pnt3d(b.pmin.x, b.pmax.y, b.pmax.z)));
  ret.fill(operator*(Pnt3d(b.pmax.x, b.pmin.y, b.pmax.z)));
  ret.fill(operator*(Pnt3d(b.pmax.x, b.pmax.y, b.pmin.z)));
  return ret;
}

// ***************************************** created by Alexis Angelidis *** //
inline BBox3d	&BBox3d::operator*=(const Mat4d &m)
{
  BBox3d	ret(m * pmin);
  
  ret.fill(m * pmax);
  ret.fill(m * Pnt3d(pmax.x, pmin.y, pmin.z));
  ret.fill(m * Pnt3d(pmin.x, pmax.y, pmin.z));
  ret.fill(m * Pnt3d(pmin.x, pmin.y, pmax.z));
  ret.fill(m * Pnt3d(pmin.x, pmax.y, pmax.z));
  ret.fill(m * Pnt3d(pmax.x, pmin.y, pmax.z));
  ret.fill(m * Pnt3d(pmax.x, pmax.y, pmin.z));
  *this = ret;
  return *this;
}

// ******************************************* coded by Angelidis Alexis *** //
inline const Mat4d	&Mat4d::identity()
{
  static const Mat4d	id(1.0, 0.0, 0.0, 0.0,
			   0.0, 1.0, 0.0, 0.0,
			   0.0, 0.0, 1.0, 0.0,
			   0.0, 0.0, 0.0, 1.0);

  return id;
}

// ***************************************** created by Alexis Angelidis *** //
inline const Mat4d	&Mat4d::symx()
{
  static const Mat4d ix(-1.0, 0.0, 0.0, 0.0,
			   0.0, 1.0, 0.0, 0.0,
			   0.0, 0.0, 1.0, 0.0,
			   0.0, 0.0, 0.0, 1.0);
  return ix;
}

// ******************************************* coded by Alexis Angelidis *** //
inline const Mat4d	&Mat4d::symy()
{
  static const Mat4d iy(1.0, 0.0, 0.0, 0.0,
			   0.0, -1.0, 0.0, 0.0,
			   0.0, 0.0, 1.0, 0.0,
			   0.0, 0.0, 0.0, 1.0);
  return iy;
}

// ******************************************* coded by Angelidis Alexis *** //
inline const Mat4d	&Mat4d::symz()
{
  static const Mat4d iz(1.0, 0.0, 0.0, 0.0,
			   0.0, 1.0, 0.0, 0.0,
			   0.0, 0.0, -1.0, 0.0,
			   0.0, 0.0, 0.0, 1.0);
  return iz;
}

// ***************************************** created by Alexis Angelidis *** //
inline const Mat4d	&Mat4d::sym(const unsigned sym)
{
  if (sym == 1)
    return symx();
  else if (sym == 2)
    return symy();
  else if (sym == 3)
    return symz();
  return identity();
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Positions the matrix this on matrix stack pname.
// ***************************************** created by Alexis Angelidis *** //
inline void	Mat4d::glLoad(const GLenum pname) const
{
  glPushAttrib(GL_TRANSFORM_BIT);
  glMatrixMode(pname);
  glLoadMatrixd(_m_m);
  glPopAttrib();
}

// ******************************************* coded by Alexis Angelidis *** //
/// DESCRIPTION: Translation.
// ***************************************** created by Angelidis Alexis *** //
struct Translation: public Transformation
{
  friend class	Mat4d;
  Vec3d	_m_v;
public:
  Translation(const Vec3d &v): _m_v(v) {}
  virtual ~Translation() {}
  Mat4d	asMatrix() const { return Mat4d(*this); }
  Translation	operator^(const double t) const
  { return Translation(_m_v * t); }
  Translation	&operator^=(const double t) { _m_v *= t; return *this; }
  Transformation	*copy() const { return new Translation(*this); }
  Pnt3d	&operator()(Pnt3d &p) const
  { p += _m_v;  return p; }
  Pnt3d	&operator()(Pnt3d &p, const double t) const
  { p += _m_v * t; return p; }
  void		operator()(Pnt3d &p, Vec3d &) const { p += _m_v;}
  void		cast(const Transformation &t);
  unsigned	nbSteps(const BBox3d &b, const double k) const;
  void		eqSymx() { _m_v(0) = -_m_v(0); }
  void		eqSymy() { _m_v(1) = -_m_v(1); }
  void		eqSymz() { _m_v(2) = -_m_v(2); }
  void		eqInverse() { _m_v = -_m_v; }
  double	det() const { return 1.0; }
  Mat4d	mlog() const;
};

// ******************************************* coded by Alexis Angelidis *** //
inline unsigned	Translation::nbSteps(const BBox3d	&,
				     const double	k) const
{
  const double	s(ceil(1.875 / k * _m_v.norm()));
  
  if (s <= 1.0)
    return 1;
  return unsigned(s);
}

// ******************************************* coded by Angelidis Alexis *** //
inline Mat4d	Translation::mlog() const
{
  return Mat4d(0.0, 0.0, 0.0, _m_v.x,
		  0.0, 0.0, 0.0, _m_v.y,
		  0.0, 0.0, 0.0, _m_v.z,
		  0.0, 0.0, 0.0, 0.0);
}

// ***************************************** created by Alexis Angelidis *** //
inline void	Translation::cast(const Transformation &t)
{
  const Translation *s(dynamic_cast<const Translation *>(&t));

  if (!s)
    {
      std::cerr << "A Translation only cast a Translation." << std::endl;
      std::exit(1);
    }
  *this = *s;
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Rotation around a pivot point. Uses class Quaternion.
// ***************************************** created by Alexis Angelidis *** //
class Rotation: public Transformation
{
  friend class	Mat4d;
protected:
  Quaternion	_m_q;
  Vec3d	_m_c;
public:
  Rotation(const Quaternion &q, const Pnt3d &c): _m_q(q), _m_c(c) {}
  Rotation(const Quaternion &q, const Vec3d &c): _m_q(q), _m_c(c) {}
  virtual ~Rotation() {}
  Mat4d	asMatrix() const { return Mat4d(*this); }
  Rotation	operator^(const double t) const
  { return Rotation(_m_q ^ t, _m_c); }
  Rotation		&operator^=(const double t) { _m_q ^= t; return *this;}
  virtual Transformation*copy() const { return new Rotation(_m_q, _m_c); }
  virtual Pnt3d	&operator()(Pnt3d &p) const
  { p -= _m_c; _m_q(p); p += _m_c; return p; }
  virtual Pnt3d	&operator()(Pnt3d &p, const double t) const
  { p -= _m_c; _m_q(p, t); p += _m_c; return p; }
  virtual void		operator()(Pnt3d &p, Vec3d &v) const
  { p -= _m_c; _m_q(p, v); p += _m_c; }
  void		cast(const Transformation &t);
  unsigned	nbSteps(const BBox3d &b, const double k) const;
  void		eqSymx() { _m_c(0) = -_m_c(0); _m_q.eqSymx(); }
  void		eqSymy() { _m_c(1) = -_m_c(1); _m_q.eqSymy(); }
  void		eqSymz() { _m_c(2) = -_m_c(2); _m_q.eqSymz(); }
  void		eqInverse() { _m_q.eqInverse(); }
  double	det() const { return 1.0; }
  Mat4d	mlog() const;
};

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Transformation that is function of $x$.
// ***************************************** created by Alexis Angelidis *** //
class UnitRotation: public Rotation
{
public:
  UnitRotation(const Quaternion &q, const Pnt3d &c): Rotation(q, c) {}
  UnitRotation(const Quaternion &q, const Vec3d &c): Rotation(q, c) {}
  virtual ~UnitRotation() {}
  virtual Transformation*copy() const { return new UnitRotation(_m_q, _m_c); }
  virtual Pnt3d	&operator()(Pnt3d &p) const
  {
    const Mat4d	m(*this);
    const Quaternion	q(_m_q ^ (1.0 / Vec3d(m * (p - _m_c)).norm()));
    
    p -= _m_c; q(p); p += _m_c; return p;
  }
  virtual Pnt3d	&operator()(Pnt3d &p, const double t) const
  {
    const Mat4d	m(*this);
    const Quaternion	q(_m_q ^ (1.0 / Vec3d(m * (p - _m_c)).norm()));

    p -= _m_c; q(p, t); p += _m_c; return p;
  }
  virtual void		operator()(Pnt3d &p, Vec3d &v) const
  {
    const Mat4d	m(*this);
    const Quaternion	q(_m_q ^ (1.0 / Vec3d(m * (p - _m_c)).norm()));

    p -= _m_c; q(p, v); p += _m_c;
  }
};

// ******************************************* coded by Angelidis Alexis *** //
inline unsigned	Rotation::nbSteps(const BBox3d	&b,
				  const double	k) const
{
  const Vec3d	v(_m_q._m_v.unit());
  Vec3d	p;
  double	l2tmp, lmax(0.0);

  for (int i(0); i < 8; ++i)
    {
      p = b.corner(i) - Pnt3d(_m_c);
      l2tmp = (p - (p * v) * v).norm2();
      if (lmax < l2tmp)
	lmax = l2tmp;
    }
  lmax = sqrt(lmax);

  const double s = ceil(1.875 / k * 2.0 * acos(_m_q._m_s) * lmax);

  if (s <= 1.0)
    return 1;
  return unsigned(s);
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Returns the logarithm matrix of this rotation.
// ******************************************* coded by Angelidis Alexis *** //
inline Mat4d	Rotation::mlog() const
{
  const double	sinTheta(_m_q._m_v.norm());
    
  if (sinTheta == 0.0)
    return Mat4d(0.0);

  const Vec3d	n((2.0 * atan2(sinTheta, _m_q._m_s) /
			   sinTheta) * _m_q._m_v);
  const Vec3d	cross(_m_c ^ n);

  return Mat4d(0.0, -n.z, n.y, cross.x,
		  n.z, 0.0, -n.x, cross.y,
		  -n.y, n.x, 0.0, cross.z,
		  0.0, 0.0, 0.0, 0.0);
}

// ***************************************** created by Alexis Angelidis *** //
inline void	Rotation::cast(const Transformation &t)
{
  const Rotation *r(dynamic_cast<const Rotation *>(&t));

  if (!r)
    {
      std::cerr << "A Rotation can only cast a Rotation." << std::endl;
      std::exit(1);
    }
  *this = *r;
}

// ******************************************* coded by Angelidis Alexis *** //
/// DESCRIPTION: Scale around center point.
// ***************************************** created by Alexis Angelidis *** //
class Scale: public Transformation
{
  friend class	Mat4d;
  double	_m_s;
  Pnt3d	_m_c;
public:
  Scale(const double s, const Pnt3d &c): _m_s(s), _m_c(c) {}
  virtual ~Scale() {}
  Mat4d	asMatrix() const { return Mat4d(*this); }
  Scale		operator^(const double t) const
  { return Scale(pow(_m_s, t), _m_c); }
  Scale	&operator^=(const double t) { _m_s = pow(_m_s, t); return *this; }
  Transformation	*copy() const { return new Scale(_m_s, _m_c); }
  Pnt3d		&operator()(Pnt3d &p) const
  { p = _m_c + (p - _m_c) * _m_s; return p; }
  Pnt3d		&operator()(Pnt3d &p, const double t) const
  { p = _m_c + (p - _m_c) * pow(_m_s, t); return p; }
  void			operator()(Pnt3d &p, Vec3d &) const
  { p = _m_c + (p - _m_c) * _m_s; }
  void		cast(const Transformation &t);
  unsigned	nbSteps(const BBox3d &b, const double k) const;
  void		eqSymx() { _m_c(0) = -_m_c(0); }
  void		eqSymy() { _m_c(1) = -_m_c(1); }
  void		eqSymz() { _m_c(2) = -_m_c(2); }
  void		eqInverse() { _m_s = 1.0 / _m_s; }
  double	det() const { return _m_s * _m_s * _m_s; }
  Mat4d	mlog() const;
};

// ******************************************* coded by Angelidis Alexis *** //
inline unsigned	Scale::nbSteps(const BBox3d	&b,
			       const double	k) const
{
  double	l2tmp, lmax(0.0);

  for (int i(0); i < 8; ++i)
    {
      l2tmp = (b.corner(i) - Pnt3d(_m_c)).norm2();
      if (lmax < l2tmp)
	lmax = l2tmp;
    }
  lmax = sqrt(lmax);

  const double s = ceil(1.875 / k * 2.0 * _m_s * log(_m_s) * lmax);

  if (s <= 1.0)
    return 1;
  return unsigned(s);
}

// ***************************************** created by Alexis Angelidis *** //
inline Mat4d	Scale::mlog() const
{
  const double logs(log(_m_s));

  return Mat4d(logs, 0.0, 0.0, -_m_c.x * logs,
		  0.0, logs, 0.0, -_m_c.y * logs,
		  0.0, 0.0, logs, -_m_c.z * logs,
		  0.0, 0.0, 0.0, 0.0);
}

// ******************************************* coded by Angelidis Alexis *** //
inline void	Scale::cast(const Transformation &t)
{
  const Scale *s(dynamic_cast<const Scale *>(&t));

  if (!s)
    {
      std::cerr << "A Scale can only cast a Scale." << std::endl;
      std::exit(1);
    }
  *this = *s;
}

// ***************************************** created by Alexis Angelidis *** //
/// DESCRIPTION: Scale around center point, along axis _m_d only.
// ******************************************* coded by Angelidis Alexis *** //
class ScaleT: public Transformation
{
  friend class	Mat4d;
  double	_m_s;
  Pnt3d	_m_c;
  Vec3d	_m_d;
public:
  ScaleT(const double s, const Pnt3d &c, const Vec3d &d):
    _m_s(s), _m_c(c), _m_d(d.unit()) {}
  virtual ~ScaleT() {}
  Transformation	*copy() const { return new ScaleT(_m_s, _m_c, _m_d); }
  Mat4d	asMatrix() const { return Mat4d(*this); }
  ScaleT	operator^(const double t) const
  { return ScaleT(pow(_m_s, t), _m_c, _m_d); }
  ScaleT	&operator^=(const double t)
  { _m_s = pow(_m_s, t); return *this; }
  Pnt3d	&operator()(Pnt3d &p) const
  { p += (_m_s - 1.0) * ((p - _m_c) * _m_d) * _m_d; return p; }
  Pnt3d	&operator()(Pnt3d &p, const double t) const
  { p += (pow(_m_s, t) - 1.0) * ((p - _m_c) * _m_d) * _m_d; return p; }
  void		operator()(Pnt3d &p, Vec3d &v) const;
  void		cast(const Transformation &t);
  unsigned	nbSteps(const BBox3d &b, const double k) const;
  void		eqSymx() { _m_c(0) = -_m_c(0); _m_d(0) = -_m_d(0); }
  void		eqSymy() { _m_c(1) = -_m_c(1); _m_d(1) = -_m_d(1);  }
  void		eqSymz() { _m_c(2) = -_m_c(2); _m_d(2) = -_m_d(2);  }
  void		eqInverse() { _m_s = 1.0 / _m_s; }
  double	det() const { return _m_s; }
  Mat4d	mlog() const;
};

// ***************************************** created by Alexis Angelidis *** //
inline unsigned	ScaleT::nbSteps(const BBox3d	&b,
				const double	k) const
{
  double	l2tmp, lmax(0.0);

  for (int i(0); i < 8; ++i)
    {
      l2tmp = (b.corner(i) - Pnt3d(_m_c)).norm2();
      if (lmax < l2tmp)
	lmax = l2tmp;
    }
  lmax = sqrt(lmax);

  const double s = ceil(1.875 / k * 2.0 * _m_s * log(_m_s) * lmax);

  if (s <= 1.0)
    return 1;
  return unsigned(s);
}

// ******************************************* coded by Alexis Angelidis *** //
/// DESCRIPTION: Logarithm of the transformation matrix.
// ***************************************** created by Angelidis Alexis *** //
inline Mat4d	ScaleT::mlog() const
{
  const double logs(log(_m_s));
  const double logsx(logs * _m_d.x), logsy(logs * _m_d.y),logsz(logs * _m_d.z);
  const double scal(Vec3d(_m_c) * _m_d);

  return 
    Mat4d(logsx * _m_d.x, logsx * _m_d.y, logsx * _m_d.z, -logsx * scal,
	     logsy * _m_d.x, logsy * _m_d.y, logsy * _m_d.z, -logsy * scal,
	     logsz * _m_d.x, logsz * _m_d.y, logsz * _m_d.z, -logsz * scal,
	     0.0, 0.0, 0.0, 0.0);
}

// ******************************************* coded by Alexis Angelidis *** //
inline void	ScaleT::cast(const Transformation &t)
{
  const ScaleT *s(dynamic_cast<const ScaleT *>(&t));

  if (!s)
    {
      std::cerr << "A ScaleT can only cast a ScaleT." << std::endl;
      std::exit(1);
    }
  *this = *s;
}

// ***************************************** created by Angelidis Alexis *** //
inline void	ScaleT::operator()(Pnt3d	&p,
				   Vec3d	&v) const
{
  p += (_m_s - 1.0) * ((p - _m_c) * _m_d) * _m_d;
  v += (1.0 / _m_s - 1.0) * (v - (v * _m_d) * _m_d);
  v.eqUnit();
}

// ******************************************* coded by Alexis Angelidis *** //
inline Mat4d::Mat4d(const Quaternion &q)
{
  const double	s(2.0 / q.norm());
  const double	xs(q(1) * s);
  const double	ys(q(2) * s);
  const double	zs(q(3) * s);
  const double	wx(q(0) * xs);
  const double	wy(q(0) * ys);
  const double	wz(q(0) * zs);
  const double	xx(q(1) * xs);
  const double	xy(q(1) * ys);
  const double	xz(q(1) * zs);
  const double	yy(q(2) * ys);
  const double	yz(q(2) * zs);
  const double	zz(q(3) * zs);

  *this = Mat4d(1.0 - (yy + zz), xy - wz, xz + wy, 0.0,
		   xy + wz, 1.0 - (xx + zz), yz - wx, 0.0,
		   xz - wy, yz + wx, 1.0 - (xx + yy), 0.0,
		   0.0, 0.0, 0.0, 1.0);
}

// ***************************************** created by Angelidis Alexis *** //
inline Mat4d::Mat4d(const Translation &t)
{
  const Vec3d	&v(t._m_v);

  *this = Mat4d(1.0, 0.0, 0.0, v.x,
		   0.0, 1.0, 0.0, v.y,
		   0.0, 0.0, 1.0, v.z,
		   0.0, 0.0, 0.0, 1.0);
}

// ******************************************* coded by Alexis Angelidis *** //
inline Mat4d::Mat4d(const Rotation &r)
{
  const Quaternion	&q(r._m_q);
  const Pnt3d		c(r._m_c);

  const double	xs(q(1) * 2.0), ys(q(2) * 2.0), zs(q(3) * 2.0);
  const double	wx(q(0) * xs), wy(q(0) * ys), wz(q(0) * zs);
  const double	xx(q(1) * xs), xy(q(1) * ys), xz(q(1) * zs);
  const double	yy(q(2) * ys), yz(q(2) * zs), zz(q(3) * zs);
  const Mat4d	mat(-yy - zz, xy - wz, xz + wy, 0.0, 
			    xy + wz, -xx - zz, yz - wx, 0.0, 
			    xz - wy, yz + wx, -xx - yy, 0.0,
			    0.0, 0.0, 0.0, 0.0);
  const Pnt3d		tmp = mat * c;

  *this = Mat4d(1.0 + mat(0,0), mat(0, 1), mat(0, 2), -tmp(0),
		   mat(1,0), 1.0 + mat(1, 1), mat(1, 2), -tmp(1),
		   mat(2,0), mat(2, 1), 1.0 + mat(2, 2), -tmp(2),
		   0.0, 0.0, 0.0, 1.0);
}

// ***************************************** created by Angelidis Alexis *** //
inline Mat4d::Mat4d(const Scale &scl)
{
  const double	s(scl._m_s);
  const Pnt3d	&c(scl._m_c);

  *this = Mat4d(s, 0.0, 0.0, (1.0 - s) * c(0),
		   0.0, s, 0.0, (1.0 - s) * c(1),
		   0.0, 0.0, s, (1.0 - s) * c(2),
		   0.0, 0.0, 0.0, 1.0);
}

// ***************************************** created by Alexis Angelidis *** //
inline Mat4d::Mat4d(const ScaleT	&scl)
{
  const double		s(scl._m_s);
  const Pnt3d		&c(scl._m_c);
  const Vec3d	&v(scl._m_d);
  Vec3d		tmp0, tmp1;

  tmp0 = ((1.0 - s) * (Vec3d(c) * v)) * v;
  tmp1 = -(1.0 - s) * v;
  *this = Mat4d
    (1.0 + tmp1(0) * v(0), tmp1(1) * v(0), tmp1(2) * v(0), tmp0(0),
     tmp1(0) * v(1), 1.0 + tmp1(1) * v(1), tmp1(2) * v(1), tmp0(1),
     tmp1(0) * v(2), tmp1(1) * v(2), 1.0 + tmp1(2) * v(2), tmp0(2),
     0.0, 0.0, 0.0, 1.0);
}

// ******************************************* coded by Alexis Angelidis *** //
#endif//__GFXTYPES_HH__

